/* Copyright (c) 2001 S.E. Grigorescu A. Jalba */

#ifndef EDGE_H
#define EDGE_H

#include "Image.h"
#include "IntKernel.h"
#include "FloatImage.h"
#include <vector> 

/*! \ingroup edge_alg
 */
void Normalize(vector< FloatImage > &v);
/*! \fn void Normalize(vector< FloatImage > &v);
  
  This function takes as input a set of images (i.e. \f$v_0\f$, \f$v_1\f$, ..., and \f$v_n\f$), considers that each set of corresponding pixels (i.e. \f$v_0(i,j)\f$, \f$v_1(i,j)\f$, ..., and \f$v_n(i,j)\f$) is a vector, and normalizes these vectors. The normalization is performed only if all of the images stored in \f$v\f$ have the same size.
*/

/*! \ingroup edge_alg
 */ 
void Magnitude(vector< FloatImage > &v, FloatImage &magnit);
/*! \fn void Magnitude(vector< FloatImage > &v, FloatImage &magnit);

  This function takes as input a set of images (i.e. \f$v_0\f$, \f$v_1\f$, ..., and \f$v_n\f$), considers that each set of corresponding pixels (i.e. \f$v_0(i,j)\f$, \f$v_1(i,j)\f$, ..., and \f$v_n(i,j)\f$) is a vector, and computes the magnitude of these vectors. The magnitude of vector \f$ [v_0(i,j), v_1(i,j), ..., v_n(i,j)]\f$ is stored in \f$magnit(i,j)\f$. The computation is performed only if all of the images stored in \f$ v \f$ have the same size.
*/

/*! \ingroup edge_alg
 */
template< class T >
void GradientMap(Image< T > &input, FloatImage &gx, FloatImage &gy);
/*! \fn template< class T > void GradientMap(Image< T > &input, FloatImage &gx, FloatImage &gy);

  Computes the gradient map of the \a input image according to the following formulas:
- on the boundaries (forward and backward differences):
\f[ 
    gx(i,0) = input(i, 1) - input(i, 0), i = 0 ... height-1 \\

    gx(i, width-1) = input(i, width-2) - input(i, width-1), i = 0 ... height-1
\f]
\f[ 
    gy(0,        i) = input(1, i) - input(0, i), i = 0 ... width-1
\f]
\f[ 
    gy(height-1, i) = input(height-2, i) - input(height-1, i), i = 0 ... width-1
\f]
- inside the image (central difference):
\f[ 
    gx(i, j) = input(i+1, j) - input(i-1, j), i = 1 ... height-2, j = 1 ... width-2
\f]
\f[ 
    gy(i, j) = input(i, j+1) - input(i, j-1), i = 1 ... width-2, j = 1 ... width-2
\f]
*/

/*! \ingroup edge_alg
 */  

template< class T >
void EdgePrewitt(Image< T > &input, int type, Image< T > &output);
/*! \fn template< class T > void EdgePrewitt(Image< T > &input, int type, Image< T > &output); 
  
  Detects the edges in the \a input image using Prewitt's edge detector and stores the result in the \a output image. The parameter \a type determines the way the output is computed. 
If \a type is 0 then the output is computed according to the following formula:
\f[ 
    output = |input \star vertical\_prewit\_kernel| + |input \star horizontal\_prewit\_kernel|
\f]
otherwise, the output is computed according to the following formula:
\f[ 
    output = \sqrt{(input \star vertical\_prewit\_kernel)^2 + (input \star horizontal\_prewit\_kernel)^2}
\f]
For more details regarding Prewitt's edge detector algorithm we refer to Sonka M., Hlavac, V. , Boyle R.:Image Processing, Understanding, and Machine Vision, 2nd edition, PWS Boston, 1999.
*/

/*! \ingroup edge_alg
 */  
template< class T >
void EdgeSobel(Image< T > &input, int type, Image< T > &output);
/*! \fn template< class T > void EdgeSobel(Image< T > &input, int type, Image< T > &output);
 
  Detects the edges in the \a input image using Sobel's edge detector and stores the result in the \a output image. The parameter \a type determines the way the output is computed. 
If \a type is 0 then the output is computed according to the following formula:
\f[ 
    output = |input \star vertical\_sobel\_kernel| + |input \star horizontal\_sobel\_kernel|
\f]
otherwise, the output is computed according to the following formula:
\f[ 
    output = \sqrt{(input \star vertical\_sobel\_kernel)^2 + (input \star horizontal\_sobel\_kernel)^2}
\f]
For more details regarding Sobel's edge detector algorithm we refer to Sonka M., Hlavac, V. , Boyle R.:Image Processing, Understanding, and Machine Vision, 2nd edition, PWS Boston, 1999. 
*/
  
/*! \ingroup edge_alg
 */  
template< class T >
void EdgeKirsch(Image< T > &input, Image< T > &output);
/*! \fn template< class T > void EdgeKirsch(Image< T > &input, Image< T > &output);

  Detects the edges in the \a input image using Kirsch's compass edge detector and stores the result in the \a output image. 

For more details regarding Kirsch's compass edge detector algorithm we refer to to Sonka M., Hlavac, V. , Boyle R.:Image Processing, Understanding, and Machine Vision, 2nd edition, PWS Boston, 1999. 
*/

/** \example test_edge.cpp */

#endif 
