/* Copyright (c) 2001 C. Grigorescu */

#ifndef FLOAT_KERNEL_H
#define FLOAT_KERNEL_H

#include "Kernel.h"

#ifndef PI
#define PI 3.14159265358979323846264
#endif

#ifndef MIN
#define MIN(a, b) ( (a<b) ? a : b )
#endif

/**
  Implements data structures and functions for manipulating different types of convolution kernels.
*/

class FloatKernel : public Kernel < float >
{
 private:
  float h_m_x, h_m_y;
  void find_limits(float sigma, float aspect, float orientation);

 public:
  /*! Creates a kernel with the specified \a name without allocating space for the kernel data. */
  FloatKernel(char *name = "") : Kernel< float >(name) {}

  /*! Creates a kernel of size \a wxh with the specified \a name and allocates space for storing the kernel data. */
  FloatKernel(int w, int h, char *name="") : Kernel< float >(w,h,name) {}

/*! Generates a centered 2D rectangular pulse (2D block) with amplitude \a max_amplitude and support in the interval \f$[center-support/2, center+support/2]\f$. On the borders of the 2D block, the kernel value is \a max_amplitude/2.
*/
  void makeRectangularPulse(float max_amplitude, int support);

/*! Generates a centered 2D triangular pulse (2D pyramid) with maximum amplitude \a max_amplitude and support in the interval \f$[center-support/2, center+support/2]\f$.
*/
  void makeTriangularPulse(float max_amplitude, int support);

/*! Generates a 2D isotropic Gaussian kernel \f$g(x,y)\f$. The parameter \a sigma determines the shape of the kernel according to the formula:

   \f[
   g(x,y) = \frac{1}{\sqrt{2 \pi} \, \sigma} exp\left(- \frac{x^2 + y^2}{2 {\sigma}^2}\right)
   \f]
 */
  void makeGauss(float sigma); 

/*! Generates a 2D oriented Gaussian kernel \f$g(x,y)\f$. The parameters \a sigma, \a aspect, and \a orientation (in degrees) determine the shape of the kernel according to the formulas:

   \f[
   g(x,y) = \frac{1}{\sqrt{2 \pi} \, \sigma} exp\left(- \frac{x_1^2 + {(aspect*y_1)}^2}{2 {\sigma}^2}\right)
   \f]
   where
   \f[    x_1 = x*cos(orientation) - y*sin(orientation) 
   \f] \f[ y_1 = x*sin(orientation) + y*cos(orientation) \f]
   
 */
  void makeGauss(float sigma, float aspect, float orientation); 

/*! Generates a 2D isotropic Difference of Gaussians kernel \f$g(x,y)\f$. The parameters \a sigma and \a ratio determine the shape of the kernel according to the formulas:

   \f[
   g(x,y) = g_1(x,y) - g_2(x,y)
   \f]
   where:
   \f[
   g_1(x,y) = \frac{1}{\sqrt{2 \pi} \sigma} exp\left(- \frac{x^2 + y^2}{2 {\sigma}^2}\right)
   \f]
   \f[
   g_2(x,y) = \frac{1}{\sqrt{2 \pi} \, ratio * \sigma} exp\left(- \frac{x^2 + y^2}{2 {(ratio*\sigma)}^2}\right)
   \f]

   The kernel has zero mean and \f$||g(x,y)|| = 1\f$.
 */
  void makeDoG(float sigma, float ratio); 

/*! Generates a 2D oriented Difference of Gaussians kernel \f$g(x,y)\f$. The parameters \a sigma, \a aspect, \a orientation (in degrees), and \a ratio determine the shape of the kernel according to the formulas:

   \f[
   g(x,y) = g_1(x,y) - g_2(x,y)
   \f]
   where:
   \f[
   g_1(x,y) = \frac{1}{\sqrt{2 \pi} \sigma} exp\left(- \frac{x_1^2 + {(aspect*y_1)}^2}{2 {\sigma}^2}\right)
   \f]
   \f[
   g_2(x,y) = \frac{1}{\sqrt{2 \pi} \, ratio * \sigma} exp\left(- \frac{x_1^2 + {(aspect*y_1)}^2}{2 {(ratio*\sigma)}^2}\right)
   \f]
   and
   \f[ x_1 = x*cos(orientation) - y*sin(orientation) \f] 
   \f[ y_1 = x*sin(orientation) + y*cos(orientation) \f]

   The kernel has zero mean and \f$||g(x,y)|| = 1\f$.
 */
  void makeDoG(float sigma, float aspect, float orientation, float ratio);

/*! Generates a 2D oriented Gabor kernel \f$g(x,y)\f$. The parameters \a sigma, \a aspect, \a orientation (in degrees), \a lambda, and \a phase (in degrees) determine the shape of the kernel according to the formulas:

   \f[
   g(x,y) = \frac{1}{\sqrt{2 \pi} \sigma} exp\left(- \frac{x_1^2 + {(aspect*y_1)}^2}{2 {\sigma}^2}\right) * 
   cos\left( 2 \pi \frac{x_1}{\lambda} + phase \right)
   \f]
   where:
   \f[    x_1 = x*cos(orientation) - y*sin(orientation) 
   \f] \f[ y_1 = x*sin(orientation) + y*cos(orientation) \f]

   The kernel has zero mean and \f$||g(x,y)|| = 1\f$.
*/
  void makeGabor(float sigma, float aspect, float orientation,
		 float lambda, float phase);

/*! Generates a 2D isotropic kernel \f$g(x,y)\f$, which is a Laplacian (second order derivative) of a Gaussian. The parameter \a sigma determines the shape of the kernel according to the formula:

   \f[
   g(x,y) = \frac{1}{\pi {\sigma}^4} \left[ 1 - \frac{x^2+y^2}{2 {\sigma}^2} \right] 
   exp\left(- \frac{x^2 + {(aspect*y)}^2}{2 {\sigma}^2}\right) 
   \f]

   The kernel has zero mean and \f$||g(x,y)|| = 1\f$.
*/
  void makeLoG(float sigma);

/*! Generates a 2D oriented kernel \f$g(x,y)\f$, which is a Laplacian (second order derivative) of a Gaussian. The parameters \a sigma , \a aspect, and \a orientation (in degrees) determine the shape of the kernel according to the formulas:

   \f[
   g(x,y) = \frac{1}{\pi {\sigma}^4} \left[ 1 - \frac{x_1^2+y_1^2}{2 {\sigma}^2} \right] 
   exp\left(- \frac{x_1^2 + {(aspect*y_1)}^2}{2 {\sigma}^2}\right) 
   \f]
   where:
   \f[    x_1 = x*cos(orientation) - y*sin(orientation) 
   \f] \f[ y_1 = x*sin(orientation) + y*cos(orientation) \f]

   The kernel has zero mean and \f$||g(x,y)|| = 1\f$.
  
 */
  void makeLoG(float sigma, float aspect, float orientation);
};

#endif
