/* Copyright (c) 2001 C. Grigorescu */
 
#ifndef _FOURIER_TRANSFORM_H_
#define _FOURIER_TRANSFORM_H_
 
#include "FloatImage.h"
 
#define HALF_SHIFT(A,size_x,size_y) \
   { float temp; \
     int i,j; \
     for (j=0; j<(size_y)/2; j++)  \
       for (i=0; i<(size_x)/2; i++)  \
         { temp = A[(size_x)*j+i]; \
           A[(size_x)*j+i] = A[(size_x)*(j+((size_y)/2))+i+(size_x)/2]; \
           A[(size_x)*(j+((size_y)/2))+i+(size_x)/2] = temp;} \
     for (j=(size_y)/2; j<(size_y); j++)  \
       for (i=0; i<(size_x)/2; i++)  \
         { temp = A[(size_x)*j+i]; \
           A[(size_x)*j+i] = A[(size_x)*(j-((size_y)/2))+i+(size_x)/2]; \
           A[(size_x)*(j-((size_y)/2))+i+(size_x)/2] = temp;} \
   }
 
/*! \ingroup img_trans
  Implements the Fast Fourier Transform (FFT) of an image. The way in which the FFT is performed depends on the presence of FFTW package (see instalation notes of TiP). If FFTW package is not installed, the class FourierTransform allows FFT only for images having both dimensions equal to powers of two. If one or both dimensions of the input image are not equal to powers of two, then the input image is padded with zeros until both dimensions become equal to powers of two and the FFT is performed on the padded image.
 
  However, if the FFTW package is installed, the FFT can be applied on images with arbitrary dimensions. In this case padding to even width and height is performed for speeding-up the transform. Another reason for padding is that TiP uses the FFT to perform convolution with floating point convolution kernels which have to be aligned before the Fourier transformation takes place. For more information about "FFTW" library (The Fastest Fourier Transform in the West), please consult http://www.fftw.org.
*/ 
class FourierTransform
{
 private:
  FloatImage real_part;
  FloatImage imag_part;
  int pid_magnitude;
  int pid_phase;
  void _fft(float *data, int nn, int isign);
 
 public:
  /*! Constructor that takes as argument the representation of an image in space domain. */
  FourierTransform(FloatImage &input);
 
  /*! Constructor which takes as argument the representation of an image in Fourier domain; the first argument represents the real part, while the second argument is the imaginary part. */ 
  FourierTransform(FloatImage &input_real, FloatImage &input_imag);
 
  /*! Performs Direct Fast Fourier Transform. */
  void fft();
 
  /*! Performs Inverse Fast Fourier Transform. */
  void ifft();
 
  /*! Displays:
    - the magnitude of the Fourier coefficients, if "log_values" = 0;
    - the logarithm of the magnitude of the Fourier coefficients, if "log_values" differs from 0. */
  void showMagnitude(int log_values);
 
  /*! Displays:
    - the phase of the Fourier coefficients, if "log_values" = 0;
    - the logarithm of the phase of the Fourier coefficients, if "log_values" differs from 0.
*/
  void showPhase(int log_values);
 
  /*! Closes the magnitude and the phase windows. */
  void closeWindows();
 
  /*! Gets the real part of the values. */
  FloatImage getReal() { return real_part;} 
 
  /*! Gets the imaginary part of the values. */
  FloatImage getImag() { return imag_part;}
};
 
void  fft(FloatImage &input, FloatImage &output_real, FloatImage &output_imag);
/*! \ingroup img_trans
  \fn fft(FloatImage &input, FloatImage &output_real, FloatImage &output_imag);
  Performs a Fast Fourier Transform (FFT) on the \a input image; the result is a complex-valued image, whose real and imaginary parts are stored as output images in \a output_real and \a output_imag, respectively. For more functionality please refer to the FourierTransform class.
*/
/* \example test_fft2.cpp */
 
void ifft(FloatImage &input_real, FloatImage &input_imag, FloatImage &output);
/*! \ingroup img_trans
  \fn ifft(FloatImage &input_real, FloatImage &input_imag, FloatImage &output);
  Performs an Inverse Fast Fourier Transform (IFFT) on a complex-valued image whose real and imaginary parts are denoted by the parameters \a input_real and \a input_imag. The real part of the output is returned in \a output.  For more functionality please refer to the FourierTransform class.
*/
 
/** \example test_fft.cpp */
/** \example test_fft2.cpp */
 
#endif                                                
