/* Copyright (c) 2001 S.E. Grigorescu */

#ifndef HISTOGRAM_H
#define HISTOGRAM_H

#include "config.h"
#include "Image.h"
#include "sys-defines.h"
#ifndef EXTERN_H
   #include "extern.h"
#endif
#ifdef HAVE_LIBPLOTTER
   #include "graph.h"
#endif

/** \ingroup hist_alg
    Implements data structures and functions for computing, manipulating and outputting the histogram of an image.
 */

template< class T > class Histogram 
{
  void compute_histogram();    
/*! \fn void compute_histogram();

  Computes the gray level histogram of the image #img using the number of bins given by #bins. If in the constructor was specified a range for which the histogram to be computed, then only those values in the specified range are taken in consideration in the histogram computation.
*/

  double min_x;
/*! Determines the lower limit of the range for which the histogram to be computed. 
*/

  double max_x;
/*! Determines the upper limit of the range for which the histogram to be computed. 
*/

  int bins; 
  /*! The number of bins.*/

  double *hist;
  /*! The buffer storing the histogram. */

  Image< T > img;
  /*! An internal image buffer in which the image given as parameter to the constructor is copied. All histogram related operations use this image rather than the image given as parameter to the constructor. */


#ifdef HAVE_LIBPLOTTER
  bool add_point(double y, double x1, double x2, int pos, Point *p);
/*! \fn bool add_point(double y, double x1, double x2, 
		 int pos, Point *p);

  This function is used by the drawing subroutines (i.e. showHistogram() and writeHistogram()) to add the visual representation of a bin to the currently plotted histogram. A bin is represented by a bar whose dimensions and position in the drawing window are determined by parameters \a y, \a x1, and \a x2. Parameter \a p is a pointer to an array of entities representing the visual representation of the bins. Parameter \a pos indicates on which position in the array pointed by \a p the current bin will be inserted.
*/

  void set_point_attr(Point &point);
/*! \fn void set_point_attr(Point &point);

  Function related to the drawing methods (i.e. showHistogram() and writeHistogram()). It is available only if Plotutils package with libplotter class library is installed (see instalation notes). 

  This function is used by the drawing subroutines to set the characteristics of \a point, which is the visual representation of a bin. It sets the type of contour and filling intensity for the bar representing a bin.
*/

  bool x_range_spec;
  bool y_range_spec;
  Grapher *grapher_disp;
  Grapher *grapher_wrt;
  int Width, Height;
  const char *bg;
  const char *top_label, *x_label, *y_label;
  const char *title_font_name;
  const char *xy_font_name;
  double title_font_size, xy_font_size;
  double final_min_x, final_max_x;
  double final_min_y, final_max_y; 
  double final_spacing_x, final_spacing_y;  
  int final_log_axis;
  bool final_transpose_axes;
  double fill_fraction;
#endif

 public:
/*! \ingroup hist_alg
    Constructor - It computes the gray level histogram of the input image using the indicated number of bins. It takes in consideration all gray levels in the input image. */ 
  Histogram(Image< T > &input, int no_bins = 100);

/*! \ingroup hist_alg
    Constructor - It computes the gray level histogram for those pixels in the input image whose gray levels are in the range [min; max] using the indicated number of bins. */
  Histogram(Image< T > &input, int no_bins, T min, T max);

/*! Destructor */ 
  ~Histogram();

  double* getHistogram() {return hist;}
/*! \fn double* getHistogram();

  Returns a pointer to the buffer storing the histogram.
*/

  void equalize(Image< T > &output);
/*! \ingroup hist_alg
    \fn void equalize(Image< T > &output);

  Performs histogram equalization of the image #img and stores the result in the \a output image. If in the constructor a range was specified for which the histogram to be computed, then the equalization is applied only to those values inside the specified range.
*/

#ifdef HAVE_LIBPLOTTER
  void setDimensions(int w = 600, int h = 600) {
    Width = w;
    Height = h;
  }
/*! \fn void setDimensions(int w = 600, int h = 600);

  Function related to the drawing methods (i.e. showHistogram() and writeHistogram()). It is available only if Plotutils package with libplotter class library is installed (see instalation notes). 

  It sets the size of the window in which the histogram is drawn.
*/

  void setBgColor(const char *color) {bg = color; }
/*! \fn void setBgColor(const char *color);

  Function related to the drawing methods (i.e. showHistogram() and writeHistogram()). It is available only if Plotutils package with libplotter class library is installed (see instalation notes). 

  It sets the background color of the window in which the histogram is drawn. The supported colors and the way of specifying them are given in the section "Specifying Colors by Name" of the documentation of the GNU Plotutils package (see http://www.gnu.org/manual/plotutils/).
*/

  void setTitle(const char *title) { top_label = title; }
/*! \fn void setTitle(const char *title);

  Function related to the drawing methods (i.e. showHistogram() and writeHistogram()). It is available only if Plotutils package with libplotter class library is installed (see instalation notes). 

  It sets the title of the histogram.
*/

  void setTitleFontName(const char *title_font) { title_font_name = title_font; }
/*! \fn void setTitleFontName(const char *title);

  Function related to the drawing methods (i.e. showHistogram() and writeHistogram()). It is available only if Plotutils package with libplotter class library is installed (see instalation notes). 

  It sets the font used for the title of the histogram. The supported fonts and the way of specifying them are given in the section "Available text fonts" of the documentation of the GNU Plotutils package (see http://www.gnu.org/manual/plotutils/).
*/

  void setTitleFontSize(double title_size) { title_font_size = title_size; }
/*! \fn void setTitleFontSize(double title_size);

  Function related to the drawing methods (i.e. showHistogram() and writeHistogram()). It is available only if Plotutils package with libplotter class library is installed (see instalation notes). 

  It sets the size of the font used for the title of the histogram. The size is specified as a fraction of the minimum dimension (width or height) of the plotting window.
*/

  void setXLabel(const char *lb) { x_label = lb; }
/*! \fn void setXLabel(const char *lb);

  Function related to the drawing methods (i.e. showHistogram() and writeHistogram()). It is available only if Plotutils package with libplotter class library is installed (see instalation notes). 

  It sets the label of the x-axis.
*/

  void setYLabel(const char *lb) { y_label = lb; }
/*! \fn void setYLabel(const char *lb);

  Function related to the drawing methods (i.e. showHistogram() and writeHistogram()). It is available only if Plotutils package with libplotter class library is installed (see instalation notes). 

  It sets the label of the y-axis.
*/

  void setXYFontName(const char *lb) { xy_font_name = lb; }
/*! \fn void setXYFontName(const char *lb);

  Function related to the drawing methods (i.e. showHistogram() and writeHistogram()). It is available only if Plotutils package with libplotter class library is installed (see instalation notes). 

  It sets the font used for the axis labels. The supported fonts and the way of specifying them are given in the section "Available text fonts" of the documentation of the GNU Plotutils package (see http://www.gnu.org/manual/plotutils/).
*/

  void setXYFontSize(double lb_size) { xy_font_size = lb_size; }
/*! \fn void setXYFontSize(double lb);

  Function related to the drawing methods (i.e. showHistogram() and writeHistogram()). It is available only if Plotutils package with libplotter class library is installed (see instalation notes). 

  It sets the size of the font used for the axis labels. The size is specified as a fraction of the minimum dimension (width or height) of the plotting window.
*/

  void setXRange(T x_min, T x_max) {
    final_min_x = x_min;
    final_max_x = x_max;
    x_range_spec = true;
  }
/*! \fn void setXRange(T x_min, T x_max);

  Function related to the drawing methods (i.e. showHistogram() and writeHistogram()). It is available only if Plotutils package with libplotter class library is installed (see instalation notes). 

  It determines which part of the histogram is drawn in the horizontal direction. In other words, it selects only a set from the computed bins to be drawn. By default, all of the bins are plotted when a displaying method is called unless a call to setXRange() was made previously.
*/

  void resetXRange() { x_range_spec = false; }
/*! \fn void resetXRange();

  Function related to the drawing methods (i.e. showHistogram() and writeHistogram()). It is available only if Plotutils package with libplotter class library is installed (see instalation notes). 

  It is complementary to the function setXRange(), determining that all bins be drawn. By default, all of the bins are plotted when a displaying method is called unless a call to setXRange() was made previously.
*/

  void setYRange(T y_min, T y_max) {
    final_min_y = y_min;
    final_max_y = y_max;
    y_range_spec = true;
  }
/*! \fn void setYRange(T y_min, T y_max);

  Function related to the drawing methods (i.e. showHistogram() and writeHistogram()). It is available only if Plotutils package with libplotter class library is installed (see instalation notes). 

  It determines which part of the histogram is drawn in the vertical direction. In other words, it selects only a part from the bars representing the bins to be drawn. By default, the bars representing the bins are entirely drawn when a displaying method is called unless a call to setYRange() was made previously.

*/

  void resetYRange() { y_range_spec = false; }
/*! \fn void resetYRange();

  Function related to the drawing methods (i.e. showHistogram() and writeHistogram()). It is available only if Plotutils package with libplotter class library is installed (see instalation notes). 

  It is complementary to the function setYRange(), determining that the entire bars representing the bins be drawn. By default, the bars representing the bins are entirely drawn when a displaying method is called unless a call to setYRange() was made previously.
*/

  void setXTicks(double xt) { final_spacing_x = xt; }
/*! \fn void setXTicks(double xt);

  Function related to the drawing methods (i.e. showHistogram() and writeHistogram()). It is available only if Plotutils package with libplotter class library is installed (see instalation notes). 

  It sets the distance between ticks on the x-axis.
*/

  void setYTicks(double yt) { final_spacing_y = yt; }
/*! \fn void setYTicks(double yt);

  Function related to the drawing methods (i.e. showHistogram() and writeHistogram()). It is available only if Plotutils package with libplotter class library is installed (see instalation notes). 

  It sets the distance between ticks on the y-axis.
*/

  void setLogMode(bool lg = true) { final_log_axis = (int)lg; }
/*! \fn setLogMode(bool lg = true);

  Function related to the drawing methods (i.e. showHistogram() and writeHistogram()). It is available only if Plotutils package with libplotter class library is installed (see instalation notes). 

  It set the axes to be log axes rather than linear axes, or vice versa. By default, the axes are linear axes.
*/

  void transposeAxes(bool tr = true) { final_transpose_axes = tr; }
/*! \fn void transposeAxes(bool tr = true);

  Function related to the drawing methods (i.e. showHistogram() and writeHistogram()). It is available only if Plotutils package with libplotter class library is installed (see instalation notes). 

  It determines if the abscissa and the ordinate are transposed. The drawing methods display on the abscissa the gray level values while on the ordinate, they display the number of pixels corresponding to each bin. By default, the axes are not transposed.
*/

  void setFillIntensity(double fill) { fill_fraction = fill; }
/*! \fn void setFillIntensity(double fill);

  Function related to the drawing methods (i.e. showHistogram() and writeHistogram()). It is available only if Plotutils package with libplotter class library is installed (see instalation notes). 

  It sets the gray intensity inside the bars representing the bins of the plotted histogram. By default, the bars are filled with black (intensity 1).
*/

  void showHistogram();
/*! \ingroup hist_alg
    \fn void showHistogram();

  Drawing method available only if Plotutils package with libplotter class library is installed (see instalation notes). 

  It opens a window and plots in it the histogram or parts of it. The way of plotting the histogram as well as which part of it is plotted are determined by a number of parameters, which can be adjusted using methods provided by the Histogram class (we call these methods "methods related to the drawing subroutines").
*/

  void writeHistogram(char *type, char *file_name);
/*! \ingroup hist_alg
    \fn void writeHistogram(char *type, char *file_name);

  Drawing method available only if Plotutils package with libplotter class library is installed (see instalation notes). 

  It plots the histogram or parts of it in the file indicated by the parameter \a file_name, in a format given by the parameter \a type. The allowed types are: gif, PS, fig, and bmp. The way of plotting the histogram as well as which part of it is plotted are determined by a number of parameters, which can be adjusted using methods provided by the Histogram class (we call these methods "methods related to the drawing subroutines"). */

  void closeWindow();
/*! \ingroup hist_alg
    \fn void closeWindow();

  Function related to the drawing methods (i.e. showHistogram() and writeHistogram()). It is available only if Plotutils package with libplotter class library is installed (see instalation notes). 

  It closes the window opened by showHistogram().
*/
#endif

};

/** \example test_hist.cpp */
#endif 

