/* Copyright (c) 2001 C. Grigorescu S.E. Grigorescu A. Jalba */

#ifndef IMAGE_H
#define IMAGE_H

#include <stdlib.h>
#include <string.h>
#include <malloc.h>
#include <signal.h> //only for kill
#include "imlib_io.h"
#include "Exceptions.h"
#include <stdio.h>

/** 
    Implements data structures and functions for manipulating all types of gray level images.
*/

template< class T > class Image 
{
  /*! \example test_image.cpp 
   */

 protected: 
  /*! Stores the width of the current image. */
  int width;

  /*! Stores the height of the current image. */
  int height; 

  /*! The image data is stored in a contiguous block of memory row by row. In order to simplify the access to each row, we built an array that stores pointers to the beginning of each row. The variable #pixels stores a pointer this array. In other words, the variable #pixels is a pointer to an array of pointers to each row in the current image. In this way, one can access the gray level in row i and column j using one of the following syntax: pixels[i][j] or pixels[0][i * #width + j]. */
  T **pixels;

  /*! Stores the name of the current image. This name is used as a file name when the image is saved on / read from the disk.*/
  char img_name[256]; 

  int pid; //only to close a window 
  
 public:
  /*! Creates an image with the specified \a name without allocating space for the image data. */
  Image(char * name = "");

  /*! Creates an image of size \f$w \times h\f$ with the specified \a name and allocates space for storing the image data. */
  Image(int w, int h, char *name = "");

  /*! Copy constructor. */
  Image(const Image<T>&);

  /*! Destructor. */
  virtual ~Image();

  /*! Returns the image width. */
  int getWidth() const { return width; }

  /*! Returns the image height. */
  int getHeight() const { return height; }

  /*! Returns the image size - \f$w \times h\f$. */
  int getSize() { return width*height; }

  /*! Returns the image name. */
  char *getName() { return img_name; }

  /*! Sets the image name. */
  void setName(const char *name);

  /*! Allocates a buffer of size \f$w \times h \times sizeof(T)\f$ for storing the image data and sets the allocated memory to "0". Sets #pixels to point out to the allocated buffer. If a buffer for storing image data was already allocated, it is freed. */
  void makeImage(int w, int h);

  /*! Frees the memory allocated to store the image data. */	
  void destroyImage();

  /*! Returns a pointer to the actual image buffer. */ 
  T *getPixels() { return pixels[0]; }

  /*! Creates an image buffer of size \f$w \times h \times sizeof(T)\f$ and initializes it with the data in the buffer pointed by \a data. */	
  void setPixels(int w, int h, T *data);

  /*! \ingroup img_st 
    Returns the minimum gray level in the current image. */
  T min();

  /*! \ingroup img_st 
    Returns the maximum gray level in the current image. */
  T max();

  /*! \ingroup img_st 
    Returns the average value of the gray levels in the current image. */
  float avg();

  /*! \ingroup img_st 
    Returns the standard deviation of the gray level distribution in the current image. */
  float dev();
  
  /*! \ingroup point_alg 
    Overloaded access operator. Allows the syntax Image[i][j] for accessing the pixel in the i-th row and j-th column.
*/
  T *operator[](int i) {
    if ((i > height) || (i < 0))
      throw Fatal_error("Index out of range"); 
    else	
      return pixels[i];
  }

  /*! \ingroup point_alg 
    Overloaded access operator. Allows the syntax Image(k) for accessing linearly the \link #pixels image buffer\endlink. If k = i * #width + j, then Image(k) is the pixel in the i-th row and j-th column.
*/
  T &operator()(int i) { 
    if ((i > getSize()) || (i < 0))
      throw Fatal_error("Index out of range"); 
    else	
      return pixels[0][i]; 
  }

/* Addressing of the internal image buffer with two indices (i-row, j-column). */
/*    T &operator()(int i, int j) {  */
/*      if ((i>getHeight()) || (i < 0)) */
/*        throw Fatal_error("Height index out of range");  */
/*      else  */
/*        if ((j>getWidth()) || (j < 0)) */
/*  	throw Fatal_error("Width index out of range");  */
/*        else */
/*  	return pixels[i][j];  */
/*    } */

  /*! \ingroup point_alg 
    Copy operator. 
*/
  Image< T > &operator=(const Image< T > &img); 

  /*! \ingroup point_alg 
    Tests if two images have the same size and the same content. 
*/
  int operator==(const Image< T > &img);

  /*! \ingroup point_alg 
    Tests if two images are different in size or content. 
*/
  int operator!=(const Image< T > &img);

  /*! \ingroup point_alg 
    Sets all the pixels in the current image to the specified value. For some "mysterious" reasons this operator doesn't work in the classes derived from the class Image. To compensate for this, another function (clearImage()) was provided to do the same job. */ 
  Image< T > &operator=(const T val);

  /*! \ingroup point_alg 
    Sets all the pixels in the current image to the specified value. 
    \sa operator=(const T val) */ 
  void clearImage(const T val);

  /*! \ingroup point_alg 
    Adds a constant to the value of each pixel in the current image. 
*/
  Image< T > &operator+=(const T val);

  /*! \ingroup point_alg
    Subtracts a constant from the value of each pixel in the current image.
*/
  Image< T > &operator-=(const T val);

  /*! \ingroup point_alg 
    Multiplies by a constant the value of each pixel in the current image. 
*/
  Image< T > &operator*=(const T val);

  /*! \ingroup point_alg
    Adds pixelwise an image to the current one. The operation is performed only if the two images have the same size. 
*/
  Image< T > &operator+=(const Image< T > & img);

  /*! \ingroup point_alg
    Subtracts pixelwise an image from the current one.  The operation is performed only if the two images have the same size. 
*/
  Image< T > &operator-=(const Image< T > & img);

  /*! \ingroup point_alg
    Multiplies pixelwise the current image by another image.  The operation is performed only if the two images have the same size. 
*/
  Image< T > &operator*=(const Image< T > & img);
  
  /*! \ingroup geom_alg 
    Pads the current image with a vertical border of size \a w and a horizontal border of size \a h. The type of padding depends on the parameter \a p. Currently defined padding types are: 
    - ZEROS - the added borders are filled with 0;
    - REFLECTION - the added borders are reflections of the initial borders;
    - WRAP - each added border is a copy of the opposite initial border.
  */
  void pad(int w, int h, PADDING_TYPE p);

  /*! \ingroup geom_alg 
    Replace the current image with a piece of it - that piece delimited by a rectangle whose top-left corner is \a (x1,y1) and bottom-right corner is \a (x2,y2). */
  void crop(int x1, int y1, int x2, int y2);

  /*! \ingroup geom_alg 
    If the current image has one or both dimensions odd valued, makes the odd valued dimension(s) even by duplicating the last row and/or column. Such a modification is, usually, necessary for image transforms such as FFT. */
  void makeEvenSized(void);
  
  /*! \ingroup io
    Closes the window opened by showImage().  */
  void closeWindow();

  /*! \ingroup io
    Reads an image from file \a name and stores it in the buffer pointed out by #pixels; the image format is "guessed" from the file name extension. */
  void readImage(char *name) { 
    T *d = read_image< T >(name, &width, &height);
    setPixels(width, height, d); 
    setName(name);    
  }

  /*! \ingroup io 
    Writes the current image to the file given by #img_name; the image format is "guessed" from the file name extension. */
  void writeImage()
  { 
    if (width * height != 0) 
      write_image(img_name, width, height, pixels[0]); 
    else
      ERROR("Empty image; nothing to write");
  }

  /*! \ingroup io 
    Writes the current image to the file \a name; the image format is "guessed" from the file name extension. */
  void writeImage(char *name)
  { 
    setName(name);
    if (width * height != 0) 
      write_image(img_name, width, height, pixels[0]); 
    else
      ERROR("Empty image; nothing to write");
  }

  /*! \ingroup io
    Opens a window and renders in it the current image. */
  int showImage()  { 
    if (width * height != 0) {
      char name[128];
      strncpy(name, img_name, 110);
      closeWindow();  
      pid = show_image< T > (pixels[0], width, height, name);	  
    }
    else {
      pid = 0;
      ERROR("Empty image; cannot display it!");
    }    
    return(pid);
  }

  // Bellow are some functions intended to be used BUT
  // not used currently for various reasons:

  // "+","-", "*" operators: as friends functions 
  //  friend Image< T > operator+< T >(const Image< T >& img1, const Image< T >& img2);
  //  friend Image< T > operator*< T >(const Image< T >& img1, const Image< T >& img2);
  //  friend Image< T > operator-< T >(const Image< T >& img1, const Image< T >& img2);
  // - not used because they allocate memory without freeing it - a garbage
  //   collection mechanism would be necessary (see Stroustroup p. 282)

  //  "+", "-", "*" operators as members of the class 
  //  Image< T > &operator+(const Image< T > &img);
  //  Image< T > &operator-(const Image< T > &img);
  //  Image< T > &operator*(const Image< T > &img);
  // - not used because they allocate memory without freeing it - a garbage
  //   collection mechanism would be necessary (see Stroustroup p. 282)
};

/*  template< class T > Image<T> operator+(const Image<T>&, const Image<T>&); */
/*  template< class T > Image<T> operator-(const Image<T>&, const Image<T>&); */
/*  template< class T > Image<T> operator*(const Image<T>&, const Image<T>&); */

#endif

