/* Copyright (c) 2001 C. Grigorescu */

#ifndef KERNEL_H
#define KERNEL_H

#include <stdlib.h>
#include <string.h>
#include <malloc.h>
#include <assert.h>
#include <iostream.h>
#include <typeinfo>
#include "imlib_io.h"
#include "Exceptions.h"
#include "Image.h"

enum BYTE_KERNEL_TYPE { 
  CROSS_3x3    = 100, 
  DISK_3x3     = 101,
  KERNEL_BYTE_UNKNOWN = 102};

enum INT_KERNEL_TYPE { 
  LAPLACIAN_1    = 200, 
  LAPLACIAN_2    = 201, 
  SOBEL_H        = 202,
  SOBEL_V        = 203,
  PREWITT_H      = 204,
  PREWITT_V      = 205,
  KIRSCH_1       = 206,
  KIRSCH_2       = 207,
  KIRSCH_3       = 208,
  KIRSCH_4       = 209,
  KIRSCH_5       = 210,
  KIRSCH_6       = 211,
  KIRSCH_7       = 212,
  KIRSCH_8       = 213,
  KERNEL_INT_UNKNOWN    = 214};

enum FLOAT_KERNEL_TYPE {
  RECTANGULAR_PULSE    = 300,
  TRIANGULAR_PULSE     = 301,
  GAUSS                = 302,
  DOG                  = 303,
  GABOR                = 304,
  LOG                  = 305,
  KERNEL_FLOAT_UNKNOWN = 306};

/**
    Implements data structures and functions for manipulating all types of convolution kernels. A convolution kernel is regarded as a particular type of image.
*/

template< class T > class Kernel : public Image < T >
{
/** \example test_kernel.cpp */
 protected:
  int kernel_type;
 
 public:
  /*! Creates a kernel with the specified \a name without allocating space for the kernel data. */
  Kernel(char *name = "") : Image< T > (name){  
    if (typeid(T) == typeid(byte)) kernel_type = KERNEL_BYTE_UNKNOWN;
    if (typeid(T) == typeid(int)) kernel_type = KERNEL_INT_UNKNOWN;
    if (typeid(T) == typeid(float)) kernel_type = KERNEL_FLOAT_UNKNOWN;
  }

  /*! Creates a kernel of size \a wxh with the specified \a name and allocates space for storing the kernel data. */
  Kernel(int w, int h, char *name = "") : Image< T >(w,h,name) {  
    if (typeid(T) == typeid(byte)) kernel_type = KERNEL_BYTE_UNKNOWN;
    if (typeid(T) == typeid(int)) kernel_type = KERNEL_INT_UNKNOWN;
    if (typeid(T) == typeid(float)) kernel_type = KERNEL_FLOAT_UNKNOWN;
  }

  /*! Reads the kernel data from file \a name and stores it in the data structure pointed out by #pixels. The file \a name should be an image file. */
  void readKernel(char *name) {
    readImage(name);
  }

  /*! Writes the current kernel as an image to the file given by #img_name.*/
  void writeKernel() {
    writeImage();
  }
  /*! Opens a window and renders in it the current kernel. */
  void showKernel() {
    showImage();
  }

/*! Reads kernel coefficients from the file specified by \a name. The format of the file should be:

  \verbatim
  #Comment
  height width
  coeff(0,0)       coeff(0,1)       ... coeff(0,width-1)
  ... 
  coeff(height-1,0) coeff(height-1,1) ... coeff(height-1,width-1)
  \endverbatim
*/
  void readKernelCoeffs(char *name);

/*! Writes kernel coefficients to the file specified by \a name; extra information can be stored inside the file as a \a comment.
  The format of the file is:

  \verbatim
  #Comment
  height width
  coeff(0,0)       coeff(0,1)       ... coeff(0,width-1)
  ... 
  coeff(height-1,0) coeff(height-1,1) ... coeff(height-1,width-1)
  \endverbatim
*/
  void writeKernelCoeffs(char *name, char *comment);

  /*! Returns the kernel type. Currently defined types are:
  - CROSS_3x3    = 100, 
  - DISK_3x3     = 101,
  - KERNEL_BYTE_UNKNOWN = 102

  - LAPLACIAN_1    = 200, 
  - LAPLACIAN_2    = 201, 
  - SOBEL_H        = 202,
  - SOBEL_V        = 203,
  - PREWITT_H      = 204,
  - PREWITT_V      = 205,
  - KIRSCH_1       = 206,
  - KIRSCH_2       = 207,
  - KIRSCH_3       = 208,
  - KIRSCH_4       = 209,
  - KIRSCH_5       = 210,
  - KIRSCH_6       = 211,
  - KIRSCH_7       = 212,
  - KIRSCH_8       = 213,
  - KERNEL_INT_UNKNOWN    = 214,

  - RECTANGULAR_PULSE    = 300,
  - TRIANGULAR_PULSE     = 301,
  - GAUSS                = 302,
  - DOG                  = 303,
  - GABOR                = 304,
  - LOG                  = 305,
  - KERNEL_FLOAT_UNKNOWN = 306. */
  int  getKernelType() { return kernel_type;};
};

#endif
