/* Copyright (c) 2001 A. Jalba */

#ifndef MORPHOP_H
#define MORPHOP_H

#include "misc.h"
#include "Image.h"
#include "ByteKernel.h"

/*! \ingroup morph_alg
 */
template< class T > 
void Erosion(Image< T > &input, ByteKernel &kern, Image< T > &output);
/*! \fn void Erosion(Image< T > &input, ByteKernel &kern, Image< T > &output);

  Performs the gray level erosion (denoted by \f$ \ominus \f$ in formulas) of the \a input image using the structuring element \a kern and stores the result in the \a output image.
*/

/*! \ingroup morph_alg
 */
template< class T > 
void Dilation(Image< T > &input, ByteKernel &kern, Image< T > &output);
/*! \fn void Dilation(Image< T > &input, ByteKernel &kern, Image< T > &output);

  Performs the gray level dilation (denoted by \f$ \oplus \f$ in formulas) of the \a input image using the structuring element \a kern and stores the result in the \a output image.
*/

/*! \ingroup morph_alg
 */
template< class T >
void Opening(Image< T > &input, ByteKernel &kern, Image< T > &output);
/*! \fn void Opening(Image< T > &input, ByteKernel &kern, Image< T > &output);

  Performs the gray level opening (denoted by \f$ \circ \f$ in formulas) of the \a input image using the structuring element \a kern and stores the result in the \a output image.
*/

/*! \ingroup morph_alg
 */
template< class T >
void Closing(Image< T > &input, ByteKernel &kern, Image< T > &output);
/*! \fn void Closing(Image< T > &input, ByteKernel &kern, Image< T > &output);

  Performs the gray level closing (denoted by \f$ \bullet \f$ in formulas) of the \a input image using the structuring element \a kern and stores the result in the \a output image. 
*/

/*! \ingroup morph_alg
 */
template< class T >
void FlatErosion(Image< T > &input, ByteKernel &kern, Image< T > &output);
/*! \fn void FlatErosion(Image< T > &input, ByteKernel &kern, Image< T > &output);

  Performs the gray level erosion of the \a input image using the flat structuring element \a kern and stores the result in the \a output image.
*/

/*! \ingroup morph_alg
 */
template< class T > 
void FlatDilation(Image< T > &input, ByteKernel &kern, Image< T > &output);
/*! \fn void FlatDilation(Image< T > &input, ByteKernel &kern, Image< T > &output);

  Performs the gray level dilation of the \a input image using the flat structuring element \a kern and stores the result in the \a output image.
*/

/*! \ingroup morph_alg
 */
template< class T >
void FlatOpening(Image< T > &input, ByteKernel &kern, Image< T > &output);
/*! \fn void FlatOpening(Image< T > &input, ByteKernel &kern, Image< T > &output);

  Performs the gray level opening of the \a input image using the flat structuring element \a kern and stores the result in the \a output image.
*/

/*! \ingroup morph_alg
 */
template< class T >
void FlatClosing(Image< T > &input, ByteKernel &kern, Image< T > &output);
/*! \fn void FlatClosing(Image< T > &input, ByteKernel &kern, Image< T > &output);

  Performs the gray level closing of the \a input image using the flat structuring element \a kern and stores the result in the \a output image.
*/

/*! \ingroup morph_alg
 */
template< class T >
void TopHat(Image< T > &input, ByteKernel &kern, Image< T > &output);
/*! \fn void TopHat(Image< T > &input, ByteKernel &kern, Image< T > &output);

  Performs the top-hat transform of the \a input image using the structuring element \a kern and stores the result in the \a output image. The top-hat transform is defined by the following formula: \f$ output = input - (input \circ kern) \f$;
*/

/*! \ingroup morph_alg
 */
template< class T >
void BottomHat(Image< T > &input, ByteKernel &kern, Image< T > &output);
/*! \fn void BottomHat(Image< T > &input, ByteKernel &kern, Image< T > &output);

  Performs the bottom-hat transform of the \a input image using the structuring element \a kern and stores the result in the \a output image. The bottom-hat transform is defined by the following formula: \f$ output = input - (input \bullet kern) \f$;
*/

/*! \ingroup morph_alg
 */
template< class T >
void HitOrMiss(Image< T > &input, ByteKernel &kern1, ByteKernel &kern2, Image< T > &output);
/*! \fn template< class T > void HitOrMiss(Image< T > &input, ByteKernel &kern1, ByteKernel &kern2, Image< T > &output);

  Performs the hit-or-miss transform of the \a input image using the structuring elements \a kern1 and \a kern2 and stores the result in the \a output image. The hit-or-miss transform is defined by the following formula: \f$ output = (input \ominus kern_1) \cap (input^c \ominus kern_2) \f$; 
where
  \f$ kern_1 \subseteq input \f$ and \f$ kern_2 \subseteq input^c \f$;
*/

/*! \ingroup morph_alg
 */
template< class T >
void TemplateMatch(Image< T > &input, ByteKernel &kern, Image< T > &output);
/*! \fn template< class T > void TemplateMatch(Image< T > &input, ByteKernel &kern, Image< T > &output);

Searches in the \a input image for occurrences of the structuring element specified by |a kern. The result is stored in the \a output image. The template match operation is defined by the following formula: \f$ output = input \odot (kern, kern^c) \f$; 
*/

/** \ingroup morph_alg
 */
template< class T > 
void Erosion3x3(Image< T > &input, int iter, CONN_TYPE connect, Image< T > &output);
/*! \fn void Erosion3x3(Image< T > &input, int iter, CONN_TYPE connect, Image< T > &output);

  Performs iteratively the erosion of the \a input image using as structural element a \f$ 3 \times 3 \f$ cross for 4-connectivity (\a connect = FOUR) or a \f$ 3 \times 3 \f$ disk for 8-connectivity (\a connect = EIGHT). The parameter \a iter specifies how many times (i.e. how many iterations) the erosion is performed. The result is stored in the \a output image.
*/

/*! \ingroup morph_alg
 */
template< class T > 
void Dilation3x3(Image< T > &input, int iter, CONN_TYPE connect, Image< T > &output);
/*! \fn void Dilation3x3(Image< T > &input, int iter, CONN_TYPE connect, Image< T > &output);
  
  Performs iteratively the dilation of the \a input image using as structural element a \f$ 3 \times 3 \f$ cross for 4-connectivity (\a connect = FOUR) or a \f$ 3 \times 3 \f$ disk for 8-connectivity (\a connect = EIGHT). The parameter \a iter specifies how many times (i.e. how many iterations) the dilation is performed. The result is stored in the \a output image.
*/

/*! \ingroup morph_alg
 */
template< class T > 
void Opening3x3(Image< T > &input, int iter, CONN_TYPE connect, Image< T > &output);
/*! \fn void Opening3x3(Image< T > &input, int iter, CONN_TYPE connect, Image< T > &output);

  Performs iteratively the opening of the \a input image using as structural element a \f$ 3 \times 3 \f$ cross for 4-connectivity (\a connect = FOUR) or a \f$ 3 \times 3 \f$ disk for 8-connectivity (\a connect = EIGHT). The parameter \a iter specifies how many times (i.e. how many iterations) the opening is performed. The result is stored in the \a output image.
*/

/*! \ingroup morph_alg
 */
template< class T >
void Closing3x3(Image< T > &input, int iter, CONN_TYPE connect, Image< T > &output);
/*! \fn void Closing3x3(Image< T > &input, int iter, CONN_TYPE connect, Image< T > &output);

  Performs iteratively the closing of the \a input image using as structural element a \f$ 3 \times 3 \f$ cross for 4-connectivity (\a connect = FOUR) or a \f$ 3 \times 3 \f$ disk for 8-connectivity (\a connect = EIGHT). The parameter \a iter specifies how many times (i.e. how many iterations) the closing is performed. The result is stored in the \a output image.
*/

/** \example test_morphop.cpp */

#endif

