/* Copyright (c) 2001 C. Grigorescu */

#ifndef _WAVELET_TRANSFORM_H_
#define _WAVELET_TRANSFORM_H_

#include "FloatImage.h"

typedef struct {
  int ncof, ioff, joff;
  float *cc, *cr;
} wavefilt;

enum WAVELET_TYPE {  HAAR = 0, DAUB_2 = 1, DAUB_3 = 2, DAUB_4 = 3, DAUB_5 = 4,
             DAUB_6 = 5, DAUB_7 = 6, DAUB_8 = 7, DAUB_9 = 8,
             BIOR_1_1 = 19, BIOR_1_3 = 20, BIOR_1_5 = 21, BIOR_2_2 = 22,
             BIOR_2_4 = 23, BIOR_2_6 = 24, BIOR_2_8 = 25, BIOR_3_1 = 26,
             BIOR_3_3 = 27, BIOR_3_5 = 28, BIOR_3_7 = 29, BIOR_3_9 = 30,
             BIOR_1_1_REC = 49, BIOR_1_3_REC = 50, BIOR_1_5_REC = 51, BIOR_2_2_REC = 52,
             BIOR_2_4_REC = 53, BIOR_2_6_REC = 54, BIOR_2_8_REC = 55, BIOR_3_1_REC = 56,
             BIOR_3_3_REC = 57, BIOR_3_5_REC = 58, BIOR_3_7_REC = 59, BIOR_3_9_REC = 60,
	     WAVELET_UNKNOWN=100};     

/** \ingroup img_trans
    Implements the Fast Wavelet Transform of an image using the pyramidal wavelet decomposition/reconstruction algorithm. 
 */

class WaveletTransform
{
 private:
  FloatImage *image;
  int wavelet_type;
  void pwtset(int wavelet_type, wavefilt &wfilt);
  void pwt(float *a, int n, int isign);
  int width, height;
  wavefilt wfilt_dec;
  wavefilt wfilt_rec;
  int filt_size;

 public:
  /*! Constructor; initializes a WaveletTransform object with the \a input image and, if specified by \a wave_type, the wavelet type. Currently the following wavelets are supported:
   - HAAR - The Haar wavelet (default);
   - DAUB_2, DAUB_3, DAUB_4, DAUB_5, DAUB_6, DAUB_7, DAUB_8, DAUB_9 - Daubechies wavelets; 
   - BIOR_1_3, BIOR_1_5, BIOR_2_2, BIOR_2_8, BIOR_3_1 - Bi-orthogonal wavelets. */
  WaveletTransform(FloatImage &input, WAVELET_TYPE wave_type = HAAR) { 
    image        = &input; 
    width        = input.getWidth();
    height       = input.getHeight();
    wavelet_type = wave_type;
    pwtset(wavelet_type, wfilt_dec);
    if (wavelet_type < 15) 
      pwtset(wavelet_type, wfilt_rec);
    else
      pwtset(wavelet_type+30, wfilt_rec);
  }

  /*! Direct Fast Wavelet Transform. The number of levels of decomposition is specified by the parameter \a levels */
  void fwt(int levels);

  /*! Inverse Fast Wavelet Transform. The number of levels used in reconstruction is specified by the parameter \a levels. */
  void ifwt(int levels);
};

void fwt(FloatImage &input, int levels, WAVELET_TYPE wave_type = HAAR);
/*! \ingroup img_trans
  \fn void fwt(FloatImage &input, int levels, WAVELET_TYPE wave_type = HAAR);
  Performs a Fast Wavelet Transform (FWT) on the \a input image using the number of levels of decomposition specified by parameter \a levels and the wavelet type specified by parameter \a wave_type. The FWT is performed "in-place", i.e. the initial \a input image will be overwritten. Currently supported wavelet types are:
   - HAAR - The Haar wavelet (default);
   - DAUB_2, DAUB_3, DAUB_4, DAUB_5, DAUB_6, DAUB_7, DAUB_8, DAUB_9 - Daubechies wavelets; 
   - BIOR_1_3, BIOR_1_5, BIOR_2_2, BIOR_2_8, BIOR_3_1 - Bi-orthogonal wavelets. 
 For more functionality regarding the wavelet transform, please refer to the WaveletTransform class. */

void ifwt(FloatImage &input, int levels, WAVELET_TYPE wave_type = HAAR);
/*! \ingroup img_trans
*/
/*! \fn void ifwt(FloatImage &input, int levels, WAVELET_TYPE wave_type = HAAR);
  Performs an Inverse Fast Wavelet Transform (IFWT) on the \a input image using the number of levels of reconstruction specified by the parameter \a levels and the wavelet type specified by the parameter \a wave_type. The IFWT is performed "in-place", i.e. the initial \a input image will be overwritten. Currently supported wavelet types are:
   - HAAR - The Haar wavelet (default);
   - DAUB_2, DAUB_3, DAUB_4, DAUB_5, DAUB_6, DAUB_7, DAUB_8, DAUB_9 - Daubechies wavelets; 
   - BIOR_1_3, BIOR_1_5, BIOR_2_2, BIOR_2_8, BIOR_3_1 - Bi-orthogonal wavelets. 
For more functionality regarding the wavelet transform, please refer to the WaveletTransform class. */

/** \example test_wavelet.cpp*/
/** \example test_wavelet2.cpp*/

#endif

