/* Copyright (c) 2001 S.E. Grigorescu A. Jalba */

#include <math.h>
#include "Edge.h"

template 
void GradientMap< int >(Image< int > &input, FloatImage &gx, FloatImage &gy);
template 
void GradientMap< byte >(Image< byte > &input, FloatImage &gx, FloatImage &gy);
template 
void GradientMap< float >(Image< float > &input, FloatImage &gx, FloatImage &gy);

template
void EdgePrewitt< int >(Image< int > &input, int type, Image< int > &output);
template
void EdgePrewitt< byte >(Image< byte > &input, int type, Image< byte > &output);
template
void EdgePrewitt< float >(Image< float > &input, int type, Image< float > &output);

template
void EdgeSobel< int >(Image< int > &input, int type, Image< int > &output);
template
void EdgeSobel< byte >(Image< byte > &input, int type, Image< byte > &output);
template
void EdgeSobel< float >(Image< float > &input, int type, Image< float > &output);

template
void EdgeKirsch< int >(Image< int > &input, Image< int > &output);
template
void EdgeKirsch< byte >(Image< byte > &input, Image< byte > &output);
template
void EdgeKirsch< float >(Image< float > &input, Image< float > &output);

void Normalize(vector< FloatImage > &v)
{
  bool flag = true;
  for(int i=1;i<v.size();i++)
    if (v[0].getWidth() != v[i].getWidth() || 
	v[0].getHeight() != v[i].getHeight())
      flag = false;
  
  if (flag) {
    FloatImage mag;
    Magnitude(v, mag);
    for(int i=0;i<mag.getSize();i++)
      mag(i) = (mag(i) > 1E-10) ? 1/mag(i) : 0;
    for(int i=0;i<v.size();i++)
      v[i] *= mag;
  }
} 

void Magnitude(vector< FloatImage > &v, FloatImage &magnit)
{
  bool flag = true;
  for(int i=1;i<v.size();i++)
    if (v[0].getWidth() != v[i].getWidth() || 
	v[0].getHeight() != v[i].getHeight())
      flag = false;

  if (flag) {
    magnit = v[0];
    magnit.clearImage(0);
    for(int i=0;i<v.size();i++)
      for(int j=0;j<v[i].getSize();j++)
	magnit(j) += v[i](j)*v[i](j);
    for(int i=0;i<magnit.getSize();i++)
      magnit(i) = sqrt(magnit(i));
  }
}

template< class T >
void GradientMap(Image< T > &input, FloatImage &gx, FloatImage &gy)
{
  int width = input.getWidth(), height = input.getHeight(), i, j;
  
  gx.makeImage(width, height);
  gy.makeImage(width, height);

  for(i=0;i<height;i++) {
    gx[i][0]=(float)(input[i][1]-input[i][0]);
    gx[i][width-1]=(float)(input[i][width-2]-input[i][width-1]);
  }

  for(i=0;i<width;i++) {
    gy[0][i]=(float)(input[1][i]-input[0][i]);
    gy[height-1][i]=(float)(input[height-2][i]-input[height-1][i]);
  }

  for(i=1;i<height-1;i++)
    for(j=1;j<width-1;j++) {
      gy[i][j]=(float)(input[i+1][j]-input[i-1][j]);
      gx[i][j]=(float)(input[i][j+1]-input[i][j-1]);
    }  
}

template< class T >
void EdgePrewitt(Image< T > &input, int type, Image< T > &output)
{
  IntKernel kh(PREWITT_H), kv(PREWITT_V);
  int width = input.getWidth(), height = input.getHeight();

  Image< T > tmp;
  tmp.makeImage(width, height);        

  for(int i=1;i<height-1;i++)
    for(int j=1;j<width-1;j++) {
      T h = 0, v = 0;
      for(int k=-1;k<=1;k++) 
	for(int m=-1;m<=1;m++) {
	  h += (T)kh[k+1][m+1]*input[i+k][j+m];
	  v += (T)kv[k+1][m+1]*input[i+k][j+m];
	}
      tmp[i][j] = (type) ? (T)sqrt(h*h+v*v) : (T)(fabs(h) + fabs(v));
    }

  output = tmp;
}

template< class T >
void EdgeSobel(Image< T > &input, int type, Image< T > &output)
{
  IntKernel kh(SOBEL_H), kv(SOBEL_V);
  int width = input.getWidth(), height = input.getHeight();

  Image< T > tmp;
  tmp.makeImage(width, height);    

  for(int i=1;i<height-1;i++)
    for(int j=1;j<width-1;j++) {
      T h = 0, v = 0;
      for(int k=-1;k<=1;k++) 
	for(int m=-1;m<=1;m++) {
	  h += (T)kh[k+1][m+1]*input[i+k][j+m];
	  v += (T)kv[k+1][m+1]*input[i+k][j+m];
	}
      tmp[i][j] = (type) ? (T)sqrt(h*h+v*v) : (T)(fabs(h) + fabs(v));
    }
  output = tmp;
}

template< class T >
void EdgeKirsch(Image< T > &input, Image< T > &output)
{
  int width = input.getWidth(), height = input.getHeight();

  Image< T > tmp;
  tmp.makeImage(width, height); 

  for(int i=1;i<height-1;i++)
    for(int j=1;j<width-1;j++) {
      T max=-(T)(1E-10); int pos;
      for(int orient=KIRSCH_1;orient<=KIRSCH_8;orient++) { 
	T val=0;
	IntKernel kern((INT_KERNEL_TYPE)orient);
	for(int k=-1;k<=1;k++) 
	  for(int m=-1;m<=1;m++) 
	    val += (T)kern[k+1][m+1]*input[i+k][j+m];
	if(val>max) {
	  max=val;
	  pos=orient;
	}
      }
      tmp[i][j] = max;
    }
  output = tmp;
}










