/* Copyright (c) 2001 C. Grigorescu */

#include <math.h>
#include <sys/wait.h>  
#include <signal.h> //only for kill
#include "misc.h"
#include "config.h"
#include "FourierTransform.h"

#ifdef HAVE_LIBFFTW
#include "fftwlib.h"
#endif

void FourierTransform::_fft(float *data, int nn, int isign)
{
  int n, mmax, m, j, istep, i;
  float wtemp, wr, wpr, wpi, wi, theta;
  float tempr, tempi;

  n = nn<<1;
  j = 1;
  for (i=1;i<n;i+=2) {
    if (j>i) {
      SWAP(data[j],data[i]);
      SWAP(data[j+1],data[i+1]);
    }
    m = n>>1;
    while((m>=2) && (j>m)) {
      j -= m;
      m >>= 1;
    }
    j+=m;
  }

  mmax = 2; 
  while( n > mmax) {
    istep = mmax<<1;
    theta = isign*(6.28318530/mmax);
    wtemp = sin(0.5*theta);
    wpr = -2.0*wtemp*wtemp;
    wpi = sin(theta);
    wr = 1.0;
    wi = 0.0;
    for (m=1;m<mmax;m+=2) {
      for (i=m;i<=n;i+=istep) {
	j=i+mmax;
	tempr = wr*data[j]-wi*data[j+1];
	tempi = wr*data[j+1]+wi*data[j];
	data[j] = data[i]-tempr;
	data[j+1] = data[i+1]-tempi;
	data[i] += tempr;
	data[i+1] += tempi;
      }
      wr = (wtemp=wr)*wpr-wi*wpi+wr;
      wi = wi*wpr+wtemp*wpi+wi;
    }
    mmax=istep;
  }
}

FourierTransform::FourierTransform(FloatImage &input)
{
#ifdef HAVE_LIBFFTW
  input.makeEvenSized();
  real_part = input;
  imag_part.makeImage(input.getWidth(),input.getHeight());
  imag_part.clearImage(0);
#else
  input.makeEvenSized();
  real_part = input;
  int w = real_part.getWidth();
  int h = real_part.getHeight();
  int p2_w = 1; int p2_h = 1;

  for(int i=0;i<16;i++) {
    if (w <= p2_w) break;
    p2_w = p2_w<<1;
  }
  for(int i=0;i<16;i++) {
    if (h <= p2_h) break;
    p2_h = p2_h<<1;
  }
  real_part.pad((p2_w - w)/2, (p2_h - h)/2, ZEROS);
  imag_part.makeImage(real_part.getWidth(),real_part.getHeight());
  imag_part.clearImage(0);
#endif
}

FourierTransform::FourierTransform(FloatImage &input_real, FloatImage &input_imag)
{
  if ((input_real.getWidth() != input_imag.getWidth()) ||
      (input_real.getHeight() != input_imag.getHeight())) {
    ERROR("Real and imaginary input images have different sizes");
  }
  else {
#ifdef HAVE_LIBFFTW
    input_real.makeEvenSized();
    input_imag.makeEvenSized();
    real_part = input_real;
    imag_part = input_imag;
#else
    int w = input_real.getWidth();
    int h = input_imag.getHeight();
    int p2_w = 1; int p2_h = 1;
    
    int flag_w = 0, flag_h =0;
    for(int i=0;i<16;i++) {
      if (w == p2_w) {
	flag_w = 1; 
	break;
      }
      p2_w = p2_w<<1;
    }
    if (!flag_w) 
      ERROR("Image width not a power of 2; results may be incorrect");
    for(int i=0;i<16;i++) {
      if (h == p2_h) {
	flag_h = 1;
	break;
      }
      p2_h = p2_h<<1;
    }
    if (!flag_h) 
      ERROR("Image height not a power of 2; results may be incorrect");
    real_part = input_real;
    imag_part = input_imag;
#endif
  }
}

void FourierTransform::fft(void)
{

#ifdef HAVE_LIBFFTW
   FloatImage temp_real = real_part;
   FloatImage temp_imag = imag_part;
   _fft2d(temp_real.getPixels(), temp_imag.getPixels(), 
	  real_part.getWidth() , real_part.getHeight(),
	  real_part.getPixels(), imag_part.getPixels());
#else
  int w =  real_part.getWidth();
  int h =  real_part.getHeight();

  float * temp = (float *) malloc(2*w* sizeof(float));
  for (int i=0;i<h;i++) {
    for(int j=0;j<w;j++) {
      temp[2*j] = real_part[i][j];
      temp[2*j+1] = imag_part[i][j];
    }
    _fft(temp-1, w, 1);
    for(int j=0;j<w;j++) {
      real_part[i][j] = temp[2*j];
      imag_part[i][j] = temp[2*j+1];
    }
  }
  free(temp);
  temp = (float *) malloc(2*h * sizeof(float));
  for(int j=0;j<w;j++) {
    for (int i=0;i<h;i++) {
      temp[2*i] = real_part[i][j];
      temp[2*i+1] = imag_part[i][j];
    }
    _fft(temp-1, h, 1);
    for(int i=0;i<h;i++) {
      real_part[i][j] = temp[2*i];
      imag_part[i][j] = temp[2*i+1];
    }
  }
  free(temp);
  for(int j=0;j<w;j++) 
    for(int i=0;i<h;i++) {
      real_part[i][j] /= w*h;
      imag_part[i][j] /= w*h;
    }
  HALF_SHIFT(real_part.getPixels(),w,h);
  HALF_SHIFT(imag_part.getPixels(),w,h);
#endif
}

void FourierTransform::ifft(void)
{
#ifdef HAVE_LIBFFTW
  FloatImage temp_real = real_part;
  FloatImage temp_imag = imag_part;
  _ifft2d(temp_real.getPixels(), temp_imag.getPixels(), 
	  real_part.getWidth() , real_part.getHeight(),
	  real_part.getPixels(), imag_part.getPixels());
#else
  int w = real_part.getWidth();
  int h = real_part.getHeight();

  HALF_SHIFT(real_part.getPixels(),w,h);
  HALF_SHIFT(imag_part.getPixels(),w,h);
  float *temp = (float *) malloc(2*w * sizeof(float));
  for (int i=0;i<h;i++) {
    for(int j=0;j<w;j++) {
      temp[2*j] = real_part[i][j];
      temp[2*j+1] = imag_part[i][j];
    }
    _fft(temp-1, w, -1);
    for(int j=0;j<w;j++) {
      real_part[i][j] = temp[2*j];
      imag_part[i][j] = temp[2*j+1];
    }
  }
  free(temp);
  temp = (float *) malloc(2*h * sizeof(float));
  for(int j=0;j<w;j++) {
    for (int i=0;i<h;i++) {
      temp[2*i] = real_part[i][j];
      temp[2*i+1] = imag_part[i][j];
    }
    _fft(temp-1, h, -1);
    for(int i=0;i<h;i++) {
      real_part[i][j] = temp[2*i];
      imag_part[i][j] = temp[2*i+1];
    }
  }
  free(temp);
#endif
}

void FourierTransform::showMagnitude(int log_values)
{
  FloatImage magnitude(real_part.getWidth(), real_part.getHeight());

  for (int j=0;j<real_part.getWidth();j++)
    for (int i=0;i<real_part.getHeight();i++) 
      magnitude[i][j] = sqrt(real_part[i][j]*real_part[i][j] +
			     imag_part[i][j]*imag_part[i][j]);
  if (log_values) {
    for (int j=0;j<real_part.getWidth();j++)
      for (int i=0;i<real_part.getHeight();i++) 
	magnitude[i][j] = log(magnitude[i][j]);
  } 
  if(pid_magnitude) {
    kill(pid_magnitude, SIGUSR1); 
    waitpid(pid_magnitude, NULL, 0);
    pid_magnitude = magnitude.showImage();
  }
  else
    pid_magnitude = magnitude.showImage();
}

void FourierTransform::showPhase(int log_values)
{
  FloatImage phase(imag_part.getWidth(), imag_part.getHeight());
  for (int j=0;j<imag_part.getWidth();j++)
    for (int i=0;i<imag_part.getHeight();i++)
      phase[i][j] = atan(imag_part[i][j]/real_part[i][j]);
  if (log_values) {
    for (int j=0;j<imag_part.getWidth();j++)
      for (int i=0;i<imag_part.getHeight();i++) 
	phase[i][j] = log(phase[i][j]);
  } 
  if(pid_phase) {
    kill(pid_phase, SIGUSR1); 
    waitpid(pid_phase, NULL, 0);
    pid_phase = phase.showImage();
  }
  else
    pid_phase = phase.showImage();
}

void FourierTransform::closeWindows()
{
  if(pid_magnitude) {
    kill(pid_magnitude, SIGUSR1); 
    waitpid(pid_magnitude, NULL, 0);
    pid_magnitude = 0;
  }
  if(pid_phase) {
    kill(pid_phase, SIGUSR1); 
    waitpid(pid_phase, NULL, 0);
    pid_phase = 0;
  }
}

void  fft(FloatImage &input, FloatImage &output_real, FloatImage &output_imag)
{
  FourierTransform a(input);
  a.fft();
  output_real = a.getReal();
  output_imag = a.getImag();
}

void ifft(FloatImage &input_real, FloatImage &input_imag, FloatImage &output)
{
  FourierTransform a(input_real, input_imag);
  a.ifft();
  output = a.getReal();
}

