/* Copyright (c) 2001 C. Grigorescu */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>
#include <config.h>

#ifdef HAVE_LIBFFTW

#include <fftw.h>
#include <rfftw.h>
#include "fftwlib.h"

/*   Wrap code for performing 2D Fourier transforms and convolution
     with in single precision floating point by using FFTW
*/

void _convolve(float *input_image_1, 
              float *input_image_2, 
              int size_x, int size_y, 
              float *convolved_image)
{

  fftw_real a[size_x][2*(size_y/2+1)], b[size_x][2*(size_y/2+1)],c[size_x][size_y];
  fftw_complex *A, *B, C[size_x][size_y/2+1];
  fftwnd_plan p, pinv;
  fftw_real scale = 1.0 / (size_x * size_y);
  int i, j, ij;

  p = rfftw2d_create_plan(size_x, size_y, 
          FFTW_REAL_TO_COMPLEX, FFTW_ESTIMATE | FFTW_IN_PLACE);
  pinv = rfftw2d_create_plan(size_x, size_y, 
          FFTW_COMPLEX_TO_REAL, FFTW_ESTIMATE);

  half_shift(input_image_1,size_x,size_y); 


  A = (fftw_complex *) &a[0][0];
  B = (fftw_complex *) &b[0][0];

  for(i=0; i<size_x; ++i)
    for(j=0; j<size_y; ++j)
      {
      a[i][j] = input_image_1[j*size_x+i];
      b[i][j] = input_image_2[j*size_x+i];
      }

  rfftwnd_one_real_to_complex(p,&a[0][0],NULL);
  rfftwnd_one_real_to_complex(p,&b[0][0],NULL);

  for(i=0;i<size_x;++i)
    for(j=0;j<(size_y/2+1);++j)
      {
      ij = i*(size_y/2+1)+j;
      C[i][j].re = (A[ij].re * B[ij].re -
                    A[ij].im * B[ij].im)*scale;

      C[i][j].im = (A[ij].re * B[ij].im +
                    A[ij].im * B[ij].re)*scale;
      }
  
  rfftwnd_one_complex_to_real(pinv,&C[0][0],&c[0][0]);

  for(i=0; i<size_x; ++i)
    for(j=0; j<size_y; ++j)
      convolved_image[j*size_x+i] = (float)c[i][j];

  half_shift(input_image_1,size_x,size_y); 

  rfftwnd_destroy_plan(p);
  rfftwnd_destroy_plan(pinv);
}

void _fft2d(float *real_input_image, float *imag_input_image,
           int size_x, int size_y, 
           float *fourier_real_part, float *fourier_imag_part)
{

  fftw_complex a[size_x*size_y];
  fftwnd_plan p;
  fftw_real scale;
  int i, j, ij; 

  scale = 1.0 / (size_x * size_y);

  p = fftw2d_create_plan(size_x, size_y, FFTW_FORWARD, FFTW_ESTIMATE | FFTW_IN_PLACE);

  half_shift(real_input_image,size_x,size_y); 
  if (imag_input_image != NULL)
    half_shift(imag_input_image,size_x,size_y); 

  for(j=0; j<size_y; ++j)
    for(i=0; i<size_x; ++i)
      {
      ij = i*size_y+j;
      a[ij].re = real_input_image[j*size_x+i];
      if (imag_input_image != NULL)
        a[ij].im = imag_input_image[j*size_x+i];
      else
        a[ij].im = 0.0;
      }

  fftwnd_one(p,&a[0],NULL);

  for(j=0; j<size_y; ++j)
    for(i=0; i<size_x; ++i)
      {
      ij = i*size_y+j;
      fourier_real_part[j*size_x+i] = (float) a[ij].re;
      fourier_imag_part[j*size_x+i] = (float) a[ij].im;
      }

  half_shift(fourier_real_part,size_x,size_y);
  half_shift(fourier_imag_part,size_x,size_y); 
  fftwnd_destroy_plan(p);
} 

void _ifft2d(float *fourier_real_image, float *fourier_imag_image,
           int size_x, int size_y, float *real_image, float *imag_image)
{
  /*  Performs a IFFT; works with FFTW package */

  fftw_complex a[size_x*size_y];
  fftwnd_plan p;
  fftw_real scale;
  int i, j, ij; 

  scale = 1.0 / (size_x * size_y);

  p = fftw2d_create_plan(size_x, size_y, FFTW_BACKWARD, FFTW_ESTIMATE | FFTW_IN_PLACE);

  half_shift(fourier_real_image,size_x,size_y);
  half_shift(fourier_imag_image,size_x,size_y);

  for(j=0; j<size_y; ++j)
    for(i=0; i<size_x; ++i)
      {
      ij = i*size_y+j;
      a[ij].re = (double) fourier_real_image[j*size_x+i];
      a[ij].im = (double) fourier_imag_image[j*size_x+i];
      }

  fftwnd_one(p,&a[0],NULL);

  for(j=0; j<size_y; ++j)
    for(i=0; i<size_x; ++i)
      {
      ij = i*size_y+j;
      real_image[j*size_x+i] = (float) a[ij].re * scale;
      imag_image[j*size_x+i] = (float) a[ij].im * scale;
      }

  half_shift(real_image,size_x,size_y);
  half_shift(imag_image,size_x,size_y);
  fftwnd_destroy_plan(p);
} 

#endif
