/* Copyright (c) 2001 C. Grigorescu */

/*********************************************************
 File: imlib_io.cpp							 

    Contains functions for reading, writing and
 displaying image files; it uses the "imlib" library 
 (in principle, all formats available for "imlib" 
 should work).

 The image buffer is allocated by the "read_image" function,
 returned as a pointer to void named "image"; it can be 
 later accessed as a linear array of type "type".

 Currently, the implemetation supports only the following
 types: 

  - BYTE_2D : "image" is a linear array of "unsigned char"
  - SHORT_2D : "image" is a linear array of "short int"
  - INT_2D : "image" is a linear array of "int"
  - FLOAT_2D : "image" is a linear array of type "float"
  - RGB_2D : "image" is a linear array of type "unsigned char" 
   containing red, green and blue values (RGB array) organized 
   as "RGBRGBRGB..."  

 ********************************************************/

#include <Imlib.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/extensions/shape.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include "imlib_io.h"
#include "Exceptions.h"

#include <signal.h>
 
#define ESC   0x09
#define ENTER 0x24

template byte  *read_image< byte >(char *filename, int *width, int *height);
template short *read_image< short >(char *filename, int *width, int *height);
template int   *read_image< int >(char *filename, int *width, int *height);
template float *read_image< float >(char *filename, int *width, int *height);
template double *read_image< double >(char *filename, int *width, int *height);
template int write_image< byte >(char *filename, int width, int height, byte *image);
template int write_image< short >(char *filename, int width, int height, short *image);
template int write_image< int >(char *filename, int width, int height, int *image);
template int write_image< float >(char *filename, int width, int height, float *image);
template int write_image< double >(char *filename, int width, int height, double *image);
template int show_image< byte >(byte *image, int width, int height, char *name);
template int show_image< short >(short *image, int width, int height, char *name);
template int show_image< int >(int *image, int width, int height, char *name);
template int show_image< float >(float *image, int width, int height, char *name);
template int show_image< double >(double *image, int width, int height, char *name);

int read_size(char *filename, int *width, int *height)
{
  ImlibData *id;
  ImlibImage *im;
  Display *disp;
 
  disp = XOpenDisplay(NULL);
  if (disp == NULL) 
    {
    printf("\n Cannot open display; is the $DISPLAY set? \n");
    return(0);
    }
  id         = Imlib_init(disp);
  im         = Imlib_load_image(id, filename);
  if (im == NULL) 
    return(0);
  *width     = im->rgb_width;
  *height    = im->rgb_height;
  Imlib_kill_image(id,im);
  XCloseDisplay(disp);
  return(1);
}

template <class T> T *read_image(char *filename, int *width, int *height)
{
  ImlibData *id;
  ImlibImage *im;
  Display *disp;
  int i;
  T *image = NULL;

  disp = XOpenDisplay(NULL);
  if (disp == NULL) 
  {
    printf("\n Cannot open display; is the $DISPLAY set? \n");
    return(0);
  }
  id         = Imlib_init(disp);
  im         = Imlib_load_image(id, filename);
  *width     = im->rgb_width;
  *height    = im->rgb_height;

  image = (T *) malloc((*width)*(*height)*sizeof(T));
  if (image == NULL) 
  {
    printf("\nCannot allocate memory for image buffer");
    return(NULL);
  }
  for(i=0;i<(*width)*(*height);i++)
    image[i] = (T) ((11*im->rgb_data[3*i] + 
		     16*im->rgb_data[3*i+1] + 
		      5*im->rgb_data[3*i+2])/32);

  Imlib_kill_image(id,im);
  XCloseDisplay(disp);

  return(image);
}

unsigned char *read_RGB_image(char *filename, int *width, int *height)
{
  ImlibData *id;
  ImlibImage *im;
  Display *disp;
  int i;
  unsigned char *rgb_image = NULL;

  disp = XOpenDisplay(NULL);
  if (disp == NULL) 
    {
    printf("\n Cannot open display; is the $DISPLAY set? \n");
    return(0);
    }
  id         = Imlib_init(disp);
  im         = Imlib_load_image(id, filename);
  *width     = im->rgb_width;
  *height    = im->rgb_height;

  rgb_image = (unsigned char *) malloc(3*(*width)*(*height)*sizeof(unsigned char));
  if (rgb_image == NULL) 
  {
    printf("\nCannot allocate memory for image buffer");
    return(NULL);
  }
  for(i=0;i<(*width)*(*height);i++) 
  {
    rgb_image[3*i]   = im->rgb_data[3*i];
    rgb_image[3*i+1] = im->rgb_data[3*i+1];
    rgb_image[3*i+2] = im->rgb_data[3*i+2];
  }

  Imlib_kill_image(id,im);
  XCloseDisplay(disp);

  return(rgb_image);
}

template <class T> int write_image(char *filename, int width, int height, T *image)
{
  ImlibData *id;
  ImlibImage *im;
  Display *disp;
  ImlibSaveInfo saveinfo;
  int i, result;
  T max, min;
  unsigned char *buf;

  disp = XOpenDisplay(NULL);
  if (disp == NULL) 
  {
    printf("\n Cannot open display; is the $DISPLAY set? \n");
    return(0);
  }
  id = Imlib_init(disp);
  if (id == NULL) return(0);

  max = image[0]; 
  min = image[0];
  for (i=0;i<width*height;i++) 
  {
    if(max < image[i]) max = image[i];
    if(min > image[i]) min = image[i];
  }

  buf = (unsigned char *) malloc(width*height*3*sizeof(unsigned char));
  if (buf == NULL) return(0);

  for(i=0;i<width*height;i++) 
  {
    buf[3*i]   = (unsigned char) (255*(image[i]-min)/(max-min));
    buf[3*i+1] = (unsigned char) (255*(image[i]-min)/(max-min));
    buf[3*i+2] = (unsigned char) (255*(image[i]-min)/(max-min));
  }
  saveinfo.color = 0;

  im = Imlib_create_image_from_data(id, buf, NULL, width, height);
  result = Imlib_save_image(id, im, filename, &saveinfo);
  Imlib_kill_image(id,im);
  XCloseDisplay(disp);
  free(buf);
  if (result == 0)
    return(0);
  else
    return(1);
}


int write_RGB_image(char *filename, int width, int height, unsigned char *image)
{
  ImlibData *id;
  ImlibImage *im;
  Display *disp;
  ImlibSaveInfo saveinfo;
  int result;

  disp = XOpenDisplay(NULL);
  if (disp == NULL) 
  {
    printf("\n Cannot open display; is the $DISPLAY set? \n");
    return(0);
  }
  id = Imlib_init(disp);
  if (id == NULL) return(0);

  im = Imlib_create_image_from_data(id, image, NULL, width, height);
  result = Imlib_save_image(id, im, filename, &saveinfo);
  Imlib_kill_image(id,im);
  XCloseDisplay(disp);
  if (result == 0)
    return(0);
  else
    return(1);
}

Display *disp;
ImlibData *id;  
ImlibImage *im;
unsigned char *buf;

void handler(int i) {
    free(buf);
    Imlib_kill_image(id,im);
    XCloseDisplay(disp);
    exit(0);	
}

template <class T> int show_image(T *image, int width, int height, char *name)
{
  // Display *disp;
  //ImlibData *id;
  XSetWindowAttributes attr;
  Window win;
  //ImlibImage *im;
  Pixmap p,m;
  XEvent ev;

  int i, pid;
  //  unsigned char *buf;

  T min, max;
  int w,h;
  int flag = 1;
 
  disp = XOpenDisplay(NULL);
  if (disp == NULL) 
  {
    printf("\n Cannot open display; is the $DISPLAY set? \n");
    return(0);
  }
  id=Imlib_init(disp);

  max = image[0]; 
  min = image[0];
  for (i=0;i<width*height;i++) 
  {
    if(max < image[i]) max = image[i];
    if(min > image[i]) min = image[i];
  }

  buf = (unsigned char *) malloc(width*height*3*sizeof(unsigned char));
  if (buf == NULL)
  {
    printf("\n Cannot allocate memory! \n");
    return(0);
  }

  for(i=0;i<width*height;i++) 
  {
    buf[3*i]   = (unsigned char) ((max-min != 0) ? (255*(image[i]-min)/(max-min)) : 0);
    buf[3*i+1] = (unsigned char) ((max-min != 0) ? (255*(image[i]-min)/(max-min)) : 0);
    buf[3*i+2] = (unsigned char) ((max-min != 0) ? (255*(image[i]-min)/(max-min)) : 0);
  }    

  im = Imlib_create_image_from_data(id, buf, NULL, width, height);
  win= XCreateWindow(disp,DefaultRootWindow(disp),0,0,width,height,
		    0,id->x.depth, InputOutput,id->x.visual,0,&attr);
  XStoreName(disp, win, name);    
  XSelectInput(disp,win,StructureNotifyMask|KeyPressMask);
  Imlib_render(id,im,width,height);
  p=Imlib_move_image(id,im);
  m=Imlib_move_mask(id,im);
  XSetWindowBackgroundPixmap(disp,win,p);
  if (m) XShapeCombineMask(disp,win,ShapeBounding,0,0,m,ShapeSet);
  XMapWindow(disp,win);
  XSync(disp,False);

  if(!(pid=fork())) {
    signal(SIGUSR1, handler);
    while (flag) {
      XNextEvent(disp,&ev);
      if (ev.type==ConfigureNotify) {
	  w=ev.xconfigure.width;h=ev.xconfigure.height;
	  Imlib_render(id,im,w,h);
	  Imlib_free_pixmap(id,p);
	  p=Imlib_move_image(id,im);
	  m=Imlib_move_mask(id,im);
	  XSetWindowBackgroundPixmap(disp,win,p);
	  if (m) XShapeCombineMask(disp,win,ShapeBounding,0,0,m,ShapeSet);
	  XClearWindow(disp,win);
	  XSync(disp,False);
	}
      if (ev.type==KeyPress) {
	if ((ev.xkey.keycode == ESC) || (ev.xkey.keycode == ENTER)) {
	  flag = 0;
	}
      }
    }
    
    free(buf);
    Imlib_kill_image(id,im);
    XCloseDisplay(disp);
    //return 0;
    exit(0);	
  }

  return pid;
}

int show_RGB_image(unsigned char *image, int width, int height, char *name)
{
  Display *disp;
  ImlibData *id;
  XSetWindowAttributes attr;
  Window win;
  ImlibImage *im;
  Pixmap p,m;
  XEvent ev;

  int pid;
  int w,h;
  int flag = 1;
 
  disp = XOpenDisplay(NULL);
  if (disp == NULL) 
  {
    printf("\n Cannot open display; is the $DISPLAY set? \n");
    return(0);
  }
  id=Imlib_init(disp);

  im = Imlib_create_image_from_data(id, image, NULL, width, height);
  win= XCreateWindow(disp,DefaultRootWindow(disp),0,0,width,height,
		    0,id->x.depth, InputOutput,id->x.visual,0,&attr);
  XStoreName(disp, win, name);    
  XSelectInput(disp,win,StructureNotifyMask|KeyPressMask);
  Imlib_render(id,im,width,height);
  p=Imlib_move_image(id,im);
  m=Imlib_move_mask(id,im);
  XSetWindowBackgroundPixmap(disp,win,p);
  if (m) XShapeCombineMask(disp,win,ShapeBounding,0,0,m,ShapeSet);
  XMapWindow(disp,win);
  XSync(disp,False);

  if(!(pid=fork())) {
    while (flag) {
      XNextEvent(disp,&ev);
      if (ev.type==ConfigureNotify) {
	  w=ev.xconfigure.width;h=ev.xconfigure.height;
	  Imlib_render(id,im,w,h);
	  Imlib_free_pixmap(id,p);
	  p=Imlib_move_image(id,im);
	  m=Imlib_move_mask(id,im);
	  XSetWindowBackgroundPixmap(disp,win,p);
	  if (m) XShapeCombineMask(disp,win,ShapeBounding,0,0,m,ShapeSet);
	  XClearWindow(disp,win);
	  XSync(disp,False);
	}
      if (ev.type==KeyPress) {
	if ((ev.xkey.keycode == ESC) || (ev.xkey.keycode == ENTER)) {
	  flag = 0;
	}
      }
    }
    
    Imlib_kill_image(id,im);
    XCloseDisplay(disp);
    return 0;
  }
  return pid;
}










