/*
 * Package:     cvtk - Computer Vision Tool Kit
 * File:        calibratecontroller.cpp
 * Description: Implementation of the controller of the calibration 
 *              tool.
 * Created:     2001-03-22
 * Author:      Sascha Lange
 * Mail:        Sascha.Lange@Uni-Osnabrueck.De
 *
 * Web:         http://cvtk.sourceforge.net
 *
 * Copyright (C) 2001 Sascha Lange
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 *
 * -------------------------- Revision History ------------------------
 * $Log: calibratecontroller.cpp,v $
 * Revision 1.16  2001/09/03 12:11:17  slange
 * finished benchmark
 *
 * Revision 1.15  2001/09/02 17:39:07  slange
 * cvs error!
 *
 * Revision 1.14  2001/09/01 13:28:41  slange
 * - working on monitor functions
 *
 * Revision 1.13  2001/08/30 17:25:32  slange
 * - added ft_interface to drive a fischertechnik robot
 * - working on demo application
 *
 * Revision 1.12  2001/08/30 13:24:30  slange
 * - added sliders to effect the specificity of the kmsegmentation
 *
 * Revision 1.11  2001/06/19 18:09:24  slange
 * - use QTimer instead of posix threads to read new images from the webcam
 * - implemented a cleanup routine to delete the webcam even in case of a call
 *   to exit(state)
 * BUG: ctrl-c causes the whole system to hang on smp - machines.
 * REASON: filedescriptor of the videodevice is not closed!
 *
 * Revision 1.10  2001/05/03 15:04:09  slange
 * - added lookup table data for orange/yellow/blue on green/black/white
 *   background to rsc
 * - fixes
 *
 * Revision 1.9  2001/05/03 13:19:44  slange
 * - efficiency improvement of frame grabbing/histogram analysis in the
 *   CController
 *
 * Revision 1.8  2001/05/03 13:04:27  slange
 * - traced down bug in CController::addFrame()
 *
 * Revision 1.7  2001/05/03 01:19:18  slange
 * - Initial revision of CRGBHistogram
 * - added template class CHistogram3Dim
 *
 * Revision 1.6  2001/05/03 00:26:29  slange
 * - first completly working draft of calibration, colorsegmentation and region
 *   finding/representation.
 * - bug fixes
 *
 * Revision 1.5  2001/04/29 00:55:49  slange
 * - finished the first complete draft of the calibration tool.
 *   - it now generates a lookup-table for values 1..64 (>> 2)
 *   - writes the entire table to stdout.
 * - fixes in k-means segmentation
 * - added button to generate lookup table in the calibrator-widget
 *
 * Revision 1.4  2001/04/28 18:05:45  slange
 * - removed patterngenerato.[h/cpp]
 * - added abstract interface CReinforcedLearning
 * - added CKMSegmentation
 *   - class definition and part of implementation
 * - added "train" button to calibratorwidget
 * - changed calibratecontroller to work with the new classes
 *
 * Revision 1.3  2001/04/27 23:15:42  slange
 * - added an UV - Histogram (histogram in an approximation of the HUV space)
 * - changed the Calibrate::CController
 *   - to work with histograms
 *   - to keep a vector of examples in memory
 *   - output file dropped
 *   - to tread the background as a (special) object
 * - fixes
 *
 * Revision 1.2  2001/04/23 18:42:52  slange
 * added copyright notice to all source files
 *
 * --------------------------------------------------------------------
 */

#include "calibratecontroller.h"
#include <iostream>
#include <qfiledialog.h>
#include "cvtkexceptions.h"
#include "rgbhistogram.h"
#include "lookuptable.h"
#include <unistd.h>

using std::invalid_argument;

cvtk::CController::CController(CImageSource* imageSource,
			       CRGBMap& mask,
			       CReinforcedLearning<3>* learner) 
  : QObject(), imageSource(imageSource), mask(mask), specFG(6),
    specBG(10), learner(learner)
{
  if (!imageSource) 
    throw invalid_argument("Null pointer to an image source.");
  ppm=CPPM(imageSource->getWidth(), imageSource->getHeight());
  hist = 0;

  emit frameCountChanged(nFr=0);      // emits can't work, signal can't
  emit objectCountChanged(nObj=0);    // be connected at this point!

  emit message("Ready to grab the backgound!", 4000);

  timer = new QTimer(this);
  connect(timer, SIGNAL(timeout()), this, SLOT(exec()));
  timer->start(0, false);
}
cvtk::CController::~CController()
{
  delete hist;
}
/* slots */
void cvtk::CController::segment()
{
  if (nFr>0) {
    insertHistInX();
    delete hist;                 // empty histogram for later use
    hist = new CRGBHistogram<256>>SHIFTLOOKUP>(); 
  }

  if (x.size() > 0) {
    vector<int> spec(nObj+1, specFG);
    spec[0] = specBG;
    emit message("I'll do some training now", 3000);
    learner->train(x, xC, spec);
    emit message("Finished the training", 1600);
  }
  else
    emit message("Show me something first!", 2000); 
}
void cvtk::CController::makeLookup()
{
  QString filename=
    QFileDialog::getSaveFileName(0,0,0,"Save LookupTable");
  if (filename.isEmpty()) {
    emit message("Not saved. No file specified.", 2000);
  }
  else {
    ofstream out (filename);
    if(!out) {
      emit message("Not saved. Couldn't open file.", 2000);
    }
    else {
      emit message("I'll calculate the lookup table.", 2000);
      CLookUpTable<SHIFTLOOKUP> lu;
      for (unsigned int r=0; r < lu.getSize(); r++)
	for (unsigned int g=0; g < lu.getSize(); g++)
	  for (unsigned int b=0; b < lu.getSize(); b++) {
	    lu.lookup(r<<SHIFTLOOKUP,g<<SHIFTLOOKUP,b<<SHIFTLOOKUP) = 
	      (unsigned char) learner->classify(Vektor3Dim((double)r,
						 (double)g,(double)b));
	  }
      out << lu; 
      emit message("Templates saved to "+filename, 2400);
    }
  }
}

void cvtk::CController::insertHistInX() {
  if (hist) {
    vector<Vektor3Dim> newX = 
      hist->getAsList(3, CRGBHistogram<256>>SHIFTLOOKUP>
		         ::LIST_PROPORTIONAL);
    // get new examples x above the threshold
    
    x.insert(x.end(), newX.begin(), newX.end());    
    // add new examples to x
    xC.insert(xC.end(), newX.size(), nObj);
    // store class information for new examples
  }
}

void cvtk::CController::newObject()
{
  if (nFr>0) {
    insertHistInX();

    delete hist;                 // empty histogram for a new object
    hist = new CRGBHistogram<256>>SHIFTLOOKUP>();                    

    emit objectCountChanged(++nObj);
    emit frameCountChanged(nFr=0);
    emit message("Show me a new object-color.", 1600);
  }
  else
    emit message("You haven't shown me that color yet!", 2400);
}
void cvtk::CController::addFrame()
{
  if (hist==0) hist=new CRGBHistogram<256>>SHIFTLOOKUP>();
  switch (nObj) {
  case 0:                       // get background -> whole image
    for (int x=0; x<ppm.getWidth(); x++)
      for (int y=0; y<ppm.getHeight(); y++) 
	if (SHIFTLOOKUP==0)
	  hist->add(ppm[y][x]);
	else {
	  RGBValue rgb;         // shift rgbValue
	  rgb.r = ppm[y][x].r >> SHIFTLOOKUP;
	  rgb.g = ppm[y][x].g >> SHIFTLOOKUP;
	  rgb.b = ppm[y][x].b >> SHIFTLOOKUP;
	  hist->add(rgb);
	}
    
    break;
  default:                      // calculate histogram of object area
    for (int x=0; x < ppm.getWidth(); x++)
      for (int y=0; y < ppm.getHeight(); y++)
	if (mask[y][x].r==255) {//if pixel in objarea, add to histogram
	  if (SHIFTLOOKUP == 0)                 
	    hist->add(ppm[y][x]);// no shift
	  else {               
	    RGBValue rgb;       // shift rgbValue
	    rgb.r = ppm[y][x].r >> SHIFTLOOKUP;
	    rgb.g = ppm[y][x].g >> SHIFTLOOKUP;
	    rgb.b = ppm[y][x].b >> SHIFTLOOKUP;
	    hist->add(rgb);
	  }
	}
    break;
  }
  
  emit frameCountChanged(++nFr);
  emit message("Succesfully grabbed a frame.", 1400);
}
void cvtk::CController::changeSpecFG(int value)
{
  specFG=value;
  emit message("Speciality of foreground: "+QString().setNum(value),
	       1500);
}
void cvtk::CController::changeSpecBG(int value)
{
  specBG=value;
  emit message("Speciality of background: "+QString().setNum(value), 
	       1500);
}
void cvtk::CController::stopExec()
{
  timer->stop();
}
void cvtk::CController::exec()
{
  RGBValue* rgb;
  RGBValue* m;
  int size;
  unsigned char bright;
  imageSource->getImage(ppm);
 if (ppm.getWidth() != mask.getWidth() || 
      ppm.getHeight() != mask.getHeight())
    throw length_error("Differnt size of image and mask.");
   
  if (nObj) {                  // overlay mask for objects
    rgb=ppm.getBuffer();
    m=mask.getBuffer();
    size=ppm.getWidth() * ppm.getHeight();
    ppm.lock(); mask.lock();
    for (int i=0; i < size ; i++, rgb++,  m++) {
      if (m->r!=0 && m->r!=255) {
	bright = (rgb->r>>2) + (rgb->b>>2) + (rgb->g>>2);
	rgb->r = rgb->b = rgb->g = bright;
      }
    }
    mask.unlock(); ppm.unlock();
  } /* end of overlay mask */
  emit imageChanged(ppm);                  // signal a changed image
}

#include "calibratecontroller.moc"












