#ifndef _calibratecontroller_h_
#define _calibratecontroller_h_

/*
 * Package:     cvtk - Computer Vision Tool Kit
 * File:        calibratecontroller.h
 * Description: Qt Controller of the calibration tool. 
 * Created:     2001-03-28
 * Author:      Sascha Lange
 * Mail:        Sascha.Lange@Uni-Osnabrueck.De
 *
 * Web:         http://cvtk.sourceforge.net
 *
 * Copyright (C) 2001 Sascha Lange
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 *
 * ------------------------- Revision History -------------------------
 * $Log: calibratecontroller.h,v $
 * Revision 1.11  2001/09/01 13:28:41  slange
 * - working on monitor functions
 *
 * Revision 1.10  2001/08/30 13:24:30  slange
 * - added sliders to effect the specificity of the kmsegmentation
 *
 * Revision 1.9  2001/06/19 18:09:24  slange
 * - use QTimer instead of posix threads to read new images from the webcam
 * - implemented a cleanup routine to delete the webcam even in case of a call
 *   to exit(state)
 * BUG: ctrl-c causes the whole system to hang on smp - machines.
 * REASON: filedescriptor of the videodevice is not closed!
 *
 * Revision 1.8  2001/05/03 15:04:09  slange
 * - added lookup table data for orange/yellow/blue on green/black/white
 *   background to rsc
 * - fixes
 *
 * Revision 1.7  2001/05/03 13:19:44  slange
 * - efficiency improvement of frame grabbing/histogram analysis in the
 *   CController
 *
 * Revision 1.6  2001/05/03 13:04:27  slange
 * - traced down bug in CController::addFrame()
 *
 * Revision 1.5  2001/04/29 00:55:49  slange
 * - finished the first complete draft of the calibration tool.
 *   - it now generates a lookup-table for values 1..64 (>> 2)
 *   - writes the entire table to stdout.
 * - fixes in k-means segmentation
 * - added button to generate lookup table in the calibrator-widget
 *
 * Revision 1.4  2001/04/28 18:05:45  slange
 * - removed patterngenerato.[h/cpp]
 * - added abstract interface CReinforcedLearning
 * - added CKMSegmentation
 *   - class definition and part of implementation
 * - added "train" button to calibratorwidget
 * - changed calibratecontroller to work with the new classes
 *
 * Revision 1.3  2001/04/27 23:15:42  slange
 * - added an UV - Histogram (histogram in an approximation of the HUV space)
 * - changed the Calibrate::CController
 *   - to work with histograms
 *   - to keep a vector of examples in memory
 *   - output file dropped
 *   - to tread the background as a (special) object
 * - fixes
 *
 * Revision 1.2  2001/04/23 18:42:52  slange
 * added copyright notice to all source files
 *
 * --------------------------------------------------------------------
 */

#include <qapplication.h>
#include <fstream.h>
#include "imagesource.h"
#include "cvtk.h"
#include "kmsegmentation.h"
#include <qtimer.h>
#include "rgbhistogram.h"

/* collects a vector of training examples x and a corresponding vector xC with
 * the class information of every x as needed by reinforced learning-/
 * segmentation-algorithms
 */
namespace cvtk {
#define SHIFTLOOKUP 2
  class CController : public QObject
  {
  Q_OBJECT
  public:
    explicit CController(CImageSource* imageSource, 
			 CRGBMap& mask, 
			 CReinforcedLearning<3>* learner);
    ~CController();
  public slots:
    void newObject();
    void addFrame();
    void stopExec();
    void segment();
    void makeLookup();
    void changeSpecFG(int);
    void changeSpecBG(int);
    void exec();                  // Thread execution function
  signals:
    void message(const char* str, int ms);
    void frameCountChanged(int n);
    void objectCountChanged(int n);
    void imageChanged(const CRGBMap& ppm);
  private:
    void insertHistInX();
    CImageSource* imageSource;
    CPPM ppm;
    CRGBMap mask;
    int nObj;
    int nFr;
    int specFG;
    int specBG;
    QTimer* timer;
    CController(const CController&);// no copies allowed
    CController& operator= (const CController&);

    vector<int> xC;              // class identification of every x
    vector<Vektor3Dim> x;        // collected examples x

    CReinforcedLearning<3>* learner;
    CRGBHistogram<256 >> SHIFTLOOKUP>* hist;
  };
}
  
#endif






