/*
 * Package:     cvtk - Computer Vision Tool Kit
 * File:        calibratecontrolleruv.cpp
 * Description: Implementation of the controller of the calibration 
 *              tool.
 * Created:     2001-04-09
 * Author:      Sascha Lange
 * Mail:        Sascha.Lange@Uni-Osnabrueck.De
 *
 * Web:         http://cvtk.sourceforge.net
 *
 * Copyright (C) 2001 Sascha Lange
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 *
 * -------------------------- Revision History ------------------------
 * $Log: calibratecontrolleruv.cpp,v $
 * Revision 1.2  2001/09/04 20:59:13  slange
 * - ready
 *
 * Revision 1.1  2001/09/04 20:00:35  slange
 * - initial checkin of uvhistogram calibration classes
 *
 * --------------------------------------------------------------------
 */

#include "calibratecontrolleruv.h"
#include <iostream>
#include <qfiledialog.h>
#include "cvtkexceptions.h"
#include "uvhistogram.h"
#include "lookuptable.h"
#include <unistd.h>

using std::invalid_argument;

cvtk::CController::CController(CImageSource* imageSource,
			       CRGBMap& mask,
			       CReinforcedLearning<2>* learner) 
  : QObject(), imageSource(imageSource), mask(mask), specFG(6),
    specBG(10), learner(learner)
{
  if (!imageSource) 
    throw invalid_argument("Null pointer to an image source.");
  ppm=CPPM(imageSource->getWidth(), imageSource->getHeight());
  hist = 0;

  emit frameCountChanged(nFr=0);      // emits can't work, signal can't
  emit objectCountChanged(nObj=0);    // be connected at this point!

  emit message("Ready to grab the backgound!", 4000);

  timer = new QTimer(this);
  connect(timer, SIGNAL(timeout()), this, SLOT(exec()));
  timer->start(0, false);
}
cvtk::CController::~CController()
{
  delete hist;
}
/* slots */
void cvtk::CController::segment()
{
  if (nFr>0) {
    insertHistInX();
    delete hist;                 // empty histogram for later use
    hist = new CUVHistogram(); 
  }

  if (x.size() > 0) {
    vector<int> spec(nObj+1, specFG);
    spec[0] = specBG;
    emit message("I'll do some training now", 3000);
    learner->train(x, xC, spec);
    emit message("Finished the training", 1600);
  }
  else
    emit message("Show me something first!", 2000); 
}
void cvtk::CController::makeLookup()
{
  QString filename=
    QFileDialog::getSaveFileName(0,0,0,"Save LookupTable");
  if (filename.isEmpty()) {
    emit message("Not saved. No file specified.", 2000);
  }
  else {
    ofstream out (filename);
    if(!out) {
      emit message("Not saved. Couldn't open file.", 2000);
    }
    else {
      emit message("I'll calculate the lookup table.", 2000);
      CLookUpTable<SHIFTLOOKUP> lu;
      for (unsigned int r=0; r < lu.getSize(); r++)
	for (unsigned int g=0; g < lu.getSize(); g++)
	  for (unsigned int b=0; b < lu.getSize(); b++) {
	    lu.lookup(r<<SHIFTLOOKUP,g<<SHIFTLOOKUP,b<<SHIFTLOOKUP) = 
	      (unsigned char) learner->classify(
		Vektor2Dim((double)((r<<SHIFTLOOKUP)-
				    (b<<SHIFTLOOKUP)+256),
			   (double)((2*(g<<SHIFTLOOKUP)-
				     (r<<SHIFTLOOKUP)-
				     (b<<SHIFTLOOKUP))/2+256)));
	  }
      out << lu; 
      emit message("Templates saved to "+filename, 2400);
    }
  }
}

void cvtk::CController::insertHistInX() {
  if (hist) {
    vector<Vektor2Dim> newX = 
      hist->getAsList(3, CUVHistogram::LIST_PROPORTIONAL);
    // get new examples x above the threshold
    
    x.insert(x.end(), newX.begin(), newX.end());    
    // add new examples to x
    xC.insert(xC.end(), newX.size(), nObj);
    // store class information for new examples
  }
}

void cvtk::CController::newObject()
{
  if (nFr>0) {
    insertHistInX();

    delete hist;                 // empty histogram for a new object
    hist = new CUVHistogram();                    

    emit objectCountChanged(++nObj);
    emit frameCountChanged(nFr=0);
    emit message("Show me a new object-color.", 1600);
  }
  else
    emit message("You haven't shown me that color yet!", 2400);
}
void cvtk::CController::addFrame()
{
  if (hist==0) hist=new CUVHistogram();
  switch (nObj) {
  case 0:                       // get background -> whole image
    for (int x=0; x<ppm.getWidth(); x++)
      for (int y=0; y<ppm.getHeight(); y++) 
	hist->add(ppm[y][x]);
    break;
  default:                      // calculate histogram of object area
    for (int x=0; x < ppm.getWidth(); x++)
      for (int y=0; y < ppm.getHeight(); y++)
	if (mask[y][x].r==255) {//if pixel in objarea, add to histogram
	  hist->add(ppm[y][x]);
	}
    break;
  }
  
  emit frameCountChanged(++nFr);
  emit message("Succesfully grabbed a frame.", 1400);
}
void cvtk::CController::changeSpecFG(int value)
{
  specFG=value;
  emit message("Speciality of foreground: "+QString().setNum(value),
	       1500);
}
void cvtk::CController::changeSpecBG(int value)
{
  specBG=value;
  emit message("Speciality of background: "+QString().setNum(value), 
	       1500);
}
void cvtk::CController::stopExec()
{
  timer->stop();
}
void cvtk::CController::exec()
{
  RGBValue* rgb;
  RGBValue* m;
  int size;
  unsigned char bright;
  imageSource->getImage(ppm);
  if (ppm.getWidth() != mask.getWidth() || 
      ppm.getHeight() != mask.getHeight())
    throw length_error("Differnt size of image and mask.");
   
  if (nObj) {                  // overlay mask for objects
    rgb=ppm.getBuffer();
    m=mask.getBuffer();
    size=ppm.getWidth() * ppm.getHeight();
    ppm.lock(); mask.lock();
    for (int i=0; i < size ; i++, rgb++,  m++) {
      if (m->r!=0 && m->r!=255) {
	bright = (rgb->r>>2) + (rgb->b>>2) + (rgb->g>>2);
	rgb->r = rgb->b = rgb->g = bright;
      }
    }
    mask.unlock(); ppm.unlock();
  } /* end of overlay mask */
  emit imageChanged(ppm);                  // signal a changed image
}

#include "calibratecontrolleruv.moc"












