/*
 * Package:     cvtk - Computer Vision Tool Kit
 * File:        calibrateuv.cpp
 * Description: Main of the calibration tool. Sets up and connects the
 *              model and the controller.
 * Created:     2001-04-09
 * Author:      Sascha Lange
 * Mail:        Sascha.Lange@Uni-Osnabrueck.De
 *
 * Web:         http://cvtk.sourceforge.net
 *
 * Copyright (C) 2001 Sascha Lange
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 
 * 02111-1307, USA.
 *
 * ------------------------- Revision History -------------------------
 * $Log: calibrateuv.cpp,v $
 * Revision 1.1  2001/09/04 20:00:35  slange
 * - initial checkin of uvhistogram calibration classes
 *
 * --------------------------------------------------------------------
 */

#include "webcam.h"
#include <iostream>
#include <fstream>
#include <unistd.h>
#include "imagewidget.h"
#include <qapplication.h>
#include "calibratewidget.h"
#include "calibratecontrolleruv.h"
#include "cvtkexceptions.h"
#include "kmsegmentation.h"

using cvtk::CPPM;
using cvtk::CRGBMap;
using cvtk::RGBValue;
using cvtk::CWebCam;
using cvtk::CImageSource;
using cvtk::CCalibrateWidget;
using cvtk::CController;
using cvtk::CImageWidget;

void printUsage(char* str)
{
  cout << "Usage: " << str << " mask.ppm\n";
}

CWebCam* cam;

void cleanup()
{
  //  cerr << "in cleanup" << endl;
  delete cam;
}

int main(int argc, char* argv[]) {
  if (argc!=2) {
    printUsage(argv[0]);
    exit(1);
  }
  std::ifstream in (argv[1]);
  if (!in) {
    cerr << "Cannot open mask file" << argv[1];
    exit(1);
  }
  QApplication app(argc, argv);

  try {
    CPPM mask;
    in >> mask;
    
    cam = new CWebCam(352,288);
    if (atexit(cleanup)!=0) {
      cerr << "Can't register cleanup function" << endl;
      delete cam;
      exit(0);
    }

    CCalibrateWidget iw(0,0);
    CKMSegmentation<2> segmenter(3, 20, .0001);
    
    CController con(cam, mask, &segmenter);
    
    app.connect(&iw, SIGNAL(newCommand()), 
		&con, SLOT(newObject()));
    app.connect(&iw, SIGNAL(addCommand()), 
		&con, SLOT(addFrame()));
    app.connect(&iw, SIGNAL(trainCommand()), 
		&con, SLOT(segment()));
    app.connect(&iw, SIGNAL(lookupCommand()), 
		&con, SLOT(makeLookup()));
    app.connect(&iw, SIGNAL(specFGChanged(int)), 
		&con, SLOT(changeSpecFG(int)));
    app.connect(&iw, SIGNAL(specBGChanged(int)), 
		&con, SLOT(changeSpecBG(int)));
    app.connect(&app, SIGNAL(lastWindowClosed()), 
		&con, SLOT(stopExec()));
    
    app.connect(&con, SIGNAL(message(const char*, int)), 
		&iw, SLOT(message(const char*, int)));
    app.connect(&con, SIGNAL(frameCountChanged(int)), 
		&iw, SLOT(setFrameCount(int)));
    app.connect(&con, SIGNAL(objectCountChanged(int)), 
		&iw, SLOT(setObjectCount(int)));
    app.connect(&con, SIGNAL(imageChanged(const CRGBMap&)), 
		&iw, SLOT(setImage(const CRGBMap&)));
     
    iw.show();
    app.setMainWidget(&iw);
    app.exec();
    //    cerr << "nach exec " << endl;
  
  }
  catch(no_file& e) {
    app.exit(0);
  }
  catch(exception& e) {
    cerr << e.what() << endl;
    app.exit(1);
  }
}






