/*
 * Package:     cvtk - Computer Vision Tool Kit
 * File:        calibratewidget.cpp
 * Description: Implementation of the widget for the calibration tool.
 * Created:     2001-03-22
 * Author:      Sascha Lange
 * Mail:        Sascha.Lange@Uni-Osnabrueck.De
 *
 * Web:         http://cvtk.sourceforge.net
 *
 * Copyright (C) 2001 Sascha Lange
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 
 * 02111-1307, USA.
 *
 * ------------------------- Revision History -------------------------
 * $Log: calibratewidget.cpp,v $
 * Revision 1.9  2001/09/04 20:00:35  slange
 * - initial checkin of uvhistogram calibration classes
 *
 * Revision 1.8  2001/09/04 15:31:27  slange
 * - corrected bugs in:
 *   - templatecontroller.cpp: use QString to emit messages
 *   - monitor.cpp: used x instead of y to check against height
 * - labeled the sliders in calibratewidget
 *
 * Revision 1.7  2001/09/01 13:28:41  slange
 * - working on monitor functions
 *
 * Revision 1.6  2001/08/30 17:25:32  slange
 * - added ft_interface to drive a fischertechnik robot
 * - working on demo application
 *
 * Revision 1.5  2001/08/30 13:24:30  slange
 * - added sliders to effect the specificity of the kmsegmentation
 *
 * Revision 1.4  2001/04/29 00:55:49  slange
 * - finished the first complete draft of the calibration tool.
 *   - it now generates a lookup-table for values 1..64 (>> 2)
 *   - writes the entire table to stdout.
 * - fixes in k-means segmentation
 * - added button to generate lookup table in the calibrator-widget
 *
 * Revision 1.3  2001/04/28 18:05:45  slange
 * - removed patterngenerato.[h/cpp]
 * - added abstract interface CReinforcedLearning
 * - added CKMSegmentation
 *   - class definition and part of implementation
 * - added "train" button to calibratorwidget
 * - changed calibratecontroller to work with the new classes
 *
 * Revision 1.2  2001/04/23 18:42:52  slange
 * added copyright notice to all source files
 *
 * --------------------------------------------------------------------
 */

#include "calibratewidget.h"

cvtk::CCalibrateWidget::CCalibrateWidget(const CRGBMap& image) 
  : QWidget(0, "Calibrate", WType_TopLevel), imageWidget(image, this, 0)
{
  initWidget();
  doResize(imageWidget.width(), imageWidget.height());
}
cvtk::CCalibrateWidget::CCalibrateWidget(int width, int height) 
  : QWidget(0, "Calibrate", WType_TopLevel), 
    imageWidget(width, height, this, 0)
{
  initWidget();
  doResize(imageWidget.width(), imageWidget.height());
}
cvtk::CCalibrateWidget::~CCalibrateWidget()
{
  delete newButton;
  delete addButton;
  delete trainButton;
  delete lookupButton;
  delete objectLabel;
  delete frameLabel;
  delete statusBar;
  delete specFGSlider;
  delete specBGSlider;
  delete fgLabel;
  delete bgLabel;
}
void cvtk::CCalibrateWidget::setImage(const CRGBMap& image)
{
  int oldWidth=imageWidget.width();
  int oldHeight=imageWidget.height();
  imageWidget.setImage(image);
  if (oldWidth != imageWidget.width() || 
      oldHeight != imageWidget.height())
    doResize(imageWidget.width(), imageWidget.height());
}
void cvtk::CCalibrateWidget::doResize(int w, int h)
{
  w = (w<220) ? 220 : w;
  h = (h<190) ? 190 : h;
  setFixedSize(w,h+150+(2*specFGSlider->sizeHint().height()));
  imageWidget.move((w-imageWidget.width())/2, 
		   (h-imageWidget.height())/2);
  newButton->setGeometry(10,h+10, (w-30)/2, 40);
  addButton->setGeometry(w/2+5, h+10, (w-30)/2, 40);
  trainButton->setGeometry(10, h+60, (w-30)/2, 40);
  lookupButton->setGeometry(w/2+5, h+60, (w-30)/2, 40);
  specFGSlider->setGeometry(110,
			    h+110, 
			    (w-120), 
			    specFGSlider->sizeHint().height());
  specBGSlider->setGeometry(110,
			    h+120+specFGSlider->sizeHint().height(),
			    (w-120),
			    specFGSlider->sizeHint().height());
  fgLabel->setGeometry(10,h+110, 90,specFGSlider->sizeHint().height());
  bgLabel->setGeometry(10,h+120+specFGSlider->sizeHint().height(), 
		       90, specBGSlider->sizeHint().height());
  statusBar->setGeometry(0,h+130+(2*specFGSlider->sizeHint().height()),
			 w,20);
}
void cvtk::CCalibrateWidget::initWidget()
{
  // Setup status bar 
  statusBar=new QStatusBar(this);
  objectLabel=new QLabel(statusBar); 
  objectLabel->setMaximumWidth(25);
  objectLabel->setAlignment(AlignRight);
  objectLabel->setNum(0);
  frameLabel= new QLabel(statusBar);
  frameLabel->setMaximumWidth(25);
  frameLabel->setAlignment(AlignRight);
  frameLabel->setNum(0);
  statusBar->addWidget(objectLabel, 1, true);
  statusBar->addWidget(frameLabel, 1, true);
  statusBar->message("CVTK Calibrate", 3000);

  // add buttons
  newButton=   new QPushButton("New Object", this);
  addButton=   new QPushButton("Grab", this);
  trainButton= new QPushButton("Train", this);
  lookupButton=new QPushButton("Lookup Table", this);
  specFGSlider=new QSlider(1,50,1,6,QSlider::Horizontal, this);
  specBGSlider=new QSlider(2,200,2,10,QSlider::Horizontal, this);
  specFGSlider->setTracking(false);
  specBGSlider->setTracking(false);
  fgLabel=     new QLabel(QString("Vordergrund"), this);
  bgLabel=     new QLabel(QString("Hintergrund"), this);

  // connect internal messages
  connect(newButton, SIGNAL(clicked()), 
	  this, SLOT(newButtonClicked()));
  connect(addButton, SIGNAL(clicked()), 
	  this, SLOT(addButtonClicked()));
  connect(trainButton, SIGNAL(clicked()), 
	  this, SLOT(trainButtonClicked()));
  connect(lookupButton, SIGNAL(clicked()), 
	  this, SLOT(lookupButtonClicked()));
  connect(specFGSlider, SIGNAL(valueChanged(int)), 
	  this, SLOT(specFGSliderChanged(int)));
  connect(specBGSlider, SIGNAL(valueChanged(int)), 
	  this, SLOT(specBGSliderChanged(int)));

  message("Let's get to work.", 3000);
}


/* basic slot implementations */

void cvtk::CCalibrateWidget::message(const char* message, int ms)
{
  if (ms==0)
    statusBar->message(message);
  else
    statusBar->message(message, ms);
}
void cvtk::CCalibrateWidget::setObjectCount(int n)
{
  objectLabel->setNum(n);
}
void cvtk::CCalibrateWidget::setFrameCount(int n)
{
  frameLabel->setNum(n);
}
void cvtk::CCalibrateWidget::newButtonClicked()
{
  emit newCommand();
}
void cvtk::CCalibrateWidget::addButtonClicked()
{
  emit addCommand();
}
void cvtk::CCalibrateWidget::trainButtonClicked()
{
  emit trainCommand();
}
void cvtk::CCalibrateWidget::lookupButtonClicked()
{
  emit lookupCommand();
}
void cvtk::CCalibrateWidget::specFGSliderChanged(int value)
{
  emit specFGChanged(value);
}
void cvtk::CCalibrateWidget::specBGSliderChanged(int value)
{
  emit specBGChanged(value);
}
#include "calibratewidget.moc"














