/*
 * Package:     cvtk - Computer Vision Tool Kit
 * File:        chaincoding.cpp
 * Description: Implementation of the region chain coding algorithm.
 * Created:     2001-04-22
 * Author:      Sascha Lange
 * Mail:        Sascha.Lange@Uni-Osnabrueck.De
 *
 * Web:         http://cvtk.sourceforge.net
 *
 * Copyright (C) 2001 Sascha Lange
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 *
 * ------------------------ Revision History --------------------------
 * $Log: chaincoding.cpp,v $
 * Revision 1.5  2001/09/01 17:40:09  slange
 * - working on monitor
 * - moved template calibration into namespace "cvtk"
 *
 * Revision 1.4  2001/08/28 13:16:48  slange
 * - traced down a bug: historylist to small!
 *
 * Revision 1.3  2001/08/14 08:28:38  slange
 * found a bug in templatemantching.cpp
 *
 * Revision 1.2  2001/08/09 23:15:08  slange
 * - reorganised chaincoding.[h/cpp] (former regions.h, regionfinder.cpp)
 * - added a destructor to color segmentation
 * - added a constructor to region list
 *
 * Revision 1.1  2001/08/09 21:47:20  slange
 * - working on the chaincoding algorithm
 * - changed regionlist from a simple typedef to a class and
 *   moved it into regionlist.h (former region.h)
 *
 * Revision 1.9  2001/08/04 20:14:20  slange
 * finished template calibration
 *
 * Revision 1.8  2001/08/04 19:41:17  slange
 * corrected getNumberOfClasses in lookuptable.h
 *
 * Revision 1.7  2001/08/04 16:43:07  slange
 * - added some mehtods needed by template calibration
 * - added mouseClick event receiver
 *
 * Revision 1.6  2001/07/11 17:51:55  slange
 * - found and corrected bug in the region algorithm
 *
 * Revision 1.5  2001/07/07 21:46:18  slange
 * - working on template matching: complete!
 * - todo: implement delete and stop after first match!
 *
 * Revision 1.4  2001/06/21 23:53:42  slange
 * - initial checkin of templatematching.[h/cpp]
 * - added method clear() to CRegionList
 *
 * Revision 1.3  2001/04/24 19:37:14  slange
 * initialise augmented states to track orange/yellow/blue on green/white/black background
 *
 * Revision 1.2  2001/04/23 18:42:52  slange
 * added copyright notice to all source files
 *
 * --------------------------------------------------------------------
 */

#include "chaincoding.h" 

/*********************** CDefaultChainCoding *************************/

cvtk::CDefaultChainCoding::
CDefaultChainCoding(CBorderTracingStrategy* traceStrat,
		    int maxRegionClass, int min, bool procInner) 
  : CChainCodingStrategy(), min(min), procInner(procInner), 
    maxRegionClass(maxRegionClass), traceStrat(traceStrat)
{
  if (!traceStrat) {
    cerr << "CDefaultChainCoding: Not a valid border tracing Strategy."
	 << endl;
    exit(1);
  }
}

cvtk::CDefaultChainCoding::~CDefaultChainCoding()
{
  delete traceStrat;
}


cvtk::CRegionList* 
cvtk::CDefaultChainCoding::codeRegions(const CRegionMap *regMap)
{
  if (!regMap)
    return 0;
  return codeRegions(regMap, new CRegionList(maxRegionClass));
}  

cvtk::CRegionList* 
cvtk::CDefaultChainCoding::codeRegions(const CRegionMap *regMap,
				       CRegionList *regions)
{
  if (regMap == 0)
    return 0;
  if (regions == 0) 
    return 0;     // exceptions !!

  if (regions->size() != maxRegionClass+1)
    regions->resize(maxRegionClass+1);

  // prepare datastructures to start algorithm
  if (regMap->getWidth() != regBounds.getWidth() ||
      regMap->getHeight() != regBounds.getHeight()) 
    regBounds.resize(regMap->getWidth(), regMap->getHeight());
  for (int x=0; x < regMap->getWidth(); x++)   // delete outer row 
    (*regMap)[0][x]=(*regMap)[regMap->getHeight()-1][x]=0;
  for (int y=0; y < regMap->getHeight(); y++)  // delete outer column
    (*regMap)[y][0]=(*regMap)[y][regMap->getWidth()-1]=0;  

  regBounds.clear();

  int y,x;
  int* buf;
  int segCl;
  bool inRegion;
  for (y=0, buf=regMap->getBuffer(); y < regMap->getHeight(); y++)
  {
    segCl = 0;
    inRegion=false;
    for (x=0; x < regMap->getWidth(); x++, buf++) { 
      switch(inRegion) {
      case true:
	if (regBounds[y][x] && segCl!=(*regMap)[y][x+1])
	  inRegion = false;
	break;
      case false:
	if (segCl!=*buf) {          // other region
	  segCl = *buf;
	  if (segCl && !regBounds[y][x]) {
	                            // not background && not yet traced
	    CRegion reg;
	    reg.start = Vektor<int, 2> (x,y);
	    reg.regionClass=segCl;
	    if (traceStrat->traceBorder(reg, *regMap, regBounds)) {
	      if (min>2) {          // calculate area only if necessary
		if (reg.size>4 && reg.getArea()>min) { 
		  regions->push_back(reg.regionClass, reg);
		}
	      }
	      else if (reg.size > min) 
		regions->push_back(reg.regionClass, reg);
	    }
	  }
	}
	if (!procInner && regBounds[y][x] 
	    && segCl==(*regMap)[y][x+1]) inRegion=true;
	                            // entered Region
	break;
      }
    }
  }
  return regions;
}

/********************** CBorderTracingStrategy ***********************/

/* Border tracing adapted from:
 *   Sonka, Hlavac, Boyle; "Image Processing, Analysis, and Machine
 *   Vision", Algorithm 5.8 (Adapted)
 */
int cvtk::CBorderTracing4::traceBorder(CRegion& reg,
				       const CRegionMap& regMap,
				       CMapWrapper<bool>& regBounds)
{
  int x=reg.start[0];
  int y=reg.start[1];
  reg.size = 1;
  int dir = 6;             // start direction for 4 AND 8-connectivity

  regBounds[y][x]=true;
  {  // try, if it's a single isolated pixel 
    int i=0;
    for (; i <= 8; i+=2)
      if (regMap[y+yDir(i)][x+xDir(i)]==reg.regionClass) break;
    if (i==8) return reg.size;     // single pixel, stop tracing;
    else {                         // found a second border pixel
      reg.size++; 
      dir = i;                     // direction of step
      reg.chaincode.push_back(dir);// push direction of step
      x+=xDir(i);
      y+=yDir(i);
      reg.second = Vektor<int, 2> (x,y);
      regBounds[y][x]=true;        // mark as inner boundary
    }	       
  } 

  // here we can assert, that there is another pixel to find!
  do {
    for(int sDir=(dir+6) % 8; ; sDir = (sDir+2) % 8) {
      if (regMap[y+yDir(sDir)][x+xDir(sDir)] == reg.regionClass) {
	x+=xDir(sDir);
	y+=yDir(sDir);
	dir=sDir;
	reg.chaincode.push_back(dir);
	reg.size++;
	regBounds[y][x]=true;
	break;
      }
    }
  } while (x != reg.second[0] || 
	   y != reg.second[1] ||         // Pn != P1
	   x-xDir(dir) != reg.start[0] ||
	   y-yDir(dir) != reg.start[1]); // Pn-1 != P0
  reg.chaincode.pop_back();
  reg.size--;
  return reg.size;
}

int cvtk::CBorderTracing8::traceBorder(CRegion& reg,
				       const CRegionMap& regMap,
				       CMapWrapper<bool>& regBounds)
{
  int x=reg.start[0];
  int y=reg.start[1];
  reg.size = 1;
  int dir = 7;             // start direction for 4 AND 8-connectivity

  regBounds[y][x]=true;
  {  // try, if it's a single isolated pixel 
    int i=0;
    for (; i <= 8; i++)
      if (regMap[y+yDir(i)][x+xDir(i)]==reg.regionClass) break;
    if (i==8) return reg.size;     // single pixel, stop tracing;
    else {                         // found a second border pixel
      reg.size++; 
      dir = i;                     // direction of step
      reg.chaincode.push_back(dir);// push direction of step
      x+=xDir(i);
      y+=yDir(i);
      reg.second = Vektor<int, 2> (x,y);
      regBounds[y][x]=true;        // mark as inner boundary
    }	       
  } 

  // here we can assert, that there is another pixel to find!
  do {
    for(int sDir=(dir+ (dir%2 ? 6:7) ) % 8; // even -> +7, odd -> +6
	; 
	sDir = (++sDir) % 8) {
      if (regMap[y+yDir(sDir)][x+xDir(sDir)] == reg.regionClass) {
	x+=xDir(sDir);
	y+=yDir(sDir);
	dir=sDir;
	reg.chaincode.push_back(dir);
	reg.size++;
	regBounds[y][x]=true;
	break;
      }
    }
  } while (x != reg.second[0] || 
	   y != reg.second[1] ||         // Pn != P1
	   x-xDir(dir) != reg.start[0] ||
	   y-yDir(dir) != reg.start[1]); // Pn-1 != P0
  reg.chaincode.pop_back();
  reg.size--;
  return reg.size;
}




