#ifndef _regions_h_
#define _regions_h_

/*
 * Package:     cvtk - Computer Vision Tool Kit
 * File:        chaincoding.h
 * Description: Defines the algorithms necessary to get a higher 
 *              representation of the segmented image (map).
 * Created:     2001-08-09
 * Author:      Sascha Lange
 * Mail:        Sascha.Lange@Uni-Osnabrueck.De
 *
 * Web:         http://cvtk.sourceforge.net
 *
 * Copyright (C) 2001 Sascha Lange
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 *
 * ------------------------ Revision History --------------------------
 * $Log: chaincoding.h,v $
 * Revision 1.5  2001/09/01 17:40:09  slange
 * - working on monitor
 * - moved template calibration into namespace "cvtk"
 *
 * Revision 1.4  2001/08/28 13:16:48  slange
 * - traced down a bug: historylist to small!
 *
 * Revision 1.3  2001/08/10 14:11:57  slange
 * working on the next build
 *
 * Revision 1.2  2001/08/09 23:15:08  slange
 * - reorganised chaincoding.[h/cpp] (former regions.h, regionfinder.cpp)
 * - added a destructor to color segmentation
 * - added a constructor to region list
 *
 * Revision 1.1  2001/08/09 21:47:20  slange
 * - working on the chaincoding algorithm
 * - changed regionlist from a simple typedef to a class and
 *   moved it into regionlist.h (former region.h)
 *
 * --------------------------------------------------------------------
 */

#include "cvtk.h"
#include "mapwrapper.h"

namespace cvtk {

  /************************ Border tracing ***************************/

  /* abstract interface
   */
  class CBorderTracingStrategy {
  public:
    virtual ~CBorderTracingStrategy() {};
    virtual int traceBorder(CRegion& reg, 
			    const CRegionMap& regionMap,
			    CMapWrapper<bool>& regBounds) =0;
  protected:
    static int xDir(int dir);
    static int yDir(int dir);
    CBorderTracingStrategy() {};
    
    friend class Monitors;
  };

  /* border tracing 4-connectivity
   */
  class CBorderTracing4 : public CBorderTracingStrategy {
  public:
    virtual int traceBorder(CRegion& reg,
			    const CRegionMap& regionMap,
			    CMapWrapper<bool>& regBounds);
  };

  /* border tracing 8-connectivity
   */
  class CBorderTracing8 : public CBorderTracingStrategy {
  public:
    virtual int traceBorder(CRegion& reg,
			    const CRegionMap& regionMap,
			    CMapWrapper<bool>& regBounds);
  };

  /********************* Chaincoding algorithm ***********************/

  class CDefaultChainCoding : public CChainCodingStrategy {
  public:
    CDefaultChainCoding(CBorderTracingStrategy* traceStrat, 
			int maxRegionClass,
			int min=3, bool procInner=false);
    virtual ~CDefaultChainCoding();

    virtual CRegionList* codeRegions(const CRegionMap*);
    virtual CRegionList* codeRegions(const CRegionMap *, CRegionList*);

  private:
    CMapWrapper<bool> regBounds;
    int min;
    bool procInner;
    unsigned int maxRegionClass;
    CBorderTracingStrategy* traceStrat;
  };

}

inline int cvtk::CBorderTracingStrategy::xDir(int dir)
{
  switch (dir) {
  case 7:
  case 0:
  case 1: return 1;
  case 3: 
  case 4: 
  case 5: return -1;
    default: return 0;
  }
}
inline int cvtk::CBorderTracingStrategy::yDir(int dir)
{
  switch (dir) {
  case 1:
  case 2:
  case 3: return 1;
  case 5:
  case 6:
  case 7: return -1;
  default: return 0;
  }
}

#endif    /* _chaincoding_h_ */ 












