#ifndef _classifiedlist_h_
#define _classifiedlist_h_

/*
 * Package:     cvtk - Computer Vision Tool Kit
 * File:        classifiedlist.h
 * Description: simple kind of multimap 
 * Created:     2001-07-17
 * Author:      Sascha Lange
 * Mail:        Sascha.Lange@Uni-Osnabrueck.De
 *
 * Web:         http://cvtk.sourceforge.net
 *
 * Copyright (C) 2001 Sascha Lange
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  
 * 02111-1307, USA.
 *
 * ------------------------ Revision History --------------------------
 * $Log: classifiedlist.h,v $
 * Revision 1.9  2001/08/10 01:23:45  slange
 * finished first pass over
 *
 * Revision 1.8  2001/08/09 14:28:57  slange
 * - added resize() to classifiedlist.h
 * - modified getArea() of CRegion to work with both,
 *   4- and 8-connectivity.
 * - 4 connectivity now coded with codes of 8-connectivity
 *
 * Revision 1.7  2001/08/07 12:40:14  slange
 * - still refactoring
 * - initial checkin of defaultfactory.cpp, objectlist.h region.h
 *
 * Revision 1.6  2001/07/29 10:27:30  slange
 * - finished object history matching!
 *
 * Revision 1.5  2001/07/28 22:52:39  slange
 * - edited erase function
 *
 * Revision 1.4  2001/07/28 22:22:43  slange
 * - found and traced down a bug
 *
 * Revision 1.3  2001/07/28 21:51:15  slange
 * - finished search algorithm
 *
 * Revision 1.2  2001/07/27 20:35:34  slange
 * - initial checkin of searchmatch.[h/cpp]
 *
 * Revision 1.1  2001/07/16 20:02:34  slange
 * - working on the history matching
 *
 * --------------------------------------------------------------------
 */

#include <vector>
#include <list>

/* A classified list is a list of buckets for several classes. 
 * To insert an object, it's class (or "key") must be specified.
 *
 * Don't overwrite the none virtual functions in a derived class.!
 */
template<class T, class CONTAINER = vector<T> >
class CClassifiedList
{
public:
  CClassifiedList (int nClasses=0);      // constructs n empty buckets
  virtual ~CClassifiedList ();

  size_t size() const;                   // get the number of buckets
  void resize(size_t size);

  CONTAINER& operator[] (int i);         // element access via
  const CONTAINER& operator[] (int i) const;// [bucket][element]
  void push_back(int c, const T& o);     // insert function
  void push_front(int c, const T& o);    // attention: front operations
  void pop_front(int c);                 // not available for vector
  void pop_back(int c);

  bool erase(int c, const T& o); // erase the first occurence

  typedef vector<CONTAINER>::iterator iterator; // iterator functions
  typedef typename CONTAINER::iterator innerIterator;
  iterator begin() { return list.begin(); };
  iterator end() { return list.end(); };

  void clear();

protected:
  vector<CONTAINER> list;
};


/************************* inline functions **************************/

template<class T, class CONTAINER>
inline void CClassifiedList<T,CONTAINER>::push_back(int c, const T& o)
{
  list[c].push_back(o);
};
template<class T, class CONTAINER>
inline void CClassifiedList<T,CONTAINER>::push_front(int c, const T& o)
{
  list[c].push_front(o);
};
template<class T, class CONTAINER>
inline void CClassifiedList<T,CONTAINER>::pop_back(int c)
{
  list[c].pop_back(o);
};
template<class T, class CONTAINER>
inline void CClassifiedList<T,CONTAINER>::pop_front(int c)
{
  list[c].pop_front(o);
};

template<class T, class CONTAINER>
inline bool CClassifiedList<T,CONTAINER>::erase(int c, const T& o)
{
  int co=0;
  for (innerIterator i=list[c].begin(); i != list[c].end(); i++, co++)
    if (*i == o) {
      list[c].erase(i);
      return true;
    }
  return false;
}
template<class T, class CONTAINER>
inline void CClassifiedList<T,CONTAINER>::resize(size_t size)
{
  list.resize(size);
}
template<class T, class CONTAINER>
inline size_t CClassifiedList<T,CONTAINER>::size() const
{
  return list.size();
};
template<class T, class CONTAINER>
inline CONTAINER& CClassifiedList<T,CONTAINER>::operator[](int i)
{
  return list[i];
};
template<class T, class CONTAINER>
inline const CONTAINER& 
CClassifiedList<T,CONTAINER>::operator[](int i) const
{
  return list[i];
};

#ifdef EXPORT_NOT_IMPLEMENTED
#include "classifiedlist.cpp"
#endif   /* EXPORT_NOT_IMPLEMENTED */

#endif




