/*
 * Package:     cvtk - Computer Vision Tool Kit
 * File:        colorsegmentation.cpp
 * Description: Implementation of the color based segmentation
 *              algorithm.
 * Created:     2001-04-19
 * Author:      Sascha Lange
 * Mail:        Sascha.Lange@Uni-Osnabrueck.De
 *
 * Web:         http://cvtk.sourceforge.net
 *
 * Copyright (C) 2001 Sascha Lange
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 
 * 02111-1307, USA.
 *
 * ------------------------ Revision History --------------------------
 * $Log: colorsegmentation.cpp,v $
 * Revision 1.12  2001/09/03 22:59:17  slange
 * - profiling
 * ! memory hole of about 32k
 *
 * Revision 1.11  2001/09/03 12:11:17  slange
 * finished benchmark
 *
 * Revision 1.10  2001/09/01 13:28:41  slange
 * - working on monitor functions
 *
 * Revision 1.9  2001/08/10 14:11:57  slange
 * working on the next build
 *
 * Revision 1.8  2001/08/09 23:15:08  slange
 * - reorganised chaincoding.[h/cpp] (former regions.h, regionfinder.cpp)
 * - added a destructor to color segmentation
 * - added a constructor to region list
 *
 * Revision 1.7  2001/08/09 17:00:54  slange
 * - reorganized colorsegmentation and the lookuptable
 * - some other corrections
 *
 * Revision 1.6  2001/08/04 16:43:07  slange
 * - added some mehtods needed by template calibration
 * - added mouseClick event receiver
 *
 * Revision 1.5  2001/06/21 23:53:42  slange
 * - initial checkin of templatematching.[h/cpp]
 * - added method clear() to CRegionList
 *
 * Revision 1.4  2001/05/03 00:26:29  slange
 * - first completly working draft of calibration, colorsegmentation and region
 *   finding/representation.
 * - bug fixes
 *
 * Revision 1.3  2001/04/24 19:37:14  slange
 * initialise augmented states to track orange/yellow/blue on green/white/black background
 *
 * Revision 1.2  2001/04/23 18:42:52  slange
 * added copyright notice to all source files
 *
 * --------------------------------------------------------------------
 */

#include "colorsegmentation.h"

/************************** construction *****************************/

cvtk::CColorSegmentation::
CColorSegmentation(const CPixelDecisionStrategy* lut,
		   int width, int height) 
  : CSegmentationStrategy(), lut(lut), width(width), height(height)
{
  if (!lut) {
    cerr << "CColorSegmentation: Invalid lut." << endl;
    exit(1);
  }
}

cvtk::CColorSegmentation::~CColorSegmentation()
{
  delete lut;
}

/********************* interface implementation **********************/

cvtk::CRegionMap* 
cvtk::CColorSegmentation::segment(const CRGBMap* image)
{
  if (!image)
    return 0;
  else
    return segment(image, new CRegionMap(image->getWidth(),
					 image->getHeight()));
}
  
cvtk::CRegionMap* 
cvtk::CColorSegmentation::segment(const CRGBMap* image, 
				  CRegionMap* regions)
{
  if (image==0) 
    return 0;
  if (regions->getWidth() != image->getWidth() ||
      regions->getHeight() != image->getHeight())
    regions->resize(image->getWidth(), image->getHeight());
  
  RGBValue* imageBuf = image->getBuffer();
  int* regionsBuf = regions->getBuffer(); 

  for (int y=0; y < image->getHeight(); ++y)
    for (int x=0; x < image->getWidth(); ++x, ++imageBuf, ++regionsBuf)
      *regionsBuf = lut->lookup(imageBuf->r, imageBuf->g, imageBuf->b);

  return regions;
}
  

  
  
