/*
 * Package:     cvtk - Computer Vision Tool Kit
 * File:        defaultfacade.cpp
 * Description: implementation of the default facade   
 * Created:     2001-08-06
 * Author:      Sascha Lange
 * Mail:        Sascha.Lange@Uni-Osnabrueck.De
 *
 * Web:         http://cvtk.sourceforge.net
 *
 * Copyright (C) 2001 Sascha Lange
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  
 * 02111-1307, USA.
 *
 * ------------------------ Revision History --------------------------
 * $Log: defaultfacade.cpp,v $
 * Revision 1.8  2001/09/04 15:31:27  slange
 * - corrected bugs in:
 *   - templatecontroller.cpp: use QString to emit messages
 *   - monitor.cpp: used x instead of y to check against height
 * - labeled the sliders in calibratewidget
 *
 * Revision 1.7  2001/09/02 17:39:07  slange
 * cvs error!
 *
 * Revision 1.6  2001/09/01 17:40:09  slange
 * - working on monitor
 * - moved template calibration into namespace "cvtk"
 *
 * Revision 1.5  2001/09/01 13:28:41  slange
 * - working on monitor functions
 *
 * Revision 1.4  2001/08/28 13:16:48  slange
 * - traced down a bug: historylist to small!
 *
 * Revision 1.3  2001/08/10 14:11:57  slange
 * working on the next build
 *
 * Revision 1.2  2001/08/07 12:40:14  slange
 * - still refactoring
 * - initial checkin of defaultfactory.cpp, objectlist.h region.h
 *
 * Revision 1.1  2001/08/06 15:59:47  slange
 * - started refactoring of the library
 * - first commit of cvtk.h cvtkfacade.h defaultfacade.cpp
 *   defaultfacade.h defaultfactory.h
 *
 * --------------------------------------------------------------------
 */

#include "defaultfacade.h"
#include "monitors.h"

/********************* Construciont / Destruction ********************/

cvtk::CDefaultFacade::
CDefaultFacade(CSegmentationStrategy* segmentation,
	       CChainCodingStrategy* chainCoding,
	       CObjectRecognitionStrategy* objectRecognition,
	       CHistoryMatchingStrategy* historyMatching,
	       COutput* output,
	       bool copyOutput)
  : CCVTKFacade(), segmentation(segmentation), 
    chainCoding(chainCoding), objectRecognition(objectRecognition),
    historyMatching(historyMatching), output(output), timeStamp(0),
    copyOutput(copyOutput)
{
  if ((segmentation || chainCoding || objectRecognition ||
       historyMatching || output) == 0) {
    cerr << "CDefaultFacade: Null pointer as argument" << endl;
    exit(1);
  }
};

cvtk::CDefaultFacade::~CDefaultFacade()
{
  delete output;
  delete historyMatching;
  delete objectRecognition;
  delete chainCoding;
  delete segmentation;
}

/******************************* pipe ********************************/

cvtk::CObjectList* cvtk::CDefaultFacade::analyse(const CRGBMap* image)
{
  CObjectList* oList = new CObjectList();
  return analyse(image, oList);
}

cvtk::CObjectList* cvtk::CDefaultFacade::analyse(const CRGBMap* image,
					   CObjectList* oList)
{
  regionList.clear();
  objectList.clear();

#ifdef DEBUG
  cerr << "Segmentation\n";
#endif
  segmentation->segment(image, &regionMap);
#ifdef DEBUG
  cerr << "ChainCoding\n";
#endif
  chainCoding->codeRegions(&regionMap, &regionList);
#ifdef DEBUG
  cerr << "TemplateMatching\n";
#endif
  objectRecognition->findObjects(&regionList, &objectList);
#ifdef DEBUG
  cerr << "HistoryMatching\n";
#endif
  historyMatching->matchHistory(&objectList, &historyList);
#ifdef DEBUG
  cerr << "setTimeStamp\n";
#endif
  output->setTimeStamp(&historyList, timeStamp++);
#ifdef DEBUG
  cerr << "calcMovement\n";
#endif
  output->calcMovement(&historyList);
#ifdef DEBUG
  cerr << "getOutput\n";
#endif
  output->getOutput(&historyList, oList);
  if (copyOutput) {
#ifdef DEBUG
    cerr << "copyOutput\n";
#endif
    this->oList.clear();
    this->oList = *oList;
  }
#ifdef DEBUG
  cerr << "ready.\n";
#endif
  return oList;
} 

/***************************** monitor *******************************/

void cvtk::CDefaultFacade::visualizeAnalysis(CRGBMap& visual)
{
  Monitors::visualizeRegionMap(regionMap, visual);
  Monitors::visualizeRegionList(regionList, visual);
  if (copyOutput)
    Monitors::visualizeObjectList(oList, visual);
  else
    Monitors::visualizeObjectList(objectList, visual);
}





