/*
 * Package:     cvtk - Computer Vision Tool Kit
 * File:        defaultfactory.cpp
 * Description: factory, that generates a pipe of the libraries 
 *              original implementation  
 * Created:     2001-08-06
 * Author:      Sascha Lange
 * Mail:        Sascha.Lange@Uni-Osnabrueck.De
 *
 * Web:         http://cvtk.sourceforge.net
 *
 * Copyright (C) 2001 Sascha Lange
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  
 * 02111-1307, USA.
 *
 * ------------------------ Revision History --------------------------
 * $Log: defaultfactory.cpp,v $
 * Revision 1.5  2001/09/04 15:31:27  slange
 * - corrected bugs in:
 *   - templatecontroller.cpp: use QString to emit messages
 *   - monitor.cpp: used x instead of y to check against height
 * - labeled the sliders in calibratewidget
 *
 * Revision 1.4  2001/09/03 12:11:17  slange
 * finished benchmark
 *
 * Revision 1.3  2001/08/28 13:16:48  slange
 * - traced down a bug: historylist to small!
 *
 * Revision 1.2  2001/08/10 14:11:57  slange
 * working on the next build
 *
 * Revision 1.1  2001/08/07 12:40:14  slange
 * - still refactoring
 * - initial checkin of defaultfactory.cpp, objectlist.h region.h
 *
 * --------------------------------------------------------------------
 */

#include <fstream>

#include "defaultfactory.h"
#include "defaultfacade.h"

#include "colorsegmentation.h"
#include "chaincoding.h"
#include "templatematching.h"
#include "historymatching.h"
#include "output.h"
#include "searchmatch.h"

cvtk::CDefaultFactory::CDefaultFactory(string lutFile, 
				       string tmplFile,
				       int minRegionSize,
				       int historyDepth)
  : CCVTKFacadeFactory(), lutFile(lutFile), tmplFile(tmplFile),
    minRegionSize(minRegionSize), historyDepth(historyDepth)
{}

cvtk::CCVTKFacade* cvtk::CDefaultFactory::getCVTKFacade()
{
  ifstream inLut(lutFile.c_str());        // open lookuptable
  if (!inLut) {
    cerr << "Cannot open lookup table file: " << lutFile.c_str() 
	 << endl;
    exit(1);
  }
  ifstream inTemp(tmplFile.c_str());      // open object templates
  if (!inTemp) {
    cerr << "Cannot open object template file: " << tmplFile.c_str() 
	 << endl;
    exit(1);
  }
  
  // construct objects which are needed to configure the layers
  CLookUpTable<2>* lut = new CLookUpTable<2>();
  inLut >> (*lut);

  CObjectTemplateList* objTempList = new CObjectTemplateList();
  inTemp >> (*objTempList);

  CSearchStrategy* probSolv =
    new CSearchStrategy(new CPriorityQueueAdaptor(new CMatchCost()),
			new CMatchGoal(),
			new CMatchExpand());

  CSegmentationStrategy* ss = new CColorSegmentation(lut);
  CChainCodingStrategy* cs = 
    new CDefaultChainCoding(new CBorderTracing4(),
			    lut->getNumberOfClasses(), 
			    minRegionSize);
  CObjectRecognitionStrategy* os =
    new CTemplateMatching(objTempList);
  CHistoryMatchingStrategy* hs = new CHistoryMatching(historyDepth,
						      probSolv);
  COutput* o = new CDefaultOutput();

  // construct and configure the layers and hide them behind the facade
  CCVTKFacade* facade = 
    new CDefaultFacade(ss, cs, os, hs, o);

  return facade;			     
}

