//
// File:        ft_interface.c
// Description: Implementation of the fischertechnik interface
// Created:     2001-02-15
// Author:      Sascha Lange
// Mail:        salange@uos.de
//
// Copyright Sascha Lange 2001
//
//------------------------ Revision History ---------------------------
// $Log: ft_interface.c,v $
// Revision 1.1  2001/08/30 17:25:32  slange
// - added ft_interface to drive a fischertechnik robot
// - working on demo application
//
// Revision 1.1  2001/03/05 12:58:01  slange
// Initial revision
//
//---------------------------------------------------------------------


#include <stdio.h>
#include <string.h>
#include <fcntl.h>
#include <pthread.h>
#include <stdlib.h>
#include <unistd.h>
#include <termios.h>
#include "ft_interface.h"

#define FT_COMMAND_DIGITAL '\xc1'
#define FT_COMMAND_ANALOG

/*
 * motors: 00|00|00|01       (= 1)      01 > links rum
 *         ^        ^                   10 > rechts rum
 *         motor4   motor1              00=11 = stop
 *
 * digis:  0|1|0|0|0|0|0|1   (=129)     Switch 1 und 7 werden gedrueckt
 */

typedef struct {
  pthread_mutex_t mutex;
  pthread_t thread_id;
  int port;
  char* device;
  unsigned char digis;
  unsigned char motors;
  motor_speed motorSpeed[4];
  struct termios oldtio;
} * _ftHandle;

pthread_mutex_t mutex = PTHREAD_MUTEX_INITIALIZER;

void* runInterface(void* arg);

ftHandle openInterface(char* device)
{
  int fd;
  _ftHandle ft;
  int i;
  struct termios newtio;
 
  if ((fd=open(device, O_RDWR | O_NOCTTY | O_NONBLOCK))<0) {
                              // Schnittstelle oeffnen
    perror("openInterface");
    return NULL;
  }

  ft = (_ftHandle) malloc(sizeof (_ftHandle));
  ft->motors=ft->digis=0;     // sicherheitshalber alle Werte loeschen
  for (i=0; i<4; i++) ft->motorSpeed[i]=SLOW;  
                              // Motoren zuerst auf langsam
  ft->port = fd;
  ft->device = device;

  tcgetattr(fd,&ft->oldtio);
  bzero(&newtio, sizeof(newtio));
  newtio=ft->oldtio;
  newtio.c_lflag = 0;

  tcflush(fd, TCIFLUSH);
  tcsetattr(fd,TCSANOW, &newtio);

  // Interface-Thread starten
  if (pthread_create(&ft->thread_id, NULL,  runInterface, 
		     (void*)ft) 
      !=0) {
    perror("thread_create");
    exit(1);
  }

  return (ftHandle) ft;
}

void closeInterface(ftHandle ft)
{
  unsigned char send[2];
  _ftHandle ftStr = (_ftHandle) ft;

  pthread_cancel(ftStr->thread_id);// Interfacethread beenden
  pthread_mutex_lock(&mutex);      // Interface locken

  send[0]=FT_COMMAND_DIGITAL;      // Motoren ausstellen
  send[1]=0;
  usleep(10000);                   // dem Interface etwas Zeit lassen
  write(ftStr->port, &send, 2);

  pthread_join(ftStr->thread_id, NULL);
                                   // auf den Thread warten (sicherer)  

  pthread_mutex_unlock(&mutex);    // Interface wieder freigeben

  tcsetattr(ftStr->port,TCSANOW, &ftStr->oldtio);
  close(ftStr->port);
  free(ftStr);
}

interface_status getInterfaceStatus(ftHandle ft)
{
  interface_status status;
  return status;
}

/*
 * bitoperationen gehen bestimmt noch schoener, aber das reicht ja 
 * wohl...
 */
void setMotor(ftHandle ft, int num, motor_direction dir, 
	      motor_speed speed)
{
  _ftHandle ftStr;

  pthread_mutex_lock(&mutex);

  ftStr = (_ftHandle) ft;
  ftStr->motors &= ~(3 << ((num-1)*2));            // maskieren  
#ifdef DEBUG
  printf("setMotor: motors after mask: %d\n", ftStr->motors);
#endif
  ftStr->motors |= dir << ((num-1)*2);             // und setzen
#ifdef DEBUG
  printf("setMotor: motors after set: %d\n", ftStr->motors);
#endif
  if (speed!=SAME)
    ftStr->motorSpeed[num-1]=speed;

  pthread_mutex_unlock(&mutex);
}

int isDigiOn(ftHandle ft, int num)
{
  int ret;
  pthread_mutex_lock(&mutex);
  ret = ((_ftHandle)ft)->digis & (1 << (num-1));
  pthread_mutex_unlock(&mutex);
  return ret;
}

void* runInterface(void * arg)
{
  _ftHandle ft;                        // Interface to run
  int speedCounter, i;
  unsigned char bitmask;
  unsigned char send[2];
  struct sched_param schedParam;
  unsigned char recieve;

  ft = (_ftHandle)arg;
  speedCounter = 1;
  pthread_setcancelstate(PTHREAD_CANCEL_ENABLE, NULL); 
                             // Thread beendbar machen
  pthread_setcanceltype(PTHREAD_CANCEL_DEFERRED, NULL);

  schedParam.sched_priority = 10; 

  /*  if (pthread_setschedparam(ft->thread_id, SCHED_RR, &schedParam) 
          != 0) {
    perror("thread_setschedparam");
    exit(1);
    } */                             // geht nur als SuperUser!

  while(1) {
    pthread_testcancel();            // noch kein Endsignal?

    send[0] = FT_COMMAND_DIGITAL;    // nur Digitale Daten auslesen
    send[1] = 0;
    bitmask = 3;                     // maskiert: 00000011

    pthread_mutex_lock(&mutex);      // Interface abschliessen

    for (i=0; i<4; i++, bitmask<<=2)// Geschwindigkeit der vier Motoren
      if((speedCounter % ft->motorSpeed[i]) == 0) 
                               // durch lnge des Impulses Modellieren;
	send[1] |= (ft->motors & bitmask); // Motor maskieren und ORen

#ifdef DEBUG
    printf("send:%d command:%d\n", send[0], send[1]);
#endif
    if (write(ft->port, send, 2)<0)  // die 2 Bytes senden
      perror("write");
    if (read(ft->port, &recieve, 1)!=1)
      perror("read");
    else
      ft->digis = recieve;
#ifdef DEBUG_RECEIVE 
    printf("Received: %d\n",recieve );;
#endif
    

    pthread_mutex_unlock(&mutex);    // Interface wieder freigeben
    usleep(10000); 

    speedCounter = (speedCounter%16)+1;
    }
  return NULL;
}

