#ifndef _histogram_h_
#define _histogram_h_

/*
 * Package:     cvtk - Computer Vision Tool Kit
 * File:        histogram.h
 * Description: One- and two-dimensional histograms.
 * Created:     2001-04-26
 * Author:      Sascha Lange
 * Mail:        Sascha.Lange@Uni-Osnabrueck.De
 *
 * Web:         http://cvtk.sourceforge.net
 *
 * Copyright (C) 2001 Sascha Lange
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 
 * 02111-1307, USA.
 *
 * ------------------------- Revision History -------------------------
 * $Log: histogram.h,v $
 * Revision 1.5  2001/08/30 13:24:30  slange
 * - added sliders to effect the specificity of the kmsegmentation
 *
 * Revision 1.4  2001/05/03 01:19:18  slange
 * - Initial revision of CRGBHistogram
 * - added template class CHistogram3Dim
 *
 * Revision 1.3  2001/04/27 23:15:42  slange
 * - added an UV - Histogram (histogram in an approximation of the HUV space)
 * - changed the Calibrate::CController
 *   - to work with histograms
 *   - to keep a vector of examples in memory
 *   - output file dropped
 *   - to tread the background as a (special) object
 * - fixes
 *
 * Revision 1.2  2001/04/27 20:51:16  slange
 * - finished implementation of CHistogram2Dim
 * - bug fixes
 *
 * Revision 1.1  2001/04/27 16:57:07  slange
 * - added 1- and 2-dimensional histograms
 * - changed typedef CRGBMap to a class definition
 * - added histogram functions to CRGBMap
 *
 * --------------------------------------------------------------------
 */

#include <iostream>
#include <vector>
#include "vektor.h"

namespace cvtk {

  template<int maxValue=256>
  class CHistogram {                     // 1-dimensional histogram
  public:
    CHistogram();

    void add(int value);
    int magnitude(int value);
    
    int& operator[] (int pos);
    const int& operator[] (int pos) const;

  protected:
    int histogram[maxValue];
  };

  template<int maxValue=256>
  class CHistogram2Dim {                // 2-dimensional histogram
  public:
    enum ListMod { LIST_PROPORTIONAL=1 };

    CHistogram2Dim();

    void add(int x, int y);
    int magnitude(int x, int y);
    
    int* operator[] (int pos);
    const int* operator[] (int pos) const;

    vector<Vektor2Dim> getAsList(int threshold = 0, int mod = 0);
    
  protected:
    int histogram[maxValue][maxValue];
  };

  template<int maxValue=256>
  class CHistogram3Dim {                // 3-dimensional histogram
  public:
    enum ListMod { LIST_PROPORTIONAL=1 };

    CHistogram3Dim();
    
    void add(int x, int y, int z);
    int magnitude(int x, int y, int z);
  
    int** operator[] (int pos);
    const int** operator[] (int pos) const;
    
    vector<Vektor3Dim> getAsList(int threshold = 0, int mod = 0);

  protected:
    int histogram[maxValue][maxValue][maxValue];
  };

}

//////////////////////// Inlined Implementation ///////////////////////

// Stream operation declarations
template<int maxValue>
ostream& operator<< (ostream&, const cvtk::CHistogram3Dim<maxValue>&);
template<int maxValue>
ostream& operator<< (ostream&, const cvtk::CHistogram2Dim<maxValue>&);
template<int maxValue>
ostream& operator<< (ostream&, const cvtk::CHistogram<maxValue>&);

// Inline functions
template<int maxValue>
inline void cvtk::CHistogram3Dim<maxValue>::add(int x, int y, int z)
{ (*this)[x][y][z]++; }
template<int maxValue> inline int 
cvtk::CHistogram3Dim<maxValue>::magnitude(int x, int y, int z)
{ return (*this)[x][y][z]; }
template<int maxValue>
inline int** cvtk::CHistogram3Dim<maxValue>::operator[] (int pos)
{ return histogram[pos]; }
template<int maxValue> inline const int** 
cvtk::CHistogram3Dim<maxValue>::operator[] (int pos) const
{ return histogram[pos]; }

template<int maxValue>
inline void cvtk::CHistogram2Dim<maxValue>::add(int x, int y)
{ (*this)[x][y]++; }
template<int maxValue>
inline int cvtk::CHistogram2Dim<maxValue>::magnitude(int x, int y)
{ return (*this)[x][y]; }
template<int maxValue>
inline int* cvtk::CHistogram2Dim<maxValue>::operator[] (int pos)
{ return histogram[pos]; }
template<int maxValue> inline const int* 
cvtk::CHistogram2Dim<maxValue>::operator[] (int pos) const
{ return histogram[pos]; }
  
template<int maxValue>
inline void cvtk::CHistogram<maxValue>::add(int value)
{ (*this)[value]++; }
template<int maxValue>
inline int cvtk::CHistogram<maxValue>::magnitude(int value)
{ return (*this)[value]; }
template<int maxValue>
inline int& cvtk::CHistogram<maxValue>::operator[] (int pos)
{ return histogram[pos]; }
template<int maxValue> inline const int& 
cvtk::CHistogram<maxValue>::operator[] (int pos) const
{ return histogram[pos]; }

#ifdef EXPORT_NOT_IMPLEMENTED
#include "histogram.cpp"
#endif  

#endif
