/*
 * Package:     cvtk - Computer Vision Tool Kit
 * File:        historymatching.cpp
 * Description: History matching algorithm 
 * Created:     2001-08-10
 * Author:      Sascha Lange
 * Mail:        Sascha.Lange@Uni-Osnabrueck.De
 *
 * Web:         http://cvtk.sourceforge.net
 *
 * Copyright (C) 2001 Sascha Lange
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  
 * 02111-1307, USA.
 *
 * ------------------------ Revision History --------------------------
 * $Log: historymatching.cpp,v $
 * Revision 1.5  2001/10/30 11:25:29  slange
 * added a linear movement model to historymatchin.[h/cpp] to
 * model position of undetected objects
 *
 * Revision 1.4  2001/09/02 17:39:07  slange
 * cvs error!
 *
 * Revision 1.3  2001/08/28 13:16:48  slange
 * - traced down a bug: historylist to small!
 *
 * Revision 1.2  2001/08/10 14:11:57  slange
 * working on the next build
 *
 * Revision 1.1  2001/08/10 00:57:05  slange
 * - moved search into namespace cvtk
 * - restructured historymatching and objecthistory
 *
 * --------------------------------------------------------------------
 */

#include "historymatching.h"
#include "searchmatch.h"

/*************************** Linear model ****************************/

cvtk::CObject 
cvtk::CLinearMovementModel::calcNextState(const CObjectHistory& hist)
{
  if (hist.history.size()<1) {
    cerr << "CLinearModel::calcNextState: History is empty!\n"
	 << endl;
    exit(1);
  }

  CObject nState = *hist.history.begin();
  // neue Position=alte Position + (Geschwindigkeit, Richtung) polar

  return nState;
}

/*************************** construction ****************************/

cvtk::CHistoryMatching::
CHistoryMatching(unsigned int sizeOfHistory,
		 CProblemSolvingStrategy* matchStrat,
		 int deleteAfterNTicks)
  : sizeOfHistory(sizeOfHistory), matchStrat(matchStrat),
    deleteAfterNTicks(deleteAfterNTicks)
{
  if (sizeOfHistory < 3) {
    cerr << "CHistoryMatching: Argument sizeOfHistory to small" 
	 << endl;
    exit(1);
  }
}

unsigned int cvtk::CHistoryMatching::uid = 0;

/********************** interface implementation *********************/

cvtk::CObjectHistoryList* 
cvtk::CHistoryMatching::matchHistory(CObjectList* objs,
				     CObjectHistoryList* hist)
{
  if (!objs)
    return 0;

  if (objs->size() != hist->size()) 
    hist->resize(objs->size());

  CObjectList::iterator oIt = objs->begin();
  CObjectHistoryList::iterator hIt = hist->begin();
  for (int templID=0; oIt != objs->end(); oIt++, hIt++, templID++) {
#ifdef DEBUG
    cerr << "Matching: Template: " << templID << endl;
#endif
    if (oIt->size() == 0) {          // no objects -> delete hists
#ifdef DEBUG
      cerr << "oIt->size()==0" << endl;
#endif
      CObjectHistoryList::innerIterator i = hIt->begin();
      while(i != hIt->end()) {
	if (i->nTicksNotSeen == deleteAfterNTicks) {
	  CObjectHistoryList::innerIterator d = i++; // this
	  hIt->erase(d);            // is list specific! doesn't work
	}                           // with vectors-> erase not def.
	else {
	  i->nTicksNotSeen++;
	  // model the new state
	  i++;
	}
      }
    }
    else if (hIt->size() == 0) {     // all detected objects are new
#ifdef DEBUG
      cerr << "hIt->size()==0" << endl;
#endif
      for (CObjectList::innerIterator obj = oIt->begin();
	   obj < oIt->end();
	   obj++) {
	CObjectHistory newObj(obj->objectClass, ++uid, sizeOfHistory);
	newObj.insert(*obj);
	hist->push_back(templID, newObj);
      }
    }
    else if (oIt->size() > 1 || hIt->size() > 1) {
#ifdef DEBUG
      cerr << "oIt->size() > 1 || hIt->size() > 1" << endl;
#endif
      // matching needed
      // idee: Abstand minimieren, dann entweder
      // 1. fuer uebrig geblibene objekte neue history anlegen oder
      // 2. uebrig gebliebene histories loeschen
      if (matchStrat) { 
	CMatchState* state = new CMatchState(oIt, hIt);
	state = dynamic_cast<CMatchState*>(matchStrat->solve(state));
	for (unsigned int i=0; i < state->matches.size(); i++) {
	  state->matches[i].hist->insert(*(state->matches[i].obj));
	  state->matches[i].hist->nTicksNotSeen=0;
	}
	for (unsigned int i=0; i < state->objUnMatched.size(); i++) {
	  CObjectHistory newObj(templID, ++uid, sizeOfHistory);
       	  newObj.insert(*(state->objUnMatched[i])); 
	  hist->push_back(templID, newObj);
	}
	for (unsigned int i=0; i < state->histUnMatched.size(); i++) {
	  if (state->histUnMatched[i]->nTicksNotSeen == 
	      deleteAfterNTicks) 
	    hist->erase(templID, *(state->histUnMatched[i])); // list
	  else {                             // specific! no vectors!
	    state->histUnMatched[i]->nTicksNotSeen++;
	    // model the state;
	  }
	}
	delete state;
      }
    }
    else {          // oIt->size() == hIt->size() == 1
#ifdef DEBUG
      cerr << "oIt->size()==hIt->size()==1" << endl;
#endif
      hIt->front().insert(oIt->front());
      hIt->front().nTicksNotSeen=0;
      // insert single detected object in single history 
    }
  }
  return hist;
}















