/*
 * Package:     cvtk - Computer Vision Tool Kit
 * File:        imagewidget.cpp
 * Description: Implementation of the ImageWidget.
 * Created:     2001-03-22
 * Author:      Sascha Lange
 * Mail:        Sascha.Lange@Uni-Osnabrueck.De
 *
 * Web:         http://www.cvtk.sourceforge.net
 *
 * Copyright (C) 2001 Sascha Lange
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 
 * 02111-1307, USA.
 *
 * ------------------------- Revision History -------------------------
 * $Log: imagewidget.cpp,v $
 * Revision 1.4  2001/08/10 14:11:57  slange
 * working on the next build
 *
 * Revision 1.3  2001/08/04 16:58:30  slange
 * added mousePressed signal to imagewidget
 *
 * Revision 1.2  2001/04/23 18:42:52  slange
 * added copyright notice to all source files
 *
 * --------------------------------------------------------------------
 */

#include "imagewidget.h"
#include <qpainter.h>
#include <qimage.h>

cvtk::CImageWidget::CImageWidget(const CRGBMap& image, 
				     QWidget* parent,
				     WFlags f)
  : QWidget(parent, "ImageWidget", f), w(-1), h(-1), buf(0)
{
  this->setMinimumSize(60, 40);
  this->setBackgroundMode(NoBackground);
  setImage(image);
}
cvtk::CImageWidget::CImageWidget(int width, int height, 
				     QWidget* parent, 
				     WFlags f)
  : QWidget(parent, "ImageWidget", f), w(-1), h(-1), buf(0)
{
  this->setMinimumSize(60, 40);
  this->setBackgroundMode(NoBackground);
  setImage(CRGBMap(1, 1));
}
cvtk::CImageWidget::~CImageWidget()
{
  free(buf);
}
void cvtk::CImageWidget::mousePressEvent(QMouseEvent* ev)
{
  emit mousePressed(ev);
}
void cvtk::CImageWidget::paintEvent(QPaintEvent* ev)
{
  CMutex::Key paintKey=paintMutex.hold();
  QPainter p(this);
  CMutex::Key dataKey=dataMutex.hold();
  p.drawPixmap(0,0,pixmap);
  dataMutex.release(dataKey);
  paintMutex.release(paintKey);
}

void cvtk::CImageWidget::setImage(const CRGBMap& image)
{
  bool resized = false;

  CMutex::Key dataKey = dataMutex.hold();           // DATA LOCKED
  if (image.getWidth() != w || image.getHeight() != h) {
    w=image.getWidth();
    h=image.getHeight();

    buf = static_cast<unsigned int*>(realloc(buf, w*h*sizeof(int)));
    this->image = 
      QImage(reinterpret_cast<uchar*>(buf), w, h, 32, NULL, 0, QImage::LittleEndian);
    resized = true;
  }

  image.lock();                                   // IMAGE LOCKED
   /* copy the 24bpp CRGBMap-buffer to a 32bpp buffer needed by the QImage */
  RGBValue* pos = image.getBuffer();
  int i, size=w*h;
  for (i=0; i<size; i++, pos++)                   // 24bpp to 32bpp
    buf[i] = pos->r << 16 | pos->g << 8 | pos->b;  
  image.unlock();                                 // IMAGE UNLOCKED
  
  pixmap.convertFromImage(this->image);

  dataMutex.release(dataKey);                     // DATA UNLOCKED

  if (resized) resize(w, h);
  else
    repaint();                                    // unfortunately update's to slow!
  // attention: using repaint will cause the calling (this) thread to do the painting!
  // that might be a problem for synchronization.
}

#include "imagewidget.moc"









