/*
 * Package:     cvtk - Computer Vision Tool Kit
 * File:        kmsegmentation.cpp
 * Description: Implementation of the k-means segmentation.
 * Created:     2001-04-19
 * Author:      Sascha Lange
 * Mail:        Sascha.Lange@Uni-Osnabrueck.De
 *
 * Web:         http://cvtk.sourceforge.net
 *
 * Copyright (C) 2001 Sascha Lange
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 *
 * ------------------------- Revision History -------------------------
 * $Log: kmsegmentation.cpp,v $
 * Revision 1.12  2001/09/02 17:39:07  slange
 * cvs error!
 *
 * Revision 1.11  2001/08/30 17:25:32  slange
 * - added ft_interface to drive a fischertechnik robot
 * - working on demo application
 *
 * Revision 1.10  2001/08/30 13:24:30  slange
 * - added sliders to effect the specificity of the kmsegmentation
 *
 * Revision 1.9  2001/05/03 13:04:27  slange
 * - traced down bug in CController::addFrame()
 *
 * Revision 1.8  2001/05/03 00:26:29  slange
 * - first completly working draft of calibration, colorsegmentation and region
 *   finding/representation.
 * - bug fixes
 *
 * Revision 1.7  2001/04/29 00:55:49  slange
 * - finished the first complete draft of the calibration tool.
 *   - it now generates a lookup-table for values 1..64 (>> 2)
 *   - writes the entire table to stdout.
 * - fixes in k-means segmentation
 * - added button to generate lookup table in the calibrator-widget
 *
 * Revision 1.6  2001/04/28 21:11:02  slange
 * - first complete implementation of kmsegmentation
 *
 * Revision 1.5  2001/04/28 18:05:45  slange
 * - removed patterngenerato.[h/cpp]
 * - added abstract interface CReinforcedLearning
 * - added CKMSegmentation
 *   - class definition and part of implementation
 * - added "train" button to calibratorwidget
 * - changed calibratecontroller to work with the new classes
 *
 * Revision 1.4  2001/04/27 16:57:07  slange
 * - added 1- and 2-dimensional histograms
 * - changed typedef CRGBMap to a class definition
 * - added histogram functions to CRGBMap
 *
 * Revision 1.3  2001/04/24 19:37:14  slange
 * initialise augmented states to track orange/yellow/blue on green/white/black background
 *
 * Revision 1.2  2001/04/23 18:42:52  slange
 * added copyright notice to all source files
 *
 * --------------------------------------------------------------------
 */

#ifndef EXPORT_NOT_IMPLEMENTED
#include "kmsegmentation.h"
#endif
#include <assert.h>
#include <cstdlib>

#define PERCENTAGE .95

#ifndef EXPORT_NOT_IMPLEMENTED
export
#endif  
template<int DIMENSION>
int CKMSegmentation<DIMENSION>::maxButNot(vector<int> v, 
					  int size, int notThis)
{
  if (size<1) return -1;
  int actMax=-1;
  int pos=-1;
  for (int i=0; i < size; i++)
    if (i!=notThis && v[i] > actMax) {
      actMax=v[i];
      pos=i;
    }
  return pos;
}

#ifndef EXPORT_NOT_IMPLEMENTED
export
#endif  
template<int DIMENSION>
CKMSegmentation<DIMENSION>::CKMSegmentation(int nCentersPerClass,
					    int maxIteration,
					    double maxError, 
					    double minChange) 
  : CReinforcedLearning<DIMENSION>(), 
    nCentersPerClass(nCentersPerClass), nClasses(0), 
    maxIteration(maxIteration), maxError(maxError), 
    minChange(minChange)
{
  for (int i=0; i < DIMENSION; i++) {   // initialise orthonormal base
    vektor_type v; 
    v[i]=1;
    baseVektors.push_back(v);
  }
};

#ifndef EXPORT_NOT_IMPLEMENTED
export
#endif
template<int DIMENSION>
int CKMSegmentation<DIMENSION>::classify(const vektor_type& x) const
{
  int cl = nearestNeighbor(x);
  return (cl >= 0) ? m[cl].c : cl;      // throw exception on cl==-1?!
}

#ifndef EXPORT_NOT_IMPLEMENTED
export
#endif
template<int DIMENSION>
void CKMSegmentation<DIMENSION>::train(const vector<vektor_type >& x,
				       const vector<int>& xC,
				       const vector<int>& spec)
{
  if (x.size() != xC.size() || x.size() < 1)
    return;                             // throw exception !!

  int correct, wrong, iteration=0;
  double change;
  double error;

  init(x,xC,spec);                      // (re)initialise the centers

  if (nClasses != static_cast<int>(spec.size())) {
    cerr << "kmsegmentation: nClasses and size of spec doesn't match"
	 << endl;
    return;                             // throw exception !!
  }

  do {
    correct=wrong=0;
    change=error=0;
#ifdef DEBUG
    cerr << "\nCKMSegmentation:train:Iteration: " 
	 << iteration+1 << endl;
#endif

    /* update for all examples x */

    int nearest;
    for (unsigned int i=0; i < x.size(); i++) {
                                      // update centers for all x
      nearest = nearestNeighbor(x[i]);// find nearest center
      if (m[nearest].c == xC[i]) {    // center has the same class as x
	m[nearest].countUpdate++;     // add x to calculate new mean   
	m[nearest].mUpdate+=x[i];    
      }
      m[nearest].hC[xC[i]]++;         // count this hit
      m[nearest].h++;
    }

    /* calculate new mean for ervery m and reset augmented state */

    unsigned int actSizeM = m.size();
    for (unsigned int i=0; i < actSizeM; i++) {
                                // calculate mean for every m
      if (m[i].countUpdate > 0) {     // update this m
	m[i].mUpdate*=(1./m[i].countUpdate);
	change+=(m[i].m-m[i].mUpdate).distance();
	m[i].m=m[i].mUpdate;
      }
#ifdef DEBUG
      cerr << "m" << i << ": " << m[i].m << " class: " << m[i].c 
	   << " total hits: " << m[i].h << " class hits: " 
	   << m[i].countUpdate << endl;
#endif
      correct+= m[i].countUpdate; // count correct/wrong classification
      wrong+= m[i].h - m[i].countUpdate;

      /* check m[i] for "part" or "delete" */

      if(checkAugmentedState(i, iteration) == STATE_DELETED) {
	i--; actSizeM--;
	continue;                       // if m deleted, don't reset
      }  

      /* reset m[i] */

      m[i].countUpdate=0;               // reset
      m[i].h = 0;                       // reset
      m[i].hC = vector<int>(nClasses);  // reset
    }
    error = ((double)wrong)/((double)wrong+correct);
#ifdef DEBUG 
    cerr << "\nChange: " << change << endl;
    cerr << "Fehler: " << error << endl;
#endif
  } while (++iteration < maxIteration && error > maxError && 
	   change > minChange);
}

#ifndef EXPORT_NOT_IMPLEMENTED
export
#endif
template<int DIMENSION>
CKMSegmentation<DIMENSION>::StateReturn 
CKMSegmentation<DIMENSION>::checkAugmentedState(int n, int iteration)
{
  if (iteration < (4*maxIteration)/6 && m[n].h > minHits[m[n].c] &&
      static_cast<double>(m[n].countUpdate)/
      static_cast<double>(m[n].h) < PERCENTAGE) {
        // too unprecise -> part
    int r = min(int((double(rand())/RAND_MAX) * DIMENSION), 
		DIMENSION-1);
    vektor_type e = baseVektors[r] * 
      ( (min(int((double(rand())/RAND_MAX)*2),1))?-1:1 );
    int c = maxButNot(m[n].hC, nClasses, m[n].c);
    CAugmentedState newM(nClasses, c, m[n].m+e);
    m.push_back(newM);
#ifdef DEBUG
    cerr << " inserted: " << newM.m << " Class: " << newM.c << endl;
#endif
  }
  else if (iteration < (4*maxIteration)/6 && m[n].h > maxHits[m[n].c]){
       // too big -> part
    int r = min(int((double(rand())/RAND_MAX) * DIMENSION), 
		DIMENSION-1);
    vektor_type e = baseVektors[r] * 
      ( (min(int((double(rand())/RAND_MAX)*2),1))?-1:1 ); // e {-1, 1}
    CAugmentedState newM(nClasses, m[n].c, m[n].m+e);
    m.push_back(newM);
#ifdef DEBUG
    cerr << " inserted: " << newM.m << " Class: " << newM.c << endl;
#endif
  } 

  if (m[n].countUpdate < minHits[m[n].c]) {    // too smal -> delete
#ifdef DEBUG
    cerr << " deleted" << endl;
#endif
    m.erase(m.begin()+n);
    return STATE_DELETED;
  }

  return STATE_UNCHANGED;
}

#ifndef EXPORT_NOT_IMPLEMENTED
export
#endif
template<int DIMENSION>
CKMSegmentation<DIMENSION>
::CAugmentedState::CAugmentedState(int nClasses,
				   int c,
				   const vektor_type& m) 
  : m(m), c(c), mUpdate(vektor_type()), hC(vector<int>(nClasses)), 
    h(0), countUpdate(0)
{};

#ifndef EXPORT_NOT_IMPLEMENTED
export
#endif
template<int DIMENSION>
ostream& CKMSegmentation<DIMENSION>::printOnStream(ostream& out) const
{
  out << "#CKMSegmentation" << endl;
  return out;
}

#ifndef EXPORT_NOT_IMPLEMENTED
export
#endif
template<int DIMENSION>
istream& CKMSegmentation<DIMENSION>::readOfStream(istream& in)
{
  return in;
}

#ifndef EXPORT_NOT_IMPLEMENTED
export
#endif
template<int DIMENSION>
void CKMSegmentation<DIMENSION>::init(const vector<vektor_type>& x,
				      const vector<int>& xC,
				      const vector<int>& spec)
{
  // spec gives a number of centers per class. now calculate the
  // average number of hits for each center. set minHits to that
  // number to set an upper threshold for the number of centers.
  // remember the highest class-number and attribute value.
  int nCl=0;
  double maxXValue=0;
  
  for (unsigned int i=0; i < xC.size(); i++) {
    if (xC[i] >= nCl) nCl = xC[i];      // find highest class-number
    if (x[i][0] > maxXValue) maxXValue=x[i][0]; // " attribute value
  }
  nCl++;

  vector<int>cCount(nCl,0);             // count hits per class
  for (unsigned int i=0; i < xC.size(); i++) 
    cCount[xC[i]]++;

  minHits.resize(nCl);                  // resize min and max Hits
  maxHits.resize(nCl);

  for (int i=0; i < nCl; i++) {         // calculate min and max
    minHits[i] = cCount[i]/spec[i];     // hits for every class
    maxHits[i] = (cCount[i]/spec[i])*2; // according to the given
#ifdef DEBUG
    cerr << "Class: " << i << " min: " << minHits[i]
	 << " max: " << maxHits[i] << endl;
#endif
  }                                     // specificity

  if (nCl > nClasses) {                 // reinitialisation necessary
    // first, adjust size of vector hC for existing centers
    for (unsigned int i=0; i < m.size(); i++) 
      m[i].hC = vector<int>(nCl);       // adjust size

    for (int i=nClasses; i < nCl; i++)  // add centers as necessary
      for (int n=0; n < nCentersPerClass; n++)
	{
	  vektor_type randomPos;
	  for (int d=0; d < DIMENSION; d++)
	    randomPos[d] = (static_cast<double>(rand())/RAND_MAX)
	                   *maxXValue;
	  CAugmentedState newM(nCl, i, randomPos);
	  m.push_back(newM);
	}
    nClasses = nCl;
  }
  assert(nClasses==nCl);
}














