#ifndef _kmsegmentation_h_
#define _kmsegmentation_h_

/*
 * Package:     cvtk - Computer Vision Tool Kit
 * File:        kmsegmentation.h
 * Description: Definition of an abstract interface to reinforced 
 *              learning algorithms and definition of a concrete 
 *              k-means segmentation algorithm
 * Created:     2001-04-28
 * Author:      Sascha Lange
 * Mail:        Sascha.Lange@Uni-Osnabrueck.De
 *
 * Web:         http://cvtk.sourceforge.net
 *
 * Copyright (C) 2001 Sascha Lange
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 *
 * -------------------------- Revision History ------------------------
 * $Log: kmsegmentation.h,v $
 * Revision 1.5  2001/08/30 13:24:30  slange
 * - added sliders to effect the specificity of the kmsegmentation
 *
 * Revision 1.4  2001/06/20 12:27:47  slange
 * Initial check in of fileis.h and fileis.cpp
 *
 * Revision 1.3  2001/04/29 00:55:49  slange
 * - finished the first complete draft of the calibration tool.
 *   - it now generates a lookup-table for values 1..64 (>> 2)
 *   - writes the entire table to stdout.
 * - fixes in k-means segmentation
 * - added button to generate lookup table in the calibrator-widget
 *
 * Revision 1.2  2001/04/28 21:11:02  slange
 * - first complete implementation of kmsegmentation
 *
 * Revision 1.1  2001/04/28 18:05:45  slange
 * - removed patterngenerato.[h/cpp]
 * - added abstract interface CReinforcedLearning
 * - added CKMSegmentation
 *   - class definition and part of implementation
 * - added "train" button to calibratorwidget
 * - changed calibratecontroller to work with the new classes
 *
 * --------------------------------------------------------------------
 */

#include <vector>
#include <iostream>
#include "vektor.h"

/////////////// abstract interface CReinforcedLearning ////////////////
template<int DIMENSION>
class CReinforcedLearning
{
public:
  virtual ~CReinforcedLearning() {};
  virtual int classify(const Vektor<double, DIMENSION>&) const=0;
                     //get classification
  virtual void train(const vector<Vektor<double, DIMENSION> >&,// x 
		     const vector<int>&,         // classes
		     const vector<int>&)=0;      // specificity

  friend ostream& operator<< <DIMENSION> (ostream&, 
					  const CReinforcedLearning&);
  friend istream& operator>> <DIMENSION> (istream&, 
					  CReinforcedLearning&);

protected:
  virtual ostream& printOnStream(ostream&) const=0;
  virtual istream& readOfStream(istream&)=0;
};


///////////////////////// k-means segmentation ////////////////////////

template<int DIMENSION>
class CKMSegmentation : public CReinforcedLearning<DIMENSION>
{
public:
  typedef Vektor<double, DIMENSION> vektor_type;
  CKMSegmentation(int nCentersPerClass = 1,// n centers per class
		  int maxIteration = 40,   // max number of iterations
		  double maxError = .0,    // stop on error < maxError
		  double minChange = .0);  // stop on change <minChange

  int classify(const vektor_type& x) const;
  void train(const vector<vektor_type>& x,
	     const vector<int>& xC,
	     const vector<int>& spec);

protected:
  ostream& printOnStream(ostream&) const;
  istream& readOfStream(istream&);

private: 
  int nCentersPerClass;
  int nClasses;
  class CAugmentedState {
  public:
    CAugmentedState(int nClasses=0, int c=0, 
		    const vektor_type& m=vektor_type());
    // normal center m (as defined by k-means algorithm with classes)
    vektor_type m;                     // coordinates of center
    int c;                             // class of this center
    // augmented state information of center m
    vektor_type mUpdate;               // helps to calculate new mean
    vector<int> hC;                    // hits by class of x
    int h;
    int countUpdate;
  };
  vector<CAugmentedState> m;
  vector<int> maxHits;                 // max x per center
  vector<int> minHits;                 // min hits per center
  
  int nearestNeighbor(const vektor_type& x) const;
  void init(const vector<vektor_type>& x, const vector<int>& xC,
	    const vector<int>& spec);
  enum StateReturn { STATE_UNCHANGED=0, STATE_DELETED=1, 
		     STATE_PARTED=2 };  
  enum StateReturn checkAugmentedState(int n, int iteration);
  
  int maxIteration;
  double maxError;
  double minChange;
  vector<vektor_type> baseVektors;
  vector<int> maxCentersPerClass;
  static int maxButNot(vector<int> v, int size, int notThis);
};

////////////////////////// inlined functions //////////////////////////

template<int DIMENSION>
inline int 
CKMSegmentation<DIMENSION>::nearestNeighbor(const vektor_type& x) const
{
  if (m.empty()) return -1;
  int nearest = 0;
  double nearestDistance = (m[0].m-x).distance(); 
  double distance;
  for (unsigned int i=0; i < m.size(); i++) {
    distance = (m[i].m-x).distance();
    if (distance < nearestDistance) {
      nearestDistance = distance;
      nearest = i;
    }
  }
  return nearest;
};

/////////////////////////// Stream operations //////////////////////////

template<int DIMENSION>
inline ostream& operator<< (ostream& out, const CReinforcedLearning<DIMENSION>& l)
{ return l.printOnStream(out); }
template<int DIMENSION>
inline istream& operator>> (istream& in, CReinforcedLearning<DIMENSION>& l)
{ return l.readOfStream(in); }

#ifdef EXPORT_NOT_IMPLEMENTED
#include "kmsegmentation.cpp"
#endif

#endif











