#ifndef _lookuptable_h_
#define _lookuptable_h_

/*
 * Package:     cvtk - Computer Vision Tool Kit
 * File:        lookuptable.h
 * Description: Class CLookUpTable defining a three dimensional lookup-table.
 * Created:     2001-04-19
 * Author:      Sascha Lange
 * Mail:        Sascha.Lange@Uni-Osnabrueck.De
 *
 * Web:         http://cvtk.sourceforge.net
 *
 * Copyright (C) 2001 Sascha Lange
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 *
 * ------------------------- Revision History -------------------------
 * $Log: lookuptable.h,v $
 * Revision 1.8  2001/08/13 15:36:40  slange
 * - build successfully
 * - initial checkin of objectlist.cpp
 * - removed graphics.h (now rgbmap.h) and ppm.cpp
 *
 * Revision 1.7  2001/08/10 14:11:57  slange
 * working on the next build
 *
 * Revision 1.6  2001/08/09 17:00:54  slange
 * - reorganized colorsegmentation and the lookuptable
 * - some other corrections
 *
 * Revision 1.5  2001/08/07 12:40:14  slange
 * - still refactoring
 * - initial checkin of defaultfactory.cpp, objectlist.h region.h
 *
 * Revision 1.4  2001/08/04 16:43:07  slange
 * - added some mehtods needed by template calibration
 * - added mouseClick event receiver
 *
 * Revision 1.3  2001/04/27 16:57:07  slange
 * - added 1- and 2-dimensional histograms
 * - changed typedef CRGBMap to a class definition
 * - added histogram functions to CRGBMap
 *
 * Revision 1.2  2001/04/23 18:42:52  slange
 * added copyright notice to all source files
 *
 * --------------------------------------------------------------------
 */

#include <iostream>
#include <vector>
#include <string.h>
#include "rgbmap.h"

namespace cvtk {

  /* to use lookup tables of different size without making
   * the colorsegmentation a template class itself, it's absolutly
   * neccessary to use an abstract interface, although this might
   * be a serious performance problem.
   */
  class CPixelDecisionStrategy {
  public:
    virtual ~CPixelDecisionStrategy() {};
    virtual unsigned char& lookup(int r, int g, int b) = 0;
    virtual unsigned char& lookup(const RGBValue& rgb) = 0;
    virtual const unsigned char& lookup(const RGBValue& rgb) const = 0;
    virtual const unsigned char& lookup(int r, int g, int b) const = 0;
    virtual int getNumberOfClasses() const = 0;
    virtual void setNumberOfClasses(int numC) = 0;
  };  
  
  template<int SHIFT>
  class CLookUpTable : public CPixelDecisionStrategy {
  public:
    CLookUpTable();
    
    virtual unsigned char& lookup(int r, int g, int b);
    virtual unsigned char& lookup(const RGBValue& rgb);
    virtual const unsigned char& lookup(const RGBValue& rgb) const;
    virtual const unsigned char& lookup(int r, int g, int b) const;
    
    unsigned int getSize() const;
    int getShift() const;
    virtual int getNumberOfClasses() const;
    virtual void setNumberOfClasses(int numC);
 
    
  protected:
    unsigned char table[256>>SHIFT][256>>SHIFT][256>>SHIFT];
    unsigned int table_size;
    int nClass;
  };
  
}

/************************* inline functions **************************/

template<int SHIFT>
inline unsigned int cvtk::CLookUpTable<SHIFT>::getSize() const 
{ 
  return table_size; 
};
template<int SHIFT>
inline int cvtk::CLookUpTable<SHIFT>::getShift() const 
{ 
  return SHIFT; 
};

template<int SHIFT>
ostream& operator<< (ostream&, const cvtk::CLookUpTable<SHIFT>&);
template<int SHIFT>
istream& operator>> (istream&, cvtk::CLookUpTable<SHIFT>&);

#ifdef EXPORT_NOT_IMPLEMENTED
#include "lookuptable.cpp"
#endif

#endif  /* _lookuptable_h_ */
















