#ifndef _mapwrapper_h_
#define _mapwrapper_h_

/*
 * Package:     cvtk - Computer Vision Tool Kit
 * File:        mapwrapper.h
 * Description: Class CMapWrapper defining a 2 dimensional map on a 
 *              plain buffer.
 * Created:     2001-03-27
 * Author:      Sascha Lange
 * Mail:        Sascha.Lange@Uni-Osnabrueck.De
 *
 * Web:         http://www.cvtk.sourceforge.net
 *
 * Copyright (C) 2001 Sascha Lange
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 
 * 02111-1307, USA.
 *
 * ------------------------- Revision History -------------------------
 * $Log: mapwrapper.h,v $
 * Revision 1.5  2001/08/09 17:00:54  slange
 * - reorganized colorsegmentation and the lookuptable
 * - some other corrections
 *
 * Revision 1.4  2001/08/07 13:42:40  slange
 * - restructured CMapWrapper's header and implementation
 * - added a constructor with a buffer as argument
 *
 * Revision 1.3  2001/04/27 16:57:07  slange
 * - added 1- and 2-dimensional histograms
 * - changed typedef CRGBMap to a class definition
 * - added histogram functions to CRGBMap
 *
 * Revision 1.2  2001/04/23 18:42:52  slange
 * added copyright notice to all source files
 *
 * --------------------------------------------------------------------
 */


#include <iostream>
#include "lockable.h"
#include "mutex.h"

/********************* CMapWrapper's signature ***********************/

/*
 * The task is to provide a covenient to use but efficent 
 * 2-dimensional wrapper for plain memory buffers used by the hardware
 * and window systems. By wrapping a map around an existing buffer, 
 * the client gains 2-dimensional acces on this buffer.
 *
 * The buffer has a reference count and will be deleted only if there's
 * no wrapper around it left.  
 *
 * The wrapper allows clients to directly access the buffer. Therefore 
 * the wrapper doesn't encapsulate data properly, it provides only
 * a convienient interface!
 *
 * Most efficency important methods like element access can't be 
 * declared to be virtual. Therefore they shouldn't be overwritten in
 * derived classes. The best way is to derieve only to add mehtods,
 * not to overwrite mehtods.
 */ 

/* ATTENTION: element access: 
 * CMapWrapper<integer> imap(4,4);
 * int x=0, y=0; imap[y][x]; !!!! 
 */

template<class T>
class CMapWrapper {
protected:
  class Buffer : public CLockable {
  public:
    int width, height;
    int refCount;                     // reference counter
    T* buf;                           // plain buffer
    T** map;                          // row-pointers

    ~Buffer();
    Buffer(int width, int height);    // construct new buffer
    Buffer(int width, int height, T* buffer);// wrap around ex. buffer

    void reSize(int width, int height);//change the size of the buffer
    void clear();                     // clear the buffer
  private:
    Buffer(const Buffer&);            // don't allow copy
    Buffer& operator=(const Buffer&);
  };

public:
  CMapWrapper();
  CMapWrapper(int width, int height); // wraps map around a buffer
  CMapWrapper(int width, int height, T* buffer); // wrap ex. buffer 
  CMapWrapper(const CMapWrapper<T>&); // copy constructor
  CMapWrapper<T>& operator= (const CMapWrapper<T>&); //
  virtual ~CMapWrapper();             // virtual destructor          

  virtual void clear();
  virtual void resize(int width, int height);

  void lock() const;                  // implements CLockableInterface
  bool isLocked() const;
  bool unlock() const;
  
  int getWidth() const;               // get dimensions of the map
  int getHeight() const;

  T* getBuffer() const;               // gain direct buffer access
  
  T* operator[] (int y);              // allow element access ]
  T* operator[] (int y) const;        // via [y][x

protected:
  Buffer* buf;
};

/* stream output operator
 */
template<class T>
ostream& operator<< (ostream& out, CMapWrapper<T>& mapwrapper);

/************************ inline functions ***************************/

template<class T>
inline void CMapWrapper<T>::lock() const 
{ 
  buf->lock(); 
} 
template<class T>
inline bool CMapWrapper<T>::isLocked() const 
{ 
  return buf->isLocked(); 
}
template<class T>
inline bool CMapWrapper<T>::unlock() const 
{ 
  return buf->unlock(); 
}

template<class T>
inline int CMapWrapper<T>::getWidth() const 
{ 
  return buf->width; 
}
template<class T>
inline int CMapWrapper<T>::getHeight() const 
{ 
  return buf->height; 
}

template<class T>
inline T* CMapWrapper<T>::getBuffer() const
{
  return buf->buf;
}

template<class T>
inline T* CMapWrapper<T>::operator[] (int y) 
{ 
  return buf->map[y]; 
}
template<class T>
inline T* CMapWrapper<T>::operator[] (int y) const 
{ 
  return buf->map[y]; 
}

#ifdef EXPORT_NOT_IMPLEMENTED
#include "mapwrapper.cpp"
#endif   /* EXPORT_NOT_IMPLEMENTED */

#endif   /* _mappwrapper_h_ */




