/*
 * Package:     cvtk - Computer Vision Tool Kit
 * File:        monitors.cpp
 * Description: methods to monitor cvtk's analyzing steps
 * Created:     2001-08-31
 * Author:      Sascha Lange
 * Mail:        Sascha.Lange@Uni-Osnabrueck.De
 *
 * Web:         http://cvtk.sourceforge.net
 *
 * Copyright (C) 2001 Sascha Lange
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  
 * 02111-1307, USA.
 *
 * ------------------------ Revision History --------------------------
 * $Log: monitors.cpp,v $
 * Revision 1.5  2001/09/04 15:31:27  slange
 * - corrected bugs in:
 *   - templatecontroller.cpp: use QString to emit messages
 *   - monitor.cpp: used x instead of y to check against height
 * - labeled the sliders in calibratewidget
 *
 * Revision 1.4  2001/09/02 18:15:35  slange
 * - error cvs
 *
 * Revision 1.3  2001/09/02 17:39:07  slange
 * cvs error!
 *
 * Revision 1.2  2001/09/01 17:40:09  slange
 * - working on monitor
 * - moved template calibration into namespace "cvtk"
 *
 * Revision 1.1  2001/09/01 13:28:41  slange
 * - working on monitor functions
 *
 * --------------------------------------------------------------------
 */

#include "monitors.h"
#include "chaincoding.h"

void cvtk::Monitors::visualizeRegionMap(const CRegionMap& segments,
					CRGBMap& visual,
					bool clearBG)
{
  static RGBValue colors[22] =
  {{0,0,0}, {255,0,0}, {0,255,0}, {0,0,255}, {255,255,0}, 
   {0,255,255}, {255,0,255}, {255,127,0}, {255,0,127}, {127,255,0},
   {0,255,127}, {127,0,255}, {0,127,255}, {255,255,127}, {127,255,255},
   {255,127,255}, {127,0,0}, {0,127,0}, {0,0,127}, {127,127,0},
   {127,0,127}, {0,127,127}};

  if (segments.getWidth() != visual.getWidth() ||
      segments.getHeight() != visual.getHeight())
    return;

  int* inBuf = segments.getBuffer();
  RGBValue* outBuf = visual.getBuffer();
  for (int y=0; y < segments.getHeight(); y++)
    for (int x=0; x < segments.getWidth(); x++, inBuf++, outBuf++) 
      if (*inBuf==0) {
	if (clearBG)
	  *outBuf=colors[*inBuf];
      }
      else if (*inBuf<22)
	*outBuf=colors[*inBuf];
      else
	*outBuf=colors[21];

}

void cvtk::Monitors::visualizeRegionList(const CRegionList& rlist,
					 CRGBMap& visual) 
{
  CRegionList& rl = const_cast<CRegionList&>(rlist);
  RGBValue white; white.r=white.g=white.b=255;
  for (CRegionList::iterator r=rl.begin(); 
       r != rl.end(); r++) 
    for (CRegionList::innerIterator s=r->begin(); 
	 s != r->end(); s++) {
      int x=s->start[0];
      int y=s->start[1];
      for (unsigned int i=0; i<s->chaincode.size(); i++) {
	visual[y][x] = white;
	x+=CBorderTracingStrategy::xDir(s->chaincode[i]);
	y+=CBorderTracingStrategy::yDir(s->chaincode[i]);
      }
    } 
}

void cvtk::Monitors::visualizeObjectList(const CObjectList& objs,
					 CRGBMap& visual)
{
  CObjectList& ol = const_cast<CObjectList&>(objs);
  RGBValue white;
  white.r=white.g=white.b=255;
  RGBValue grey;
  grey.r=grey.g=grey.b=180;
  for (CObjectList::iterator c=ol.begin();
       c != ol.end(); c++)
    for (CObjectList::innerIterator o=c->begin();
	 o != c->end(); o++) {
      
      // position-cross
      drawLine(visual, o->center[0]-5, o->center[1]-5,
	       o->center[0]+5, o->center[1]+5, white);
      drawLine(visual, o->center[0]-5, o->center[1]-4,
	       o->center[0]+4, o->center[1]+5, white);
      drawLine(visual, o->center[0]-4, o->center[1]-5,
	       o->center[0]+5, o->center[1]+4, white);

      drawLine(visual, o->center[0]-5, o->center[1]+5,
	       o->center[0]+5, o->center[1]-5, white);
      drawLine(visual, o->center[0]-5, o->center[1]+4,
	       o->center[0]+4, o->center[1]-5, white);
      drawLine(visual, o->center[0]-4, o->center[1]+5,
	       o->center[0]+5, o->center[1]-4, white);

      // orientation
      if (o->orientation>=0) {
	double radial = o->orientation*M_PI/180.;
	Vektor<int, 2> orientation=o->center;
	
	orientation[0]+=static_cast<int>(30.*cos(radial));
	orientation[1]+=static_cast<int>(30.*sin(radial));

	drawLine(visual, o->center[0], o->center[1],
		 orientation[0], orientation[1], grey);
      }
      
      // direction
      if (o->direction>=0 && o->velocity>=0) {
	double radial = o->direction*M_PI/180.;
	Vektor<int, 2> direction=o->center;

	direction[0]+=static_cast<int>(5.*o->velocity*cos(radial));
	direction[1]+=static_cast<int>(5.*o->velocity*sin(radial));

	drawLine(visual, o->center[0], o->center[1],
		 direction[0], direction[1], grey);
	
      }
    }
}

void cvtk::Monitors::drawLine(CRGBMap& visual, int x1, int y1, 
			      int x2, int y2, RGBValue color)
{
    
  int x=x1,y=y1, error, delta, schwelle, dx, dy, inc_x, inc_y;

  dx = x2-x;
  dy = y2-y;

  if (dx>0) inc_x=1; else inc_x=-1;
  if (dy>0) inc_y=1; else inc_y=-1;

  if (abs(dy) < abs(dx)) {
    error = -abs(dx);
    delta = 2* abs(dy);
    schwelle = 2*error;
    while (x != x2) {
      if (y<0 || x<0 || y>=visual.getHeight() || x>=visual.getWidth())
	return;
      visual[y][x] = color;
      x+=inc_x;
      error+=delta;
      if (error>0) { y+=inc_y; error+=schwelle; }
    }
  }
  else {
    error = -abs(dy);
    delta = 2* abs(dx);
    schwelle = 2*error;
    while (y != y2) {
      if (y<0 || x<0 || y>=visual.getHeight() || x>=visual.getWidth())
	return;
      visual[y][x] = color;
      y+=inc_y;
      error+=delta;
      if (error>0) { x+=inc_x; error+=schwelle; }
    }
  }

  if (y<0 || x<0 || y>=visual.getHeight() || x>=visual.getWidth())
    return;
  visual[y][x]=color;
}











