/*
 * Package:     cvtk - Computer Vision Tool Kit
 * File:        objecttemplate.cpp
 * Description: Implementation of the object-template datastructure. 
 * Created:     2001-06-21
 * Author:      Sascha Lange
 * Mail:        Sascha.Lange@Uni-Osnabrueck.De
 *
 * Web:         http://cvtk.sourceforge.net
 *
 * Copyright (C) 2001 Sascha Lange
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  
 * 02111-1307, USA.
 *
 * ------------------------ Revision History --------------------------
 * $Log: objecttemplate.cpp,v $
 * Revision 1.5  2001/08/04 14:50:29  slange
 * added stream input functions to object templates
 *
 * Revision 1.4  2001/07/06 02:04:18  slange
 * - working on the template matching: completed center matching
 *
 * Revision 1.3  2001/07/05 16:19:50  slange
 * - added tolerance to object tempplate definition
 *
 * Revision 1.2  2001/07/04 23:00:04  uid51551
 * - working on the templatematching
 * - calculation of the region center
 *
 * Revision 1.1  2001/06/21 15:57:46  slange
 * initial checkin of objecttemplate.[h/cpp ]
 *
 * --------------------------------------------------------------------
 */

#include "objecttemplate.h"
#include <string>

/*
 * function to log a syntax error and exit
 */
inline static void syntaxError(string error)
{
  cerr << error << endl;
  exit(1);
}

ostream& operator<< (ostream& out, const CObjectTemplateList& list)
{
  for (CObjectTemplateList::const_iterator i=list.begin();
       i != list.end(); i++) {
    out << (*i);
  }
  return out;
}  
istream& operator>> (istream& in, CObjectTemplateList& list)
{
  string line;
  while(in>>line) {
    if (line.compare(string("object"),0,6)==0) {
#ifdef DEBUG
      cerr << "found object template" << endl;
#endif
      CObjectTemplate objT;
      in >> objT;
      list.push_back(objT);
    }
  }
  return in;
}
ostream& operator<< (ostream& out, const CObjectTemplate& objT)
{
  out << "object" << endl << objT.id << endl;
  out << objT.center;
  if (objT.axis.colClass!=0) {
    out << objT.axis;
    for (vector<CMark>::const_iterator i=objT.marks.begin();
	 i != objT.marks.end(); i++) {
      out << *i;
    }
  }
  out << "end" << endl;
  return out;
} 
istream& operator>> (istream& in, CObjectTemplate& objT)
{
  string line;
  /* read center */
  if(!(in >> objT.id))
    syntaxError(">>ObjectTemplate: expected id!");
  in >> line; 
  if (line.compare(string("center"), 0,6) != 0)
    syntaxError(">>ObjectTemplate: syntax error near \""+
                line+"\". expected \"center\"");
#ifdef DEBUG
  cerr << "found center" << endl;
#endif
  in >> objT.center;
  in >> line; 
  if (line.compare(string("end"), 0,3) == 0) return in;
  if (line.compare(string("axis"), 0,4) != 0)
    syntaxError(">>ObjectTemplate: syntax error near \""+
                line+"\". expected \"axis\" or \"end\"");
#ifdef DEBUG
  cerr << "found axis" << endl;
#endif
  in >> objT.axis;
  while(in>>line) {
    if (line.compare(string("end"),0,3)==0) return in;
    if (line.compare(string("mark"), 0,4) != 0)
    syntaxError(">>ObjectTemplate: syntax error near \""+
                line+"\". expected \"mark\" or \"end\"");
#ifdef DEBUG
    cerr << "found mark" << endl;
#endif
    CMark mark;
    in >> mark;
    objT.marks.push_back(mark);
  }                       
                         // reached end of stream too early
  syntaxError(">>CObjectTemplate: unexpected end of stream");  
  return in;             // never reached 
}
ostream& operator<< (ostream& out, const CCenter& center)
{
  out << "center " << center.colClass << " " << center.area << " "
      << center.areaTol << endl;
  return out;
}
istream& operator>> (istream& in, CCenter& center)
{
  if (!(in >> center.colClass)) 
    syntaxError(">>Center: failed to read colClass");
  if (!(in >> center.area)) {
    if (in.eof()) syntaxError(">>Center: unexpected end of stream");
    else  {
      in.clear();
      center.area = 0;
    }
  }
  else 
    if (!(in >> center.areaTol))
      syntaxError(">>Center: failed to read tolerance of area");
  return in;
}
ostream& operator<< (ostream& out, const CAxis& axis)
{
  out << "axis " << axis.colClass << " " << axis.distanceToCenter 
      << " " << axis.distanceTol << " " << axis.area << " "
      << axis.areaTol << endl;
  return out;
}
istream& operator>> (istream& in, CAxis& axis)
{
  if (!(in >> axis.colClass)) 
    syntaxError(">>Axis: failed to read colClass");
  if (!(in >> axis.distanceToCenter))
    syntaxError(">>Axis: failed to read distance");
  if (!(in >> axis.distanceTol))
    syntaxError(">>Axis: failed to read tolerance of distance");
  if (!(in >> axis.area)) {
    if (in.eof()) syntaxError(">>Axis: unexpected end of stream");
    else {
      axis.area = 0;
      in.clear();
    }
  }
  else 
    if (!(in >> axis.areaTol))
      syntaxError(">>Axis: failed to read tolerance of area");
  return in;
}
ostream& operator<< (ostream& out, const CMark& mark)
{
  out << "mark " << mark.colClass << " " << mark.distanceToCenter 
      << " " << mark.distanceTol << " " << mark.direction << " "
      << mark.directionTol << " " << mark.area << " "
      << mark.areaTol << endl;
  return out;
}
istream& operator>> (istream& in, CMark& mark)
{
  if (!(in >> mark.colClass)) 
    syntaxError(">>Mark: failed to read colClass");
  if (!(in >> mark.distanceToCenter))
    syntaxError(">>Mark: failed to read distance");
  if (!(in >> mark.distanceTol))
    syntaxError(">>Mark: failed to read tolerance of distance");
  if (!(in >> mark.direction))
    syntaxError(">>Mark: failed to read direction");
  if (!(in >> mark.directionTol))
    syntaxError(">>Mark: failed to read tolerance of direction");

  if (!(in >> mark.area)) {
    if (in.eof()) syntaxError(">>Mark: unexpected end of stream");
    else {
      mark.area = 0;
      in.clear();
    }
  }
  else 
    if (!(in >> mark.areaTol))
      syntaxError(">>Mark: failed to read tolerance of area");
  return in;
}
