#ifndef _regionlist_h_
#define _regionlist_h_

/*
 * Package:     cvtk - Computer Vision Tool Kit
 * File:        regionlist.h
 * Description: Deklaration of the regions
 * Created:     2001-08-07
 * Author:      Sascha Lange
 * Mail:        Sascha.Lange@Uni-Osnabrueck.De
 *
 * Web:         http://cvtk.sourceforge.net
 *
 * Copyright (C) 2001 Sascha Lange
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  
 * 02111-1307, USA.
 *
 * ------------------------ Revision History --------------------------
 * $Log: regionlist.h,v $
 * Revision 1.3  2001/08/10 01:50:06  slange
 * trying to build again
 *
 * Revision 1.2  2001/08/09 23:15:08  slange
 * - reorganised chaincoding.[h/cpp] (former regions.h, regionfinder.cpp)
 * - added a destructor to color segmentation
 * - added a constructor to region list
 *
 * Revision 1.1  2001/08/09 21:47:20  slange
 * - working on the chaincoding algorithm
 * - changed regionlist from a simple typedef to a class and
 *   moved it into regionlist.h (former region.h)
 *
 * Revision 1.3  2001/08/09 17:00:54  slange
 * - reorganized colorsegmentation and the lookuptable
 * - some other corrections
 *
 * Revision 1.2  2001/08/09 14:28:57  slange
 * - added resize() to classifiedlist.h
 * - modified getArea() of CRegion to work with both,
 *   4- and 8-connectivity.
 * - 4 connectivity now coded with codes of 8-connectivity
 *
 * Revision 1.1  2001/08/07 12:40:14  slange
 * - still refactoring
 * - initial checkin of defaultfactory.cpp, objectlist.h region.h
 *
 * --------------------------------------------------------------------
 */

#include "vektor.h"
#include "classifiedlist.h"

/* region, coded as (Freeman) chaincode
 * for a description of the coding and the area calculation algorithm
 * see Sonka, Hlavac, Boyle: "Image Processing, Analysis and Machine
 * Vision" pp. 142 and pp. 236
 * for the definition of the center and its calculation see
 * Horn: "Robot Vision" pp. 48 (adapted)
 */  

namespace cvtk {

  /* A single region with center, size, class and border information
   */
  class CRegion {
  public:
    CRegion();
    Vektor<int, 2> start;
    Vektor<int, 2> second;
    Vektor<int, 2> middle;
    vector<int> chaincode;
    int size;
    int regionClass;
    
    int getArea() const;
    Vektor<int, 2> getCenter() const;
  
  private:
    int area;
    Vektor<int, 2> center;
  };

  /* classified region list, with additional region sepecific
   * functionallity
   */
  class CRegionList : public CClassifiedList<CRegion> {
  public:
    CRegionList(int nClasses=0);
    const CRegion& findNearest(const Vektor<int, 2>&) const;
  };

};

/************************ inline functions ***************************/

inline cvtk::CRegion::CRegion() 
  : size(0), area(0), center(-1,-1)  // region is not valid until 
{}                                   // center is set to positive

/* Calculate both, area and center, in one cycle through the 
 * chain code when asked for one of them.
 */
inline int cvtk::CRegion::getArea() const 
{
  if (area)
    return area;
  
  int vertPos=0;
  int horiPos=0;
  const_cast<CRegion*>(this)->center = Vektor<int, 2> (0,0);
  
  int a = area;
  
  // walk through the chaincode and integrate (see Horn for a desc.)
  for (unsigned int i=0; i < chaincode.size(); i++) 
    switch(chaincode[i]) {
    case 0: a-=vertPos; center[0]-=vertPos*horiPos++; break;
    case 1:
      a-=vertPos;
      center[0]-=vertPos*horiPos;
      center[1]+=vertPos++*horiPos++;
      break;
    case 2: center[1]+=horiPos*vertPos++; break;
    case 3: 
      a+=vertPos;
      center[0]+=horiPos*vertPos;
      center[1]+=horiPos--*vertPos++;
      break;
    case 4: a+=vertPos; center[0]+=vertPos*horiPos--; break;
    case 5: 
      a+=vertPos;
      center[0]+=vertPos*horiPos;
      center[1]-=horiPos--*vertPos--;
      break;
    case 6: center[1]-=horiPos*vertPos--; break;
    case 7: 
      a-=vertPos;
      center[0]-=vertPos*horiPos;
      center[1]-=horiPos++*vertPos--;
      break;
    };
  
  if (a!=0) {                     // avoid division by zero
    center[0]/=a;
    center[1]/=a;
    const_cast<CRegion*>(this)->center+=start;
  }
  else 
    const_cast<CRegion*>(this)->center=start;
  const_cast<CRegion*>(this)->area=a;
  return area;
}

inline Vektor<int, 2> cvtk::CRegion::getCenter() const
{
  if (!area)                       // if not already calculated 
    getArea();                     // calculate values
  
  return center;
}

#endif      /* _regionlist_h_ */

