/*
 * Package:     cvtk - Computer Vision Tool Kit
 * File:        rgbmap.cpp
 * Description: Implementation of the CRGBMap.
 * Created:     2001-04-27
 * Author:      Sascha Lange
 * Mail:        Sascha.Lange@Uni-Osnabrueck.De
 *
 * Web:         http://cvtk.sourceforge.net
 *
 * Copyright (C) 2001 Sascha Lange
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 
 * 02111-1307, USA.
 *
 * ------------------------- Revision History -------------------------
 * $Log: rgbmap.cpp,v $
 * Revision 1.6  2001/08/13 15:36:40  slange
 * - build successfully
 * - initial checkin of objectlist.cpp
 * - removed graphics.h (now rgbmap.h) and ppm.cpp
 *
 * Revision 1.5  2001/05/03 15:04:09  slange
 * - added lookup table data for orange/yellow/blue on 
 *   green/black/white background to rsc
 * - fixes
 *
 * Revision 1.4  2001/05/03 13:04:27  slange
 * - traced down bug in CController::addFrame()
 *
 * Revision 1.3  2001/04/27 23:15:42  slange
 * - added an UV - Histogram (histogram in an approximation of the 
 *   HUV space)
 * - changed the Calibrate::CController
 *   - to work with histograms
 *   - to keep a vector of examples in memory
 *   - output file dropped
 *   - to tread the background as a (special) object
 * - fixes
 *
 * Revision 1.2  2001/04/27 20:51:16  slange
 * - finished implementation of CHistogram2Dim
 * - bug fixes
 *
 * Revision 1.1  2001/04/27 16:57:07  slange
 * - added 1- and 2-dimensional histograms
 * - changed typedef CRGBMap to a class definition
 * - added histogram functions to CRGBMap
 *
 * --------------------------------------------------------------------
 */

#include "rgbmap.h"
#include "assert.h"
#include <ctype.h>

/****************************** CRGBMap ******************************/

cvtk::CRGBMap::CRGBMap(int width=0, int height=0) 
  : CMapWrapper<RGBValue>(width, height)
{};

cvtk::CRGBMap::CRGBMap(const CRGBMap& rgbMap) 
  : CMapWrapper<RGBValue>(rgbMap)
{};

/******************************* CPPM ********************************/

cvtk::CPPM::CPPM(PPMFormat type) : CRGBMap(), type(type)
{}

cvtk::CPPM::CPPM(const CPPM& ppm) : CRGBMap(ppm), type(ppm.getFormat())
{}

cvtk::CPPM::CPPM(const CRGBMap& rgbMap, PPMFormat type)
  : CRGBMap(rgbMap), type(type)
{}

cvtk::CPPM::CPPM(int width, int height, PPMFormat type)
  : CRGBMap(width, height), type(type)
{};

cvtk::CPPM& cvtk::CPPM::operator= (const CPPM& ppm)
{
  *dynamic_cast<CRGBMap*>(this) = ppm; // Call the operator of the base
  type = ppm.type;                     // class and set the type.
  return *this;
}

/* member access functions 
 */
void cvtk::CPPM::setFormat(PPMFormat format)         
{ type = format; }

cvtk::CPPM::PPMFormat cvtk::CPPM::getFormat() const
{ return type; }

/************************* Stream functions **************************/

namespace cvtk {
  static void skip(istream& in);
};

/* ugly output function for PPM files.
 */
ostream& cvtk::operator<< (ostream& out, const CPPM& ppm)
{
  out << "P" << ppm.getFormat();
  out << "\n# CREATOR: CVTK Version 0.1    Class CPPM\n";
  out << ppm.getWidth() << ' ' 
      << ppm.getHeight() << ' ' 
      << ppm.getMaxColorValue() << endl;
  if (ppm.getFormat() == CPPM::ASCII) {
    for (int y=0; y < ppm.getHeight(); y++)
      for (int x=0; x < ppm.getWidth(); x++)
	out << ppm[y][x] << ((x==ppm.getWidth()-1)? "\n" : "  ");
  }
  else
    out << "Output for type P" 
	<< ppm.getFormat() << " not yet implemented.\n";
  return out;
}

/* ugly parsing function for PPM files. Allows comments in the header.
 * Throws exceptions.
 */
istream& cvtk::operator>> (istream& in, CPPM& ppm)
{
  char c;
  skip(in);                        // skip any comments and spaces!
  if(!in.get(c)) exit(1);          // throw format exception!!!
  if (c!='P' && c!='p') exit(1);   // throw format exception!!!

  int width, height, maxCol, type;
  skip(in);
  in >> type;
  skip(in);
  in >> width;
  skip(in);
  in >> height;
  skip(in);
  in >> maxCol;                     // maxCol ignored!
  ppm.type = CPPM::PPMFormat(type);
  if (width != ppm.getWidth() || height != ppm.getHeight()) { 
    ppm.buf->reSize(width, height); // VERY CLOSE to implementation!
  }
  if (ppm.type == CPPM::ASCII) {
    for (int y=0; y < height; y++)
      for (int x=0; x < width; x++) {
	skip(in);
	in >> ppm[y][x];            // comments within a single 
      }                             // rgbValue not allowed!
  }
  else {
    cout << "Read ppm of type P" << ppm.type 
	 << " not yet implemented.\n";
    exit(1);
  }
  return in;
}

/* skips spaces and recursivly multi-line comments on a given stream.
 */ 
static void cvtk::skip(istream& in)
{
  char c;
  do {
    if (!in.get(c)) return;
  } while(isspace(c) || c=='\n');
  if (c=='#') {                  // start of a comment
    do {                         // look for the end of the line
      if (!in.get(c)) return;
    } while(c!='\n');
    skip(in);                    // skip again at start of next line.
  }
  else                           // c was neither comment(#) nor space
    in.putback(c);               // so better put it back, 
}                                // before returning

/*
CHistogram<256> cvtk::CRGBMap::getBrightnessHistogram()
{
  CHistogram<> hist;
  int brightness;
  RGBValue* value;

  for (int x=0; x < getWidth(); x++)
    for (int y=0; y < getHeight(); y++) {
      value = &(*this)[y][x];
      brightness = (value->r + value->g + value->b) / 3;
      hist.add(brightness);
    }
  return hist;
}

CUVHistogram* cvtk::CRGBMap::getUVHistogram()
{
  CUVHistogram* hist = new CUVHistogram();
  for (int x=0; x < getWidth(); x++)
    for (int y=0; y < getHeight(); y++)
      hist->add((*this)[y][x]);
  return hist;
}

CRGBHistogram<256>* cvtk::CRGBMap::getRGBHistogram()
{
  CRGBHistogram<256>* hist = new CRGBHistogram<256>();
  for (int x=0; x < getWidth(); x++)
    for (int y=0; y < getHeight(); y++)
      hist->add((*this)[y][x]);
  return hist;
}*/











