#include "search.h"

/*
 * Package:     cvtk - Computer Vision Tool Kit
 * File:        search.cpp
 * Description: Implementation of a general search shell
 * Created:     2001-07-26
 * Author:      Sascha Lange
 * Mail:        Sascha.Lange@Uni-Osnabrueck.De
 *
 * Web:         http://cvtk.sourceforge.net
 *
 * Copyright (C) 2001 Sascha Lange
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  
 * 02111-1307, USA.
 *
 * ------------------------ Revision History --------------------------
 * $Log: search.cpp,v $
 * Revision 1.4  2001/08/10 00:57:05  slange
 * - moved search into namespace cvtk
 * - restructured historymatching and objecthistory
 *
 * Revision 1.3  2001/07/28 21:51:15  slange
 * - finished search algorithm
 *
 * Revision 1.2  2001/07/27 20:35:34  slange
 * - initial checkin of searchmatch.[h/cpp]
 *
 * Revision 1.1  2001/07/27 00:23:07  slange
 * working on the general search shell's implementation
 * 
 * --------------------------------------------------------------------
 */


/************************ CSearchStrategy ****************************/

cvtk::CSearchStrategy::
CSearchStrategy(CProblemStateQueue* stateQueue,
		CGoalTestStrategy* goalTestStrat,
		CExpandStrategy* expandStrat)
  : CProblemSolvingStrategy(), stateQueue(stateQueue), 
    goalTestStrat(goalTestStrat), expandStrat(expandStrat)
{};

cvtk::CSearchStrategy::~CSearchStrategy()
{
  stateQueue->clear();
  delete stateQueue;
  delete goalTestStrat;
  delete expandStrat;
}

cvtk::CProblemState* 
cvtk::CSearchStrategy::solve(CProblemState* problem)
{
  vector<CProblemState*> expanded;
  stateQueue->clear();
  stateQueue->enqueue(problem);         // starting point of search
  
  while(true) {                  
    if (stateQueue->isEmpty())          // no state left -> failure
      return 0;

    problem=stateQueue->dequeue();      // get next state to expand

    if (goalTestStrat->isGoal(problem)) // if reached goal, finish!
      return problem;
                           
    expandStrat->expand(problem, &expanded);// expand node
    stateQueue->enqueue(expanded);      // and enqueue resulting nodes
    expanded.clear();                   // cleanup for next expansion 
  }
  
  return 0;
}

/******************** CEvaluatedProblemStateQueue ********************/

cvtk::CEvaluatedProblemStateQueue
::CEvaluatedProblemStateQueue(CEvaluationStrategy* evalStrat)
  : CProblemStateQueue(), evalStrat(evalStrat)
{};
cvtk::CEvaluatedProblemStateQueue::~CEvaluatedProblemStateQueue()
{
  delete evalStrat;
}

/*********************** CPrioriytQueueAdaptor ***********************/

cvtk::CPriorityQueueAdaptor::
CPriorityQueueAdaptor(CEvaluationStrategy* evalStrat)
  : CEvaluatedProblemStateQueue(evalStrat)
{};

cvtk::CPriorityQueueAdaptor::~CPriorityQueueAdaptor()
{
  clear();
};

cvtk::CProblemState* cvtk::CPriorityQueueAdaptor::front() const
{
  return queue.top().state;
}
cvtk::CProblemState* cvtk::CPriorityQueueAdaptor::dequeue()
{
  CProblemState* ps=queue.top().state;
  queue.pop();
  return ps;
}
void cvtk::CPriorityQueueAdaptor::enqueue(CProblemState* ps)
{
  CEvaluatedProblemState* eps 
    = dynamic_cast<CEvaluatedProblemState*>(ps);  // throw exception!
  eps->setCost(evalStrat->evaluate(eps));
  queue.push(CPPair(eps->getCost(), eps));
}
void cvtk::CPriorityQueueAdaptor::
enqueue(const vector<CProblemState*>& ps)
{
  for (unsigned int i=0; i <ps.size(); i++)
    enqueue(ps[i]);
}
bool cvtk::CPriorityQueueAdaptor::isEmpty() const
{
  return queue.empty();
}
void cvtk::CPriorityQueueAdaptor::clear()
{
  while(!isEmpty())
    delete dequeue();
}

/******************** CAStarEvaluationStrategy ***********************/

cvtk::CAStarEvaluationStrategy::
CAStarEvaluationStrategy(CEvaluationStrategy* g,
			 CEvaluationStrategy* h)
  : g(g), h(h)
{};

cvtk::CAStarEvaluationStrategy::~CAStarEvaluationStrategy()
{
  delete g;
  delete h;
}









