#ifndef _search_h_
#define _search_h_

/*
 * Package:     cvtk - Computer Vision Tool Kit
 * File:        search.h
 * Description: Deklaration of a general search shell
 * Created:     2001-07-26
 * Author:      Sascha Lange
 * Mail:        Sascha.Lange@Uni-Osnabrueck.De
 *
 * Web:         http://cvtk.sourceforge.net
 *
 * Copyright (C) 2001 Sascha Lange
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  
 * 02111-1307, USA.
 *
 * ------------------------ Revision History --------------------------
 * $Log: search.h,v $
 * Revision 1.5  2001/08/10 00:57:05  slange
 * - moved search into namespace cvtk
 * - restructured historymatching and objecthistory
 *
 * Revision 1.4  2001/07/28 21:51:15  slange
 * - finished search algorithm
 *
 * Revision 1.3  2001/07/27 20:35:34  slange
 * - initial checkin of searchmatch.[h/cpp]
 *
 * Revision 1.2  2001/07/27 00:23:07  slange
 * working on the general search shell's implementation
 *
 * Revision 1.1  2001/07/26 18:04:46  slange
 * - initial checkin of search.h
 * 
 * --------------------------------------------------------------------
 */

#include <vector>
#include <queue>

/* Deklaration of a general search shell and A* as described in 
 * Russel, Norvig "Artificial Intelligence - A Modern Approach"
 */

namespace cvtk {

  /* abstract interface to problem-solving strategies
   */
  class CProblemState {
  public:
    virtual ~CProblemState() {};
  protected:
    CProblemState() {};
  };

  class CProblemSolvingStrategy {
  public:
    virtual CProblemState* solve(CProblemState*) = 0;
    virtual ~CProblemSolvingStrategy() {};
  protected:
    CProblemSolvingStrategy() {};
  };

  /* general search shell
   */
  class CProblemStateQueue {
  public:
    virtual CProblemState* front() const =0;
    virtual CProblemState* dequeue()=0;
    virtual void enqueue(CProblemState*)=0;
    virtual void enqueue(const vector<CProblemState*>&)=0;
    virtual bool isEmpty() const =0;
    virtual void clear() =0;
    virtual ~CProblemStateQueue() {};
  protected:
    CProblemStateQueue() {};
  };

  class CGoalTestStrategy {
  public:
    virtual bool isGoal(const CProblemState*) const=0;
    virtual ~CGoalTestStrategy() {};
  protected:
    CGoalTestStrategy() {};
  };

  class CExpandStrategy {
  public:
    virtual void expand (CProblemState*, 
			 vector<CProblemState*>*) const=0;
    virtual ~CExpandStrategy() {};
  protected:
    CExpandStrategy() {};
  };

  class CSearchStrategy : public CProblemSolvingStrategy {
  public:
    CSearchStrategy(CProblemStateQueue*, 
		    CGoalTestStrategy*,
		    CExpandStrategy*);
    virtual CProblemState* solve(CProblemState*);
    virtual ~CSearchStrategy();
  protected:
    CProblemStateQueue* stateQueue;
    CGoalTestStrategy* goalTestStrat;
    CExpandStrategy* expandStrat;
  };

  /* general cost search
   */
  class CEvaluatedProblemState : public CProblemState {
  public:
    virtual double getCost() const;
    virtual void setCost(double);
    virtual ~CEvaluatedProblemState() {};
  protected:
    double cost;
    CEvaluatedProblemState() : CProblemState(), cost(0) {};
  };
  
  class CEvaluationStrategy {
  public:
    virtual double evaluate(const CProblemState*) const =0;
    virtual ~CEvaluationStrategy() {};
  protected:
    CEvaluationStrategy() {};
  };

  class CEvaluatedProblemStateQueue : public CProblemStateQueue {
  public:
    virtual ~CEvaluatedProblemStateQueue();
  protected:
    CEvaluatedProblemStateQueue(CEvaluationStrategy*);
    CEvaluationStrategy* evalStrat;
  };

  /*  Implementation, adapting the stl priority_queue
   */
  class CPriorityQueueAdaptor : public CEvaluatedProblemStateQueue {
  public:
    CPriorityQueueAdaptor(CEvaluationStrategy*);
    virtual CProblemState* front() const;
    virtual CProblemState* dequeue();
    virtual void enqueue(CProblemState*);
    virtual void enqueue(const vector<CProblemState*>&);
    virtual bool isEmpty() const;
    virtual void clear();
    virtual ~CPriorityQueueAdaptor();
  protected:
    class CPPair {
    public:
      CPPair(double priority, CEvaluatedProblemState* state)
	: priority(priority), state(state) {};
      bool operator> (const CPPair& right) const { 
	return this->priority > right.priority; }
      bool operator< (const CPPair& right) const {
	return this->priority < right.priority; }
      double priority;
      CEvaluatedProblemState* state;
    };
    priority_queue<CPPair, vector<CPPair>, greater<CPPair> > queue;
  };

  /* A-Star specific
   */

  // Implementing the Evaluation Strategy using recursive composition
  class CAStarEvaluationStrategy : CEvaluationStrategy {
  public:
    CAStarEvaluationStrategy(CEvaluationStrategy*, 
			     CEvaluationStrategy*);
    ~CAStarEvaluationStrategy();
    virtual double evaluate(const CProblemState*) const;
  protected:
    CEvaluationStrategy* g;
    CEvaluationStrategy* h;
  };  

}

/************************* inline functions **************************/

inline
double cvtk::CAStarEvaluationStrategy::
evaluate(const CProblemState* p) const
{
  return g->evaluate(p) + h->evaluate(p);
}

inline double cvtk::CEvaluatedProblemState::getCost() const
{
  return cost;
}
inline void cvtk::CEvaluatedProblemState::setCost(double cost)
{
  this->cost=cost;
}
  
#endif
