#include "searchmatch.h"

/*
 * Package:     cvtk - Computer Vision Tool Kit
 * File:        searchmatch.cpp
 * Description: Implementation of the problem specific search 
 * Created:     2001-07-27
 * Author:      Sascha Lange
 * Mail:        Sascha.Lange@Uni-Osnabrueck.De
 *
 * Web:         http://cvtk.sourceforge.net
 *
 * Copyright (C) 2001 Sascha Lange
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  
 * 02111-1307, USA.
 *
 * ------------------------ Revision History --------------------------
 * $Log: searchmatch.cpp,v $
 * Revision 1.4  2001/08/10 00:57:05  slange
 * - moved search into namespace cvtk
 * - restructured historymatching and objecthistory
 *
 * Revision 1.3  2001/07/28 22:22:43  slange
 * - found and traced down a bug
 *
 * Revision 1.2  2001/07/28 21:51:15  slange
 * - finished search algorithm
 *
 * Revision 1.1  2001/07/27 20:35:34  slange
 * - initial checkin of searchmatch.[h/cpp]
 *
 * --------------------------------------------------------------------
 */


/************************** CMatchState ******************************/

/* Build up the pointer lists, on which the expanding algorithm
 * works.
 */
cvtk::CMatchState::CMatchState(CObjectList::iterator objs,
			       CObjectHistoryList::iterator hists)
  : CEvaluatedProblemState(), objs(objs), hists(hists)
{
  for (CObjectList::innerIterator i=objs->begin();
       i != objs->end(); i++)
    objToMatch.push_back(&(*i));
  for (CObjectHistoryList::innerIterator i=hists->begin();
       i != hists->end(); i++)
    histToMatch.push_back(&(*i));
}
cvtk::CMatchState::CMatchState(const CMatchState& state)
  :  objs(state.objs), hists(state.hists),
     objToMatch(state.objToMatch), objUnMatched(state.objUnMatched),
     histToMatch(state.histToMatch), 
     histUnMatched(state.histUnMatched), matches(state.matches)
{
  cost=state.cost;
}
cvtk::CMatchState& 
cvtk::CMatchState::operator= (const CMatchState& state)
{
  cost=state.cost;
  objs=state.objs;
  hists=state.hists;
  objToMatch=state.objToMatch;
  objUnMatched=state.objUnMatched;
  histToMatch=state.histToMatch;
  histUnMatched=state.histUnMatched;
  matches=matches;
  return *this;
}

/****************** Evaluation / Goal Functions **********************/

bool cvtk::CMatchGoal::isGoal(const CProblemState* state) const
{
  const CMatchState* s = dynamic_cast<const CMatchState*>(state);
  return s->objToMatch.empty() && s->histToMatch.empty();
}

double cvtk::CMatchCost::evaluate(const CProblemState* state) const
{
  const CMatchState* s = dynamic_cast<const CMatchState*>(state);
  double cost=0.;
  vector<CMatch>::const_iterator i;  
  for (i = s->matches.begin(); i != s->matches.end(); i++) {
    cost+=(i->obj->center
	   -i->hist->history.front().center).distance();
  }

  return cost;  
}

double cvtk::CMatchHeuristic::evaluate(const CProblemState*) const
{
  return 0;
}

/************************* CMatchExpand ******************************/

void cvtk::CMatchExpand::expand(CProblemState* state,
				vector<CProblemState*>* exp) const
{
  CMatchState* s = dynamic_cast<CMatchState*>(state);
  if (s->objToMatch.empty() && s->histToMatch.empty()) // just to be
    return;                        // sure. can't occure if gt correct.
  if (s->objToMatch.empty()) {            // no further unmatched 
    CMatchState* e = new CMatchState(*s); // objects but histories
    while(!e->histToMatch.empty()) {     
      e->histUnMatched.push_back(e->histToMatch.back());
      e->histToMatch.pop_back();
    }
    exp->push_back(dynamic_cast<CProblemState*>(e));
    return;
  }

  CObject* obj = s->objToMatch.back();    // pop next object to match
  s->objToMatch.pop_back();
  if (s->objToMatch.size()+1>s->histToMatch.size()) {// more objects ->
    CMatchState* e = new CMatchState(*s); // this object may be left
    e->objUnMatched.push_back(obj);        // unmatched
    exp->push_back(dynamic_cast<CProblemState*>(e));
  }
  for(unsigned int i=0; i < s->histToMatch.size(); i++) {
    CObjectHistory* hist = &(*(s->histToMatch.front()));
    s->histToMatch.pop_front();           // remove matched history
    CMatchState* e = new CMatchState(*s); // to copy new node e from s
    s->histToMatch.push_back(hist);       // and put history back
    e->matches.push_back(CMatch(obj, hist)); // save match
    exp->push_back(e);
  }
}






