/*
 * Package:     cvtk - Computer Vision Tool Kit
 * File:        template.cpp
 * Description: Main of the template calibration tool. Sets up and 
 *              connects the model and the controller.
 * Created:     2001-08-04
 * Author:      Sascha Lange
 * Mail:        Sascha.Lange@Uni-Osnabrueck.De
 *
 * Web:         http://cvtk.sourceforge.net
 *
 * Copyright (C) 2001 Sascha Lange
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 
 *  02111-1307, USA.
 *
 * ------------------------- Revision History -------------------------
 * $Log: template.cpp,v $
 * Revision 1.5  2001/09/03 12:11:17  slange
 * finished benchmark
 *
 * Revision 1.4  2001/09/02 17:39:07  slange
 * cvs error!
 *
 * Revision 1.3  2001/09/01 17:40:09  slange
 * - working on monitor
 * - moved template calibration into namespace "cvtk"
 *
 * Revision 1.2  2001/08/04 16:43:07  slange
 * - added some mehtods needed by template calibration
 * - added mouseClick event receiver
 *
 * Revision 1.1  2001/08/04 12:03:09  slange
 * - initial checkin of templatecontroller.[h/cpp] and template.cpp
 * - finished GUI of the template calibration tool
 *
 * --------------------------------------------------------------------
 */

#include "webcam.h"
#include <iostream>
#include <fstream>
#include <unistd.h>
#include "imagewidget.h"
#include <qapplication.h>
#include "templatewidget.h"
#include "templatecontroller.h"
#include "cvtkexceptions.h"
#include "cvtk.h"
#include "lookuptable.h"
#include "colorsegmentation.h"
#include "chaincoding.h"

using cvtk::CPPM;
using cvtk::CRGBMap;
using cvtk::RGBValue;
using cvtk::CWebCam;
using cvtk::CLookUpTable;
using cvtk::CImageSource;
using cvtk::CImageWidget;
using cvtk::CTemplateWidget;
using cvtk::CController;
using cvtk::CColorSegmentation;
using cvtk::CDefaultChainCoding;
using cvtk::CBorderTracing4;

void printUsage(char* str)
{
  cout << "Usage: " << str << " lookuptable.lut\n";
}

CWebCam* cam;

void cleanup()
{
  //  cerr << "in cleanup" << endl;
  delete cam;
}

int main(int argc, char* argv[]) {
  if (argc!=2) {
    printUsage(argv[0]);
    exit(1);
  }


  std::ifstream in (argv[1]);
  if (!in) {
    cerr << "Cannot open lookup table" << argv[1];
    exit(1);
  }

  QApplication app(argc, argv);
  try {
    CLookUpTable<2>* lut = new CLookUpTable<2>(); // could be to big
                                               // to create on stack
    in >> (*lut);
    
    // register cam cleanup function 
    cam = new CWebCam(352,288);
    if (atexit(cleanup)!=0) {
      cerr << "Can't register cleanup function" << endl;
      delete cam;
      exit(0);
    }

    CTemplateWidget iw(0,0);
    CColorSegmentation colSeg(lut);
    CDefaultChainCoding regFinder(new CBorderTracing4(),
				  lut->getNumberOfClasses());

    CController con(cam, &colSeg, &regFinder);
    
    app.connect(&iw, SIGNAL(newCommand()), &con, SLOT(newTemplate()));
    app.connect(&iw, SIGNAL(grabCommand()), &con, SLOT(grabFrame()));
    app.connect(&iw, SIGNAL(mouseClick(QMouseEvent*)),
		&con, SLOT(mouseClick(QMouseEvent*)));
    app.connect(&iw, SIGNAL(saveCommand()),
		&con, SLOT(saveTemplates()));
    app.connect(&app, SIGNAL(lastWindowClosed()),
		&con, SLOT(stopExec()));
    
    app.connect(&con, SIGNAL(message(const char*, int)),
		&iw, SLOT(message(const char*, int)));

    app.connect(&con, SIGNAL(imageChanged(const CRGBMap&)),
		&iw, SLOT(setImage(const CRGBMap&)));
     
    iw.show();
    app.setMainWidget(&iw);
    app.exec();

  }
  catch(no_file& e) {
    app.exit(0);
  }
  catch(exception& e) {
    cerr << e.what() << endl;
    app.exit(1);
  }
}
