/*
 * Package:     cvtk - Computer Vision Tool Kit
 * File:        templatecontroller.cpp
 * Description: Implementation of the controller of the calibration tool.
 * Created:     2001-08-04
 * Author:      Sascha Lange
 * Mail:        Sascha.Lange@Uni-Osnabrueck.De
 *
 * Web:         http://cvtk.sourceforge.net
 *
 * Copyright (C) 2001 Sascha Lange
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 *
 * ------------------------- Revision History -------------------------
 * $Log: templatecontroller.cpp,v $
 * Revision 1.9  2001/09/04 15:31:27  slange
 * - corrected bugs in:
 *   - templatecontroller.cpp: use QString to emit messages
 *   - monitor.cpp: used x instead of y to check against height
 * - labeled the sliders in calibratewidget
 *
 * Revision 1.8  2001/09/01 17:40:09  slange
 * - working on monitor
 * - moved template calibration into namespace "cvtk"
 *
 * Revision 1.7  2001/08/04 20:14:20  slange
 * finished template calibration
 *
 * Revision 1.6  2001/08/04 20:01:29  slange
 * corrected bug: region list was to small
 *
 * Revision 1.5  2001/08/04 19:25:49  slange
 * - finished template controller
 *
 * Revision 1.4  2001/08/04 16:43:07  slange
 * - added some mehtods needed by template calibration
 * - added mouseClick event receiver
 *
 * Revision 1.3  2001/08/04 14:50:29  slange
 * added stream input functions to object templates
 *
 * Revision 1.2  2001/08/04 14:11:57  slange
 * working on the template calibration's functionallity
 *
 * Revision 1.1  2001/08/04 12:03:09  slange
 * - initial checkin of templatecontroller.[h/cpp] and template.cpp
 * - finished GUI of the template calibration tool
 *
 * --------------------------------------------------------------------
 */

#include "templatecontroller.h"
#include <iostream>
#include <qfiledialog.h>
#include "cvtkexceptions.h"
#include "rgbhistogram.h"
#include "lookuptable.h"
#include <unistd.h>
#include "monitors.h"

using std::invalid_argument;

cvtk::CController::CController(CImageSource* imageSource,
			       CSegmentationStrategy* segmentation,
			       CChainCodingStrategy* regionFinder) 
  : QObject(), imageSource(imageSource), segmentation(segmentation),
    regionFinder(regionFinder), segments(), templateList(), templ(),
    templateID(0), markType(0), grabbing(true)
{
  if (!imageSource) 
    throw invalid_argument("Null pointer to the image source.");
  if (!segmentation) 
    throw invalid_argument("Null pointer to the segmentation.");
  if (!regionFinder) 
    throw invalid_argument("Null pointer to the region finder.");

  regionList = CRegionList();
  ppm=CPPM(imageSource->getWidth(), imageSource->getHeight());
  templ.id=templateID++;

  emit message("Ready to grab a frame!", 4000);

  timer = new QTimer(this);
  connect(timer, SIGNAL(timeout()), this, SLOT(exec()));
  timer->start(0, false);
}

cvtk::CController::~CController()
{
  delete timer;
}

/* slots */
void cvtk::CController::newTemplate()
{
  templateList.push_back(templ);
  templ = CObjectTemplate();
  templ.id = templateID++;
  markType=0;
  emit message("Template added. Grab a frame.",2000);
}
void cvtk::CController::grabFrame()
{  
  if (grabbing) {
    grabbing = false;
    emit message("Frame grabbed. Click on center.", 2000);
  }
  else {
    grabbing = true;
  }
}
void cvtk::CController::mouseClick(QMouseEvent* ev) 
{
  if (grabbing) {
    emit message("Grab a frame before clicking on marks.", 2000);
  }
  else {
    Vektor<int, 2> click(ev->x(), ev->y());
    CRegion reg=regionList.findNearest(click);
    switch (markType++) {
    case 0:           // clicked on center
      center = reg.getCenter();
      templ.center.colClass = reg.regionClass;
      templ.center.area     = reg.getArea();
      templ.center.areaTol  = (templ.center.area*4)/10;
      emit message(QString("Center ")
		   +QString().setNum(templ.center.colClass)
		   +QString(" ")
		   +QString().setNum(templ.center.area)
		   +QString(" ")
		   +QString().setNum(templ.center.areaTol), 2000);
      break;
    case 1:           // clicked on axis
      templ.axis.colClass = reg.regionClass;
      templ.axis.area     = reg.getArea();
      templ.axis.areaTol  = (templ.axis.area*4)/10;
      templ.axis.distanceToCenter = static_cast<int>(
	(center-reg.getCenter()).distance());
      templ.axis.distanceTol=(templ.axis.distanceToCenter*4)/10;
      orientation = static_cast<int>((reg.getCenter()-center).angle());
      emit message(QString("Axis ")
		   +QString().setNum(templ.axis.colClass)
		   +QString(" ")
		   +QString().setNum(templ.axis.area)
		   +QString(" ")
		   +QString().setNum(templ.axis.areaTol)
		   +QString(" ")
		   +QString().setNum(templ.axis.distanceToCenter)
		   +QString(" ")
		   +QString().setNum(templ.axis.distanceTol), 2000);
      break;
    default:          // clicked on a mark
      CMark mark;
      mark.colClass = reg.regionClass;
      mark.area     = reg.getArea();
      mark.areaTol  = (mark.area*4)/10;
      mark.distanceToCenter = static_cast<int>((center-reg.getCenter())
					       .distance());
      mark.distanceTol=(mark.distanceToCenter*4)/10;      
      mark.directionTol= (90*4)/10;
      mark.direction=static_cast<int>((reg.getCenter()
				       -center).angle());
      mark.direction= (mark.direction > orientation) 
	? mark.direction-orientation 
	: 360+mark.direction-orientation;
      templ.marks.push_back(mark);     
      emit message(QString("Mark ")
		   +QString().setNum(mark.colClass)
		   +QString(" ")
		   +QString().setNum(mark.area)
		   +QString(" ")
		   +QString().setNum( mark.areaTol)
		   +QString(" ")
		   +QString().setNum(mark.distanceToCenter)
		   +QString(" ")
		   +QString().setNum(mark.distanceTol)
		   +QString(" ")
		   +QString().setNum(mark.direction)
		   +QString(" ")
		   +QString().setNum(mark.directionTol), 2000);
      break;
    }
  }
}
void cvtk::CController::saveTemplates()
{
  bool oldGrabState=grabbing;
  grabbing=false;
  QString filename=
    QFileDialog::getSaveFileName(0,0,0,"Save Templates");
  if (filename.isEmpty()) {
    emit message("Not saved. No file specified.", 2000);
  }
  else {
    ofstream out (filename);
    if(!out) {
      emit message("Not saved. Couldn't open file.", 2000);
    }
    else {
      out << templateList; 
      emit message("Templates saved to "+filename, 2400);
    }
  }
  grabbing=oldGrabState;
}
void cvtk::CController::stopExec()
{
  timer->stop();
}
void cvtk::CController::exec()
{
  if (grabbing) {
    regionList.clear();
    imageSource->getImage(ppm);
    segmentation->segment(&ppm, &segments);
    regionFinder->codeRegions(&segments, &regionList);
    ppm.lock();
    Monitors::visualizeRegionMap(segments, ppm);
    Monitors::visualizeRegionList(regionList, ppm);
    ppm.unlock();
  }

  emit imageChanged(ppm);                  // signal a changed image
}

#include "templatecontroller.moc"

