/*
 * Package:     cvtk - Computer Vision Tool Kit
 * File:        templatematching.cpp
 * Description: Implementation of the object-template matching
 *              algorithm 
 * Created:     2001-06-21
 * Author:      Sascha Lange
 * Mail:        Sascha.Lange@Uni-Osnabrueck.De
 *
 * Web:         http://cvtk.sourceforge.net
 *
 * Copyright (C) 2001 Sascha Lange
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  
 * 02111-1307, USA.
 *
 * ------------------------ Revision History --------------------------
 * $Log: templatematching.cpp,v $
 * Revision 1.11  2001/09/04 15:31:27  slange
 * - corrected bugs in:
 *   - templatecontroller.cpp: use QString to emit messages
 *   - monitor.cpp: used x instead of y to check against height
 * - labeled the sliders in calibratewidget
 *
 * Revision 1.10  2001/08/10 14:11:57  slange
 * working on the next build
 *
 * Revision 1.9  2001/08/09 23:51:08  slange
 * reorganised templatematching.[h/cpp]
 *
 * Revision 1.8  2001/07/16 20:02:34  slange
 * - working on the history matching
 *
 * Revision 1.7  2001/07/15 22:19:13  slange
 * - added pendel.cpp, file to demonstrate library.
 *
 * Revision 1.6  2001/07/07 21:46:18  slange
 * - working on template matching: complete!
 * - todo: implement delete and stop after first match!
 *
 * Revision 1.4  2001/07/06 10:58:29  slange
 * - working on template matching: completed axis matching
 *
 * Revision 1.3  2001/07/06 02:04:18  slange
 * - working on the template matching: completed center matching
 *
 * Revision 1.2  2001/07/04 23:00:04  uid51551
 * - working on the templatematching
 * - calculation of the region center
 *
 * Revision 1.1  2001/06/21 23:53:42  slange
 * - initial checkin of templatematching.[h/cpp]
 * - added method clear() to CRegionList
 *
 * --------------------------------------------------------------------
 */

#include "templatematching.h"
#include <cmath>

cvtk::CTemplateMatching::~CTemplateMatching()
{
  delete templates;
}

cvtk::CTemplateMatching::
CTemplateMatching(CObjectTemplateList* templates,
		  int flags)
  : templates(templates)
{
  if(!templates->size()) {
    cerr << "TemplateMatching: Empty template list." << endl;
    exit(1);
  }
  deleteAfterMatch = flags & DELETE_AFTER_MATCH;
  stopAfterMatch = flags & STOP_AFTER_MATCH;
}

bool cvtk::CTemplateMatching::matchMarks(const CObjectTemplate& t, 
					 CRegionList * rlist,
					 CObjectList* objList,
					 vector<int>& match,
					 unsigned int pos)
{
  int or=-1;   
  CRegion& center = (*rlist)[t.center.colClass][match[0]];
  if (t.axis.colClass > 0) {         // calculate orientation of axis
    CRegion& axis = (*rlist)[t.axis.colClass][match[1]];
    or=static_cast<int>((axis.getCenter()-center.getCenter()).angle());
  }
  if (pos == t.marks.size()) {       // found a match !
    objList->push_back(t.id, CObject(center.getCenter(), or , t.id));
    if (deleteAfterMatch) {          // delete matched regions
      (*rlist)[t.center.colClass]
	.erase((*rlist)[t.center.colClass].begin()+match[0]);
      if (t.axis.colClass>0)
	(*rlist)[t.axis.colClass]
	  .erase((*rlist)[t.axis.colClass].begin()+match[1]);
      for (unsigned int i=2; i < match.size(); i++)
	(*rlist)[t.marks[i-2].colClass]
	  .erase((*rlist)[t.marks[i-2].colClass].begin()+match[i]);
    }
#ifdef DEBUG
    cerr << "Matched template: " << t.id << ": " << center.getCenter()
	 << " orientation: " << or << endl;
#endif
    return true;
  }
  else {
    bool matched=false;
    for (unsigned int m=0;
	 m < (*rlist)[t.marks[pos].colClass].size();
	 m++) {
      CRegion& mark = (*rlist)[t.marks[pos].colClass][m];

      /* calculate angle between axis and |center mark| */
      int direction = static_cast<int>((mark.getCenter()
					-center.getCenter()).angle());
      direction = (direction > or) ? direction-or : 360+direction-or; 

      if ( (t.marks[pos].area == 0 ||
	    abs(t.marks[pos].area-mark.getArea()) 
	    <= t.marks[pos].areaTol)           // match area
	   && (abs(static_cast<int>
		  ((mark.getCenter()-center.getCenter()).distance())
		  -t.marks[pos].distanceToCenter) 
	       <=t.marks[pos].distanceTol)     // match distance
	   && (abs(direction-t.marks[pos].direction)
	       <=t.marks[pos].directionTol)) { // match angle
	match[pos+2]=m;
	matched|=matchMarks(t, rlist, objList, match, pos+1);
	if (stopAfterMatch && matched)
	  return true;
      }
    }
    return matched;
  }
}

bool cvtk::CTemplateMatching::matchAxis(const CObjectTemplate& t,
					CRegionList* rlist,
					CObjectList* objList,
					vector<int>& match)
{
  if (t.axis.colClass==0)
    return matchMarks(t, rlist, objList, match, 0);
  else {
    bool matched = false;
    CRegion& center = (*rlist)[t.center.colClass][match[0]];
    for (unsigned int a=0;
	 a < (*rlist)[t.axis.colClass].size();
	 a++) {
      CRegion& axis = (*rlist)[t.axis.colClass][a];
      
      if ( (t.axis.area == 0 ||
	    abs(t.axis.area-axis.getArea()) <= t.axis.areaTol)
	   && abs(static_cast<int>
		  ((axis.getCenter()-center.getCenter()).distance())
		  -t.axis.distanceToCenter) <= t.axis.distanceTol) {
	match[1] = a;
	matched |= matchMarks(t, rlist, objList, match, 0);
	if (stopAfterMatch && matched)  // stop after first match!
	  return true;
      }
    }
    return matched;
  }
}

bool cvtk::CTemplateMatching::matchCenter(const  CObjectTemplate& t,
					  CRegionList* rlist,
					  CObjectList* objList,
					  vector<int>& match)
{
  bool matched = false;
  for (unsigned int c=0;
       c < (*rlist)[t.center.colClass].size(); 
       c++) {                         // try every possible center
    CRegion& center = (*rlist)[t.center.colClass][c];
    if (t.center.area == 0 ||
	abs(t.center.area-center.getArea()) <= t.center.areaTol) {
#ifdef DEBUG
      cerr << "found matching center: " 
	   << t.center.colClass << endl;
#endif
      match[0] = c;
      matched |= matchAxis(t, rlist, objList, match);
      if (stopAfterMatch && matched)
	return true;
    }
  }
  return matched;
}

cvtk::CObjectList* 
cvtk::CTemplateMatching::findObjects(CRegionList* rlist)
{
  if (!rlist)
    return 0;
  return findObjects(rlist, new CObjectList(templates->size()));
}

cvtk::CObjectList* 
cvtk::CTemplateMatching::findObjects(CRegionList* rlist,
				     CObjectList* objList)
{
  if (rlist == 0) 
    return 0;
  if (objList == 0)
    return 0;
  if (objList->size() != templates->size())
    objList->resize(templates->size());

  for (CObjectTemplateList::const_iterator t = templates->begin(); 
       t != templates->end(); 
       t++) {                           // match every template
    vector<int> match(t->marks.size()+2);
#ifdef DEBUG
    cerr << "findObjects: template " << t->id << endl;
#endif
    matchCenter(*t, rlist, objList, match);// recursive matching of
  }                                        // template
  return objList;
}











