#ifndef _vektor_h_
#define _vektor_h_

/*
 * Package:     cvtk - Computer Vision Tool Kit
 * File:        vektor.h
 * Description: Template class of an n-dimensional vector with 
 *              fundamental vector operations
 *              arithmetics.
 * Created:     2001-03-28
 * Author:      Sascha Lange
 * Mail:        Sascha.Lange@Uni-Osnabrueck.De
 *
 * Web:         http://www.cvtk.sourceforge.net
 *
 * Copyright (C) 2001 Sascha Lange
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 *
 * ------------------------ Revision History --------------------------
 * $Log: vektor.h,v $
 * Revision 1.8  2001/08/07 12:40:14  slange
 * - still refactoring
 * - initial checkin of defaultfactory.cpp, objectlist.h region.h
 *
 * Revision 1.7  2001/07/07 11:47:38  slange
 * - added angle to Vektor
 *
 * Revision 1.6  2001/07/04 23:00:04  uid51551
 * - working on the templatematching
 * - calculation of the region center
 *
 * Revision 1.5  2001/04/28 18:05:45  slange
 * - removed patterngenerato.[h/cpp]
 * - added abstract interface CReinforcedLearning
 * - added CKMSegmentation
 *   - class definition and part of implementation
 * - added "train" button to calibratorwidget
 * - changed calibratecontroller to work with the new classes
 *
 * Revision 1.4  2001/04/27 20:51:16  slange
 * - finished implementation of CHistogram2Dim
 * - bug fixes
 *
 * Revision 1.3  2001/04/27 16:57:07  slange
 * - added 1- and 2-dimensional histograms
 * - changed typedef CRGBMap to a class definition
 * - added histogram functions to CRGBMap
 *
 * Revision 1.2  2001/04/23 18:42:52  slange
 * added copyright notice to all source files
 *
 * --------------------------------------------------------------------
 */

#include <stdarg.h>
#include <vector>
#include <string.h>
#include <cmath>

/************************** Vektor signature *************************/

/* Declaration of a basic mathematical vector. Initialize with type
 * and dimension.
 */
template <class T, int DIMENSION>
class Vektor
{
public:
  Vektor();                     // construct the 0 - vector
  Vektor(T first, ...);         // specify the entries of the vector
  explicit Vektor(T* a);        // construct vector from an array
  ~Vektor();                    // destructor

  // default copy - constructor is suitable (for the moment)

  T& operator[] (int i);        // unsecure element access
  T& operator[] (int i) const;

  Vektor& operator==(const Vektor& comp); // equal operator

  Vektor& operator-= (const Vektor& subtr);// basic vector operations
  Vektor operator- (const Vektor& subtr) const;
  Vektor& operator+= (const Vektor& sum);
  Vektor operator+ (const Vektor& sum) const;
  Vektor& operator*= (const T skalar);
  Vektor operator* (const T skalar) const;

  T distanceSquare();           // euklidian norm
  double distance();            // euklidian norm ^2 (faster)
  double angle();               // angle to x-axis (counterclockwise)

private:
  T vek[DIMENSION];
};

/* some often used vectors
 */
typedef Vektor<double, 1> Vektor1Dim;
typedef Vektor<double, 2> Vektor2Dim;
typedef Vektor<double, 3> Vektor3Dim;

/************************* inline functions **************************/

/* The class is totally inlined, because it's too basic to realise
 * member access, construction, etc. with function calls!
 */

template <class T, int DIMENSION>
inline Vektor<T, DIMENSION>::Vektor() {
  memset(vek, 0, sizeof(vek[0]) * DIMENSION); // set every entry to 0
}

template <class T, int DIMENSION>
inline Vektor<T, DIMENSION>::Vektor(T first, ...) 
{
  vek[0] = first;
  va_list ap;                   // read the argument list      
  va_start(ap, first);
  for (int i=1; i < DIMENSION; i++) 
    vek[i] = va_arg(ap, T);     // that's not type save!!!
  va_end(ap);      
}

template <class T, int DIMENSION>
inline Vektor<T, DIMENSION>::Vektor(T* a)
{
  memcpy(v, vek, sizeof(T) * DIMENSION); // copy the argument into vek
}

template <class T, int DIMENSION>
inline Vektor<T, DIMENSION>::~Vektor() // empty (for the moment)
{}

template <class T, int DIMENSION>
inline T& Vektor<T, DIMENSION>::operator[] (int i) 
{
  return vek[i]; 
}

template <class T, int DIMENSION>
inline T& Vektor<T, DIMENSION>::operator[] (int i) const 
{ 
  return (*const_cast<Vektor<T,DIMENSION>*>(this))[i];
}

template <class T, int DIMENSION> 
inline Vektor<T, DIMENSION>& 
Vektor<T, DIMENSION>::operator==(const Vektor& comp)
{
  for (int i=0; i < DIMENSION; ++i) // compare every element
    if (vek[i]!=comp[i]) return false;
  return true;
}

template <class T, int DIMENSION>
inline Vektor<T, DIMENSION>& 
Vektor<T, DIMENSION>::operator-= (const Vektor& subtr)
{
  for (int i=0; i < DIMENSION; i++)
    vek[i]-=subtr[i];
  return *this;
}
template <class T, int DIMENSION>
inline Vektor<T, DIMENSION> 
Vektor<T, DIMENSION>::operator- (const Vektor& subtr) const
{
  Vektor r=*this;
  return r-=subtr;
}

template <class T, int DIMENSION>
inline Vektor<T, DIMENSION>& 
Vektor<T, DIMENSION>::operator+= (const Vektor& summ)
{
  for (int i=0; i < DIMENSION; i++)
    vek[i]+=summ[i];
  return *this;
}
template <class T, int DIMENSION>
inline Vektor<T, DIMENSION> 
Vektor<T, DIMENSION>:: operator+ (const Vektor& sum) const
{
  Vektor r=*this;
  return r+=sum;
}

template <class T, int DIMENSION>
inline Vektor<T, DIMENSION>& 
Vektor<T, DIMENSION>::operator*= (const T skalar)
{
  for (int i=0; i <DIMENSION; i++)
    vek[i]*=skalar;
  return *this;
}
template <class T, int DIMENSION>
inline Vektor<T, DIMENSION> 
Vektor<T, DIMENSION>::operator* (const T skalar) const
{
  Vektor r=*this;
  return r*=skalar;
}

template <class T, int DIMENSION>
inline T Vektor<T, DIMENSION>::distanceSquare()
{
  T buf=0;
  for (int i=0; i < DIMENSION; i++) // calculate x1^2+x2^2+...
    buf += vek[i]*vek[i];                           
  return buf;
}
template <class T, int DIMENSION>
inline double Vektor<T, DIMENSION>::distance()
{
  return sqrt(static_cast<double>(distanceSquare()));
};

template <class T, int DIMENSION>
inline double Vektor<T, DIMENSION>::angle()
{
  if (DIMENSION != 2) {
    cerr << "Vektor.angle(): only implemented for 2 dimensions" 
	 << endl;
    return 0.;
  }
  double angle = 360*asin(vek[0]/distance())/(2*M_PI); // in degrees
  return (vek[1]<0) ? 270+angle : 90-angle;  // caused by the libraries
};                                           // implemetation of asin


/* simple stream in- and output operators
 */ 
template <class T, int DIMENSION>
inline istream& operator>> (istream& in, Vektor<T, DIMENSION>& v)
{
  for (int i=0; i<DIMENSION; i++)
    in >> v[i];
  return in;
};
template <class T, int DIMENSION>
inline ostream& operator<< (ostream& out, 
			    const Vektor<T, DIMENSION>& v)
{
  for (int i=0; i<DIMENSION; i++)
    out << v[i] << " ";
  return out;
};

#endif    /* _vektor_h_ */








