/*
 *  Program: binarize
 *
 *  Thresholds pixel values to produce a bit image.
 */

/* From the Vista library: */
#include <vista/Vlib.h>
#include <vista/file.h>
#include <vista/mu.h>
#include <vista/option.h>
#include <vista/VImage.h>

/* From the standard C library: */
#include <stdio.h>

/* Later in this file: */
extern VImage BinarizeImage ();

int main (argc, argv)
    int argc;
    char *argv[];
{
    /* Command line options: */
    static VDouble threshold = 0.0;
    static VOptionDescRec options[] = {
	{ "threshold", VDoubleRepn, 1, (VPointer) & threshold,
	      VOptionalOpt, NULL, "Value at which pixel is set to 1" }
    };

    FILE *in_file, *out_file;
    VAttrList list;
    VAttrListPosn posn;
    VImage src, result;
    int nimages = 0;

    /* Parse command line arguments and identify files: */
    VParseFilterCmd (VNumber (options), options, argc, argv,
		     & in_file, & out_file);

    /* Read the input file: */
    list = VReadFile (in_file, NULL);
    if (! list)
	exit (1);

    /* For each attribute read... */
    for (VFirstAttr (list, & posn); VAttrExists (& posn); VNextAttr (& posn)) {

	/* If it's an image, replace it with a thresholded version: */
	if (VGetAttrRepn (& posn) != VImageRepn)
	    continue;
	VGetAttrValue (& posn, NULL, VImageRepn, & src);
	result = BinarizeImage (src, NULL, VAllBands, threshold);
	if (! result)
	    exit (1);
	VSetAttrValue (& posn, NULL, VImageRepn, result);
	VDestroyImage (src);
	nimages++;
    }

    /* Write out the results: */
    if (! VWriteFile (out_file, list))
	exit (1);

    fprintf (stderr, "%s: Thresholded %d image(s).\n", argv[0], nimages);
    return 0;
}


/*
 *  BinarizeImage
 *
 *  Convert pixel values to 0 or 1 according to whether they are less
 *  than a given threshold.
 */

VImage BinarizeImage (src, dest, band, threshold)
    VImage src;
    VImage dest;
    VBand band;
    double threshold;
{
    int nbands, npixels, i;
    VPointer src_pixels;
    VBit *dest_pixels;

    /* Locate the block of source image pixels: */
    if (! VSelectBand ("BinarizeImage", src, band, & npixels, & src_pixels))
	return NULL;

    /* Grumble about a threshold outside the source pixel range: */
    if (threshold < VPixelMinValue (src) || threshold > VPixelMaxValue (src))
	VWarning (
    "BinarizeImage: Threshold (%g) is outside source pixel range [%g,%g]",
		  threshold, VPixelMaxValue (src), VPixelMaxValue (src));

    /* Check the destination image, which should have VBit pixels.
       If it is NULL, then create one of the appropriate size and type. */
    nbands = (band == VAllBands) ? VImageNBands (src) : 1;
    dest = VSelectDestImage ("BinarizeImage", dest, nbands,
			     VImageNRows (src), VImageNColumns (src),
			     VBitRepn);
    if (! dest)
	return NULL;

    /* Locate the block of destination image pixels: */
    dest_pixels = & VPixel (dest, 0, 0, 0, VBit);

    /* Define a macro, parameterized by source pixel type,
       that can be instantiated for each possible type: */
#define Binarize(type)						\
    {								\
	type *src_pp = (type *) src_pixels;			\
	type t = threshold;					\
								\
	for (i = 0; i < npixels; i++)				\
	*dest_pixels++ = (*src_pp++ >= t);			\
    }

    /* Instantiate the macro once for each source pixel type: */
    switch (VPixelRepn (src)) {
    case VBitRepn:	Binarize (VBit);	break;
    case VUByteRepn:	Binarize (VUByte);	break;
    case VSByteRepn:	Binarize (VSByte);	break;
    case VShortRepn:	Binarize (VShort);	break;
    case VLongRepn:	Binarize (VLong);	break;
    case VFloatRepn:	Binarize (VFloat);	break;
    case VDoubleRepn:	Binarize (VDouble);	break;
    }

#undef Binarize

    /* Let the destination inherit any attributes of the source image: */
    VCopyImageAttrs (src, dest);
    return dest;
}
