/*
 *  Program: show
 *
 *  Display an image under X Windows.
 */

/* From the Vista library: */
#include <vista/VX.h>
#include <vista/file.h>
#include <vista/option.h>
#include <vista/VImage.h>

/* From the standard C library: */
#include <stdio.h>

/* Later in this file: */
extern void OpenCallback ();
extern void QuitCallback ();
extern void InputCallback ();
extern VBoolean DisplayImage ();

/* The image currently displayed: */
VImage image = NULL;


/*
 *  main
 *
 *  Program entry point.
 */

int main (argc, argv)
    int argc;
    char *argv[];
{
    /* Default resource settings: */
    static VStringConst resources[] = {
	"*messageAreaNLines: 2",
	"*fileBox.textColumns: 50",
	"*fileBox.pattern: *.v",
	"*imageView.cursor: crosshair",
	NULL
    };

    FILE *in_file;

    /* Parse command line options and open the input file: */
    VXInit ("Show", resources, & argc, argv);
    VParseFilterCmd (0, NULL, argc, argv, & in_file, NULL);

    /* Read and display an image from the input file: */
    if (! DisplayImage (in_file))
	exit (1);	/* no image could be found */
    fclose (in_file);

    /* Create the File menu: */
    VXAddMenu ("File",
	       "Open...", "Ctrl<Key>O", OpenCallback, NULL,
	       "Quit", "Ctrl<Key>Q", QuitCallback, NULL,
	       NULL);

    /* Add a hander for mouse clicks on the image: */
    VXAddInputCallback (VXIbuttonPress, InputCallback, NULL);

    /* Process events: */
    VXAppMainLoop ();
}


/*
 *  OpenCallback
 *
 *  Callback for Open menu item.
 */

/*ARGSUSED*/
void OpenCallback (client_data)
    VPointer client_data;
{
    VStringConst filename;
    FILE *in_file;

    /* Pop up a dialog box to get the new file's name: */
    filename = VXPopupFileBox ("Open Image File");
    if (! filename)
	return;		/* Cancel button was clicked */

    /* Open it and display the image it contains: */
    in_file = fopen (filename, "r");
    if (in_file) {
	DisplayImage (in_file);
	fclose (in_file);
    } else VWarning ("Unable to open %s.", filename);
    VFree ((VPointer) filename);
}


/*
 *  QuitCallback
 *
 *  Callback for Quit menu item.
 */

/*ARGSUSED*/
void QuitCallback (client_data)
    VPointer client_data;
{
    exit (0);
}


/*
 *  InputCallback
 *
 *  Respond to a mouse click in the image.
 */

void InputCallback (input_data, client_data)
    VXInputData input_data;
    VPointer client_data;
{
    VDouble pixel;

    /* Only respond to the left button: */
    if (input_data->value != 1)
	return;

    /* Report the pixel clicked upon: */
    pixel = VGetPixel (image, 0, input_data->row, input_data->column);
    VXDisplayMessage (TRUE, "Row %d, column %d: %g",
		      input_data->row, input_data->column, pixel);

    /* Outline the pixel: */
    VXClearLines ();
    VXDrawLine (input_data->row - 0.5, input_data->column - 0.5,
		input_data->row - 0.5, input_data->column + 1.5);
    VXDrawLine (input_data->row - 0.5, input_data->column + 1.5,
		input_data->row + 1.5, input_data->column + 1.5);
    VXDrawLine (input_data->row + 1.5, input_data->column + 1.5,
		input_data->row + 1.5, input_data->column - 0.5);
    VXDrawLine (input_data->row + 1.5, input_data->column - 0.5,
		input_data->row - 0.5, input_data->column - 0.5);
}


/*
 *  DisplayImage
 *
 *  Load and display a new image.
 */

VBoolean DisplayImage (in_file)
    FILE *in_file;
{
    VAttrList list;
    VImage *images;
    int nimages, i;

    /* Read the image from the specified file: */
    nimages = VReadImages (in_file, & list, & images);
    if (nimages == 0)
	return FALSE;	    /* no image was found */

    /* Discard unneeded stuff: */
    VDestroyAttrList (list);
    if (nimages > 1) {
	VWarning ("Showing only the first of %d images", nimages);
	for (i = 1; i < nimages; i++)
	    VDestroyImage (images[i]);
    }

    /* Note the new image while discarding any previous one: */
    if (image)
	VDestroyImage (image);
    image = images[0];
    VFree ((VPointer) images);

    /* Display the new image: */
    VXClearLines ();
    VXDisplayMessage (TRUE, "");
    VXSetImage (image, 0, 1.0, 0, 0);

    return TRUE;
}
