/***************************************************************************
* Filename: histogram-extrema-thresholding.c 
* Copyright (C) 2000-2001 Daniel X. Pape. 
* 
* This operator thresholds an image's gray value scale into several
* intervals, where each interval define homogenous image regions. This
* is done based on an evaluation of the histogram, and attempts to use
* the histogram minima as the threshold values.
* 
***************************************************************************
* Revision History: 
* Wednesday, January 19, 2000 - Original 
***************************************************************************
* This program is free software; you can redistribute it and/or modify 
* it under the terms of the GNU General Public License as published by 
* the Free Software Foundation; either version 2 of the License, or 
* (at your option) any later version. 
* 
* This program is distributed in the hope that it will be useful, 
* but WITHOUT ANY WARRANTY; without even the implied warranty of 
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the 
* GNU General Public License for more details. 
* 
* You should have received a copy of the GNU General Public License 
* along with this program; if not, write to the Free Software 
* Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA 
***************************************************************************/


#include "ipxPluginInterface.h"

#include "libgimp/gimp.h"


#include "localOperator.h"
#include "ipxUtils.h"

#include <math.h>
#include <stdlib.h>

#define IPX_PLUGIN_DATE "v0.3.2 - July 09, 2001"

#ifndef NDEBUG
#include <stdio.h>
#endif

static gint G, B, b, Tmax, T;
static guchar *C, *NC, *PC;
static gchar *R;


GimpPlugInInfo PLUG_IN_INFO =
{
  NULL,    /* init_proc */
  NULL,    /* quit_proc */
  query,   /* query_proc */
  run,     /* run_proc */
};

void
query ()
{
  static GimpParamDef args[] =
  {
    { GIMP_PDB_INT32, "run_mode", "Interactive, non-interactive" },
    { GIMP_PDB_IMAGE, "image", "Input image (unused)" },
    { GIMP_PDB_DRAWABLE, "drawable", "Input drawable" },
  };

  static GimpParamDef *return_vals = NULL;
  static int nargs = sizeof (args) / sizeof (args[0]);
  static int nreturn_vals = 0;

  gimp_install_procedure ("plug_in_ipx-histogram-extrema-thresholding",
			  "Histogram Extrema-thresholding",
			  "This plugin calculates ...",
			  "Daniel X. Pape",
			  "Daniel X. Pape",
              IPX_PLUGIN_DATE,
			  "<Image>/Filters/IPX/Histogram-Extrema-Threshold",
			  "RGB*, GRAY*",
			  GIMP_PLUGIN,
			  nargs, nreturn_vals,
			  args, return_vals);
}



void
run (gchar *name, gint nparams, GimpParam *param, gint *nreturn_vals,
GimpParam **return_vals)
{
    
    static GimpParam values[1];
    GimpDrawable *drawable;
    GimpRunModeType run_mode;
    GimpPDBStatusType status = GIMP_PDB_SUCCESS;
    
    run_mode = param[0].data.d_int32;
    
    *nreturn_vals = 1;
    *return_vals = values;
    
    values[0].type = GIMP_PDB_STATUS;
    values[0].data.d_status = status;
    
    switch (run_mode)
    {
        case GIMP_RUN_INTERACTIVE:
                /*  Possibly retrieve data  */
/*             gimp_get_data ("plug_in_ipx-extreme-value", &bvals); */
            
                /*  First acquire information with a dialog  */
/*             if (! extreme-value_dialog ()) */
/*                 return; */
            break;
            
        case GIMP_RUN_NONINTERACTIVE:
                /*  Make sure all the arguments are there!  */
            if (nparams != 3)
                status = GIMP_PDB_CALLING_ERROR;
            if (status == GIMP_PDB_SUCCESS)
            {
            }
            break;
            
        case GIMP_RUN_WITH_LAST_VALS:
                /*  Possibly retrieve data  */
/*             gimp_get_data ("plug_in_ipx-extreme-value", &bvals); */
            break;
            
        default:
            break;
    }
    
    
        /*  Get the specified drawable  */
    drawable = gimp_drawable_get (param[2].data.d_drawable);
    
        /*  Make sure that the drawable is gray or RGB color  */
    if (gimp_drawable_is_rgb (drawable->id) ||
        gimp_drawable_is_gray (drawable->id))
    {
        gimp_tile_cache_ntiles(2*(drawable->width / gimp_tile_width() + 1));

        doPointProcessing(drawable, 2);
        
        if (run_mode != GIMP_RUN_NONINTERACTIVE)
            gimp_displays_flush ();
        
        
            /*  Store data  */
/*         if (run_mode == GIMP_RUN_INTERACTIVE) */
/*             gimp_set_data ("plug_in_ipx-extreme-value", &bvals, sizeof (bvals));     */
    }
    else
    {
        status = GIMP_PDB_EXECUTION_ERROR;
    }
    
    gimp_drawable_detach (drawable);
        
    values[0].data.d_status = status;
}

void initOperator(void)
{
#ifndef NDEBUG
    printf("initOperator: begin\n");
#endif
    
    G = 256; /* num gray values */
    B = 9; /* must be odd */ /* this will be a GUI slider... */
    b = (B-1)/2;
    Tmax = G; /* 2 <= Tmax <= G */
    NC = (guchar *) calloc(Tmax+1, sizeof(guchar));
    NC[Tmax] = G-1;
    PC = (guchar *) calloc(Tmax, sizeof(guchar));
    C = (guchar *) calloc(G, sizeof(guchar));
    R = (gchar *) calloc(G, sizeof(gchar));

#ifndef NDEBUG
    printf("initOperator: end\n");
#endif
}

void cleanupOperator(void)
{
#ifndef NDEBUG
    printf("cleanupOperator: begin\n");
#endif
    
    free(NC);
    free(PC);
    free(C);
    free(R);

#ifndef NDEBUG
    printf("cleanupOperator: end\n");
#endif
}


/* ************************************************************************ */
void preProcessPointPass(GimpPixelRgn *srcPR, gint currentPassNumber)
{
#ifndef NDEBUG
    printf("preProcessPointPass: begin\n");
#endif

    if(currentPassNumber == 1) /* second pass */
    {
        gint z, s, t;

        for(z=1; z<G; z++)
            C[z] += C[z-1];

        for(z=0; z<G; z++)
        {
            if(z <= b || z >= G-b)
                R[z] = 0;
            else
            {
                gint CB = 0;
                for(t=z-b; t<z+b; t++)
                    CB += C[t];
                R[z] = C[z] - (CB/B);
            }
        }

        NC[0] = 0;
        s = 1;
        t = 0;
        
        for(z=b+1; z<G-b; z++)
        {
            if(R[z] < 0 && R[z-1] >= 0)
            {
                NC[s++] = z-1;
            }
            else if(R[z] > 0 && R[z-1] <= 0)
            {
                PC[t++] = z-1;
                if(t > Tmax)
                {
#ifndef NDEBUG
                    printf ("uhoh - shouldn't have taken this long.\n");
#endif
                    exit(1); /* problem? */
                }
            }
        }

        NC[s] = G; T = t-1;
#ifndef NDEBUG
        printf("T=%d\n", T);
#endif
    }
    
#ifndef NDEBUG
    printf("preProcessPointPass: end\n");
#endif
}


/* ************************************************************************ */
guchar processPoint(guchar point, gint currentPassNumber)
{
    if(currentPassNumber == 0) /* first pass */
    {
        /* here we are just computing the histogram */
        C[point]++;
        return point;
    }
    else if(currentPassNumber == 1) /* second pass */
    {
        gint s;
        for(s=0; s<T; s++)
            if(point >= NC[s] && point < NC[s+1])
                return PC[s];
    }

        // shouldn't get here...
    return point;
}



/* ******************************** unused ******************************** */

void preProcessWindowPass(GimpPixelRgn *srcPR, gint currentPassNumber) {}
guchar processWindow(guchar **window, gint currentPassNumber) {return 0;}
void postProcessWindowPass(GimpPixelRgn *srcPR, gint currentPassNumber) {}

void postProcessPointPass(GimpPixelRgn *srcPR, gint currentPassNumber)
{
#ifndef NDEBUG
    printf("postProcessPointPass: begin\n");
#endif
#ifndef NDEBUG
    printf("postProcessPointPass: end\n");
#endif
}
