/***************************************************************************
* Filename: localOperator.c 
* Copyright (C) 2000-2001 Daniel X. Pape. 
* 
* This file contains the common functions that the plugins call in order
* to support the "doProcessing" framework.
* 
***************************************************************************
* Revision History: 
* Thursday, July 16, 1998 - Original 
***************************************************************************
* This program is free software; you can redistribute it and/or modify 
* it under the terms of the GNU General Public License as published by 
* the Free Software Foundation; either version 2 of the License, or 
* (at your option) any later version. 
* 
* This program is distributed in the hope that it will be useful, 
* but WITHOUT ANY WARRANTY; without even the implied warranty of 
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the 
* GNU General Public License for more details. 
* 
* You should have received a copy of the GNU General Public License 
* along with this program; if not, write to the Free Software 
* Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA 
***************************************************************************/


#include "localOperator.h"
#include "localOperatorP.h"

#include "ipxPluginInterface.h"

#include "libgimp/gimp.h"

#include <stdlib.h>
#include <math.h>

#include <sys/types.h>
#include <signal.h>
#include <unistd.h>

#include <stdio.h>
#include <assert.h>


/* global variables that some plugins need */
gint windowSize, halfWindowSize, numPixelsInWindow;

/* globals that localOperator (and some plugins) uses */

gint gWidth, gHeight;
gint gBytesPerPixel;

gint gX1, gY1, gX2, gY2;
gint gMaskWidth,gMaskHeight; 


/* ************************************************************************ */
void setWindowSize(gint ws)
{
    windowSize = ws;                        /* eg: 3 for a 3x3 pixel window */
    halfWindowSize = (int)(windowSize / 2); /* eg: 1 for a 3x3 pixel window */
    numPixelsInWindow = windowSize * windowSize;
}
    

/* ************************************************************************ */
void
prepare_row_windowed(GimpPixelRgn *pixel_rgn, guchar *data,
                     int x, int y, int w)
{
    gint b;

        /* skip leading edge pixel */
    guchar *d = data + (halfWindowSize * pixel_rgn->bpp); 

    assert(pixel_rgn != NULL && data != NULL && w != 0);

    if (y < 0)
    {
        gimp_pixel_rgn_get_row (pixel_rgn, d, x, 0, w);
    }
    else if (y >= (int)pixel_rgn->h)
    {
        gimp_pixel_rgn_get_row (pixel_rgn, d, x, (pixel_rgn->h - 1), w);
    }
    else
    {
        gimp_pixel_rgn_get_row (pixel_rgn, d, x, y, w);
    }
    
        /*  Fill in edge pixels  */
    for (b = 0; b < (gint)pixel_rgn->bpp; b++)
    {
        d[b - (int)pixel_rgn->bpp] = d[b];
        d[w * pixel_rgn->bpp + b] = d[(w - 1) * pixel_rgn->bpp + b];
    }

}

void
prepare_row (GimpPixelRgn *pixel_rgn, guchar *data,
             int x, int y, int w)
{
    guchar *d = data;

    assert(pixel_rgn != NULL && data != NULL && w != 0);

    if (y < 0)
    {
        gimp_pixel_rgn_get_row (pixel_rgn, d, x, 0, w);
    }
    else if (y >= (int)pixel_rgn->h)
    {
        gimp_pixel_rgn_get_row (pixel_rgn, d, x, (pixel_rgn->h - 1), w);
    }
    else
    {
        gimp_pixel_rgn_get_row (pixel_rgn, d, x, y, w);
    }
    
}


/* ************************************************************************ */
guchar ** initRowBuffers(void)
{
    int i;
    guchar **rows;

    rows = (guchar **) malloc (windowSize * sizeof(guchar *));
    assert(rows != NULL);

    for(i=0; i<windowSize; i++)
    {
        rows[i] = (guchar *) malloc((gMaskWidth + 2*halfWindowSize) *
                                    sizeof(guchar)*gBytesPerPixel);
        assert(rows[i] != NULL);
    }
    return rows;
}


/* ************************************************************************ */
guchar ** initWindowBuffer(void)
{
    int i;
    guchar **window;

    window = (guchar **) malloc(windowSize * sizeof(guchar *));
    assert(window != NULL);

    for(i=0; i<windowSize; i++)
    {
        window[i] = (guchar *) malloc(windowSize *
                                      sizeof(guchar) * gBytesPerPixel);
        assert(window[i] != NULL);
    }
    return window;
}


/* ************************************************************************ */
void initPixelRegions(GimpDrawable *drawable,
                      GimpPixelRgn *srcPR,
                      GimpPixelRgn *tmpPR,
                      GimpPixelRgn *dstPR, int numPasses)
{
    gimp_pixel_rgn_init(srcPR, drawable, gX1,gY1, gX2, gY2, FALSE,FALSE);
    gimp_pixel_rgn_init(dstPR, drawable, gX1,gY1, gX2, gY2, TRUE,TRUE);
    if(numPasses > 1)
        gimp_pixel_rgn_init(tmpPR, drawable, gX1,gY1, gX2, gY2, FALSE,FALSE);

            /* the tmpPR "dirty" and "shadow" arguments used to be
           FALSE,TRUE, but I found that that would sometimes cause
           pixels around the edge of a region to be messed up. I changed
           it to FALSE,FALSE (against the recommendation in Kevin
           Turner's "Writing a GIMP Plug-In" guide), and it seems to
           work fine now. Except that you can't Undo properly...*/
}


/* ************************************************************************ */
gint fillRowBuffers(guchar **rows, GimpPixelRgn *currentSrcPR)
{
    gint i;
    gint a = -halfWindowSize;

    assert(rows != NULL);
    assert(currentSrcPR != NULL);
    
    for(i=0; i<windowSize; i++)
    {
        prepare_row_windowed(currentSrcPR, rows[i], gX1,gY1+(a++), gMaskWidth);
    }   
    return (gY1 + (a-1));
}


/* ************************************************************************ */
gint updateRowBuffers(guchar **rows, GimpPixelRgn *currentSrcPR,
                      gint last_row_retrieved)
{
    guchar *tmp;
    gint i;
    
        /*  shuffle the row pointers  */
    tmp = rows[0];
    for(i=1; i<windowSize; i++)
        rows[i-1] = rows[i];        
    rows[i-1] = tmp;
    
        /*  prepare the next row  */
    prepare_row_windowed(currentSrcPR, rows[i-1], gX1,
                         ++last_row_retrieved, gMaskWidth);

    return last_row_retrieved;
}


/* ************************************************************************ */
void getWindowFromRows(gint col, guchar **window, guchar **rows)
{
    gint i, j, a;
    
    for(i=0; i<windowSize; i++)
    {
        a = -halfWindowSize;
        for(j=0; j<windowSize; j++)
        {
            assert(col + (a)*gBytesPerPixel +
                   (halfWindowSize*gBytesPerPixel) <=
                   ((gMaskWidth + 2*halfWindowSize) *
                   sizeof(guchar)*gBytesPerPixel) - 1);

            window[i][j] =
                rows[i][
                    col + /* which column we are on */
                    (a++)*gBytesPerPixel + /* which channel we want */
                    (halfWindowSize*gBytesPerPixel) /* skip edge pixel */
                ];
        }
    }
}


/* ************************************************************************ */
void doPointProcessing(GimpDrawable *drawable, int numPasses)
{
    GimpPixelRgn srcPR, tmpPR, dstPR;
    guchar new_pixel_value;
//    gint alpha;

    guchar *src_row_buf, *dst_row_buf, *d;
                
    gint row_index, col_index, currentPass;

#if 0
    printf("Waiting... (pid %d)\n", getpid());
    kill(getpid(), SIGSTOP);
#endif

        /* Get the input area. This is the bounding box of the selection
         * in the image (or the entire image if there is no
         * selection). */
    
    gimp_drawable_mask_bounds (drawable->id, &gX1, &gY1, &gX2, &gY2);
    gMaskWidth = gX2-gX1;
    gMaskHeight = gY2-gY1;

    assert(gMaskWidth > 0 && gMaskHeight > 0);
        
    gimp_progress_init ("Processing...");
    
    gWidth = drawable->width;
    gHeight = drawable->height;
    gBytesPerPixel = drawable->bpp;
//    alpha = gimp_drawable_has_alpha (drawable->id);

        /* call specific plugin code */
    initOperator();
    
        /*  allocate row buffers and window array */
    src_row_buf = dst_row_buf = NULL;
    src_row_buf = (guchar *) malloc(gMaskWidth*sizeof(guchar)*gBytesPerPixel);
    assert(src_row_buf != NULL);
    dst_row_buf = (guchar *) malloc(gMaskWidth*sizeof(guchar)*gBytesPerPixel);
    assert(dst_row_buf != NULL);

        /*  initialize the pixel regions  */
    initPixelRegions(drawable, &srcPR, &tmpPR, &dstPR, numPasses);
    
    for(currentPass = 0; currentPass < numPasses; currentPass++)
    {
        GimpPixelRgn *currentSrcPR, *currentDstPR;

        /* setup pixel regions */
        
        if(currentPass == 0)
            currentSrcPR = &srcPR;
        else
            currentSrcPR = &tmpPR;

        if(numPasses > 1 && currentPass < numPasses - 1)
            currentDstPR = &tmpPR; 
        else
            currentDstPR = &dstPR;

            /* call specific plugin code */
        preProcessPointPass(currentSrcPR, currentPass);
        
            /* loop through the rows  */
        for (row_index = gY1; row_index < gY2; row_index++)
        {
            prepare_row(currentSrcPR, src_row_buf, gX1, row_index, gMaskWidth);

            d = dst_row_buf;

                /* loop over pixels in this row */
            for (col_index = 0; col_index < gMaskWidth * gBytesPerPixel;
                 col_index++)
            {
                    /* call specific plugin code */
                new_pixel_value =
                    processPoint(src_row_buf[col_index], currentPass);
                *d++ = new_pixel_value;
            }
        
                /*  store the dst row */
            gimp_pixel_rgn_set_row(currentDstPR, dst_row_buf,
                                   gX1, row_index, gMaskWidth);

            if ((row_index % 5) == 0)
                gimp_progress_update (((double)row_index + (currentPass*gMaskHeight)) /
                                      ((double)gMaskHeight * numPasses) );
        }

            /* call specific plugin code */
        postProcessPointPass(currentSrcPR, currentPass);
                
    }

        /*  update the region  */
    gimp_drawable_flush (drawable);
    gimp_drawable_merge_shadow (drawable->id, TRUE);
    gimp_drawable_update (drawable->id, gX1, gY1, gMaskWidth, gMaskHeight);


    /* clean up */

        /* call specific plugin code */
    cleanupOperator();

          
    free(src_row_buf);
    free(dst_row_buf);
}


/* ************************************************************************ */
void doWindowProcessing(GimpDrawable *drawable, int numPasses)
{
    GimpPixelRgn srcPR, tmpPR, dstPR;
    guchar new_pixel_value;
//    gint alpha;

    guchar *dst_row_buf, *d;
    
    guchar **rows = NULL;
    guchar **window = NULL;
                
    gint i, last_row_retrieved;
    gint row_index, col_index, currentPass;

#if 0
    printf("Waiting... (pid %d)\n", getpid());
    kill(getpid(), SIGSTOP);
#endif

        /* Get the input area. This is the bounding box of the selection
         * in the image (or the entire image if there is no
         * selection). */
    
    gimp_drawable_mask_bounds (drawable->id, &gX1, &gY1, &gX2, &gY2);
    gMaskWidth = gX2-gX1;
    gMaskHeight = gY2-gY1;

    assert(gMaskWidth > 0 && gMaskHeight > 0);
        
    gimp_progress_init ("Processing...");
    
    gWidth = drawable->width;
    gHeight = drawable->height;
    gBytesPerPixel = drawable->bpp;
//    alpha = gimp_drawable_has_alpha (drawable->id);

        /* call specific plugin code */
    initOperator();
    
        /*  allocate row buffers and window array */
    rows = initRowBuffers(); 
    window = initWindowBuffer();
    dst_row_buf = NULL;
    dst_row_buf = (guchar *) malloc(gMaskWidth*sizeof(guchar)*gBytesPerPixel);
    assert(dst_row_buf != NULL);

        /*  initialize the pixel regions  */
    initPixelRegions(drawable, &srcPR, &tmpPR, &dstPR, numPasses);
    
    for(currentPass = 0; currentPass < numPasses; currentPass++)
    {
        GimpPixelRgn *currentSrcPR, *currentDstPR;

        /* setup pixel regions */
        
        if(currentPass == 0)
            currentSrcPR = &srcPR;
        else
            currentSrcPR = &tmpPR;

        if(numPasses > 1 && currentPass < numPasses - 1)
            currentDstPR = &tmpPR; 
        else
            currentDstPR = &dstPR;

            /* call specific plugin code */
        preProcessWindowPass(currentSrcPR, currentPass);

            /* load the first "windowSize" rows into the row buffers */
        last_row_retrieved = fillRowBuffers(rows, currentSrcPR);
    

        /*  loop through the rows  */
        for (row_index = gY1; row_index < gY2; row_index++)
        {
            d = dst_row_buf;

                /* loop over pixels in this row */
            for (col_index = 0; col_index < gMaskWidth * gBytesPerPixel;
                 col_index++)
            {
                getWindowFromRows(col_index, window, rows);

                    /* call specific plugin code */
                new_pixel_value = processWindow(window, currentPass);
                *d++ = new_pixel_value;
            }
        
                /*  store the dst row */
            gimp_pixel_rgn_set_row(currentDstPR, dst_row_buf,
                                   gX1, row_index, gMaskWidth);

            last_row_retrieved =
                updateRowBuffers(rows, currentSrcPR, last_row_retrieved);

            if ((row_index % 5) == 0)
                gimp_progress_update (((double)row_index + (currentPass*gMaskHeight)) /
                                      ((double)gMaskHeight * numPasses) );
        }

            /* call specific plugin code */
        postProcessWindowPass(currentSrcPR, currentPass);

    }

        /*  update the region  */
    gimp_drawable_flush (drawable);
    gimp_drawable_merge_shadow (drawable->id, TRUE);
    gimp_drawable_update (drawable->id, gX1, gY1, gMaskWidth, gMaskHeight);


    /* clean up */
          
    for(i=0; i<windowSize; i++)
        free(rows[i]);
    free(rows);

    for(i=0; i<windowSize; i++)
        free(window[i]);
    free(window);

    free(dst_row_buf);
    
}

