/*
 *  $Id: Canny.c,v 1.6 1994/06/30 23:03:24 pope Exp pope $
 *
 *  This file contains routines for performing Canny edge detection.
 */

/*
 *  Copyright 1993, 1994 University of British Columbia
 *
 *  Permission to use, copy, modify, distribute, and sell this software and its
 *  documentation for any purpose is hereby granted without fee, provided that
 *  the above copyright notice appears in all copies and that both that
 *  copyright notice and this permission notice appear in supporting
 *  documentation. UBC makes no representations about the suitability of this
 *  software for any purpose. It is provided "as is" without express or
 *  implied warranty.
 *
 *  Author: Daniel Ko, UBC Laboratory for Computational Intelligence
 */

/* From the Vista library: */
#include <vista/Vlib.h>
#include <vista/VImage.h>
#include <vista/mu.h>
#include <vista/os.h>

/* From the standard C libaray: */
#include <math.h>

/* File identification string: */
VRcsId ("$Id: Canny.c,v 1.6 1994/06/30 23:03:24 pope Exp pope $");

/* Macros: */
#define Sign(x) ((x) > 0 ? 1 : -1)

/* Function prototypes: */
static double EstimateNoise (VImage, double, double);
static void NonMaximalSuppress (VImage dest, VImage xygrad, VImage mag,
				double, VImage *ori);


/*
 *  VCanny
 *
 *  Detects edges in the input image using the local maxima of the
 *  gradient magnitude after Gaussian smoothing. A rotationally
 *  symmetric Gaussian is used.
 *
 *  Note: the "band" parameter must specify a single band
 *  (i.e., cannot be VAllBands).
 */

VImage VCanny (VImage src, VImage dest, VBand band, double sigma,
	       VImage *ori, double *noise)
{
    VImage xgrad, ygrad, xygrad, mag, src_float;
    VFloat *pp, amax;
    int i;
    
    /* Check to ensure that "band" != VAllBands and "band" exists: */
    if (band == VAllBands) {
	VWarning ("VCanny: Cannot operate on all bands of the image");
	return NULL;
    }
    if (band < 0 || band >= VImageNBands (src)) {
	VWarning ("VCanny: Band %d referenced in image of %d bands",
		  band, VImageNBands (src));
	return NULL;
    }
    
    /* Make sure that "sigma" is positive: */
    if (sigma <= 0) {
	VWarning ("VCanny: Sigma %g is not a positive number", sigma);
	return NULL;
    }
    
    /* Note that "src" and "dest" can be the same without affecting the
       result of this routine. */
    
    /* Validate/create "dest" image: */
    dest = VSelectDestImage ("VCanny", dest, 1, VImageNRows (src),
			     VImageNColumns (src), VUByteRepn);
    if (! dest)
	return NULL;
    VFillImage (dest, VAllBands, 0.0);
    
    /* Check "ori": */
    if (ori) {

	/* Validate/create "*ori" image: */
	*ori = VSelectDestImage ("VCanny", *ori, 1, VImageNRows (src),
				 VImageNColumns (src), VFloatRepn);
	if (! *ori)
	    return NULL;
	VFillImage (*ori, VAllBands, 0.0);

	/* Set band interpretation of orientation map: */
	VSetAttr (VImageAttrList (*ori), VComponentInterpAttr, VBandInterpDict,
		  VLongRepn, (VLong) VBandInterpOrientation);
    }
    
    /* Convert "src" to VFloatRepn if necessary: */
    if (VPixelRepn (src) != VFloatRepn) {
	src_float = VConvertImageRange (src, NULL, band, VFloatRepn);
	if (! src_float)
	    return NULL;
	band = 0;
    } else src_float = src;
    
    /* Find the gradient (after Gaussian smoothing) in x and y directions: */
    i = 6.0 * sigma + 1;
    if ((i & 1) == 0)
	i++;
    xgrad = VGaussianConvolveImage (src_float, NULL, band, sigma, i,
				    VFilterGaussianDx);
    ygrad = VGaussianConvolveImage (src_float, NULL, band, sigma, i,
				    VFilterGaussianDy);
    
    /* Free "src_float" if necessary: */
    if (src_float != src)
	VDestroyImage (src_float);
    
    /* Combine "xgrad" and "ygrad" to a single complex image: */
    xygrad = VBuildComplexImage (xgrad, 0, ygrad, 0, NULL, VFloatRepn);
    if (! xygrad)
	return NULL;

    /* Free "ygrad": */
    VDestroyImage (ygrad);

    /* Calculate gradient magnitude: */
    mag = VImageMagnitude (xygrad, xgrad, 0);  /* Reuse "xgrad". */
    if (! mag) {
	VDestroyImage (xygrad);
	VDestroyImage (xgrad);
	return NULL;
    }
    
    /* Determine the range of gradient magnitude values: */
    pp = VPixelPtr (mag, 0, 0, 0);
    amax = *pp;
    for (i = VImageNPixels (mag); i > 0; i--, pp++)
	if (*pp > amax)
	    amax = *pp;

    /* Estimate noise: */
    if (noise)
	*noise = EstimateNoise (mag, (double) amax, *noise) *
	    (VRepnMaxValue (VUByteRepn) - 1) / amax + 1.0;

    /* Non-maximal suppression: */
    NonMaximalSuppress (dest, xygrad, mag, (double) amax, ori);
    
    /* Free "xygrad" and "mag" (i.e., "xgrad"): */
    VDestroyImage (xygrad);
    VDestroyImage (mag);

    /* Copy image attributes form source image: */
    VCopyImageAttrs (src, dest);
    VSetAttr (VImageAttrList (dest), VComponentInterpAttr, VBandInterpDict,
	      VLongRepn, (VLong) VBandInterpGradient);
    
    return dest;
}


/*
 *  EstimateNoise
 *
 *  Estimate image noise as a basis for determining low and high thresholds
 *  to be used for hysteresis thresholding. The value returned is the 
 *  gradient magnitude, r, such that fraction q of the pixels have gradient
 *  magnitude less than r.
 */

static double EstimateNoise (VImage mag, double amax, double q)
{
    int i, t, histogram[1000];
    VFloat *pp;
    
    if (q < 0.0 || q > 1.0) {
	VWarning ("VCanny: Noise estimation fraction (%g) is outside [0,1]",
		  q);
	q = 0.6;
    }

    /* Initialize histogram: */
    VZero (histogram, VNumber (histogram));

    /* Build histogram: */
    for (i = VImageNPixels (mag), pp = VPixelPtr (mag, 0, 0, 0); i > 0; i--)
	histogram[(int) (*pp++ / amax * (VNumber (histogram) - 1))]++;

    /* Find the index "i" such that percentile % of the pixels have indices 
       (in the histogram) less than "i" (roughly speaking): */
    for (i = 0, t = q * VImageNPixels (mag);
	 i < VNumber (histogram) && t >= 0; t -= histogram[i++]) ;

    /* Determine low threshold: */
    return (VFloat) i  * amax / VNumber (histogram);
}


/*
 *  NonMaximalSuppress
 *
 *  For each pixel, if the gradient magnitude is a local
 *  (3x3 neighbourhood) maxima in the gradient orientation, then
 *  output the gradient magnitude (appropriately scaled) and
 *  optionally the gradient orientation, otherwise output zero.
 */

static void NonMaximalSuppress (VImage dest, VImage xygrad, VImage mag,
				double amax, VImage *ori)
{
    int x,y;
    float g;               /* center pixel gradient magnitude value */
    float g1, g2;          /* interpolated gradient magnitude values */
    float ga, gb, gc, gd;  /* nearest pixels' gradient magnitude values */
    float ux, uy;          /* components of edge normal vector */
    float dx, dy;
    float nf;		   /* output pixel normalizing factor */
    
    nf = (VRepnMaxValue (VUByteRepn) - 1) / amax;

    /* For each pixel (except those on the image boundary), check if 
       the gradient magnitude is a local (3x3 neighbourhood) maxima in 
       the gradient orientation: */
    for (x = 1; x < VImageNColumns (mag) - 1; x++) {
	for (y = 1; y < VImageNRows (mag) - 1; y++) {

	    dx = VPixel (xygrad, 0, y, x, VFloat);
	    dy = VPixel (xygrad, 1, y, x, VFloat);
	    
	    if (dx == 0.0f && dy == 0.0f)
		continue;

	    if (fabs (dy) > fabs (dx)) {
		    
		ux = fabs (dx) / fabs (dy);
		uy = 1.0f;
		    
		gb = VPixel (mag, 0, y - 1, x, VFloat);
		gd = VPixel (mag, 0, y + 1, x, VFloat);
		    
		if (Sign (dx) != Sign (dy)) {
		    ga = VPixel (mag, 0, y - 1, x - 1, VFloat);
		    gc = VPixel (mag, 0, y + 1, x + 1, VFloat);
		} else {
		    ga = VPixel (mag, 0, y - 1, x + 1, VFloat);
		    gc = VPixel (mag, 0, y + 1, x - 1, VFloat);
		}
		    
	    } else {
		    
		ux = fabs (dy) / fabs (dx);
		uy = 1.0f;

		gb = VPixel (mag, 0, y, x + 1, VFloat);
		gd = VPixel (mag, 0, y, x - 1, VFloat);

		if (Sign (dx) != Sign (dy)) {
		    ga = VPixel (mag, 0, y + 1, x + 1, VFloat);
		    gc = VPixel (mag, 0, y - 1, x - 1, VFloat);
		} else {
		    ga = VPixel (mag, 0, y - 1, x + 1, VFloat);
		    gc = VPixel (mag, 0, y + 1, x - 1, VFloat);
		}
	    }
		
	    g1 = (ux * ga) + (uy - ux) * gb;
	    g2 = (ux * gc) + (uy - ux) * gd;
	    g = sqrt (dx * dx + dy * dy);
	    
	    if (g > g1 && g >= g2) {
		
		/* Store magnitude in output image while mapping [0,amax]
		   to [1,255]: */
		VPixel (dest, 0, y, x, VUByte) =
		    VPixel (mag, 0, y, x, VFloat) * nf + 1;
		if (ori)
		    VPixel (*ori, 0, y, x, VFloat) = atan2 (dy, dx);
		
	    }
	}
    }
}
