/*
 *  $Id: Complex.c,v 1.5 1994/06/15 23:14:49 pope Exp pope $
 *
 *  Routines for manipulating images with complex pixel values.
 */

/*
 *  Copyright 1993, 1994 University of British Columbia
 *
 *  Permission to use, copy, modify, distribute, and sell this software and its
 *  documentation for any purpose is hereby granted without fee, provided that
 *  the above copyright notice appears in all copies and that both that
 *  copyright notice and this permission notice appear in supporting
 *  documentation. UBC makes no representations about the suitability of this
 *  software for any purpose. It is provided "as is" without express or
 *  implied warranty.
 *
 *  Author: Arthur Pope, UBC Laboratory for Computational Intelligence
 */

/* From the Vista library: */
#include <vista/Vlib.h>
#include <vista/os.h>
#include <vista/VImage.h>

/* From the standard C library: */
#include <math.h>

/* File identification string: */
VRcsId ("$Id: Complex.c,v 1.5 1994/06/15 23:14:49 pope Exp pope $");


/*
 *  VBuildComplexImage
 *
 *  Creates a complex image from separate image bands of real and imaginary
 *  components.
 */

VImage VBuildComplexImage (VImage r_src, VBand r_band,
			   VImage i_src, VBand i_band,
			   VImage dest, VRepnKind pixel_repn)
{
    VImage src, result;
    int nbands, band;
    size_t size;

    /* Check the image bands provided: */
    if (! r_src && ! i_src) {
	VWarning ("VBuildComplexImage: "
		  "Neither real nor imaginary component given");
	return NULL;
    }
    if (r_src && r_band != VAllBands && r_band >= VImageNBands (r_src)) {
	VWarning ("VBuildComplexImage: "
		  "Band %d referenced in image of %d band(s)",
		  r_band, VImageNBands (r_src));
	return NULL;
    }
    if (i_src && i_band != VAllBands && i_band >= VImageNBands (i_src)) {
	VWarning ("VBuildComplexImage: "
		  "Band %d referenced in image of %d band(s)",
		  i_band, VImageNBands (i_src));
	return NULL;
    }
    if (r_src && i_src) {
	if (VImageNRows (r_src) != VImageNRows (i_src) ||
	    VImageNColumns (r_src) != VImageNColumns (i_src)) {
	    VWarning ("VBuildComplexImage: "
		      "Real and imaginary images have different sizes");
	    return NULL;
	}
	if ((r_band == VAllBands ? VImageNBands (r_src) : 1) !=
	    (i_band == VAllBands ? VImageNBands (i_src) : 1)) {
	    VWarning ("VBuildComplexImage: "
		      "Different numbers of real and imaginary bands given");
	    return NULL;
	}
    }
    
    /* Check or create the destination image: */
    if (r_src) {
	src = r_src;
	nbands = (r_band == VAllBands) ? VImageNBands (r_src) : 1;
    } else {
	src = i_src;
	nbands = (i_band == VAllBands) ? VImageNBands (i_src) : 1;
    }
    nbands *= 2;
    result = VSelectDestImage ("VBuildComplexImage", dest, nbands,
			       VImageNRows (src), VImageNColumns (src),
			       pixel_repn);
    if (! result)
	return NULL;
    if (r_src == dest || i_src == dest)
	result = VCreateImage (nbands, VImageNRows (src),
			       VImageNColumns (src), pixel_repn);

    /* Convert the real and imaginary bands to the destination pixel
       representation, if necessary: */
    if (r_src && VPixelRepn (r_src) != pixel_repn) {
	r_src = VConvertImageRange (r_src, NULL, r_band, pixel_repn);
	if (! r_src)
	    goto Fail;
	r_band = 0;	
    }
    if (i_src && VPixelRepn (i_src) != pixel_repn) {
    	i_src = VConvertImageRange (i_src, NULL, i_band, pixel_repn);
	if (! i_src) {
Fail:	    if (result != dest)
		VDestroyImage (result);
            return NULL;
        }
	i_band = 0;	
    }

    /* Copy the real and imaginary bands into the destination image: */
    size =
	VImageNRows (result) * VImageNColumns (result) * VPixelSize (result);
    if (r_band == VAllBands)
	r_band = 0;
    if (i_band == VAllBands)
	i_band = 0;
    for (band = 0; band < nbands; band += 2, r_band++, i_band++) {
	if (r_src)
	    memcpy (VPixelPtr (result, band, 0, 0),
		    VPixelPtr (r_src, r_band, 0, 0), size);
	else VFillImage (result, band, 0.0);
	if (i_src)
	    memcpy (VPixelPtr (result, band + 1, 0, 0),
		    VPixelPtr (i_src, i_band, 0, 0), size);
	else VFillImage (result, band + 1, 0.0);
    }

    /* Give the destination image a component interpretation attribute
       of "Complex": */
    VImageNComponents (result) = 2;
    VSetAttr (VImageAttrList (result), VComponentInterpAttr, VBandInterpDict,
	      VLongRepn, VBandInterpComplex);
    VImageNColors (result) = VImageNViewpoints (result) = 1;
    VImageNFrames (result) = nbands / 2;

    /* Return the result: */
    if (dest && (r_src == dest || i_src == dest)) {
	VCopyImage (result, dest, VAllBands);
	VDestroyImage (result);
	return dest;
    } else return result;
}


/*
 *  VImageMagnitude
 *
 *  Compute the magnitude of a complex image.
 *  Src is a 2-component, 2n-band image with any pixel type.
 *  Dest is a 1-component, n band image, the same size and pixel type as src.
 *  Each destination pixel is set to sqrt (src0^2 + src1^2), where src0 and 
 *  src1 are the correponding pixels of source components 0 and 1.
 */

VImage VImageMagnitude (VImage src, VImage dest, VBand band)
{
    int nbands, src_band, dest_band;
    int i, npixels = VImageNRows (src) * VImageNColumns (src);
    VDouble t0, t1;
    VBoolean clipped = FALSE;

    #define Magnitude(type, clip) {					    \
	type *src_pp0, *src_pp1, *dest_pp;				    \
									    \
	/* Track position in src with pointer for each band: */		    \
	src_pp0 = & VPixel (src, src_band, 0, 0, type);			    \
	src_pp1 = & VPixel (src, src_band + 1, 0, 0, type);		    \
									    \
	/* Track position in dest with one pointer: */			    \
	dest_pp = & VPixel (dest, dest_band, 0, 0, type);		    \
									    \
	/* Compute the Magnitude at each dest pixel: */			    \
	for (i = npixels; i > 0; i--) {    				    \
	    t0 = *src_pp0++;						    \
	    t1 = *src_pp1++;						    \
	    t0 = sqrt (t0 * t0 + t1 * t1);				    \
	    clip;							    \
	    *dest_pp++ = t0;						    \
        }								    \
    }
		
    #define Clip(high)						    	    \
	if (t0 > high) {						    \
	    t0 = high;							    \
	    clipped = TRUE;						    \
	}
		
    #define Nothing

    /* Ensure that the source image has two components and enough bands: */
    if (VImageNComponents (src) != 2) {
	VWarning ("VImageMagnitude: Source image must have two components");
	return NULL;
    }
    if (band == VAllBands) {
	src_band = 0;
	nbands = VImageNBands (src) / 2;
    } else  if (band >= 0 && band + 1 < VImageNBands (src)) {
	src_band = band;
	nbands = 1;
    } else {
	VWarning ("VImageMagnitude: Band %d referenced in image of %d band(s)",
		  band, VImageNBands (src));
	return NULL;
    }

    /* Check or create the destination image: */
    /* (Source and destination images will always be different.) */
    dest = VSelectDestImage ("VImageMagnitude", dest, nbands,
			     VImageNRows (src), VImageNColumns (src),
			     VPixelRepn (src));
    if (! dest)
	return NULL;

    /* Compute the magnitude at each pixel: */
    for (dest_band = 0; dest_band < nbands; dest_band++, src_band += 2)
	switch (VPixelRepn (src)) {
	    
	case VUByteRepn:
	    Magnitude (VUByte, Clip (255.0));
	    break;
	    
	case VSByteRepn:
	    Magnitude (VSByte, Clip (127.0));
	    break;
	    
	case VShortRepn:
	    Magnitude (VShort, Clip (32767.0));
	    break;
	    
	case VLongRepn:
	    Magnitude (VLong, Clip (VRepnMaxValue (VLongRepn)));
	    break;
	    
	case VFloatRepn:
	    Magnitude (VFloat, Nothing);
	    break;
	    
	case VDoubleRepn:
	    Magnitude (VDouble, Nothing);
	    
	default:
	    VError ("VImageMagnitude: %s images not supported",
		    VPixelRepnName (src));
	}
    
    if (clipped)
	VWarning ("VImageMagnitude: Destination pixel(s) out of range");

    /* Adopt the band interpretation information from the source image but
       delete any component interpretation attribute: */
    if (nbands > 1) {
	if (VImageAttrList (dest))
	    VDestroyAttrList (VImageAttrList (dest));
	VImageAttrList (dest) = VCopyAttrList (VImageAttrList (src));
	VImageNFrames (dest) = VImageNFrames (src);
	VImageNViewpoints (dest) = VImageNViewpoints (src);
	VImageNColors (dest) = VImageNColors (src);
	VImageNComponents (dest) = 1;
    } else VCopyImageAttrs (src, dest);
    VSetAttr (VImageAttrList (dest), VComponentInterpAttr, VBandInterpDict,
	      VLongRepn, (VLong) VBandInterpIntensity);

    return dest;

    #undef Magnitude
    #undef Clip
    #undef Nothing
}


/*
 *  VImagePhase
 *
 *  Compute the phase of a complex image.
 *  Src is a 2-component, 2n-band image with any pixel type.
 *  Dest is a 1-component, n band image, the same size as src.
 *  Its pixel type is Double if src is Double; otherwise it's Float.
 *  Each destination pixel is set to atan2 (src1, src0), where src0 and 
 *  src1 are the correponding pixels of source components 0 and 1.
 */

VImage VImagePhase (VImage src, VImage dest, VBand band)
{
    VRepnKind repn;
    int nbands, src_band, dest_band;
    int i, npixels = VImageNRows (src) * VImageNColumns (src);

    #define Phase(src_type, dest_type) {				    \
	src_type *src_pp0, *src_pp1;					    \
	dest_type *dest_pp;						    \
									    \
	/* Track position in src with pointer for each band: */		    \
	src_pp0 = & VPixel (src, src_band, 0, 0, src_type);		    \
	src_pp1 = & VPixel (src, src_band + 1, 0, 0, src_type);		    \
									    \
	/* Track position in dest with one pointer: */			    \
	dest_pp = & VPixel (dest, dest_band, 0, 0, dest_type);		    \
									    \
	/* Compute the phase at each dest pixel: */			    \
	for (i = npixels; i > 0; i--, src_pp0++, src_pp1++, dest_pp++)      \
	    if (*src_pp0 == 0 && *src_pp1 == 0)				    \
		*dest_pp = 0;						    \
	    else *dest_pp = atan2 ((double) *src_pp1, (double) *src_pp0);   \
    }
		
    /* Ensure that the source image has two components and enough bands: */
    if (VImageNComponents (src) != 2) {
	VWarning ("VImagePhase: Source image must have two components");
	return NULL;
    }
    if (band == VAllBands) {
	src_band = 0;
	nbands = VImageNBands (src) / 2;
    } else if (band >= 0 && band + 1 < VImageNBands (src)) {
	src_band = band;
	nbands = 1;
    } else {
	VWarning ("VImagePhase: Band %d referenced in image of %d band(s)",
		  band, VImageNBands (src));
	return NULL;
    }

    /* Check or create the destination image: */
    repn = (VPixelRepn (src) == VDoubleRepn) ? VDoubleRepn : VFloatRepn;
    dest = VSelectDestImage ("VImagePhase", dest, nbands,
			     VImageNRows (src), VImageNColumns (src), repn);
    if (! dest)
	return NULL;

    /* Compute the phase at each pixel: */
    for (dest_band = 0; dest_band < nbands; dest_band++, src_band += 2)
	switch (VPixelRepn (src)) {
	    
	case VUByteRepn:
	    Phase (VUByte, VFloat);
	    break;
	    
	case VSByteRepn:
	    Phase (VSByte, VFloat);
	    break;
	    
	case VShortRepn:
	    Phase (VShort, VFloat);
	    break;
	    
	case VLongRepn:
	    Phase (VLong, VFloat);
	    break;
	    
	case VFloatRepn:
	    Phase (VFloat, VFloat);
	    break;
	    
	case VDoubleRepn:
	    Phase (VDouble, VDouble);
	    break;
	    
	default:
	    VError ("VImagePhase: %s images not supported",
		    VPixelRepnName (src));
	}

    /* Adopt the band interpretation information from the source image but
       delete any component interpretation: */
    if (nbands > 1) {
	if (VImageAttrList (dest))
	    VDestroyAttrList (VImageAttrList (dest));
	VImageAttrList (dest) = VCopyAttrList (VImageAttrList (src));
	VImageNFrames (dest) = VImageNFrames (src);
	VImageNViewpoints (dest) = VImageNViewpoints (src);
	VImageNColors (dest) = VImageNColors (src);
	VImageNComponents (dest) = 1;
    } else VCopyImageAttrs (src, dest);
    VSetAttr (VImageAttrList (dest), VComponentInterpAttr, VBandInterpDict,
	      VLongRepn, (VLong) VBandInterpOrientation);

    return dest;

    #undef Phase
}
