/*
 *  $Id: Fft.c,v 1.5 1994/06/15 23:14:49 pope Exp pope $
 *
 *  This file contains routines for computing Fourier transforms.
 */

/*
 *  Copyright 1993, 1994 University of British Columbia
 *
 *  Permission to use, copy, modify, distribute, and sell this software and its
 *  documentation for any purpose is hereby granted without fee, provided that
 *  the above copyright notice appears in all copies and that both that
 *  copyright notice and this permission notice appear in supporting
 *  documentation. UBC makes no representations about the suitability of this
 *  software for any purpose. It is provided "as is" without express or
 *  implied warranty.
 *
 *  Author: Arthur Pope, UBC Laboratory for Computational Intelligence
 */

/* From the Vista library: */
#include <vista/Vlib.h>
#include <vista/os.h>
#include <vista/VImage.h>

/* From the standard C library: */
#include <math.h>

/* File identification string: */
VRcsId ("$Id: Fft.c,v 1.5 1994/06/15 23:14:49 pope Exp pope $");

/* Later in this file: */
static void VMultiDimFFTFloat (int, int *, VFloat *, VFloat *,
			       VBoolean, double);
static void VMultiDimFFTDouble (int, int *, VDouble *, VDouble *,
				VBoolean, double);


/*
 *  VImageFFT
 *
 *  Computes the Fourier transform of an image.
 *  The source image must be a complex image of Float or Double 
 *  pixels, with row and column sizes that are powers of 2. 
 *  The destination image, if provided, must have the same properties.
 *  Each destination pixel is divided by norm.
 *
 *  Note: This uses ffs() to find the least significant bit on in an
 *	  integer. It isn't part of ANSI C, but it's widely available.
 */

VImage VImageFFT (VImage src, VImage dest, VBooleanPromoted inverse,
		  double norm)
{
    int nrows, ncolumns, dims[2], band;

    /* Ensure that the source image has appropriate properties: */
    if (VPixelRepn (src) != VFloatRepn && VPixelRepn (src) != VDoubleRepn) {
	VWarning ("VImageFFT: Source image must have"
		  "VFloat or VDouble pixels");
	return NULL;
    }
    if (VImageNComponents (src) != 2) {
	VWarning ("VImageFFT: Source image must have two components");
	return NULL;
    }
    nrows = VImageNRows (src);
    ncolumns = VImageNColumns (src);
    if ((1 << (ffs (nrows) - 1)) != nrows) {
	VWarning ("VImageFFT: Source image must have 2**n rows");
	return NULL;
    }
    if ((1 << (ffs (ncolumns) - 1)) != ncolumns) {
	VWarning ("VImageFFT: Source image must have 2**n columns");
	return NULL;
    }

    /* Check or create the destination image: */
    dest = VSelectDestImage ("VImageFFT", dest, VImageNBands (src),
			     nrows, ncolumns, VPixelRepn (src));
    if (dest == NULL)
	return NULL;

    /* Initialize the destination image to the contents of the source image: */
    VCopyImage (src, dest, VAllBands);

    /* Transform it: */
    dims[0] = nrows;
    dims[1] = ncolumns;
    for (band = 0; band < VImageNBands (src); band += 2)
	switch (VPixelRepn (src)) {
	    
	case VFloatRepn:
	    VMultiDimFFTFloat (2, dims,
			       (VFloat *) VPixelPtr (dest, band, 0, 0),
			       (VFloat *) VPixelPtr (dest, band + 1, 0, 0),
			       inverse, norm);
	    break;
	    
	case VDoubleRepn:
	    VMultiDimFFTDouble (2, dims,
				(VDouble *) VPixelPtr (dest, band, 0, 0),
				(VDouble *) VPixelPtr (dest, band + 1, 0, 0),
				inverse, norm);
	    break;

	default:
	    break;
	}

    return dest;
}


/*
 *  VMultiDimFFTFloat, VMultiDimFFTDouble
 *
 *  Compute the Fourier transform of multidimensional Float or Double data.
 *
 *  ndim is the number of dimensions, and dims[0 ... ndim-1] is the size
 *  of each.
 *
 *  r_data and i_data contain the data's real and imaginary components.
 *  Each is a multidimensional array in which elements are stored in
 *  consecutive locations with the rightmost index of the array increasing
 *  most rapidly. A two-dimensional array, for example, is stored as rows.
 *
 *  If inverse is FALSE, r_data and i_data are replaced with their
 *  Fourier transform; if it is TRUE, they are replaced with their 
 *  inverse Fourier transform.
 *
 *  This implementation is taken from "Numerical Recipes in C", pp. 467-470.
 */

static void VMultiDimFFTFloat (int ndim, int *dims,
			       VFloat *r_data, VFloat *i_data,
			       VBoolean inverse, double norm)
{
    int i1, i2, i3, i2rev, i3rev, ip2, ip3, ifp1, ifp2;
    int ibit, idim, k1, k2, n, nprev, nrem, nelements;
    VFloat tempi, tempr;
    VDouble sign, theta, wi, wpi, wpr, wr, wtemp;
    
    #define SWAP(a,b) tempr=(a);(a)=(b);(b)=tempr
    
    /* Total up the number of elements in the data array: */
    for (idim = 0, nelements = 1; idim < ndim; idim++)
	nelements *= dims[idim];

    sign = (inverse ? -2.0 : 2.0) * M_PI;

    /* Main loop over the dimensions: */
    for (idim = ndim - 1, nprev = 1; idim >= 0; idim--, nprev *= n) {
	n = dims[idim];
	nrem = nelements / (n * nprev);
	ip2 = nprev * n;
	ip3 = ip2 * nrem;
	i2 = i2rev = 0;

	/* Swap bit reversal pairs: */
	while (1) {
	    if (i2 < i2rev) {
		for (i1 = i2; i1 < i2 + nprev; i1++) {
		    for (i3 = i1; i3 < ip3; i3 += ip2) {
			i3rev = i2rev + i3 - i2;
			SWAP (r_data[i3], r_data[i3rev]);
			SWAP (i_data[i3], i_data[i3rev]);
		    }
		}
	    }
	    if ((i2 += nprev) > ip2)
		break;
	    ibit = ip2 >> 1;
	    while (ibit >= nprev && (i2rev & ibit)) {
		i2rev -= ibit;
		ibit >>= 1;
	    }
	    i2rev += ibit;
	}
	for (ifp1 = nprev; ifp1 < ip2; ifp1 <<= 1) {
	    ifp2 = ifp1 << 1;
	    theta = sign / (ifp2 / nprev);
	    wtemp = sin (0.5 * theta);
	    wpr = -2.0 * wtemp * wtemp;
	    wpi = sin (theta);
	    wr = 1.0;
	    wi = 0.0;
	    for (i3 = 0; i3 < ifp1; i3 += nprev) {
		for (i1 = i3; i1 < i3 + nprev; i1++) {
		    for (i2 = i1; i2 < ip3; i2 += ifp2) {
			k1 = i2;
			k2 = k1 + ifp1;
			tempr = wr * r_data[k2] - wi * i_data[k2];
			tempi = wr * i_data[k2] + wi * r_data[k2];
			r_data[k2] = r_data[k1] - tempr;
			i_data[k2] = i_data[k1] - tempi;
			r_data[k1] += tempr;
			i_data[k1] += tempi;
		    }
		}
		wr = (wtemp = wr) * wpr - wi * wpi + wr;
		wi = wi * wpr + wtemp * wpi + wi;
	    }
	}
    }

    /* Divide through by the normalizing constant: */
    if (norm != 1.0)
	for (i1 = 0; i1 < nelements; i1++) {
	    *r_data++ /= norm;
	    *i_data++ /= norm;
	}

    #undef SWAP
}

static void VMultiDimFFTDouble (int ndim, int *dims,
				VDouble *r_data, VDouble *i_data,
				VBoolean inverse, double norm)
{
    int i1, i2, i3, i2rev, i3rev, ip2, ip3, ifp1, ifp2;
    int ibit, idim, k1, k2, n, nprev, nrem, nelements;
    VDouble tempi, tempr;
    VDouble sign, theta, wi, wpi, wpr, wr, wtemp;
    
    #define SWAP(a,b) tempr=(a);(a)=(b);(b)=tempr
    
    /* Total up the number of elements in the data array: */
    for (idim = 0, nelements = 1; idim < ndim; idim++)
	nelements *= dims[idim];

    sign = (inverse ? -2.0 : 2.0) * M_PI;

    /* Main loop over the dimensions: */
    for (idim = ndim - 1, nprev = 1; idim >= 0; idim--, nprev *= n) {
	n = dims[idim];
	nrem = nelements / (n * nprev);
	ip2 = nprev * n;
	ip3 = ip2 * nrem;
	i2 = i2rev = 0;

	/* Swap bit reversal pairs: */
	while (1) {
	    if (i2 < i2rev) {
		for (i1 = i2; i1 < i2 + nprev; i1++) {
		    for (i3 = i1; i3 < ip3; i3 += ip2) {
			i3rev = i2rev + i3 - i2;
			SWAP (r_data[i3], r_data[i3rev]);
			SWAP (i_data[i3], i_data[i3rev]);
		    }
		}
	    }
	    if ((i2 += nprev) > ip2)
		break;
	    ibit = ip2 >> 1;
	    while (ibit >= nprev && (i2rev & ibit)) {
		i2rev -= ibit;
		ibit >>= 1;
	    }
	    i2rev += ibit;
	}
	for (ifp1 = nprev; ifp1 < ip2; ifp1 <<= 1) {
	    ifp2 = ifp1 << 1;
	    theta = sign / (ifp2 / nprev);
	    wtemp = sin (0.5 * theta);
	    wpr = -2.0 * wtemp * wtemp;
	    wpi = sin (theta);
	    wr = 1.0;
	    wi = 0.0;
	    for (i3 = 0; i3 < ifp1; i3 += nprev) {
		for (i1 = i3; i1 < i3 + nprev; i1++) {
		    for (i2 = i1; i2 < ip3; i2 += ifp2) {
			k1 = i2;
			k2 = k1 + ifp1;
			tempr = wr * r_data[k2] - wi * i_data[k2];
			tempi = wr * i_data[k2] + wi * r_data[k2];
			r_data[k2] = r_data[k1] - tempr;
			i_data[k2] = i_data[k1] - tempi;
			r_data[k1] += tempr;
			i_data[k1] += tempi;
		    }
		}
		wr = (wtemp = wr) * wpr - wi * wpi + wr;
		wi = wi * wpr + wtemp * wpi + wi;
	    }
	}
    }

    /* Divide through by the normalizing constant: */
    if (norm != 1.0)
	for (i1 = 0; i1 < nelements; i1++) {
	    *r_data++ /= norm;
	    *i_data++ /= norm;
	}

    #undef SWAP
}
