/*
 *  $Id: GaussConv.c,v 1.5 1994/06/15 23:14:49 pope Exp pope $
 *
 *  This file contains routines for performing Gaussian convolution.
 */

/*
 *  Copyright 1993, 1994 University of British Columbia
 *
 *  Permission to use, copy, modify, distribute, and sell this software and its
 *  documentation for any purpose is hereby granted without fee, provided that
 *  the above copyright notice appears in all copies and that both that
 *  copyright notice and this permission notice appear in supporting
 *  documentation. UBC makes no representations about the suitability of this
 *  software for any purpose. It is provided "as is" without express or
 *  implied warranty.
 *
 *  Author: Daniel Ko, Art Pope, UBC Laboratory for Computational Intelligence
 */

/* From the Vista library: */
#include "vista/Vlib.h"
#include "vista/mu.h"
#include "vista/os.h"
#include "vista/VImage.h"

/* File identification string: */
VRcsId ("$Id: GaussConv.c,v 1.5 1994/06/15 23:14:49 pope Exp pope $");


/*
 *  VGaussianConvolveImage
 *
 *  Convolve an image with a Gaussian filter, or a filter that is the
 *  partial derivate of a Gaussian in the X or Y direction.
 *  The 2D filter is implemented as two 1D filters, each with the specified
 *  standard deviation and length.
 */

VImage VGaussianConvolveImage (VImage src, VImage dest, VBand band,
			       double sigma, int filter_size,
			       VFilterKind filter_kind)
{
    VImage src_float, result, tmp, filters[3];
    VRepnKind dest_pixel_repn, conv_pixel_repn;
    int i;
    static VConvolvePadMethod pad_methods[3] =
	{ VConvolvePadBorder, VConvolvePadBorder, VConvolvePadBorder };
    static int shifts[3] = { 0, 0, 0 };

    /* Ensure that the standard deviation and filter size are positive: */
    if (sigma <= 0.0) {
	VWarning ("VGaussianConvolveImage: Sigma (%g) is not positive", sigma);
	return NULL;
    }
    if (filter_size <= 0) {
	VWarning ("VGaussianConvolveImage: Filter size (%d) is not positive",
		  filter_size);
	return NULL;
    }

    /* Check that the filter type is correctly specified: */
    switch (filter_kind) {
    case VFilterGaussian:	break;
    case VFilterGaussianDx:	break;
    case VFilterGaussianDy:	break;
    default:
	VWarning ("VGaussianConvolveImage: Unknown filter type (%d)",
		  filter_kind);
	return NULL;
    }

    /* Determine the appropriate destination pixel type. Convolving a bit
       or ubyte image with a derivative of a Gaussian produces an sbyte image;
       otherwise the destination has the same type as src. */
    if ((filter_kind == VFilterGaussianDx ||
	 filter_kind == VFilterGaussianDy) &&
	(VPixelRepn (src) == VBitRepn || VPixelRepn (src) == VUByteRepn))
	dest_pixel_repn = VSByteRepn;
    else dest_pixel_repn = VPixelRepn (src);

    /* Convert the source image to Float or Double for the sake of
       efficiency: */
    if (! VIsFloatPtRepn (VPixelRepn (src))) {
	if (! (src_float = VConvertImageRange (src, NULL, band, VFloatRepn)))
	    return NULL;
	band = VAllBands;
    } else src_float = src;
    conv_pixel_repn = VPixelRepn (src_float);

    /* Make separate row and column filters: */
    filters[0] = NULL;
    filters[1] = VCreateImage (1, filter_size, 1, VDoubleRepn);
    filters[2] = VCreateImage (1, 1, filter_size, VDoubleRepn);
    (filter_kind == VFilterGaussianDy ? VGaussianD1Kernel : VGaussianKernel)
	(filter_size, VImageData (filters[1]), sigma);
    (filter_kind == VFilterGaussianDx ? VGaussianD1Kernel : VGaussianKernel)
	(filter_size, VImageData (filters[2]), sigma);
    if (filter_kind == VFilterGaussianDx)
	for (i = 0; i < filter_size; i++)
	    VPixel (filters[2], 0, 0, i, VDouble) =
		- VPixel (filters[2], 0, 0, i, VDouble);

    /* Convolve src with the two filters: */
    result =
	VConvolveImageSep (src_float, (conv_pixel_repn == dest_pixel_repn) ?
			   dest : NULL, band, filters, pad_methods, shifts);

    /* Convert the result to the apprporate destination type, if necessary: */
    if (result && conv_pixel_repn != dest_pixel_repn) {
	tmp = result;
	result = VConvertImageRange (tmp, dest, VAllBands, dest_pixel_repn);
	VDestroyImage (tmp);
    }
    if (src_float != src)
	VDestroyImage (src_float);
    VDestroyImage (filters[1]);
    VDestroyImage (filters[2]);
    return result;
}
