/*
 *  $Id: Gradient.c,v 1.5 1995/05/01 23:21:17 pope Exp pope $
 *
 *  Routines for computing the gradient of an image.
 */

/*
 *  Copyright 1993, 1994 University of British Columbia
 *
 *  Permission to use, copy, modify, distribute, and sell this software and its
 *  documentation for any purpose is hereby granted without fee, provided that
 *  the above copyright notice appears in all copies and that both that
 *  copyright notice and this permission notice appear in supporting
 *  documentation. UBC makes no representations about the suitability of this
 *  software for any purpose. It is provided "as is" without express or
 *  implied warranty.
 *
 *  Author: Arthur Pope, UBC Laboratory for Computational Intelligence
 */

/* From the Vista library: */
#include <vista/Vlib.h>
#include <vista/os.h>
#include <vista/VImage.h>

/* File identification string: */
VRcsId ("$Id: Gradient.c,v 1.5 1995/05/01 23:21:17 pope Exp pope $");


/*
 *  Some macros for used in computing the gradient.
 */

#define Gradient(src_type, dest_type, work_type, clip) {		    \
    src_type *src_pp0, *src_pp1, *src_pp2;				    \
    dest_type *dest_pp0, *dest_pp1;					    \
    work_type t;							    \
									    \
    /* Track position in src with pointer for each row: */		    \
    src_pp0 = & VPixel (src, src_band, 0, 1, src_type); /* row above */	    \
    src_pp1 = & VPixel (src, src_band, 1, 0, src_type); /* center row */    \
    src_pp2 = & VPixel (src, src_band, 2, 1, src_type); /* row below */	    \
									    \
    /* Track position in dest with one pointer for each band: */	    \
    dest_pp0 = & VPixel (dest, dest_band, 1, 1, dest_type);	/* hor */  \
    dest_pp1 = & VPixel (dest, dest_band + 1, 1, 1, dest_type); /* vert */   \
									    \
    /* For rows 1 through nrows-1 of the images... */			    \
    for (i = VImageNRows (src) - 2; i >= 0; i--)			    \
        for (j = 0; j < VImageNColumns (src); j++) {			    \
									    \
	    /* Compute horizontal and vertical gradients: */		    \
	    t = ((work_type) *(src_pp1 + 2) - (work_type) *src_pp1) /	    \
		(work_type) 2;						    \
	    clip;							    \
	    *dest_pp0++ = t;						    \
	    t = ((work_type) *src_pp0 - (work_type) *src_pp2) /		    \
		(work_type) 2;						    \
	    clip;							    \
	    *dest_pp1++ = t;						    \
									    \
	    src_pp0++;							    \
	    src_pp1++;							    \
	    src_pp2++;							    \
	}								    \
									    \
    /* Zero the first and last columns of dest: */			    \
    dest_pp0 = & VPixel (dest, dest_band, 1, 0, dest_type);		    \
    dest_pp1 = & VPixel (dest, dest_band + 1, 1, 0, dest_type);		    \
    for (i = 1; i < VImageNRows (dest); i++) {				    \
	*(dest_pp0 - 1) = *dest_pp0 = *(dest_pp1 - 1) = *dest_pp1 = 0;	    \
	dest_pp0 += VImageNColumns (dest);				    \
	dest_pp1 += VImageNColumns (dest);				    \
    }									    \
}

#define Clip(low, high)							    \
    if (t < low) {							    \
	t = low;							    \
	clipped = TRUE;						      	    \
    } else if (t > high) {						    \
	t = high;							    \
	clipped = TRUE;							    \
    }
		
#define Nothing


/*
 *  ASSUMPTION: Floating point 0.0 is binary all-bits-zero.
 *
 *  This is true of the IEEE standard floating point format, but it
 *  isn't assured by the ANSI C standard.
 */

#define ZeroRow(image,band,row,n)					    \
    memset (VPixelPtr (image, band, row, 0), 0, n)


/*
 *  VImageGradient
 *
 *  Compute the vertical and horizontal components of the intensity gradient
 *  at each pixel of an image.
 *  Src must be a single component image.
 *  Dest will be a two-component image of the same size as src, with band 0
 *  containing the horizontal component of the intensity gradient, and band 1
 *  the vertical component.
 *  If src is a Bit or UByte image, then dest is an SByte image; otherwise,
 *  dest has the same pixel type as src.
 */

VImage VImageGradient (VImage src, VImage dest, VBand band)
{
    int nbands, src_band, dest_band;
    VRepnKind repn;
    int i, j;
    size_t n;
    VBoolean clipped = FALSE;

    /* Check properties of the source image: */
    if (VImageNComponents (src) != 1) {
	VWarning ("VImageGradient: Source image must have one component");
	return NULL;
    }
    if (band == VAllBands) {
	src_band = 0;
	nbands = VImageNBands (src) * 2;
    } else if (band >= 0 && band < VImageNBands (src)) {
	src_band = band;
	nbands = 2;
    } else {
	VWarning ("VImageGradient: Band %d referenced in image of %d band(s)",
		  band, VImageNBands (src));
	return NULL;
    }

    /* Check or create the destination image: */
    /* (Source and destination images ought to be different.) */
    switch (VPixelRepn (src)) {

    case VBitRepn:
    case VUByteRepn:
	repn = VSByteRepn;
	break;

    default:
	repn = VPixelRepn (src);
    }
    dest = VSelectDestImage ("VImageGradient", dest, nbands,
			     VImageNRows (src), VImageNColumns (src), repn);
    if (! dest)
	return NULL;

    /* For each destination band: */
    for (dest_band = 0; dest_band < nbands; dest_band += 2, src_band++) {

	/* Compute the gradient at all but the first and last rows: */
	switch (VPixelRepn (src)) {
	    
	case VBitRepn:
	    Gradient (VBit, VSByte, VShort, Clip (-128, 127));
	    break;
	    
	case VUByteRepn:
	    Gradient (VUByte, VSByte, VShort, Clip (-128, 127));
	    break;
	    
	case VSByteRepn:
	    Gradient (VSByte, VSByte, VShort, Clip (-128, 127));
	    break;
	    
	case VShortRepn:
	    Gradient (VShort, VShort, VLong, Clip (-32768, 32767));
	    break;
	    
	case VLongRepn:
	    Gradient (VLong, VLong, VLong, Nothing);
	    break;
	    
	case VFloatRepn:
	    Gradient (VFloat, VFloat, VFloat, Nothing);
	    break;
	    
	case VDoubleRepn:
	    Gradient (VDouble, VDouble, VDouble, Nothing);
	    break;

	default:
	    break;
	}

	/* Zero out the first and last rows: */
	n = VImageNColumns (dest) * VPixelSize (dest);
	ZeroRow (dest, dest_band, 0, n);
	ZeroRow (dest, dest_band, VImageNRows (dest) - 1, n);
	ZeroRow (dest, dest_band + 1, 0, n);
	ZeroRow (dest, dest_band + 1, VImageNRows (dest) - 1, n);
    }

    /* Adopt the band interpretation information from the source image but
       supply a component interpretation attribute of "gradient": */
    if (nbands > 1) {
	if (VImageAttrList (dest))
	    VDestroyAttrList (VImageAttrList (dest));
	VImageAttrList (dest) = VCopyAttrList (VImageAttrList (src));
	VImageNFrames (dest) = VImageNFrames (src);
	VImageNViewpoints (dest) = VImageNViewpoints (src);
	VImageNColors (dest) = VImageNColors (src);
    } else VCopyImageAttrs (src, dest);
    VImageNComponents (dest) = 2;
    VSetAttr (VImageAttrList (dest), VComponentInterpAttr, VBandInterpDict,
	      VLongRepn, (VLong) VBandInterpGradient);

    if (clipped)
	VWarning ("VImageGradient: Destination pixel(s) out of range");

    return dest;
}
