/*
 *  $Id: GradientSeq.c,v 1.1 1994/06/10 01:32:53 pope Exp pope $
 *
 *  Computes the gradient of an image sequence.
 */

/*
 *  Copyright 1994 University of British Columbia
 *
 *  Permission to use, copy, modify, distribute, and sell this software and its
 *  documentation for any purpose is hereby granted without fee, provided that
 *  the above copyright notice appears in all copies and that both that
 *  copyright notice and this permission notice appear in supporting
 *  documentation. UBC makes no representations about the suitability of this
 *  software for any purpose. It is provided "as is" without express or
 *  implied warranty.
 *
 *  Author: Arthur Pope, UBC Laboratory for Computational Intelligence
 */

/* From the Vista library: */
#include <vista/Vlib.h>
#include <vista/os.h>
#include <vista/VImage.h>

/* File identification string: */
VRcsId ("$Id: GradientSeq.c,v 1.1 1994/06/10 01:32:53 pope Exp pope $");


/*
 *  VImageGradientSeq
 *
 *  Computes spatiotemporal derivative of an image sequence using
 *  central differences with a specified kernel.
 *
 *  Suitable kernels are [    -1  0  1	   ] /  2
 *		     and [ 1  -8  0  8  -1 ] / 12
 *
 *  The destination image has k - 1 fewer frames than the source image, where
 *  k is the kernel size.
 *
 *  Restrictions: This version supports Float images only.
 */

VImage VImageGradientSeq (VImage src, VImage dest, VBand band,
			  int kernel_size, VDouble kernel[])
{
    int half, last_src_band, src_band, dest_band, i, j, k;
    int last_row, last_column;
    VDouble t;
    VFloat *src_pp, *dest_pp;

    /* Require an odd-sized kernel: */
    if (! (kernel_size & 1) || kernel_size < 3) {
	VWarning ("VImageGradientSeq: Invalid kernel size (%d)", kernel_size);
	return NULL;
    }
    half = kernel_size / 2;

    /* Check properties of the source image: */
    if (VPixelRepn (src) != VFloatRepn) {
	VWarning ("VImageGradientSeq: Source image must have float pixels");
	return NULL;
    }
    if (VImageNRows (src) < kernel_size ||
	VImageNColumns (src) < kernel_size ||
	VImageNFrames (src) < kernel_size) {
	VWarning ("VImageGradientSeq: Source image too small for kernel");
	return NULL;
    }
    if (VImageNFrames (src) != VImageNBands (src)) {
	VWarning ("VImageGradientSeq: Source image must be motion sequence");
	return NULL;
    }
    src_band = (band == VAllBands) ? half : band;
    if (src_band < half || src_band > VImageNBands (src) - half - 1) {
	VWarning ("VImageGradientSeq: "
		  "Band %d referenced in image of %d bands",
		  band, VImageNBands (src));
	return NULL;
    }
    last_src_band =
	(band == VAllBands) ? (VImageNBands (src) - half - 1) : band;
    last_row = VImageNRows (src) - half - 1;
    last_column = VImageNColumns (src) - half - 1;

    /* Check or create the destination image: */
    dest = VSelectDestImage ("VImageGradientSeq", dest,
			     (last_src_band - src_band + 1) * 3,
			     VImageNRows (src), VImageNColumns (src),
			     VFloatRepn);
    if (! dest)
	return NULL;

    /* Zero out the border of the destination image: */
    VFillImage (dest, VAllBands, 0.0);

    /* For each source band: */
    for (dest_band = 0; src_band <= last_src_band; src_band++) {

	/* Fill a destination band with horizontal derivatives: */
	for (i = half; i <= last_row; i++) {
	    dest_pp = VPixelPtr (dest, dest_band, i, half);
	    for (j = half; j < last_column; j++) {
		src_pp = VPixelPtr (src, src_band, i, j - half);
		for (k = 0, t = 0.0; k < kernel_size; k++)
		    t += *src_pp++ * kernel[k];
		*dest_pp++ = t;
	    }
	}
	dest_band++;

	/* Fill a destination band with vertical derivatives: */
	for (i = half; i <= last_row; i++) {
	    dest_pp = VPixelPtr (dest, dest_band, i, half);
	    for (j = half; j < last_column; j++) {
		src_pp = VPixelPtr (src, src_band, i - half, j);
		for (k = kernel_size - 1, t = 0.0; k >= 0; k--) {
		    t += *src_pp * kernel[k];
		    src_pp += VImageNColumns (src);
		}
		*dest_pp++ = t;
	    }
	}
	dest_band++;

	/* Fill a destination band with temporal derivatives: */
	for (i = half; i <= last_row; i++) {
	    dest_pp = VPixelPtr (dest, dest_band, i, half);
	    for (j = half; j < last_column; j++) {
		for (k = 0, t = 0.0; k < kernel_size; k++)
		    t += VPixel (src, src_band - half + k, i, j, VFloat) *
			kernel[k];
		*dest_pp++ = t;
	    }
	}
	dest_band++;
    }

    /* Adopt the band interpretation information from the source image but
       supply a component interpretation attribute of "gradient": */
    VCopyImageAttrs (src, dest);
    VSetBandInterp (dest,
		    VBandInterpNone, VImageNBands (dest) / 3,
		    VBandInterpNone, 1,
		    VBandInterpNone, 1,
		    VBandInterpGradient, 3);

    return dest;
}
