/*
 * $Id: ImageToPS.c,v 1.5 1994/06/15 23:14:49 pope Exp pope $
 *
 * This file contains routines for converting Vista images to
 * PostScript.
 *
 */

/*
 *  Copyright 1993, 1994 University of British Columbia
 *
 *  Permission to use, copy, modify, distribute, and sell this software and its
 *  documentation for any purpose is hereby granted without fee, provided that
 *  the above copyright notice appears in all copies and that both that
 *  copyright notice and this permission notice appear in supporting
 *  documentation. UBC makes no representations about the suitability of this
 *  software for any purpose. It is provided "as is" without express or
 *  implied warranty.
 *
 *  Author: Daniel Ko, UBC Laboratory for Computational Intelligence
 */

/* From the Vista library: */
#include "vista/Vlib.h"
#include "vista/mu.h"
#include "vista/os.h"
#include "vista/postscript.h"
#include "vista/VImage.h"

/* From the standard C library: */
#include <math.h>

/* File identification string: */
VRcsId ("$Id: ImageToPS.c,v 1.5 1994/06/15 23:14:49 pope Exp pope $");


/*
 *  Number of characters of image data placed on each PostScript line.
 */

#define charactersPerLine 60


/*
 *  VImageBandToPS
 *
 *  Convert a VImage band to PostScript as a gray-scale image.
 */

VBoolean VImageBandToPS (FILE *f, VImage image, VBand band)
{
    VFloat aspect = 1.0;
    VImage ubyte_image = NULL;
    int npixels, i, nchars;
    VUByte *pp;
    VBoolean result = FALSE;

    /* Check to ensure that "band" exists: */
    if (band < 0 || band >= VImageNBands(image)) {
	VWarning ("VImageBandToPS: Band %d referenced in image of %d band(s)",
		  band, VImageNBands (image));
	goto Done;
    }

    /* Get image attributes: */
    npixels = VImageNRows (image) * VImageNColumns (image);
    VGetAttr (VImageAttrList (image), VPixelAspectRatioAttr, NULL,
	      VFloatRepn, & aspect);

    /* Represent the image in unsigned 8-bit pixels: */
    if (VPixelRepn (image) != VUByteRepn) {
	ubyte_image = VConvertImageRange (image, NULL, band, VUByteRepn);
	if (ubyte_image == NULL)
	    goto Done;
	pp = VPixelPtr (ubyte_image, 0, 0, 0);
    } else pp = VPixelPtr (image, band, 0, 0);
    
    /* Output PostScript code to render image: */
    fprintf (f,
	     "gs\n"
	     "%%%%BeginData: %d Hex Lines\n"
	     "%d %d %d %g image_gray\n",
	     (2 * npixels + charactersPerLine - 1) / charactersPerLine + 1,
	     VImageNColumns (image), VImageNRows (image),
	     VImageNRows (image), VImageNColumns (image) * aspect);

    /* Output the pixel values in hex, in row-major order: */
    for (i = nchars = 0; i < npixels; i++) {
	fprintf (f, "%02x", (unsigned int) *pp++);
	nchars += 2;
	if (nchars >= charactersPerLine) {
	    fputc ('\n', f);
	    nchars = 0;
	}
    }
    if (nchars != 0)
	fputc ('\n', f);
    fputs ("%%EndData\ngr\n", f);
    result = TRUE;

Done:
    if (ubyte_image)
	VDestroyImage (ubyte_image);
    return result;
}


/*
 * VRGBImageToPS
 *
 * Convert Vista image to PostScript color image format.
 */

VBoolean VRGBImageToPS (FILE *f, VImage image, VBand band)
{
    VFloat aspect = 1.0;
    int npixels, i, nchars;
    VImage ubyte_images[3];
    VUByte *pp[3];
    VBoolean result = FALSE;

    ubyte_images[0] = ubyte_images[1] = ubyte_images[2] = NULL;

    /* Require a single-component RGB image: */
    if (VImageNComponents (image) != 1 ||
	VImageColorInterp (image) != VBandInterpRGB) {
	VWarning ("VRGBImageToPS: Image isn't single-component RGB");
	goto Done;
    }
    
    /* Check the band argument: */
    if (band < 0 || band + 3 > VImageNBands (image)) {
	VWarning ("VRGBImageToPS: Band %d referenced in image of %d band(s)",
		  band, VImageNBands (image));
	goto Done;
    }

    /* Get image attributes: */
    npixels = VImageNRows (image) * VImageNColumns (image);
    VGetAttr (VImageAttrList (image), VPixelAspectRatioAttr, NULL,
	      VFloatRepn, & aspect);

    /* Represent the image with unsigned 8-bit pixels: */
    if (VPixelRepn (image) != VUByteRepn)
	for (i = 0; i < 3; i++) {
	    if (! (ubyte_images[i] =
		   VConvertImageRange (image, NULL, band + i, VUByteRepn)))
		goto Done;
	    pp[i] = VPixelPtr (ubyte_images[i], 0, 0, 0);
	}
    else
	for (i = 0; i < 3; i++)
	    pp[i] = VPixelPtr (image, band + i, 0, 0);

    /* Print PostScript code to print image: */
    fprintf (f,
	     "gs\n"
	     "%%%%BeginData: %d Hex Lines\n"
	     "%d %d %d %g image_color\n",
	     (6 * npixels + charactersPerLine - 1) / charactersPerLine + 1,
	     VImageNColumns (image), VImageNRows (image),
	     VImageNRows (image), VImageNColumns (image) * aspect);

    /* Print the pixel values in hex, in row-major order: */
    nchars = 0;
    for (i = nchars = 0; i < npixels; i++) {
	fprintf (f, "%02x%02x%02x", (unsigned int) *pp[0]++,
		 (unsigned int) *pp[1]++, (unsigned int) *pp[2]++);
	nchars += 6;
	if (nchars >= charactersPerLine) {
	    fputc ('\n', f);
	    nchars = 0;
	}
    }
    if (nchars != 0)
	fputc ('\n', f);
    fputs ("%%EndData\ngr\n", f);
    result = TRUE;

Done:
    for (i = 0; i < 3; i++)
	if (ubyte_images[i])
	    VDestroyImage (ubyte_images[i]);
    return result;
}


/*
 *  VGradientImageToPS
 *
 *  Render a gradient image as a flow field.
 */

VBoolean VGradientImageToPS (FILE *f, VImage image, VBand band,
			     double arrow_width)
{
    int i, j;
    double u, v, mag, angle, limit;

    /* limit is a rough upper limit on the length of arrow allowed: */
    limit = VMax (VImageNRows (image), VImageNColumns (image)) * 2.0;

    /* Require a two-component gradient image: */
    if (VImageNComponents (image) != 2 ||
	VImageComponentInterp (image) != VBandInterpGradient) {
	VWarning ("VGradientImageToPS: Image isn't two-component gradient");
	return FALSE;
    }
    
    /* Check the band argument: */
    if (band < 0 || band + 2 > VImageNBands (image)) {
	VWarning ("VGradientImageToPS:"
		  "Band %d referenced in image of %d band(s)",
		  band, VImageNBands (image));
	return FALSE;
    }

    /* Draw a little arrow for each pixel with a well-defined gradient: */
    for (i = 1; i < VImageNRows (image); i++)
	for (j = 1; j < VImageNColumns (image); j++) {
	    u = VGetPixel (image, band, i, j);
	    v = VGetPixel (image, band + 1, i, j);
	    mag = sqrt (u * u + v * v);
	    if (mag == 0.0 || mag > limit)
		continue;
	    angle = atan2 (v, u);	/* CCW from X axis */
	    angle += 0.5 * M_PI;	/* CCW from row axis */
	    fprintf (f, "%g %g %g %g %g ar\n",
		     mag, arrow_width, angle * 180.0 / M_PI,
		     i + 0.5, j + 0.5);
	}

    return TRUE;
}
