/*
 *  $Id: Kernel.c,v 1.1 1994/06/10 01:33:00 pope Exp pope $
 *
 *  This file contains routines for generating kernels.
 */

/*
 *  Copyright 1994 University of British Columbia
 *
 *  Permission to use, copy, modify, distribute, and sell this software and its
 *  documentation for any purpose is hereby granted without fee, provided that
 *  the above copyright notice appears in all copies and that both that
 *  copyright notice and this permission notice appear in supporting
 *  documentation. UBC makes no representations about the suitability of this
 *  software for any purpose. It is provided "as is" without express or
 *  implied warranty.
 *
 *  Author: Art Pope, UBC Laboratory for Computational Intelligence
 */

/* From the Vista library: */
#include "vista/Vlib.h"
#include "vista/mu.h"
#include "vista/os.h"
#include "vista/VImage.h"

/* From the standard C library: */
#include <math.h>

/* File identification string: */
VRcsId ("$Id: Kernel.c,v 1.1 1994/06/10 01:33:00 pope Exp pope $");


/*
 *  VGaussianKernel
 *
 *  Generate a 1D Gaussian kernel.
 */

VBoolean VGaussianKernel (int ncoeffs, VDouble coeffs[], double s)
{
    double sum, x, b = -1.0 / (2.0 * s * s);
    VDouble *p1, *p2, *pend = coeffs + ncoeffs;

    /* Setup depends on whether the number of coefficients asked for is
       odd or even: */
    if (ncoeffs & 1) {
	p1 = & coeffs[ncoeffs / 2];
	*p1-- = 1.0;
	p2 = p1 + 2;
	x = 1.0;
	sum = 0.5;
    } else {
	p1 = & coeffs[ncoeffs / 2];
	p2 = p1 + 1;
	x = 0.5;
	sum = 0.0;
    }

    /* Fill the vector with coefficients, then make them sum to 1.0: */
    while (p2 < pend) {
	sum += *p1-- = *p2++ = exp (x * x * b);
	x += 1.0;
    }
    sum += sum;
    for (p1 = coeffs; p1 < pend; p1++)
	*p1 /= sum;

    return TRUE;
}


/*
 *  VGaussianD1Kernel
 *
 *  Generate a kernel for the first derivative of a 1D Gaussian.
 */

VBoolean VGaussianD1Kernel (int ncoeffs, VDouble coeffs[], double s)
{
    double x, a = 1.0 / (sqrt (2.0 * M_PI) * s * s * s);
    double b = -1.0 / (2.0 * s * s);
    VDouble *p1, *p2, *pend = coeffs + ncoeffs;

    /* Setup depends on whether the number of coefficients asked for is
       odd or even: */
    if (ncoeffs & 1) {
	p1 = & coeffs[ncoeffs / 2];
	*p1-- = 0.0;
	p2 = p1 + 2;
	x = 1.0;
    } else {
	p1 = & coeffs[ncoeffs / 2];
	p2 = p1 + 1;
	x = 0.5;
    }

    /* Fill the vector with coefficients: */
    while (p2 < pend) {
	*p2++ = - (*p1-- = a * x * exp (x * x * b));
	x += 1.0;
    }

    return TRUE;
}


/*
 *  VGaussianSplineKernel
 *
 *  Generate a 1D Gaussian kernel with its tails splined to zero.
 *  The kernel is Gaussian over the domain [-2s,2s].
 *  Over [-4s,-2s) and over (2s,4s], it splines to zero.
 *  Beyond -4s and 4s it is zero.
 *  Thus ncoeffs should be at least 8s + 1 to get a smoothly splined
 *  kernel.
 *
 *  Based on Mark Nitzberg, David Mumford, Takahiro Shiota, `Filtering,
 *  Segmentation and Depth', Lecture Notes in Computer Science 662,
 *  Springer-Verlag, 1993.
 */

VBoolean VGaussianSplineKernel (int ncoeffs, VDouble coeffs[], double s)
{
    double sum, x, t, b = -1.0 / (2.0 * s * s);
    double k2 = 2.0 * s, k4 = 4.0 * s;
    VDouble *p1, *p2, *pend = coeffs + ncoeffs;

    /* Setup depends on whether the number of coefficients asked for is
       odd or even: */
    if (ncoeffs & 1) {
	p1 = & coeffs[ncoeffs / 2];
	*p1-- = 1.0;
	p2 = p1 + 2;
	x = 1.0;
	sum = 0.5;
    } else {
	p1 = & coeffs[ncoeffs / 2];
	p2 = p1 + 1;
	x = 0.5;
	sum = 0.0;
    }

    /* Fill the vector with coefficients, then make them sum to 1.0: */
    while (p2 < pend) {
	if (x <= k2)
	    t = exp (x * x * b);
	else if (x <= k4) {
	    t = 4.0 - x / s;
	    t *= 1.0 / (16.0 * M_E * M_E) * t * t * t;
	} else t = 0.0;
	sum += *p1-- = *p2++ = t;
	x += 1.0;
    }
    sum += sum;
    for (p1 = coeffs; p1 < pend; p1++)
	*p1 /= sum;

    return TRUE;
}
