/*
 *  $Id: Link.c,v 1.5 1995/05/01 23:22:57 pope Exp pope $
 *
 *  The routines in this file assemble a set of linked edge pixels from
 *  an image.
 */

/*
 *  Copyright 1993, 1994 University of British Columbia
 *
 *  Permission to use, copy, modify, distribute, and sell this software and its
 *  documentation for any purpose is hereby granted without fee, provided that
 *  the above copyright notice appears in all copies and that both that
 *  copyright notice and this permission notice appear in supporting
 *  documentation. UBC makes no representations about the suitability of this
 *  software for any purpose. It is provided "as is" without express or
 *  implied warranty.
 *
 *  Author: David Lowe, UBC Laboratory for Computational Intelligence
 */

/* From the Vista library: */
#include "vista/Vlib.h"
#include "vista/os.h"
#include "vista/VEdges.h"
#include "vista/VImage.h"

/* File identification string: */
VRcsId ("$Id: Link.c,v 1.5 1995/05/01 23:22:57 pope Exp pope $");

/*------------------------ Data Structures --------------------------------*/

/* The points along an edge have their x,y coordinates and gradients stored
   first in the PointData array.  This array is expanded as needed to
   contain all the points in the current edge. Once all the data for an 
   edge has been accumulated, the edge is added to the edges data 
   structure.
*/
typedef struct PointStruct {
    VFloat x;
    VFloat y;
    VFloat gradient;
} PointRec, *Point;

static Point PointData;	/* Array of points in current edge */
static int PointCount;	/* Number of points so far in current edge */
static int MaxPoints;	/* Maximum PointCount before array must be expanded */

/* These global variables are defined for efficiency on inner loops. */
static int LowThreshold, HighThreshold;		/* Thresholding parameters. */
static VUByte **ImagePixels;			/* Current image array. */
static int ImageRows, ImageColumns;		/* Image array dimensions. */

#define ABS(x) (((x) < 0) ? -(x) : (x))

/* Forward function references. */
static void TrackEdge(int, int, VEdges, int);
static void StorePoint(int, int);
static int NextPoint(int *, int *);


/*---------------------- Link up Zero Crossings ---------------------------*/

/* Scan the image and link adjacent non-zero points into lists.
   This is applied only to the first band of the image.
   Note that the image is modified as scanned edge points are zeroed. 
*/
VEdges VLinkImage( VImage image, int low, int high, int minlength)
{
    int r, c;
    VEdges edges;

    /* Check that image is of type UByte. */
    if (VPixelRepn(image) != VUByteRepn) {
	VWarning ("VLinkImage: Image must be of type UByte.");
	return NULL;
    }

    /* Set up these global variables for inner loop efficiency. */
    LowThreshold = low;
    HighThreshold = high;
    ImagePixels = VPixelArray(image, VUByte)[0];  /* Select band 0. */
    ImageRows = VImageNRows(image);
    ImageColumns = VImageNColumns(image);

    edges = VCreateEdges(ImageRows, ImageColumns, 0, 3);

    PointData = NULL;
    PointCount = 0;
    MaxPoints = 0;

    for (r = 0; r < ImageRows; r++)
	for (c = 0; c < ImageColumns; c++)
	    /* Start tracking from points above high threshold. */
	    if ((int) ImagePixels[r][c] >= HighThreshold)
		TrackEdge(r, c, edges, minlength);

    VFree(PointData);
    return edges;
}


/* Start tracking an edge from the given position.  As each point is added
   to an edge, zero its pixel in the image.  After the end of the edge
   is reached, then reverse points in the array and start tracking from
   the beginning in the other direction.  Finally, the new edge is added to
   the VEdges structure.
*/
static void TrackEdge(int row, int col, VEdges edges, int minlength)
{
    Point p1, p2;
    VFloat temp;
    int i, r = row, c = col;
    VBoolean closed;

    /* Put first point in PointData array. */
    StorePoint(r, c);

    /* Track to end of connected points in first direction. */
    while (NextPoint(&r, &c))
	StorePoint(r, c);

    /* Remove points from top of stack that are below high threshold. */
    while (PointData[--PointCount].gradient < HighThreshold) ;
    PointCount++;

    /* Reverse list of points by exchanging pairs of points. */
    for (i = 0; i < PointCount / 2; i++) {
	p1 = & PointData[i];
	p2 = & PointData[PointCount - i - 1];
	temp = p1->x;  p1->x = p2->x;  p2->x = temp;
	temp = p1->y;  p1->y = p2->y;  p2->y = temp;
	temp = p1->gradient;  p1->gradient = p2->gradient;
	p2->gradient = temp;
    }

    /* Start tracking from original point in opposite direction. */
    r = row,
    c = col;
    while (NextPoint(&r, &c))
	StorePoint(r, c);

    /* Remove points from top of stack that are below high threshold. */
    while (PointData[--PointCount].gradient < HighThreshold) ;
    PointCount++;

    /* Add this edge to the set of edges. */
    if (PointCount >= minlength) {
	/* The edge is closed if the first and last points are adjacent */
	closed = (ABS(PointData[0].x - PointData[PointCount - 1].x) < 2  &&
	          ABS(PointData[0].y - PointData[PointCount - 1].y) < 2);
	VAddEdge (edges, NULL, PointCount, (VFloat *) PointData, closed, TRUE);
    }
    PointCount = 0;
}


/* Store the point with the given row and column coordinates at the next
   location in the PointData array, and zero this pixel in the image. */
static void StorePoint(int r, int c)
{
    Point pnt;

    /* Make sure that PointData is large enough to hold this point.
       If necessary, expand and reallocate the array. */
    if (PointCount >= MaxPoints) {
	MaxPoints += 1000;
	PointData = VRealloc (PointData, MaxPoints * sizeof (PointRec));
    }
    pnt = & PointData[PointCount++];

    /* Convert from (row,column) to (x,y) coordinates. */
    pnt->x = 0.5 + (VFloat) c;
    pnt->y = -0.5 + (VFloat) (ImageRows - r);
    pnt->gradient = (VFloat) ImagePixels[r][c];

    /* Zero the pixel in the image. */
    ImagePixels[r][c] = 0;
}


/* Given the location of an image point, find the best edge continuation
     from this point to one of its neighbors.  Return the position
     of the new point. */
static int NextPoint(int *rp, int *cp)
{
    int i, r, c;

    /* These arrays give the row and column offsets for each of the eight
       neighbors of a point, starting with those that are 4-connected. */
    static int roff[8] = {1, 0, -1, 0, 1, 1, -1, -1};
    static int coff[8] = {0, 1, 0, -1, 1, -1, -1, 1};

    /* Check each of the neighbors for a pixel above the low threshold. */
    for (i = 0; i < 8; i++) {
	r = *rp + roff[i];
	c = *cp + coff[i];
	if (r >= 0  &&  c >= 0  &&  r < ImageRows  &&  c < ImageColumns)
	    if ((int) ImagePixels[r][c] >= LowThreshold) {
		*rp += roff[i];
		*cp += coff[i];
		return TRUE;
	    }
    }
    return FALSE;
}
