/*
 *  $Id: OptFlowWLS.c,v 1.1 1994/06/10 01:33:05 pope Exp pope $
 *
 *  This file implements optical flow estimation by weighted least-squares
 *  fit of local first-order constraints to a model of constant velocity in
 *  each small spatial neighbourhood.
 */

/*
 *  Copyright 1994 University of British Columbia
 *
 *  Permission to use, copy, modify, distribute, and sell this software and its
 *  documentation for any purpose is hereby granted without fee, provided that
 *  the above copyright notice appears in all copies and that both that
 *  copyright notice and this permission notice appear in supporting
 *  documentation. UBC makes no representations about the suitability of this
 *  software for any purpose. It is provided "as is" without express or
 *  implied warranty.
 *
 *  Adapted by Arthur Pope, UBC Laboratory for Computational Intelligence,
 *  from implementation by John Barron, University of Western Ontario.
 */

/* From the Vista library: */
#include "vista/Vlib.h"
#include "vista/mu.h"
#include "vista/os.h"
#include "vista/VImage.h"

/* From the standard C library: */
#include <assert.h>
#include <math.h>

/* File identification string: */
VRcsId ("$Id: OptFlowWLS.c,v 1.1 1994/06/10 01:33:05 pope Exp pope $");

/* Later in this file: */
static void ComputeEigenvectors (float [2][2], float [2], float [2][2]);


/*
 *  VOptFlowWLS
 *
 *  Compute optical flow using a weighted least-squares fit of local
 *  first-order constraints to a model of constant velocity in each
 *  small spatial neighborhood.
 */

VImage VOptFlowWLS (VImage src, VImage dest, int method, double smoothing,
		    double threshold, double noise[2], double prior_variance)
{
    VImage result = NULL, src_float = NULL;
    VImage smooth = NULL, grad = NULL, filters[3];
    int smooth_size, nframes, margin, frame, i, j, i2, j2;
    float dI[3], det, mag = 1.0, t, weight[5][5], m[2][2], m_inv[2][2];
    float b[2], d[2], e[2][2], v[2];
    VBoolean success = FALSE;
    static VConvolvePadMethod pad_methods[3] =
	{ VConvolvePadTrim, VConvolvePadNone, VConvolvePadNone };
    static int shifts[3] = { 0, 0, 0 };
    static float coeff[5] = { 0.0625, 0.25, 0.375, 0.25, 0.0625 };
    static VDouble kernel[5] = { 1/12.0, -8/12.0, 0, 8/12.0, -1/12.0 };

    /* Smoothing is to a radius of about 3 standard deviations: */
    smooth_size = 6 * smoothing + 1;
    if ((smooth_size & 1) == 0)
	smooth_size++;
    filters[0] = filters[1] = filters[2] = NULL;

    /* Validate the source image: */
    if (VImageNFrames (src) != VImageNBands (src)) {
	VWarning ("VOptFlowWLS: Source image must be a pure motion sequence");
	return FALSE;
    }
    nframes = VImageNFrames (src) - smooth_size - VNumber (kernel) + 2;
    if (nframes < 1) {
	VWarning ("VOptFlowWLS: Source image needs %d frames for smoothing=%g",
		  smooth_size + VNumber (kernel) - 1, smoothing);
	return FALSE;
    }

    /* Validate/create destination image: */
    if (src == dest) {
	VWarning ("VOptFlowWLS: Source and destination must be different");
	return NULL;
    }
    result = VSelectDestImage ("VOptFlowLS", dest, nframes * 2,
			       VImageNRows (src), VImageNColumns (src),
			       VFloatRepn);
    if (! result)
	goto Done;
    t = 0.0;
    VFillImage (result, VAllBands, HUGE_VAL / t);	/* infinity */

    /* Convert the source image to Float if necessary: */
    if (VPixelRepn (src) != VFloatRepn) {
	src_float = VConvertImageCopy (src, NULL, VAllBands, VFloatRepn);
	if (! src_float)
	     goto Done;
    } else src_float = src;

    /* Smooth the source image: */
    filters[0] = VCreateImage (smooth_size, 1, 1, VDoubleRepn);
    filters[1] = VCreateImage (1, smooth_size, 1, VDoubleRepn);
    filters[2] = VCreateImage (1, 1, smooth_size, VDoubleRepn);
    VGaussianKernel (smooth_size, VImageData (filters[0]), smoothing);
    memcpy (VImageData (filters[1]), VImageData (filters[0]),
	    VImageSize (filters[0]));
    memcpy (VImageData (filters[2]), VImageData (filters[0]),
	    VImageSize (filters[0]));
    smooth = VConvolveImageSep (src_float, NULL, VAllBands,
				filters, pad_methods, shifts);
    if (! smooth)
	goto Done;

    /* Compute derivatives wrt X, Y, and T: */
    grad = VImageGradientSeq (smooth, NULL, VAllBands,
			      VNumber (kernel), kernel);
    if (! grad)
	goto Done;
    assert (VImageNFrames (grad) == nframes);
    margin = VNumber (kernel) / 2;

    /* Initialize a neighborhood weighting kernel: */
    for (i = 0; i < 5; i++)
	for (j = 0; j < 5; j++)
	    weight[i][j] = coeff[i] * coeff[j];

    /* Compute flow at each pixel: */
    for (frame = 0; frame < nframes; frame++)
	for (i = margin; i < VImageNRows (grad) - margin - 1; i++)
	    for (j = margin; j < VImageNColumns (grad) - margin - 1; j++) {

		if (method == 0) {
		    dI[0] = VPixel (grad, frame * 3, i, j, VFloat);
		    dI[1] = VPixel (grad, frame * 3 + 1, i, j, VFloat);
		    dI[2] = VPixel (grad, frame * 3 + 2, i, j, VFloat);
		    if (dI[2] == 0.0)
			continue;
		    mag = dI[0] * dI[0] + dI[1] * dI[1];
		    if (sqrt (mag) < threshold)
			continue;
		    t = - dI[2] / mag;
		    VPixel (result, frame * 2, i, j, VFloat) = t * dI[0];
		    VPixel (result, frame * 2 + 1, i, j, VFloat) = t * dI[1];
		    continue;
		}

		/* Compute elements of M by summing over a 5x5 neighborhood: */
		m[0][0] = m[0][1] = m[1][1] = b[0] = b[1] = 0.0;
		for (i2 = -2; i2 < 2; i2++)
		    for (j2 = -2; j2 < 2; j2++) {
			dI[0] = VPixel (grad, frame * 3, i + i2, j + j2,
					VFloat);
			dI[1] = VPixel (grad, frame * 3 + 1, i + i2, j + j2,
					VFloat);
			dI[2] = VPixel (grad, frame * 3 + 2, i + i2, j + j2,
					VFloat);
			if (noise[0] != 0.0)
			    mag = noise[0] * (dI[0] * dI[0] + dI[1] * dI[1]) +
				noise[1];
			t = weight [i2 + 2] [j2 + 2] / mag;
			m[0][0] += t * dI[0] * dI[0];
			m[1][1] += t * dI[1] * dI[1];
			m[0][1] += t * dI[0] * dI[1];
			b[0] += t * dI[0] * dI[2];
			b[1] += t * dI[1] * dI[2];
		    }
		m[1][0] = m[0][1];  /* M is symmetric */
		if (prior_variance != 0.0) {
		    m[0][0] += 1.0 / prior_variance;
		    m[1][1] += 1.0 / prior_variance;
		}

		/* Invert M: */
		det = m[0][0] * m[1][1] - m[1][0] * m[0][1]; /* det (M) */
		m_inv[0][0] = m[1][1] / det;
		m_inv[0][1] = m_inv[1][0] = - m[0][1] / det;
		m_inv[1][1] = m[0][0] / det;

		/* Solve for the velocity vector: */
		v[0] = - (m_inv[0][0] * b[0] + m_inv[0][1] * b[1]);
		v[1] = - (m_inv[1][0] * b[0] + m_inv[1][1] * b[1]);

		/* Compute the eigenvalues and eigenvectors of M:
		   (Eigenvectors occupy columns of e) */
		ComputeEigenvectors (m, d, e);

		/* Sort the eigenvalues/vectors so that largest is first: */
		if (d[0] < d[1]) {
		    #define Swap(a,b) t = a; a = b; b = t
		    Swap (d[0], d[1]);
		    Swap (e[0][0], e[0][1]);
		    Swap (e[1][0], e[1][1]);
		    #undef Swap
		}

		if (method == 1) {

		    /* To estimate full velocity vector, spread of M must
		       be small: */
		    if (d[1] >= threshold && det > 0.0) {
			VPixel (result, frame * 2, i, j, VFloat) = v[0];
			VPixel (result, frame * 2 + 1, i, j, VFloat) = v[1];
		    }

		} else {

		    /* To estimate normal velocity vector, spread of M need
		       be small in one direction only: */
		    if (d[0] >= threshold && fabs (det) > 0.00000001) {
			VPixel (result, frame * 2, i, j, VFloat) =
			    (v[0] * e[0][0] + v[1] * e[1][0]) * e[0][0];
			VPixel (result, frame * 2 + 1, i, j, VFloat) =
			    (v[0] * e[0][0] + v[1] * e[1][0]) * e[1][0];
		    }

		}
	    }
    success = TRUE;

Done:
    if (src_float != src)
	VDestroyImage (src_float);
    for (i = 0; i < 3; i++)
	if (filters[i])
	    VDestroyImage (filters[i]);
    if (smooth)
	VDestroyImage (smooth);
    if (grad)
	VDestroyImage (grad);
    if (success) {
	VCopyImageAttrs (src, result);
	VSetBandInterp (result,
			VBandInterpNone, nframes,	/* motion frames */
			VBandInterpNone, 1,		/* viewpoints */
			VBandInterpNone, 1,		/* color channels */
			VBandInterpGradient, 2);	/* vector components */
    } else if (result && result != dest)
	VDestroyImage (result);
    return result;
}


/*
 *  ComputeEigenvectors
 *
 *  Uses Jacobi rotations to compute all eigenvalues and eigenvectors
 *  of a 2x2 real symmetrix matrix. Based on Numerical Recipes
 *  implementation.
 */

static void ComputeEigenvectors (float a[2][2], float d[2], float e[2][2])
{
    float theta, tau, t, s, g, h, c;

    #define ROTATE(i)				\
	g = e[i][0];				\
	h = e[i][1];				\
	e[i][0] = g - s * (h + g * tau);	\
	e[i][1] = h + s * (g - h * tau);

    /* Initialize V to the identity matrix: */
    e[0][0] = e[1][1] = 1.0;
    e[0][1] = e[1][0] = 0.0;

    /* Initialize D to the diagonals of A: */
    d[0] = a[0][0];
    d[1] = a[1][1];

    /* If the off-diagonal element is zero, we're done: */
    if (a[0][1] == 0.0)
	return;

    /* Rotations are required. */
    g = 100.0 * fabs (a[0][1]);
    h = d[1] - d[0];
    if ((float) fabs (h) + g == (float) fabs (h))
	t = a[0][1] / h;
    else {
	theta = 0.5 * h / a[0][1];
	t = 1.0 / (fabs (theta) + sqrt (1.0 + theta * theta));
	if (theta < 0)
	    t = -t;
    }
    c = 1.0 / sqrt (1.0 + t * t);
    s = t * c;
    tau = s / (1.0 + c);
    h = t * a[0][1];
    d[0] -= h;
    d[1] += h;
    ROTATE (0);
    ROTATE (1);

#if 0
    /* Check the eigenvalues and eigenvectors obtained. */
    /* The eigenvectors should be orthogonal: */
    theta = acos ((e[0][0] * e[0][1] + e[1][0] * e[1][1]) /
		  (sqrt (e[0][0] * e[0][0] + e[1][0] * e[1][0]) *
		   sqrt (e[0][1] * e[0][1] + e[1][1] * e[1][1])));
    theta *= 180.0 / M_PI;
    if (theta < 89.5 || theta > 90.5)
	VError ("ComputeEigenvectors: Angle = %g", theta);

    /* Eigenvector test: */
    g = a[0][0] * e[0][0] + a[0][1] * e[1][0] - d[0] * e[0][0];
    h = a[1][0] * e[0][0] + a[1][1] * e[1][0] - d[0] * e[1][0];
    t = sqrt (g * g + h * h);
    if (t > 0.1)
	VError ("ComputeEigenvectors: Col 0 = %g", t);
    g = a[0][0] * e[0][1] + a[0][1] * e[1][1] - d[1] * e[0][1];
    h = a[1][0] * e[0][1] + a[1][1] * e[1][1] - d[1] * e[1][1];
    t = sqrt (g * g + h * h);
    if (t > 0.1)
	VError ("ComputeEigenvectors: Col 1 = %g", t);
#endif

    #undef ROTATE
}
