/*
 * $Id: PostScript.c,v 1.5 1994/06/15 23:14:49 pope Exp pope $
 *
 * This file contains routines for producing PostScript files.
 */

/*
 *  Copyright 1993, 1994 University of British Columbia
 *
 *  Permission to use, copy, modify, distribute, and sell this software and its
 *  documentation for any purpose is hereby granted without fee, provided that
 *  the above copyright notice appears in all copies and that both that
 *  copyright notice and this permission notice appear in supporting
 *  documentation. UBC makes no representations about the suitability of this
 *  software for any purpose. It is provided "as is" without express or
 *  implied warranty.
 *
 *  Author: Daniel Ko, Art Pope, UBC Laboratory for Computational Intelligence
 */

/* From the Vista library: */
#include "vista/Vlib.h"
#include "vista/file.h"
#include "vista/mu.h"
#include "vista/option.h"
#include "vista/os.h"
#include "vista/postscript.h"

/* From the standard C library: */
#include <ctype.h>
#include <math.h>
#include <pwd.h>
#include <sys/types.h>
#include <sys/utsname.h>

/* File identification string: */
VRcsId ("$Id: PostScript.c,v 1.5 1994/06/15 23:14:49 pope Exp pope $");


/*
 *  Parse table of PostScript command-line options.
 */

/* Default page height, width, margins, etc.: */
enum {
    pHeight, pWidth, tMargin, bMargin, lMargin, rMargin,
    fWidth, capHeight, nMeasures
};
static VStringConst measures[nMeasures] = {
    "11in",	/* default page height */
    "8.5in",	/* default page width */
    "1in",	/* default top margin */
    "1in",	/* default bottom margin */
    "1in",	/* default left margin */
    "1in",	/* default right margin */
    "0pt",	/* default width of framing rectangle line */
    "14pt"	/* default caption text height */
};

static VBoolean landscape;	/* landscape orientation */
static VBoolean frame_found;
static VString caption = NULL;

static VOptionDescRec options[] = {
    { "landscape", VBooleanRepn, 1, & landscape, VOptionalOpt, NULL,
	  "Landscape orientation" },
    { "height", VStringRepn, 1, & measures[pHeight], VOptionalOpt, NULL,
	  "Page height" },
    { "width", VStringRepn, 1, & measures[pWidth], VOptionalOpt, NULL,
	  "Page width" },
    { "top", VStringRepn, 1, & measures[tMargin], VOptionalOpt, NULL,
	  "Top margin width" },
    { "bottom", VStringRepn, 1, & measures[bMargin], VOptionalOpt, NULL,
	  "Bottom margin width" },
    { "left", VStringRepn, 1, & measures[lMargin], VOptionalOpt, NULL,
	  "Left margin width" },
    { "right", VStringRepn, 1, & measures[rMargin], VOptionalOpt, NULL,
	  "Right margin width" },
    { "frame", VStringRepn, 1, & measures[fWidth], & frame_found, NULL,
	  "Frame line width" },
    { "caption", VStringRepn, 1, & caption, VOptionalOpt, NULL,
	  "Caption text" },
    { "capheight", VStringRepn, 1, & measures[capHeight], VOptionalOpt, NULL,
	  "Caption font height" }
};


/*
 *  VParsePSOptions
 *
 *  Parse PostScript specific options from command line.
 */

VBoolean VParsePSOptions (int *argc, char **argv, VPSDescRec *ps)
{
    VBoolean r;
    float t;

    /* Parse command-line options: */
    if (! VParseCommand (VNumber (options), options, argc, argv))
	return FALSE;

    /* Convert specified units to PostScript units: */
    r = VParsePSDistance (measures[pHeight], "height", & ps->page_height);
    r &= VParsePSDistance (measures[pWidth], "width", & ps->page_width);
    r &= VParsePSDistance (measures[tMargin], "top", & ps->top_margin);
    r &= VParsePSDistance (measures[bMargin], "bottom", & ps->bottom_margin);
    r &= VParsePSDistance (measures[lMargin], "left", & ps->left_margin);
    r &= VParsePSDistance (measures[rMargin], "right", & ps->right_margin);
    if (frame_found)
	r &= VParsePSDistance (measures[fWidth], "frame", & ps->frame_width);
    r &= VParsePSDistance (measures[capHeight], "capheight",
			   & ps->caption_height);
    if (! r)
	return FALSE;

    /* Set the fields of ps: */
    ps->eps = FALSE;
    ps->npages = 0;
    ps->prog_name = argv[0];
    if (ps->landscape = landscape) {

	/* For landscape orientation, rotate margin and page dimensions: */
	t = ps->page_width;
	ps->page_width = ps->page_height;
	ps->page_height = t;

	t = ps->top_margin;
	ps->top_margin = ps->left_margin;
	ps->left_margin = ps->bottom_margin;
	ps->bottom_margin = ps->right_margin;
	ps->right_margin = t;
    }
    if (! frame_found)
	ps->frame_width = -1.0f;
    ps->caption = caption;

    return TRUE;
}


/*
 *  VParsePSDistance
 *
 *  Parse a distance measure (e.g. 8.5in), converting it to PostScript units.
 */

VBoolean VParsePSDistance (VStringConst str, VStringConst keyword,
			   float *value)
{
    const char *cp;
    char buf[100];
    int i, len = strlen (str), ch1, ch2;
    float mult;
    static struct {
	char abbrev[2];
	float mult;
    } units[] = {
        { "pt", 0.996 },		/* 1 point */
        { "pc", 11.96 },		/* 1 pica = 12 points */
        { "in", 72.0 },			/* 1 inch = 72.27 points */
        { "bp", 1.0 },			/* 72 big points = 1 inch */
        { "cm", 28.4 },			/* 2.54 cm = 1 inch */
        { "mm", 2.84 },			/* 10 mm = 1 cm */
        { "dd", 1.07 },			/* 1157 didot points = 1238 pt */
        { "cc", 12.8 },			/* 1 cicero = 12 dd */
        { "sp", 1.52E-5 }		/* 65536 scaled points = 1 pt */
    };

    if (len > sizeof (buf)) {
	fprintf (stderr, "Option -%s value too long\n", keyword);
	return FALSE;
    }

    /* Take two trailing letters as a unit name: */
    if (len > 2 && isalpha (str[len - 2]) && isalpha (str[len - 1])) {
	ch1 = tolower (str[len - 2]);
	ch2 = tolower (str[len - 1]);
	len -= 2;
	memcpy (buf, str, len);
	buf[len] = 0;
	cp = buf;
    } else {
	ch1 = DefaultPSUnits[0];
	ch2 = DefaultPSUnits[1];
	cp = str;
    }

    /* Look up the unit name: */
    for (i = 0; i < VNumber (units); i++)
	if (units[i].abbrev[0] == ch1 && units[i].abbrev[1] == ch2)
	    break;
    if (i == VNumber (units)) {
	fprintf (stderr, "Option -%s units (%c%c) not recognized\n",
		 keyword, ch1, ch2);
	return FALSE;
    }
    mult = units[i].mult;

    /* Parse the prefix as a decimal floating point number: */
    *value = strtod (cp, (char **) & cp) * mult;
    if (cp && *cp) {
	fprintf (stderr, "Option -%s has incorrect value %s\n", keyword, str);
	return FALSE;
    }
    return TRUE;
}


/*
 *  VReportValidPSOptions
 *
 *  Print, to stderr, a summary of PostScript program options.
 */

void VReportValidPSOptions (void)
{
    VReportValidOptions (VNumber (options), options);
}


/*
 *  VBeginPSDoc
 */

void VBeginPSDoc (FILE *f, VPSDescRec *ps, VBooleanPromoted eps,
		  VStringConst title)
{
    time_t now;
    struct passwd *who;
    struct utsname utsname;

    ps->eps = eps;
    ps->bb_set = FALSE;
    ps->bb_left = ps->bb_bottom = DBL_MAX;
    ps->bb_right = ps->bb_top = - DBL_MAX;
    fprintf (f, "%%!PS-Adobe-3.0%s\n", ps->eps ? " EPSF-3.0" : "");
    fputs ("%%BoundingBox: (atend)\n", f);
    fprintf (f, "%%%%Creator: %s\n", ps->prog_name);
    time (& now);
    fprintf (f, "%%%%CreationDate: %s", ctime (& now));
    if (who = getpwuid (getuid ()))
	fprintf(f, "%%%%For: %s@%s (%s)\n",
		who->pw_name, (uname (& utsname) == -1) ?
		"<unknown host>" : utsname.nodename, who->pw_gecos);
    fprintf (f, "%%%%Orientation: %s\n",
	     ps->landscape ? "Landscape" : "Portrait");
    if (! ps->eps)
	fputs ("%%PageOrder: Ascend\n", f);
    fprintf (f, "%%%%Pages: %s\n", ps->eps ? "0" : "atend");
    if (title)
	fprintf (f, "%%%%Title: %s\n", title);
}


/*
 *  VEndPSDoc
 */

void VEndPSDoc (FILE *f, VPSDescRec *ps)
{
    fputs ("%%Trailer\n", f);

    if (! ps->bb_set) {
	ps->bb_left = ps->left_margin;
	ps->bb_bottom = ps->bottom_margin;
	ps->bb_top = ps->page_height - ps->top_margin;
	ps->bb_right = ps->page_width - ps->right_margin;
    }

    /* For landscape orientation, rotate bounding box: */
    if (ps->landscape)
	fprintf (f, "%%%%BoundingBox: %d %d %d %d\n",
		 (int) floor (ps->page_height - ps->bb_top),
		 (int) floor (ps->bb_left),
		 (int) ceil (ps->page_height - ps->bb_bottom),
		 (int) ceil (ps->bb_right));
    else fprintf (f, "%%%%BoundingBox: %d %d %d %d\n",
		  (int) floor (ps->bb_left),
		  (int) floor (ps->bb_bottom),
		  (int) ceil (ps->bb_right),
		  (int) ceil (ps->bb_top));

    if (! ps->eps)
	fprintf (f, "%%%%Pages: %d\n", ps->npages);

    fputs ("%%EOF\n", f);
}


/*
 *  VEndPSHeader
 */

void VEndPSHeader (FILE *f, VPSDescRec *ps)
{
    fputs ("%%EndComments\n", f);
}


/*
 *  VBeginPSPage
 */

void VBeginPSPage (FILE *f, VPSDescRec *ps)
{
    ps->npages++;
    if (ps->eps && ps->npages == 2)
	VWarning ("VBeginPSPage: Multipage Encapsulated PostScript document "
		  "being produced");
    if (! ps->eps)
	fprintf (f, "%%%%Page: %d %d\n", ps->npages, ps->npages);
    fputs ("/VistaEnteredState save def\n", f);
}


/*
 *  VEndPSPage
 */

void VEndPSPage (FILE *f, VPSDescRec *ps)
{
    fputs ("VistaEnteredState restore\n", f);
    if (! ps->eps) {
	fputs ("showpage\n", f);
	fputs ("%%PageTrailer\n", f);
    }
}


/*
 *  VIncludePSFile
 */

VBoolean VIncludePSFile (FILE *f, VStringConst inc_file)
{
    FILE *inc_f;
    char line[1024];

    /* Open the file to be included ("-" indicates stdin): */
    if (! (inc_f = VOpenInputFile (inc_file, FALSE)))
	return FALSE;

    /* Copy from inc_f to f until EOF: */
    while (fgets (line, sizeof(line), inc_f) != NULL)
	fputs (line, f);

    /* Clean up: */
    fclose (inc_f);
    return TRUE;
}


/*
 *  VIncludePSProcSet
 */

VBoolean VIncludePSProcSet (FILE *f, VPSDescRec *ps, VStringConst filename,
			    VStringConst proc_set_name, VStringConst version,
			    VStringConst revision)
{
    fprintf (f, "%%%%BeginResource: procset %s %s %s\n",
	     proc_set_name, version, revision);

    if (! VIncludePSFile (f, filename))
	return FALSE;

    fputs ("%%EndResource\n", f);

    return TRUE;
}


/*
 *  VExtendPSBoundingBox
 *
 *  Extend the bounding box to encompass a specified rectangle on the page.
 */

void VExtendPSBoundingBox (VPSDescRec *ps, double left, double bottom,
			   double right, double top)
{
    ps->bb_set = TRUE;
    ps->bb_left = VMin (ps->bb_left, left);
    ps->bb_bottom = VMin (ps->bb_bottom, bottom);
    ps->bb_right = VMax (ps->bb_right, right);
    ps->bb_top = VMax (ps->bb_top, top);
}


/*
 *  VBeginVPSDoc
 */

VBoolean VBeginVPSDoc (FILE *f, VPSDescRec *ps, VBooleanPromoted eps,
		       VStringConst title)
{
    VBeginPSDoc (f, ps, eps, title);

    fputs ("%%DocumentNeededResources: font " VPSCaptionFont "\n", f);
    VEndPSHeader (f, ps);

    fputs ("%%BeginProlog\n", f);
    if (! VIncludeVPSProcSet (f, ps))
	return FALSE;
    fputs ("%%EndProlog\n", f);

    fputs ("%%BeginSetup\n", f);
    fputs ("%%IncludeResource: font " VPSCaptionFont "\n", f);
    fputs ("%%EndSetup\n", f);

    return TRUE;
}


/*
 *  VEndVPSDoc
 */

void VEndVPSDoc (FILE *f, VPSDescRec *ps)
{
    VEndPSDoc (f, ps);
}


/*
 *  VIncludeVPSProcSet
 *
 *  Include the "vista_procs" procedure set.
 */

VBoolean VIncludeVPSProcSet (FILE *f, VPSDescRec *ps)
{
    return VIncludePSProcSet (f, ps, VPSProcSetFilename, VPSProcSetName,
			      VPSProcSetVersion, VPSProcSetRevision);
}


/*
 *  VBeginVPSPage
 *
 *  Draw frame and caption if requested.
 *  Transform the user space to one with the following properties:
 *   o aspect ratio = 1
 *   o origin at top left hand corner
 *   o a point is specified by (row, column)
 *   o row increases downward
 *   o column increases to the right
 *   o row extends from 0 to <nrows>
 *   o column extends from 0 to <ncolumns>
 *   o occupies the largest feasible rectangle that is still
 *     completely within the usable area in the page.
 *     - a feasible rectangle has width to height ratio equals
 *	 <ncolumns>/<nrows>
 *     - the usable area is whole page less margins less
 *	 frame less caption
 *     - the feasible rectangle is centered within the usable area
 */

void VBeginVPSPage (FILE *f, VPSDescRec *ps, double nrows, double ncolumns)
{
    float xmin, xmax, ymin, ymax, rscale, cscale, scale;
    float ytl, xtl, xsize, ysize, capy, fw = VMax (1.0f, ps->frame_width);

    VBeginPSPage (f, ps);
    if (! ps->eps) {
	fputs ("%%PageResources: font " VPSCaptionFont "\n", f);
	fputs ("%%BeginPageSetup\n", f);
    }
    fputs ("VistaDict begin\n", f);

    /* First take care of landscape orientation: */
    if (ps->landscape)
	fprintf (f, "%g 0 translate 90 rotate\n", ps->page_height);

    /* Determine the dimensions of the usable area: */
    xmin = ps->left_margin;
    xmax = ps->page_width - ps->right_margin;
    ymin = ps->bottom_margin;
    ymax = ps->page_height - ps->top_margin;
    if (ps->frame_width > 0.0f) {
	xmin += fw;
	xmax -= fw;
	ymin += fw;
	ymax -= fw;
    }
    if (ps->caption)
	ymin += (VPSAboveCaption + VPSBelowCaption + 1.0f) *
	    ps->caption_height;

    /* Determines the largest feasible rectangle: */
    /* (Scales are expressed in number of PS units per row or column) */
    rscale = (ymax - ymin) / (float) nrows;
    cscale = (xmax - xmin) / (float) ncolumns;
    scale = VMin (rscale, cscale);
    xsize = scale * ncolumns;
    ysize = scale * nrows;

    /* Locate top left hand corner: */
    xtl = (xmin + xmax - xsize) / 2.0f;
    ytl = (ymin + ymax + ysize) / 2.0f;

    /* If requested, draw a frame around the place the image will go: */
    if (ps->frame_width >= 0.0f) {
	fprintf (f, "%g %g %g %g %g frame\n",
		 xtl - fw / 2.0f, ytl - ysize - fw / 2.0f,
		 xsize + fw, ysize + fw, ps->frame_width);
	VExtendPSBoundingBox (ps, xtl - fw, ytl - ysize - fw,
			      xtl + xsize + fw, ytl + fw);
    } else VExtendPSBoundingBox (ps, xtl, ytl - ysize, xtl + xsize, ytl);

    /* Render the caption: */
    if (ps->caption) {
	capy = ytl - ysize - (VPSAboveCaption + 1.0f) * ps->caption_height;
	if (ps->frame_width > 0.0f)
	    capy -= fw;
	fprintf (f, "%g %g (%s) /%s %g caption\n",
		 capy, xtl + xsize / 2.0f, ps->caption,
		 VPSCaptionFont, ps->caption_height);
	/* ==> The following doesn't allow for the width of the caption. */
	VExtendPSBoundingBox (ps, xtl,
			      capy - VPSBelowCaption * ps->caption_height,
			      xtl + xsize, ytl);
    }

    /* Transform PostScript user space: */
    fprintf (f, "%g %g translate 270 rotate %g %g scale\n",
	     xtl, ytl, scale, scale);

    if (! ps->eps)
	fputs ("%%EndPageSetup\n", f);
}


/*
 *  VEndVPSPage
 */

void VEndVPSPage (FILE *f, VPSDescRec *ps)
{
    fputs ("end\n", f);
    VEndPSPage (f, ps);
}
