/*
 *  $Id: Scale.c,v 1.5 1994/06/15 23:14:49 pope Exp pope $
 *
 *  This file contains routines for scaling images.
 */

/*
 *  Copyright 1993, 1994 University of British Columbia
 *
 *  Permission to use, copy, modify, distribute, and sell this software and its
 *  documentation for any purpose is hereby granted without fee, provided that
 *  the above copyright notice appears in all copies and that both that
 *  copyright notice and this permission notice appear in supporting
 *  documentation. UBC makes no representations about the suitability of this
 *  software for any purpose. It is provided "as is" without express or
 *  implied warranty.
 *
 *  Author: Daniel Ko, UBC Laboratory for Computational Intelligence
 */

/* From the Vista library: */
#include "vista/Vlib.h"
#include "vista/mu.h"
#include "vista/os.h"
#include "vista/VImage.h"

/* From the standard C library: */
#include <math.h>

/* File identification string: */
VRcsId ("$Id: Scale.c,v 1.5 1994/06/15 23:14:49 pope Exp pope $");

/* Later in this file: */
static VImage BiLinearScale (VImage, VImage, VBand, int, int);


/*
 *  VScaleImage
 *
 *  Scales the row- and column-dimensions of an image by separate
 *  real-valued factors using bilinear interpolation.
 *  If necessary, the image to be scaled will be reduced to an 
 *  appropriate size before bilinear interpolation.
 *
 */

VImage VScaleImage (VImage src, VImage dest, VBand band,
		    double row_scale, double column_scale)
{
    int src_nrows, src_ncolumns, dest_nrows, dest_ncolumns;
    VImage tmp_image, result;
    int npixels, row_reduce_factor, column_reduce_factor;
    VPointer src_first; 

    /* Validates row_scale and column_scale. */
    if (row_scale <= 0.0 || column_scale <= 0.0) {
	VWarning ("VScaleImage: Scaling factors are not positive");
	return NULL;
    }

    /* Checks existence of the specified band(s) */
    if (! VSelectBand ("VScaleImage", src, band, & npixels, & src_first))
	return NULL;

    /* Extracts data from source image */
    src_nrows    = VImageNRows(src);
    src_ncolumns = VImageNColumns(src);

    /* Calculates dimensions of destination image */
    dest_nrows = src_nrows * row_scale;
    dest_ncolumns = src_ncolumns * column_scale;

    /* If dimensions of source and destination are the same, just copy: */
    if (src_nrows == dest_nrows && src_ncolumns == dest_ncolumns)
	return VCopyImage (src, dest, band);

    if (dest_nrows == 0 || dest_ncolumns == 0) {
	VWarning ("VScaleImage: Destination image would have zero dimension");
	return NULL;
    } 

    /* Calculates integral reduction factors */
    row_reduce_factor = 1.0 / row_scale;
    row_reduce_factor = (row_reduce_factor == 0 ? 1 : row_reduce_factor);

    column_reduce_factor = 1.0 / column_scale;
    column_reduce_factor =
	(column_reduce_factor == 0 ? 1 : column_reduce_factor);
    
    if (row_reduce_factor == 1 && column_reduce_factor == 1) {

	/* Reduction is not necessary. Do bilinear interpolation directly: */
	result = BiLinearScale (src, dest, band, dest_nrows, dest_ncolumns);

    } else {

	/* Performs reduction first to make sure that the scaling factors for
	   bilinear interpolation are as close to 1 as possible: */
	tmp_image = VReduceImage(src, NULL, band, 
				 row_reduce_factor, column_reduce_factor);

	if (VImageNRows(tmp_image) == dest_nrows &&
	    VImageNColumns(tmp_image) == dest_ncolumns ) {

	  /* The reduced image already has the appropriate dimensions.
	     Don't perform bilinear interpolation. Just copy: */
	    result = VCopyImage (tmp_image, dest, VAllBands);

	} else {

	    /* Preforms bilinear interpolation: */
	    result = BiLinearScale (tmp_image, dest, band, 
				    dest_nrows, dest_ncolumns);
	}
	VDestroyImage(tmp_image);
    }
    
    return result;
}


/*
 *  BiLinearScale
 *
 *  Perform scaling in the row- and column-dimensions of an image
 *  using bilinear interpolation.
 *
 */

static VImage BiLinearScale (VImage src, VImage dest, VBand band, 
			     int dest_nrows, int dest_ncolumns)
{
    int src_nrows, src_ncolumns, src_nbands, dest_nbands; 
    int src_curr_band, dest_curr_band;          
    VRepnKind src_repn, dest_repn;
    int i, j, k;
    float row_gap, column_gap, row_offset, column_offset,
          row_pos, column_pos, x, y;
    int npixels;
    VPointer src_first;
    
    /* Checks existence of the specified band(s) */
    if (! VSelectBand ("VScaleImage", src, band, & npixels, & src_first))
	return NULL;

    /* Extract data from source image */
    src_nrows    = VImageNRows(src);
    src_ncolumns = VImageNColumns(src);
    src_nbands   = VImageNBands(src);
    src_repn     = VPixelRepn(src);
    
    /* If dimensions of source and destination are the same, just copy: */
    if (src_nrows == dest_nrows && src_ncolumns == dest_ncolumns)
	return VCopyImage (src, dest, band);
    
    if (dest_nrows == 0 || dest_ncolumns == 0) {
	VWarning ("VScaleImage: Destination image would have zero dimension");
	return NULL;
    }
    
    /* Checks the destination image, or create one if it is NULL: */
    dest_nbands = (band == VAllBands) ? src_nbands : 1;
    dest_repn = src_repn;
    dest = VSelectDestImage ("VScaleImage", dest, dest_nbands, dest_nrows,
                             dest_ncolumns, dest_repn);
    if (! dest)
	return NULL;
    
    /* Set up scaling parameters: */
    row_gap = (float) src_nrows / dest_nrows;
    column_gap = (float) src_ncolumns / dest_ncolumns;
    row_offset = row_gap / 2.0f - 0.5f;
    column_offset = column_gap / 2.0f - 0.5f;
    
    /*
     * Macros for getting the coordinates of the 4 source pixels surrounding
     * a particular destination pixel:
     */ 
    #define Top(r)    ( (int) (r) )
    #define Bottom(r) ( (int) (r) + 1 )
    #define Left(c)   ( (int) (c) )
    #define Right(c)  ( (int) (c) + 1 )
    
    /*
     * Returns the value of a specified source pixel. If the specified
     * pixel does not exist, then return zero.
     */
    #define GetSrcPixel(image, band, row, column, type)                      \
        ( row >= src_nrows || column >= src_ncolumns )                       \
        ? (type) 0                                                           \
        : VPixel(image, band, row, column, type)   

    /*
     * Performs bilinear interpolation:
     */
    #define Interpolation(type)                                              \
        {                                                                    \
            type tl_pixel, tr_pixel, bl_pixel, br_pixel;                     \
            type interpolation;                                              \
                                                                             \
            tl_pixel = GetSrcPixel(src, src_curr_band,                       \
                                   Top(row_pos), Left(column_pos), type);    \
            tr_pixel = GetSrcPixel(src, src_curr_band,                       \
                                   Top(row_pos), Right(column_pos), type);   \
            bl_pixel = GetSrcPixel(src, src_curr_band,                       \
                                   Bottom(row_pos), Left(column_pos), type); \
            br_pixel = GetSrcPixel(src, src_curr_band,                       \
                                   Bottom(row_pos), Right(column_pos), type);\
            interpolation = (type) ( x * y * tl_pixel +                      \
                               (1 - x) * y * tr_pixel +                      \
                               x * (1 - y) * bl_pixel +                      \
                               (1 - x) * (1 - y) * br_pixel );               \
            VPixel(dest, dest_curr_band, j, k, type) = interpolation;        \
         } 


    /* Determines the value of each pixel in the destination image: */
    for (i = 0; i < dest_nbands; i++) {
	dest_curr_band = (band == VAllBands ? i : 0);
	src_curr_band  = (band == VAllBands ? i : band);
	
	for (j = 0; j < dest_nrows; j++)
	    for (k = 0; k < dest_ncolumns; k++) {
		
		/* Determines the position of a destination pixel
		   with respect to a source pixel: */
		
		row_pos = j * row_gap + row_offset;
		if (row_pos < 0)
		    row_pos = 0;
		else if (row_pos > src_nrows-1)
		    row_pos = src_nrows-1;
		
		column_pos = k * column_gap + column_offset;
		if (column_pos < 0)
		    column_pos = 0;
		else if (column_pos > src_ncolumns-1)
		    column_pos = src_ncolumns-1;
		
		x = Right(column_pos) - column_pos;
		y = Bottom(row_pos) - row_pos;
		
		switch (src_repn) {
		case VBitRepn:		Interpolation(VBit);    break;
		case VUByteRepn:	Interpolation(VUByte);	break;
		case VSByteRepn:	Interpolation(VSByte);	break;
		case VShortRepn:	Interpolation(VShort);	break;
		case VLongRepn:		Interpolation(VLong);   break;
		case VFloatRepn:	Interpolation(VFloat);	break;
		case VDoubleRepn:	Interpolation(VDouble);	break;
		default: break;
		}
	    }
    }

    VCopyImageAttrs (src, dest);
    return dest;
    
    #undef Top
    #undef Bottom
    #undef Left
    #undef Right
    #undef GetSrcPixel
    #undef Interpolation
}


/*
 *  VReduceImage
 *
 *  Reduces the row- and column-dimensions of an image
 *  by separate integer-valued factors.
 */

VImage VReduceImage (VImage src, VImage dest, VBand band, 
		     int row_scale, int column_scale)
{
    int src_nrows, src_ncolumns, src_nbands,
        dest_nrows, dest_ncolumns, dest_nbands;
    int src_curr_band, dest_curr_band; 
    int i, j, k, x, y;
    VRepnKind src_repn, dest_repn;
    int npixels;
    VPointer src_first;

    /* Validates row_scale and column_scale: */
    if (row_scale <= 0 || column_scale <= 0) {
	VWarning ("VReduceImage: Scaling factors are not positive");
	return NULL;
    }
    
    /* If scales are 1, just copy: */
    if (row_scale == 1 && column_scale == 1)
	return VCopyImage (src, dest, band);
    
    /* Checks existence of the specified band(s) */
    if (! VSelectBand ("VReduceImage", src, band, & npixels, & src_first))
	return NULL;
    
    /* Extracts data from source image */
    src_nrows    = VImageNRows(src);
    src_ncolumns = VImageNColumns(src);
    src_nbands   = VImageNBands(src);
    src_repn     = VPixelRepn(src);
    
    /* Determines the dimensions of the destination image */
    dest_nrows    = src_nrows / row_scale;       /* floor is used */
    dest_ncolumns = src_ncolumns / column_scale; /* floor is used */
    
    if (dest_nrows == 0 || dest_ncolumns == 0) {
	VWarning ("VReduceImage: Destination image would have zero dimension");
	return NULL;
    }
    
    /* Checks the destination image, or create one if it is NULL: */
    dest_nbands = (band == VAllBands) ? src_nbands : 1;
    dest_repn = src_repn;
    dest = VSelectDestImage ("VReduceImage", dest, dest_nbands, dest_nrows,
			     dest_ncolumns, dest_repn);
    if (! dest)
	return NULL;
    
    /*
     * 
     * Determines the block of source pixels to which a particular
     * destination pixel maps and takes the average of that block of pixels 
     * to be the value of the destination pixel.
     *
     */

    /* Version for interger-valued source pixels: */
    #define IntAverage(type)                                                 \
        {                                                                    \
            VDouble average = 0;                                             \
            VLong npix = row_scale * column_scale;                           \
            type xy_pixel;                                                   \
            for (x = j*row_scale; x < (j+1)*row_scale; x++)                  \
              for (y = k*column_scale; y < (k+1)*column_scale; y++) {        \
                xy_pixel = VPixel(src, src_curr_band, x, y, type);           \
                average += ( (VDouble) xy_pixel ) / npix;                    \
	      }                                                              \
            VPixel(dest, dest_curr_band, j, k, type) = rint (average);	     \
	}

    /* Version for real-valued source pixels: */
    #define RealAverage(type)                                                \
        {                                                                    \
            VDouble average = 0;                                             \
            VLong npix = row_scale * column_scale;                           \
            type xy_pixel;                                                   \
            for (x = j*row_scale; x < (j+1)*row_scale; x++)                  \
              for (y = k*column_scale; y < (k+1)*column_scale; y++) {        \
                xy_pixel = VPixel(src, src_curr_band, x, y, type);           \
                average += ( (VDouble) xy_pixel ) / npix;                    \
	      }                                                              \
            VPixel(dest, dest_curr_band, j, k, type) = (type) average;       \
	}

    /* Determines value of each pixel in the destination image: */
    for (i = 0; i < dest_nbands; i++) {
	dest_curr_band = (band == VAllBands ? i : 0);
	src_curr_band = (band == VAllBands ? i : band);
	
	for (j = 0; j < dest_nrows; j++)
	    for (k = 0; k < dest_ncolumns; k++) {
		switch (dest_repn) {
		case VBitRepn:       IntAverage(VBit);          break;
		case VUByteRepn:     IntAverage(VUByte);        break;
		case VSByteRepn:     IntAverage(VSByte);        break;
		case VShortRepn:     IntAverage(VShort);        break;
		case VLongRepn:      IntAverage(VLong);         break;
		case VFloatRepn:     RealAverage(VFloat);       break;
		case VDoubleRepn:    RealAverage(VDouble);      break;
		default: break;
		}
	    }
    }

    VCopyImageAttrs (src, dest);
    return dest;

    #undef IntAverage
    #undef RealAverage
}


/*
 * VExpandImage
 *
 * Expands the row- and column-dimensions of an image
 * by separate integer-valued factors.
 *
 */

VImage VExpandImage (VImage src, VImage dest, VBand band,
		     int row_scale, int column_scale)
{
    int src_nrows, src_ncolumns, src_nbands,
        dest_nrows, dest_ncolumns, dest_nbands;
    int src_curr_band, dest_curr_band; 
    int i, j, k;
    VRepnKind src_repn, dest_repn;
    int npixels;
    VPointer  src_first;

    /* Validates row_scale and column_scale. */
    if (row_scale <= 0 || column_scale <= 0) {
	VWarning ("VExpandImage: Scaling factors are not positive");
	return NULL;
    }
    
    /* If scales are 1, just copy: */
    if (row_scale == 1 && column_scale == 1)
	return VCopyImage (src, dest, band);
    
    /* Checks existence of the specified band(s) */
    if (! VSelectBand ("VExpandImage", src, band, & npixels, & src_first))
	return NULL;
    
    /* Extracts data from source image */
    src_nrows    = VImageNRows(src);
    src_ncolumns = VImageNColumns(src);
    src_nbands   = VImageNBands(src);
    src_repn     = VPixelRepn(src);

    /* Determines the dimensions of the destination image */
    dest_nrows    = src_nrows * row_scale;
    dest_ncolumns = src_ncolumns * column_scale;
    
    /* Checks the destination image, or create one if it is NULL */
    dest_nbands = (band == VAllBands) ? src_nbands : 1;
    dest_repn = src_repn;
    dest = VSelectDestImage ("VExpandImage", dest, dest_nbands, dest_nrows,
                             dest_ncolumns, dest_repn);
    if (! dest)
	return NULL;

    /*
     *
     * Determines the source pixel to which a particular destination pixel
     * maps and assigns the value of the source pixel to the destination pixel.
     *
     */

    #define AssignPixel(type) \
    { \
      VPixel(dest, dest_curr_band, j, k, type) = \
	  VPixel(src, src_curr_band, (j/row_scale), (k/column_scale), type); \
    }


    /* Determines the value of each pixel in the destination image. */
    for (i = 0; i < dest_nbands; i++) {
	dest_curr_band = (band == VAllBands ? i : 0);
	src_curr_band = (band == VAllBands ? i : band);

	for (j = 0; j < dest_nrows; j++)
	    for (k = 0; k < dest_ncolumns; k++) {
           
		switch (dest_repn) {
		case VBitRepn:       AssignPixel(VBit);          break;
		case VUByteRepn:     AssignPixel(VUByte);        break;
		case VSByteRepn:     AssignPixel(VSByte);        break;
		case VShortRepn:     AssignPixel(VShort);        break;
		case VLongRepn:      AssignPixel(VLong);         break;
		case VFloatRepn:     AssignPixel(VFloat);        break;
		case VDoubleRepn:    AssignPixel(VDouble);       break;
		default: break;
		}
	    }
    }

    VCopyImageAttrs (src, dest);
    return dest;

    #undef AssignPixel
}
