/*
 *  $Id: SegEdges.c,v 1.1 1994/06/10 01:33:20 pope Exp pope $
 *
 *  This file implements a routine that recursively splits edges segments
 *  into straight line segments. 
 */

/*
 *  Copyright 1994 University of British Columbia
 *
 *  Permission to use, copy, modify, distribute, and sell this software and its
 *  documentation for any purpose is hereby granted without fee, provided that
 *  the above copyright notice appears in all copies and that both that
 *  copyright notice and this permission notice appear in supporting
 *  documentation. UBC makes no representations about the suitability of this
 *  software for any purpose. It is provided "as is" without express or
 *  implied warranty.
 *
 *  Author: Arthur Pope, UBC Laboratory for Computational Intelligence
 */

/* From the Vista library: */
#include "vista/Vlib.h"
#include "vista/mu.h"
#include "vista/os.h"
#include "vista/VEdges.h"

/* From the standard C library: */
#include <math.h>

/* File identification string: */
VRcsId ("$Id: SegEdges.c,v 1.1 1994/06/10 01:33:20 pope Exp pope $");

/* State of an edge segmentation recursion: */
typedef struct {
    double accuracy;
    int granularity;
    VFloat **points;		/* array of segment's edge points */
    int *breakpoints;		/* list of breakpoints selected */
    int nbreakpoints;		/* current length of list */
} SegEdges;

/* Later in this file: */
static float SplitSegment (SegEdges *, int, int);
static int MaxPoint (SegEdges *, int, int, float *);
static void FitLine (int, VFloat **, VFloat [2], VFloat [2]);


/*
 *  VSegEdgesIntoLines
 *
 *  Breaks edges into straight-line segments, essentially by recursively
 *  subdividing at the point of maximum deviation from a straight line
 *  joining the edgepoints. Based on a method described in D.G. Lowe,
 *  `Three-dimensional object recognition from single two-dimensional
 *  images', Art. Intel. 31 (1987), pp. 355-395.
 *
 *  After subdividing it returns one of the following, according to the
 *  value of product:
 *	0 - the subdivided edges including internal points
 *	1 - the subdivided edges with internal points omitted
 *	2 - straight lines fit to the subdivided edges by least squares
 *	    optimization
 */

VEdges VSegEdgesIntoLines (VEdges src, double accuracy, int granularity,
			   double magnitude, int product)
{
    VEdges result;
    VEdge e, new_e;
    int max_npoints, i, j, npoints;
    int pf_size = VNPointFields (src) * sizeof (VFloat);
    int ef_size = VNEdgeFields (src) * sizeof (VFloat);
    VFloat *endpoints, *edge_fields, t;    
    SegEdges seg;

    seg.accuracy = accuracy * accuracy;		/* square is used in calcs */
    seg.granularity = granularity;

    /* Create an edge set for the results: */
    result = VCreateEdges (VEdgesNRows (src), VEdgesNColumns (src),
			   VNEdgeFields (src), VNPointFields (src));

    /* Allocate enough space to store a list of breakpoints for the
       longest edge segment: */
    for (e = VFirstEdge (src), max_npoints = 0;
	 VEdgeExists (e); e = VNextEdge (e))
	max_npoints = VMax (max_npoints, VEdgeNPoints (e));
    seg.breakpoints = VMalloc (sizeof (int) * max_npoints);
    seg.breakpoints[0] = 0;

    /* Segment each edge: */
    for (e = VFirstEdge (src); VEdgeExists (e); e = VNextEdge (e)) {
	seg.points = VEdgePointArray (e);
	seg.nbreakpoints = 1;
	SplitSegment (& seg, 0, VEdgeNPoints (e) - 1);
	seg.breakpoints[seg.nbreakpoints] = VEdgeNPoints (e) - 1;
	for (i = 0; i < seg.nbreakpoints; i++) {

	    /* Prune the segment if its average magnitude is too low: */
	    if (magnitude > 0.0 && VNPointFields (src) > 2) {
		t = 0.0;
		for (j = seg.breakpoints[i]; j <= seg.breakpoints[i + 1]; j++)
		    t += fabs (seg.points[j][2]);
		if (t / (j - seg.breakpoints[i]) < magnitude)
		    continue;
	    }
	    
	    /* Add the segment to the results edge set: */
	    npoints = seg.breakpoints[i + 1] - seg.breakpoints[i] + 1;
	    if (product == 0) {

		/* Retain all points within each segment: */
		new_e = VAddEdge (result, VEdgeFields (e), npoints,
				  seg.points[seg.breakpoints[i]], FALSE, TRUE);

		continue;
	    }

	    /* Just retain two endpoints: */
	    endpoints = VMalloc (2 * pf_size);
	    memcpy (endpoints, seg.points[seg.breakpoints[i]], pf_size);
	    memcpy (((char *) endpoints) + pf_size,
		    seg.points[seg.breakpoints[i + 1]], pf_size);
	    if (product == 2 && npoints > 2) {

		/* Fit a straight line to the segment points: */
		FitLine (npoints, & seg.points[seg.breakpoints[i]], 
			 endpoints, endpoints + VNPointFields (src));
	    }
	    edge_fields = VMalloc (ef_size);
	    memcpy (edge_fields, VEdgeFields (e), ef_size);
	    VAddEdge (result, edge_fields, 2, endpoints, FALSE, FALSE);
	}
    }
    VFree (seg.breakpoints);

    /* Copy attributes from src to result: */
    VDestroyAttrList (VEdgesAttrList (result));
    VEdgesAttrList (result) = VCopyAttrList (VEdgesAttrList (src));

    return result;
}


/*
 *  SplitSegment
 *
 *  Recursively split an edge segment into shorter segments as long as
 *  any shorter segment has a better length/deviation ratio.
 *
 *  Returns the maximum length/deviation ratio attained for the segment.
 */

static float SplitSegment (SegEdges *seg, int first, int last)
{
    int maxp, save_nbreakpoints = seg->nbreakpoints;
    float sig, sig1, sig2, maxsig;

    /* If the segment is too short, don't split it further: */
    if ((last - first) <= seg->granularity)
	return 0.0;

    /* Find the point of maximum deviation: */
    maxp = MaxPoint (seg, first, last, & sig);

    /* Create subsegments with recursive calls: */
    sig1 = SplitSegment (seg, first, maxp);
    seg->breakpoints[seg->nbreakpoints++] = maxp;
    sig2 = SplitSegment (seg, maxp, last);
    maxsig = VMax (sig1, sig2);

    /* If the best of lower segments is better than the current segment
       then retain the breakpoints defining the lower segments,
       otherwise drop them: */
    if (maxsig > sig) {
	return maxsig;
    } else {
	seg->nbreakpoints = save_nbreakpoints;
	return sig;
    }
}


/*
 *  MaxPoint
 *
 *  Return the point with maximum perpendicular distance from the line
 *  joining first and last. Also returns the square of the length/deviation
 *  ratio via *sigp.
 */

static int MaxPoint (SegEdges *seg, int first, int last, float *sigp)
{
    float x0, y0, dx, dy, dev, px, py, t, maxdev;
    int maxp, i;
    
    x0 = seg->points[first][0];
    y0 = seg->points[first][1];
    dx = seg->points[last][0] - x0;
    dy = seg->points[last][1] - y0;
    
    /* Locate the point of maximum deviation from a straight line through
       the first and last points. */
    for (i = maxp = first + 1, maxdev = 0.0; i < last; i++) {
	px = seg->points[i][0] - x0;
	py = seg->points[i][1] - y0;

	/* Compute squared distance to the line: */
	t = (dx * px + dy * py) / (dx * dx + dy * dy);
	px -= dx * t;
	py -= dy * t;
	dev = px * px + py * py;

	if (dev > maxdev) {
	    maxdev = dev;
	    maxp = i;
	}
    }
    
    /* The maximum deviation is assumed to be at least a couple of pixels
       in size to account for limitations on measurement accuracy. */
    if (maxdev < seg->accuracy)
	maxdev = seg->accuracy;

    /* Significance equals length over maximum deviation. We return its
       square, which equals length squared over maxdev value above. */
    *sigp = (dx * dx + dy * dy) / maxdev;
    return maxp;
}


/*
 *  FitLine
 *
 */

static void FitLine (int npoints, VFloat **points,
		     VFloat endpt0[2], VFloat endpt1[2])
{
    int i;
    VFloat x, y, xm, ym, x2, y2, xy, xm2, ym2, a, b, c, cost, sint, rho, t;

    /* Compute mean of coordinates and sum of their products: */
    xm = ym = x2 = y2 = xy = 0.0;
    for (i = 0; i < npoints; i++) {
	xm += x = points[i][0];
	ym += y = points[i][1];
	x2 += x * x;
	y2 += y * y;
	xy += x * y;
    }
    xm /= npoints;
    ym /= npoints;
	
    /* Compute parameters of the fitted line: */
    xm2 = xm * xm;
    ym2 = ym * ym;
    a = 2.0 * (xy - npoints * xm * ym);
    b = x2 - y2 - npoints * (xm2 - ym2);
    c = sqrt (a * a + b * b);
    cost = sqrt ((b + c) / (2.0 * c));
    if (fabs (cost) < 0.001) {
	cost = 0.0;
	sint = 1.0;
    } else sint = a / (2.0 * c * cost);
 
    /* Project the endpoints onto the line: */
    rho = -xm * sint + ym * cost;
    t = cost * points[0][0] + sint * points[0][1];
    endpt0[0] = t * cost - rho * sint;
    endpt0[1] = t * sint + rho * cost;
    t = cost * points[npoints - 1][0] + sint * points[npoints - 1][1];
    endpt1[0] = t * cost - rho * sint;
    endpt1[1] = t * sint + rho * cost;
}
