/*
 *  $Id: UbcIff.c,v 1.7 1994/06/15 23:14:49 pope Exp pope $
 *
 *  This file contains routines for allocating and freeing memory, with
 *  error checking.
 */

/*
 *  Copyright 1993, 1994 University of British Columbia
 *
 *  Permission to use, copy, modify, distribute, and sell this software and its
 *  documentation for any purpose is hereby granted without fee, provided that
 *  the above copyright notice appears in all copies and that both that
 *  copyright notice and this permission notice appear in supporting
 *  documentation. UBC makes no representations about the suitability of this
 *  software for any purpose. It is provided "as is" without express or
 *  implied warranty.
 *
 *  Author: Arthur Pope, UBC Laboratory for Computational Intelligence
 */

#ifdef SupportUbcIff

/* From the Vista library: */
#include "vista/Vlib.h"
#include "vista/os.h"
#include "vista/VImage.h"
#include "vista/mu.h"

/* From the standard C library: */
#include <assert.h>

/* File identification string: */
VRcsId ("$Id: UbcIff.c,v 1.7 1994/06/15 23:14:49 pope Exp pope $");

/* From UBC image.h (which can't be included because of name conflicts): */
#define	ImVersion 2		/* IFF version number			*/
#define	ImNParms  6		/* # of image parameters in header	*/
#define ImHdrSize 1024		/* physical header size in file		*/

static VStringConst ImHdrFmt =
    "Imagefile version-%d: nrows = %D ncols = %D bpp = %hd signed = %hd "
    "positive = %hd\n";


/*
 *  VReadUbcIff
 * 
 *  Read an image in UBC Image File Format, from a stream.
 *
 *  The only pixel sizes currently supported are those that map directly
 *  onto VImage pixel representations.
 */

VImage VReadUbcIff (FILE *f)
{
    int version, i;
    long nrows, ncolumns;
    short bpp, sign, positive;
    VRepnKind pixel_repn;
    VImage image;
    size_t size;
    char *cp, buf[ImHdrSize + 1];

    /* Read in the UBC IFF header: */
    i = fread (buf, 1, ImHdrSize, f);
    if (i == 0) {
	VWarning ("VReadUbcIff: Read from stream failed");
	return NULL;
    }
    if (i < ImHdrSize ||
	sscanf (buf, ImHdrFmt, & version, & nrows, & ncolumns, & bpp, & sign,
		& positive) != ImNParms) {
	VWarning ("VReadUbcIff: Unrecognizable UBC IFF header");
	return NULL;
    }
    if (version != ImVersion) {
	VWarning ("VReadUbcIff: Obsolete UBC IFF header");
	return NULL;
    }

    /* We presently don't support negative images, but there seem to be
       few of them. */
    if (! positive)
	VWarning ("VReadUbcIff: Negative image will be considered positive");

    /* Map bpp and sign attributes to a VImage pixel representation: */
    if (bpp == 1)
	pixel_repn = VBitRepn;
    else if (bpp == 8)
	pixel_repn = sign ? VSByteRepn : VUByteRepn;
    else if (bpp == 16 && sign)
	pixel_repn = VShortRepn;
    else if (bpp == 32 && sign)
	pixel_repn = VLongRepn;
    else {
	VWarning ("VReadUbcIff: %s image of %d bpp not supported",
		  sign ? "Signed" : "Unsigned", bpp);
	return NULL;
    }

    /* Create the image: */
    if (! (image = VCreateImage (1, (int) nrows, (int) ncolumns, pixel_repn)))
	return NULL;

    /* Make any annotation a "comment" attribute of the VImage: */
    cp = strchr (buf, '\n');
    assert (cp);
    if (*++cp) {
	buf[ImHdrSize] = 0;
	VSetAttr (VImageAttrList (image), VCommentAttr, NULL,
		  VStringRepn, VNewString (cp));
    }

    /* Read the pixel values with one big read: */
    size = nrows * ncolumns;
    if (pixel_repn == VBitRepn)
	size = (size + 7) / 8;
    else size *= bpp / 8;
    if ((i = fread (VImageData (image), 1, size, f)) != size) {
	VWarning ("VReadUbcIff: "
		  "Could only read %d out of %d bytes of image data",
		  i, size);
Fail:	VDestroyImage (image);
	return NULL;
    }

    /* Unpack the pixel values from the little-endian order used by IFF: */
    size = VImageSize (image);
    if (! VUnpackData (pixel_repn, nrows * ncolumns, VImageData (image),
		       VLsbFirst, & size, & VImageData (image), NULL))
	goto Fail;

    /* Return the result: */
    return image;
}


/*
 *  VWriteUbcIff
 *
 *  Write an image to a stream in UBC Image File Format.
 *
 *  Only integer pixel representations are currently supported.
 */

VBoolean VWriteUbcIff (FILE *f, VImage image, VBand band)
{
    int sign;
    char *cp, *comment, buf[ImHdrSize];
    size_t length;
    VPointer ptr;
    VBoolean alloced, success;

    /* The image must have the specified band: */
    if (band < 0 || band >= VImageNBands (image)) {
	VWarning ("VWriteUbcIff: Band %d referenced in image of %d bands",
		  band, VImageNBands (image));
	return FALSE;
    }

    /* Floating-point pixel values aren't supported: */
    if (VPixelRepn (image) == VFloatRepn ||
	VPixelRepn (image) == VDoubleRepn) {
	VWarning ("VWriteUbcIff: Floating-point pixel values not supported");
	return FALSE;
    }

    /* Compose the header: */
    sign = VPixelRepn (image) == VSByteRepn ||
	   VPixelRepn (image) == VShortRepn ||
	   VPixelRepn (image) == VLongRepn;
    sprintf (buf, ImHdrFmt, ImVersion, VImageNRows (image),
	     VImageNColumns (image), VPixelPrecision (image), sign, 1);

    /* Include any "comment" attribute: */
    cp = strchr (buf, '\n');
    assert (cp);
    cp++;
    length = ImHdrSize - (cp - buf);
    if (VGetAttr (VImageAttrList (image), VCommentAttr, NULL,
		  VStringRepn, & comment) == VAttrFound) {
	strncpy (cp, comment, length);
	buf[ImHdrSize] = 0;
    } else memset (cp, 0, length);
    if (fwrite (buf, 1, ImHdrSize, f) != ImHdrSize) {
	VWarning ("VWriteUbcIff: Write to stream failed");
	return FALSE;
    }

    /* Pack the pixel values into the little-endian order used by IFF: */
    if (! VPackData (VPixelRepn (image),
		     VImageNRows (image) * VImageNColumns (image),
		     VPixelPtr (image, band, 0, 0), VLsbFirst,
		     & length, & ptr, & alloced))
	return FALSE;

    /* Output the pixel values: */
    success = fwrite ((char *) ptr, 1, (int) length, f) == length;

    /* Free any buffer allocated to hold packed pixels: */
    if (alloced)
	VFree (ptr);

    if (! success)
	VWarning ("VWriteUbcIff: Write to stream failed");
    return success;
}

#endif
