/*
 *  $Id: ZeroC.c,v 1.6 1994/06/15 23:14:49 pope Exp pope $
 *
 *  This file contains routines for detecting zero crossings in images.
 */

/*
 *  Copyright 1993, 1994 University of British Columbia
 *
 *  Permission to use, copy, modify, distribute, and sell this software and its
 *  documentation for any purpose is hereby granted without fee, provided that
 *  the above copyright notice appears in all copies and that both that
 *  copyright notice and this permission notice appear in supporting
 *  documentation. UBC makes no representations about the suitability of this
 *  software for any purpose. It is provided "as is" without express or
 *  implied warranty.
 *
 *  Authors: David Lowe, Arthur Pope,
 *	     UBC Laboratory for Computational Intelligence
 */

/* From the Vista library: */
#include "vista/Vlib.h"
#include "vista/os.h"
#include "vista/VImage.h"

/* File identification string: */
VRcsId ("$Id: ZeroC.c,v 1.6 1994/06/15 23:14:49 pope Exp pope $");


/*
 *  ZeroC
 *
 *  This macro computes zero crossings within one band of an image.
 */

#define ZeroC(src_type, dest_type, work_type)				\
{									\
    src_type val, bval, rval;						\
    dest_type *dest_pp = & VPixel (result, dest_band, 0, 0, dest_type);	\
    work_type t, dm = VIsFloatPtRepn (VPixelRepn (result)) ?		\
	1.0 : VPixelMaxValue (result);					\
									\
    for (row = 0; row < nrows; row++)					\
	for (column = 0; column < ncolumns; column++) {			\
									\
	    /* Fetch the current source pixel, and pixels adjacent 	\
	       to the right and below: */				\
	    val = VPixel (src, src_band, row, column, src_type);	\
	    bval = VPixel (src, src_band, row + 1, column, src_type);	\
	    rval = VPixel (src, src_band, row, column + 1, src_type);	\
									\
	    /* Check for sign change below or to the right of the	\
	       current pixel, and set gradient magnitude: */		\
	    if (val >= 0) {						\
		if (bval < 0) {						\
		    t = val - bval;					\
									\
		    /* Choose max if there is also gradient to right: */\
		    if (rval < 0 && (work_type) (val - rval) > t)	\
			t = val - rval;					\
		} else t = (rval < 0) ? val - rval : 0;			\
	    } else {  /* val is < 0 */					\
		if (bval >= 0) {					\
		    t = bval - val;					\
		    if (rval >= 0 && (work_type) (rval - val) > t)	\
			t = rval - val;					\
		} else t = (rval >= 0) ? rval - val : 0;		\
	    }								\
	    *dest_pp++ = (t > dm) ? dm : t;				\
	}								\
}


/*  
 *  VZeroCrossings
 *
 *  Scan through the image marking all zero-crossing pixels with the local
 *  gradient and all other pixels zero.  A zero crossing is marked whenever 
 *  there is a sign change between a pixel and its right neighbor or lower
 *  neighbor.  The gradient for the pixel is the (maximum) difference
 *  across one of these sign changes.  The result is one row and column
 *  smaller than original region.
 */

VImage VZeroCrossings (VImage src, VImage dest, VBand band)
{
    int nbands, nrows, ncolumns, src_band, dest_band, row, column;
    VRepnKind dest_repn;
    VImage result;

    /* Ensure that the source image has the specified band: */
    if (band == VAllBands) {
	src_band = 0;
	nbands = VImageNBands (src);
    } else {
	if (! VSelectBand ("VZeroCrossings", src, band, NULL, NULL))
	    return NULL;
	src_band = band;
	nbands = 1;
    }

    /* Determine what properties the destination image should have: */
    nrows = VImageNRows (src) - 1;
    ncolumns = VImageNColumns (src) - 1;
    switch (VPixelRepn (src)) {

    case VBitRepn:
    case VUByteRepn:
	VWarning ("VZeroCrossings: Source image has unsigned pixels");
	return NULL;

    case VSByteRepn:
	dest_repn = VUByteRepn;
	break;

    case VShortRepn:
	dest_repn = VLongRepn;
	break;

    default:
	dest_repn = VPixelRepn (src);
    }

    /* If the destination image exists, check it; if not, create it: */
    result = VSelectDestImage ("VZeroCrossings", dest, nbands,
			       nrows, ncolumns, dest_repn);
    if (result == NULL)
	return NULL;
    if (src == result)
	result = VCreateImage (nbands, nrows, ncolumns, dest_repn);

    /* For each band to be processed: */
    for (dest_band = 0; dest_band < nbands; dest_band++, src_band++)

	switch (VPixelRepn (src)) {

	case VSByteRepn:
 	    ZeroC (VSByte, VUByte, unsigned short);
	    break;

	case VShortRepn:
 	    ZeroC (VShort, VLong, unsigned long);
	    break;

	case VLongRepn:
	    ZeroC (VLong, VLong, unsigned long);
	    break;

	case VFloatRepn:
	    ZeroC (VFloat, VFloat, VFloat);
	    break;

	case VDoubleRepn:
	    ZeroC (VDouble, VDouble, VDouble);
	    break;

	default:
	    break;
	}

    /* Give the result a copy of the src's image attributes: */
    if (src == dest) {
	VCopyImagePixels (result, dest, VAllBands);
	VDestroyImage (result);
	result = dest;
    } else VCopyImageAttrs (src, result);

    /* Except label the result as a gradient image: */
    if (VImageNComponents (result) == 1)
	VSetAttr (VImageAttrList (result), VComponentInterpAttr,
		  VBandInterpDict, VLongRepn, VBandInterpGradient);

    return result;
}
