/*
 *  $Id: v1to2.c,v 1.4 1994/06/15 23:14:49 pope Exp pope $
 * 
 *  The v1to2 program converts a data file from V1 format to V2.
 *
 *  An attribute such as this in a V1 file:
 *
 *	image: {
 *		data: 0
 *		nrows: 256
 *		etc.
 *	}
 *
 *  becomes this in a V2 file:
 *
 *	name: image {
 *		data: 0
 *		length: n
 *		nrows: 256
 *		etc.
 *	}
 *
 *  Something that gets read as attribute list from V1 files must be converted
 *  to a bundle in V2 files while adding the appropriate length attribute.
 *  Lengths are computed by counting backward from the end of the binary data
 *  section.
 */

/*
 *  Copyright 1993, 1994 University of British Columbia
 *
 *  Permission to use, copy, modify, distribute, and sell this software and its
 *  documentation for any purpose is hereby granted without fee, provided that
 *  the above copyright notice appears in all copies and that both that
 *  copyright notice and this permission notice appear in supporting
 *  documentation. UBC makes no representations about the suitability of this
 *  software for any purpose. It is provided "as is" without express or
 *  implied warranty.
 *
 *  Author: Art Pope, UBC Laboratory for Computational Intelligence
 */

/* From the Vista library: */
#include "vista/Vlib.h"
#include "vista/file.h"
#include "vista/mu.h"
#include "vista/option.h"
#include "vista/os.h"
#include "vista/VList.h"

/* File identification string: */
VRcsId ("$Id: v1to2.c,v 1.4 1994/06/15 23:14:49 pope Exp pope $");

/* Observations made by watching complaints from VReadFile: */
VBoolean is_V1_file = FALSE;		/* is V1 (not V2) format file */
VBoolean has_data = FALSE;		/* EOF isn't right after delimiter */

/* Binary data loaded from the input file: */
typedef struct {
    int len;				/* length of data block */
    char buf[4096];			/* block of data */
} DataBlock;

VList data_list;			/* list of data blocks */
long data_length;			/* total length of data */

/* Record of objects: */
int nobjects = 0;			/* number with binary data */
VBoolean multi_band_image = FALSE;	/* whether big image encountered */

/* Later in this file: */
void TraverseList (VAttrList);
void ReadBinaryData (FILE *);
void WriteBinaryData (FILE *);
void MyWarningHandler (VStringConst);


/*
 *  main
 *
 *  Program entry point.
 */

int main(int argc, char *argv[])
{
    FILE *in_file, *out_file;
    VAttrList list;

    /* Parse command line arguments: */
    VParseFilterCmd (0, NULL, argc, argv, & in_file, & out_file);
    
    /* Install a warning handler that suppresses complaints about
       file version, no EOF: */
    VSetWarningHandler (MyWarningHandler);

    /* Read the input file: */
    if (! (list = VReadFile (in_file, NULL)))
	exit (EXIT_FAILURE);

    if (is_V1_file) {

	/* Load the binary data into memory: */
	ReadBinaryData (in_file);

	/* Traverse the attribute list, adding a length attribute wherever
	   there is a data attribute: */
	TraverseList (list);
	if (data_length != 0)
	    VWarning ("File contains some binary data not accounted for");

    } else VWarning ("File isn't in V1 format");


    /* Write the output file: */
    if (! VWriteFile (out_file, list))
	exit (EXIT_FAILURE);

    /* Append the binary data: */
    if (is_V1_file)
	WriteBinaryData (out_file);

    /* Issue a warning for any multiband image: */
    if (multi_band_image)
	VWarning ("File contains multi-band image(s) -- "
		  "check their attributes");

    fprintf(stderr, "%s: Converted %d object%s.\n",
	    argv[0], nobjects, nobjects == 1 ? "" : "s");
    return EXIT_SUCCESS;
}


void TraverseList (VAttrList list)
{
    VAttrListPosn posn, subposn;
    VAttrList sublist;
    VBundle b;
    VLong data;

    /* Traverse the list from back to front, recursing on attribute list
       values: */
    for (VLastAttr (list, & posn); VAttrExists (& posn); VPrevAttr (& posn)) {
	if (VGetAttrRepn (& posn) != VAttrListRepn)
	    continue;
	VGetAttrValue (& posn, NULL, VAttrListRepn, & sublist);
	TraverseList (sublist);

	/* If a list contains a data attribute, turn it into a VBundle so
	   that it gets written out as a typed object instead of an attribute
	   list: */
	if (VLookupAttr (sublist, VDataAttr, & subposn)) {
	    b = VCreateBundle (VGetAttrName (& posn), sublist, 0, NULL);
	    VSetAttrValue (& posn, NULL, VBundleRepn, b);
	}
    }

    /* If the list includes a data attribute, insert a length attribute
       as well: */
    if (VLookupAttr (list, VDataAttr, & posn) &&
	VGetAttrValue (& posn, NULL, VLongRepn, & data)) {
	VInsertAttr (& posn, TRUE, VLengthAttr, NULL, VLongRepn,
 		     (VLong) data_length - data);
	data_length = data;
	nobjects++;
    }

    /* If the list includes an nbands attribute reporting > 1 band, issue
       a warning message about multiband images: */
    if (VLookupAttr (list, "nbands", & posn) &&
	VGetAttrValue (& posn, NULL, VLongRepn, & data) &&
	data > 1)
	multi_band_image = TRUE;
}


void ReadBinaryData (FILE *in_file)
{
    DataBlock *b;

    data_list = VListCreate ();
    data_length = 0;
    while (1) {
	b = VNew (DataBlock);
	b->len = fread (b->buf, 1, sizeof (b->buf), in_file);
	if (b->len == 0)
	    break;
	VListAppend (data_list, b);
	data_length += b->len;
    }

    if (has_data  != (data_length > 0))
	VWarning ("ReadBinaryData: Internal error");

    /* Some files have had NEWLINEs appended to them by emacs. Delete them: */
    b = VListLast (data_list);
    if (b && b->buf[b->len - 1] == '\n') {
	VWarning ("Deleted trailing newline");
	b->len--;
	data_length--;
    }
}


void WriteBinaryData (FILE *out_file)
{
    DataBlock *b;

    for (b = VListFirst (data_list); b; b = VListNext (data_list))
	if (fwrite (b->buf, 1, b->len, out_file) != b->len)
	    VError ("Write to stream failed");
}


void MyWarningHandler (VStringConst msg)
{
    if (strstr (msg, "Obsolete data file -- pipe it thru v1to2.")) {
	is_V1_file = TRUE;
	return;
    }
    if (is_V1_file && strstr (msg, "File continues beyond expected EOF.")) {
	has_data = TRUE;
	return;
    }
    fprintf (stderr, "%s", msg);
}
