/*
 *  $Id: vblur.c,v 1.1 1994/06/10 01:36:42 pope Exp pope $
 *
 *  The vblur program, which degrades an image by reducing contrast, blurring,
 *  and adding noise.
 */

/*
 *  Copyright 1994 University of British Columbia
 *
 *  Permission to use, copy, modify, distribute, and sell this software and its
 *  documentation for any purpose is hereby granted without fee, provided that
 *  the above copyright notice appears in all copies and that both that
 *  copyright notice and this permission notice appear in supporting
 *  documentation. UBC makes no representations about the suitability of this
 *  software for any purpose. It is provided "as is" without express or
 *  implied warranty.
 *
 *  Author: Arthur Pope, UBC Laboratory for Computational Intelligence
 */

/* From the Vista library: */
#include "vista/Vlib.h"
#include "vista/mu.h"
#include "vista/option.h"
#include "vista/os.h"
#include "vista/VImage.h"

/* From the standard C library: */
#include <math.h>

/* File identification string: */
VRcsId ("$Id: vblur.c,v 1.1 1994/06/10 01:36:42 pope Exp pope $");

/* Later in this file: */
static void DegradeImage (VImage);

/* Command line options: */
static VDouble contrast = 0.0, blur = 1.0, noise = 1.0;
static VOptionDescRec options[] = {
    { "contrast", VDoubleRepn, 1, & contrast, VOptionalOpt, NULL,
	  "Positive to increase; negative to decrease" },
    { "blur", VDoubleRepn, 1, & blur, VOptionalOpt, NULL,
	  "Std dev of Gaussian blurring" },
    { "noise", VDoubleRepn, 1, & noise, VOptionalOpt, NULL,
	  "Std dev of additive Gaussian noise" }
};


/*
 *  Program entry point.
 */

int main (int argc, char *argv[])
{
    FILE *in_file, *out_file;
    VAttrList list;
    VAttrListPosn posn;
    VImage image;
    int nimages = 0;

    /* Parse command line arguments and identify the input and output files: */
    VParseFilterCmd (VNumber (options), options, argc, argv,
		     & in_file, & out_file);

    /* Read the input file: */
    if (! (list = VReadFile (in_file, NULL)))
	exit (EXIT_FAILURE);

    /* Process each image: */
    for (VFirstAttr (list, & posn); VAttrExists (& posn); VNextAttr (& posn)) {
	if (VGetAttrRepn (& posn) != VImageRepn)
	    continue;
	VGetAttrValue (& posn, NULL, VImageRepn, & image);
	DegradeImage (image);
	nimages++;
    }

    /* Write the output file: */
    if (! VWriteFile (out_file, list))
	exit (EXIT_FAILURE);

    fprintf (stderr, "%s: Processed %d image%s.\n",
	     argv[0], nimages, nimages == 1 ? "" : "s");
    return EXIT_SUCCESS;
}


static void DegradeImage (VImage image)
{
    VImage tmp1, tmp2;

    /* Reduce the contrast of the source pixels: */
    if (contrast != 0.0 &&
	! VAdjustImage (image, image, VAllBands, 0.0, contrast))
	exit (EXIT_FAILURE);
    
    /* Convolve the image with a Gaussian filter: */
    if (blur > 0.0 &&
	! VGaussianConvolveImage (image, image, VAllBands, blur,
				  6 * blur + 1, VFilterGaussian))
	exit (EXIT_FAILURE);

    /* Add Gaussian noise: */
    if (noise > 0.0) {
	if (! (tmp1 = VConvertImageCopy (image, NULL,
					 VAllBands, VDoubleRepn)) ||
	    ! (tmp2 = VCreateImageLike (tmp1)) ||
	    ! VNormalNoiseImage (tmp2, VAllBands, 0.0, noise) ||
	    ! VImageOpI (tmp1, tmp1, VAllBands, VImageOpAdd,
			 tmp2, VAllBands, NULL, NULL) ||
	    ! VConvertImageCopy (tmp1, image, VAllBands, VPixelRepn (image)))
	    exit (EXIT_FAILURE);
	VDestroyImage (tmp1);
	VDestroyImage (tmp2);
    }

    /* Convert the image back to its origin pixel representation: */
}
