/*
 *  $Id: vcanny.c,v 1.5 1995/05/01 23:27:59 pope Exp pope $
 *
 *  The Canny edge detector.
 */

/*
 *  Copyright 1993, 1994 University of British Columbia
 *
 *  Permission to use, copy, modify, distribute, and sell this software and its
 *  documentation for any purpose is hereby granted without fee, provided that
 *  the above copyright notice appears in all copies and that both that
 *  copyright notice and this permission notice appear in supporting
 *  documentation. UBC makes no representations about the suitability of this
 *  software for any purpose. It is provided "as is" without express or
 *  implied warranty.
 *
 *  Author: Daniel Ko, UBC Laboratory for Computational Intelligence
 */

/* From the Vista library: */
#include <vista/Vlib.h>
#include <vista/mu.h>
#include <vista/option.h>
#include <vista/os.h>
#include <vista/VImage.h>

/* File identification string: */
VRcsId ("$Id: vcanny.c,v 1.5 1995/05/01 23:27:59 pope Exp pope $");

/* Where to put the results: */
enum {
    ResultAfter,
    ResultInsert,
    ResultReplace
};

VDictEntry ResultDict[] = {
    { "after", ResultAfter },
    { "insert", ResultInsert },
    { "replace", ResultReplace },
    { NULL }
};


/*
 *  main
 *
 *  Program entry point.
 */

int main (int argc, char **argv)
{
    static VDouble sigma = 1.0, noise_opt = 0.3;
    static VBoolean ori_opt = FALSE;
    static VLong result_opt = ResultReplace;
    static VOptionDescRec options[] = {
	{ "sigma", VDoubleRepn, 1, & sigma, VOptionalOpt, NULL,
	      "Standard deviation of the Gaussian" },
	{ "noise", VDoubleRepn, 1, & noise_opt, VOptionalOpt, NULL,
	      "Fraction of pixels considered edge noise" },
	{ "orientation", VBooleanRepn, 1, & ori_opt, VOptionalOpt, NULL,
	      "Whether to record gradient orientation" },
	{ "result", VLongRepn, 1, & result_opt, VOptionalOpt, ResultDict,
	      "Where to place results" }
    };
    FILE *in_file, *out_file;
    VAttrList list;
    VImage src, result, ori;
    VAttrListPosn posn, next;
    int nimages = 0;
    double noise;
    VLong low;
    
    /* Parse command line arguments: */
    VParseFilterCmd (VNumber (options), options, argc, argv,
		     & in_file, & out_file);
    
    /* Read source image(s): */
    if (! (list = VReadFile (in_file, NULL)))
	exit (EXIT_FAILURE);
    
    /* For each image: */
    for (VFirstAttr (list, & posn); VAttrExists (& posn); posn = next) {
	next = posn;
	VNextAttr (& next);

	if (VGetAttrRepn (& posn) != VImageRepn)
	    continue;
	VGetAttrValue (& posn, NULL, VImageRepn, & src);
	nimages++;

	/* Only band 0 is processed: */
	if (VImageNBands (src) > 1)
	    VWarning ("Only processing band 0 of %d-band source image",
		      VImageNBands (src));

	/* Detect edges in the image: */
	ori = NULL;
	noise = noise_opt;
	result = VCanny (src, NULL, 0, sigma,
			 ori_opt ? & ori : NULL, & noise);
	if (! result)
	    exit (EXIT_FAILURE);

	/* Record hysteresis thresholds derived from noise estimates: */
	low = (VLong) VMax (noise, 1.0);
	VSetAttr (VImageAttrList (result), "low_threshold", NULL,
		  VLongRepn, low);
	VSetAttr (VImageAttrList (result), "high_threshold", NULL,
		  VLongRepn, low * 2);
		  
	/* Store the results back in list of file objects: */
	switch (result_opt) {

	case ResultAfter:
	    if (ori_opt)
		VInsertAttr (& posn, TRUE, "edge_ori", NULL, VImageRepn, ori);
	    VInsertAttr (& posn, TRUE, "edges", NULL, VImageRepn, result);
	    break;

	case ResultInsert:
	    VAppendAttr (VImageAttrList (src), "edges", NULL, VImageRepn,
			 result);
	    if (ori_opt)
		VAppendAttr (VImageAttrList (src), "edge_ori", NULL,
			     VImageRepn, ori);
	    break;
	    
	case ResultReplace:
	    VSetAttrValue (& posn, NULL, VImageRepn, result);
	    if (ori_opt)
		VInsertAttr (& posn, TRUE, "edge_ori", NULL, VImageRepn, ori);
	    VDestroyImage (src);
	}
    }	

    /* Write the results to the output file: */
    if (VWriteFile (out_file, list))
	fprintf (stderr, "%s: Detected edges in %d image%s.\n",
		 argv[0], nimages, nimages == 1 ? "" : "s");
    
    return EXIT_SUCCESS;
}
