/*
 *  $Id: vconvert.c,v 1.6 1994/06/15 23:14:49 pope Exp pope $
 *
 *  The vconvert program, which converts images from one pixel representation
 *  to another.
 */

/*
 *  Copyright 1993, 1994 University of British Columbia
 *
 *  Permission to use, copy, modify, distribute, and sell this software and its
 *  documentation for any purpose is hereby granted without fee, provided that
 *  the above copyright notice appears in all copies and that both that
 *  copyright notice and this permission notice appear in supporting
 *  documentation. UBC makes no representations about the suitability of this
 *  software for any purpose. It is provided "as is" without express or
 *  implied warranty.
 *
 *  Author: Arthur Pope, UBC Laboratory for Computational Intelligence
 */

/* From the Vista library: */
#include "vista/Vlib.h"
#include "vista/mu.h"
#include "vista/option.h"
#include "vista/os.h"
#include "vista/VImage.h"

/* File identification string: */
VRcsId ("$Id: vconvert.c,v 1.6 1994/06/15 23:14:49 pope Exp pope $");

/* We support these methods of converting pixel values: */
enum {
    rangeMethod,		/* map source range to destination range */
    copyMethod,			/* just copy it without mapping */
    opt1Method,			/* map actual range of source pixel values */
    opt2Method,			/*   to full destination range */
    opt3Method,
    linearMethod		/* explicitly specified linear mapping */
};

/* Keywords for representing conversion methods: */
static VDictEntry method_dict[] = {
    { "range", rangeMethod }, 
    { "copy", copyMethod },
    { "opt1", opt1Method }, 
    { "opt2", opt2Method }, 
    { "opt3", opt3Method },
    { "linear", linearMethod },
    { NULL }
};

int main (int argc, char **argv)
{
    static VLong repn = VUByteRepn;
    static VLong method = rangeMethod;
    static VDouble a = 1.0, b = 0.0;
    static VOptionDescRec options[] = {
	{ "repn", VLongRepn, 1, & repn, VOptionalOpt, VNumericRepnDict,
	      "Desired pixel representation" },
        { "map", VLongRepn, 1, & method, VOptionalOpt, method_dict,
	      "Pixel value mapping" },
        { "a", VDoubleRepn, 1, & a, VOptionalOpt, NULL,
	      "Multiplication factor for -map linear" },
        { "b", VDoubleRepn, 1, & b, VOptionalOpt, NULL,
	      "Addition term for -map linear" }
    };
    FILE *in_file, *out_file;
    VAttrList list;
    VImage src, result;
    int nimages = 0;
    VAttrListPosn posn;

    /* Parse command line arguments: */
    VParseFilterCmd (VNumber (options), options, argc, argv,
		     & in_file, & out_file);

    /* Read source image(s): */
    if (! (list = VReadFile (in_file, NULL)))
	exit (EXIT_FAILURE);

    /* Convert each source image to the specified representation: */
    for (VFirstAttr (list, & posn); VAttrExists (& posn); VNextAttr (& posn)) {
	if (VGetAttrRepn (& posn) != VImageRepn)
	    continue;
	VGetAttrValue (& posn, NULL, VImageRepn, & src);

	switch (method) {

	case rangeMethod:
	    result = VConvertImageRange (src, NULL, VAllBands, (int) repn);
	    break;

	case copyMethod:
	    result = VConvertImageCopy (src, NULL, VAllBands, (int) repn);
	    break;

	case linearMethod:
	    result = VConvertImageLinear (src, NULL, VAllBands, (int) repn,
					  (double) a, (double) b);
	    break;

	default:
	    result = VConvertImageOpt (src, NULL, VAllBands, (int) repn,
				       method - opt1Method + 1);
	}
	VSetAttrValue (& posn, NULL, VImageRepn, result);
	VDestroyImage (src);
	nimages++;
    }

    /* Write the results to the output file: */
    if (! VWriteFile (out_file, list))
	exit (EXIT_FAILURE);
    
    fprintf (stderr, "%s: Converted %d image%s to %s representation.\n",
	     argv[0], nimages, nimages == 1 ? "" : "s",
	     VRepnName ((int) repn));
    return EXIT_SUCCESS;
}
