/*
 *  $Id: vfft.c,v 1.5 1994/06/15 23:14:49 pope Exp pope $
 *
 *  The vfft program computes the forward or inverse Fourier transform of
 *  an image.
 */

/*
 *  Copyright 1993, 1994 University of British Columbia
 *
 *  Permission to use, copy, modify, distribute, and sell this software and its
 *  documentation for any purpose is hereby granted without fee, provided that
 *  the above copyright notice appears in all copies and that both that
 *  copyright notice and this permission notice appear in supporting
 *  documentation. UBC makes no representations about the suitability of this
 *  software for any purpose. It is provided "as is" without express or
 *  implied warranty.
 *
 *  Author: Arthur Pope, UBC Laboratory for Computational Intelligence
 */

/* From the Vista library: */
#include "vista/Vlib.h"
#include "vista/mu.h"
#include "vista/option.h"
#include "vista/os.h"
#include "vista/VImage.h"

/* From the standard C library: */
#include <math.h>

/* File identification string: */
VRcsId ("$Id: vfft.c,v 1.5 1994/06/15 23:14:49 pope Exp pope $");


int main (int argc, char *argv[])
{
    static VDouble norm;
    static VBoolean inverse, norm_found;
    static VOptionDescRec options[] = {
        { "inverse", VBooleanRepn, 1, & inverse, VOptionalOpt, NULL,
	      "Compute inverse transform" },
        { "norm", VDoubleRepn, 1, & norm, & norm_found, NULL,
	      "Normalizing constant" }
    };
    FILE *in_file, *out_file;
    VAttrList list;
    VAttrListPosn posn;
    VImage src, result;
    VRepnKind repn;
    int nimages = 0;

    /* Parse command line arguments and identify the input and output files: */
    VParseFilterCmd (VNumber (options), options, argc, argv,
		     & in_file, & out_file);

    /* Read the input file: */
    if (! (list = VReadFile (in_file, NULL)))
	exit (EXIT_FAILURE);

    /* Transform each image: */
    for (VFirstAttr (list, & posn); VAttrExists (& posn); VNextAttr (& posn)) {
	if (VGetAttrRepn (& posn) != VImageRepn)
	    continue;
	VGetAttrValue (& posn, NULL, VImageRepn, & src);

	/* Ensure that the image has 2 components of Float or Double pixels: */
	switch (VImageNComponents (src)) {

	case 1:
	    repn = VPixelRepn (src) == VDoubleRepn ? VDoubleRepn : VFloatRepn;
	    result = VBuildComplexImage (src, VAllBands, NULL, 0, NULL, repn);
	    if (! result)
		exit (EXIT_SUCCESS);
	    VDestroyImage (src);
	    src = result;
	    break;

	case 2:
	    if (VPixelRepn (src) == VFloatRepn ||
		VPixelRepn (src) == VDoubleRepn)
		break;
	    result = VConvertImageRange (src, NULL, VAllBands, VFloatRepn);
	    if (! result)
		exit (EXIT_FAILURE);
	    VDestroyImage (src);
	    src = result;
	    break;

	default:
	    VError ("Source image must have 1 or 2 components");
	}

	/* If no normalizing constant was specified on the command line,
	   use sqrt (nrows * ncolumns): */
	if (! norm_found)
	    norm = sqrt ((double) VImageNRows (src) * VImageNColumns (src));

	if (! VImageFFT (src, src, inverse, (double) norm))
	    exit (EXIT_FAILURE);
	VSetAttrValue (& posn, NULL, VImageRepn, src);
	nimages++;
    }

    /* Write the output file: */
    if (! VWriteFile (out_file, list))
	exit (EXIT_FAILURE);

    fprintf (stderr, "%s: Transformed %d image%s.\n",
	     argv[0], nimages, nimages == 1 ? "" : "s");
    return EXIT_SUCCESS;
}
