/*
 * $Id: vlayout.c,v 1.5 1994/06/15 23:14:49 pope Exp pope $
 *
 *  A program for laying out several Encapsulated PostScript documents
 *  on a single page.
 */

/*
 *  Copyright 1993, 1994 University of British Columbia
 *
 *  Permission to use, copy, modify, distribute, and sell this software and its
 *  documentation for any purpose is hereby granted without fee, provided that
 *  the above copyright notice appears in all copies and that both that
 *  copyright notice and this permission notice appear in supporting
 *  documentation. UBC makes no representations about the suitability of this
 *  software for any purpose. It is provided "as is" without express or
 *  implied warranty.
 *
 *  Author: Daniel Ko, UBC Laboratory for Computational Intelligence
 */

/* From the Vista library: */
#include "vista/Vlib.h"
#include "vista/file.h"
#include "vista/mu.h"
#include "vista/option.h"
#include "vista/os.h"
#include "vista/postscript.h"
#include "vista/VList.h"

/* File identification string: */
VRcsId ("$Id: vlayout.c,v 1.5 1994/06/15 23:14:49 pope Exp pope $");

/* Command line options: */
static VArgVector in_filenames;		/* list of input files */
static VString out_filename;		/* output file */
static VBoolean eps = FALSE;		/* whether EPS or PS format */
static VString gutter_string = "0.5in";
static VShort nrows, ncolumns = 2;	/* numbers of rows, columns */

static VBoolean in_found, out_found;

static VOptionDescRec options[] = {
    { "in", VStringRepn, 0, & in_filenames, & in_found, NULL,
	 "Names of input files" },
    { "out", VStringRepn, 1, & out_filename, & out_found, NULL,
	 "Name of output file" },
    { "eps", VBooleanRepn, 1, & eps, VOptionalOpt, NULL,
	  "Whether to produce Encapsulated PostScript" },
    { "gutter", VStringRepn, 1, & gutter_string, VOptionalOpt, NULL,
	 "Space between rows and between columns" },
    { "ncolumns", VShortRepn, 1, & ncolumns, VOptionalOpt, NULL,
	 "Number of columns" }
};

/* Page layout parameters: */
static VPSDescRec ps;			/* standard PostScript parameters */
static float gutter;			/* whitespace between rows, columns */
static float box_height, box_width;	/* space for each EPS document */

/* Output file: */
static FILE *outfile;

/* Some PostScript DSC comments are removed from incorporated
   EPS documents because they often confuse software that hasn't learned
   to look for %%BeginDocument / %%EndDocument pairs. */
static VStringConst dsc_trim_list[] = {
    "Page:", "Pages:", "EOF"
};

/* Later in this file: */
static void IncludeEPSFile (VStringConst, float, float);
static void SetupPage (void);
static VBoolean SubstrEqu (VStringConst, VStringConst);


/*
 *  Entry point.
 */

int main (int argc, char **argv)
{
    float llx, lly;
    int i;

    if (! VParseCommand (VNumber (options), options, & argc, argv) ||
	! VParsePSOptions (& argc, argv, & ps) ||
	! VIdentifyFiles (VNumber (options), options, "in",
			 & argc, argv, 0    /* stdin */) ||
	! VIdentifyFiles (VNumber (options), options, "out",
			 & argc, argv, -1) ||
	! VParsePSDistance (gutter_string, "gutter", & gutter))
	goto Usage;
    if (argc > 1) {
	VReportBadArgs (argc, argv);
Usage:  VReportUsage (argv[0], VNumber (options), options, NULL);
	VReportValidPSOptions ();
	exit (EXIT_FAILURE);
    }

    nrows = (in_filenames.number + ncolumns - 1) / ncolumns;

    /* Output the PostScript document and page headers: */
    SetupPage ();

    /* Determine the space available for each EPS document: */
    box_height = (ps.page_height - ps.bottom_margin - ps.top_margin -
	(nrows - 1) * gutter) / nrows;
    box_width = (ps.page_width - ps.left_margin - ps.right_margin -
	(ncolumns - 1) * gutter) / ncolumns;
    if (box_height <= 10.0f || box_width < 10.0f)
	VError ("Margins leave too little room for EPS documents");

    /* Include each EPS document: */
    llx = ps.left_margin;
    lly = ps.page_height - ps.top_margin - box_height;
    for (i = 0; i < in_filenames.number; ) {
	IncludeEPSFile (((VStringConst *) in_filenames.vector)[i], llx, lly);
	if (++i % ncolumns == 0) {
	    llx = ps.left_margin;
	    lly -= (box_height + gutter);
	} else llx += (box_width + gutter);
    }

    fputs ("end\n", outfile);
    VEndPSPage (outfile, & ps);
    VEndVPSDoc (outfile, & ps);

    /* Provide feedback: */
    fprintf (stderr, "%s: %d EPS file%s placed on page.\n",
	     argv[0], in_filenames.number,
	     in_filenames.number == 1 ? "" : "s");

    return EXIT_SUCCESS;
}


/*
 *  SetupPage
 *
 *  Set up PostScript page environment.
 */

static void SetupPage (void)
{
    float fw, ymin = ps.bottom_margin;

    /* Output the PostScript document and page headers: */
    if (! VBeginVPSDoc (outfile = VOpenOutputFile (out_filename, TRUE),
			& ps, eps, (strcmp (out_filename, "-") == 0) ?
			out_filename : "<stdout>"))
	exit (EXIT_FAILURE);
    VBeginPSPage (outfile, & ps);
    if (! eps) {
	fputs ("%%PageResources: font " VPSCaptionFont "\n", outfile);
	fputs ("%%BeginPageSetup\n", outfile);
    }
    fputs ("VistaDict begin\n", outfile);

    /* Take care of landscape orientation: */
    if (ps.landscape)
	fprintf (outfile, "%g 0 translate 90 rotate\n", ps.page_height);

    /* Draw caption: */
    if (ps.caption) {
	ymin += VPSBelowCaption * ps.caption_height;
	fprintf (outfile, "%g %g (%s) /%s %g caption\n",
		 ymin, ps.left_margin +
		 (ps.page_width - ps.right_margin - ps.left_margin) / 2.0f,
		 ps.caption, VPSCaptionFont, ps.caption_height);
	ymin += (VPSAboveCaption + 1.0f) * ps.caption_height;
	VExtendPSBoundingBox (& ps, ps.left_margin, ps.bottom_margin,
			      ps.page_width - ps.right_margin, ymin);
    }

    /* Draw a frame: */
    fw = (ps.frame_width >= 0.0f) ? VMax (1.0f, ps.frame_width) : 0.0f;
    if (ps.frame_width >= 0.0f) {
	fprintf (outfile, "%g %g %g %g %g frame\n",
		 ps.left_margin + fw / 2.0f, ymin + fw / 2.0f,
		 ps.page_width - ps.left_margin - ps.right_margin - fw,
		 ps.page_height - ps.bottom_margin - ps.top_margin - fw,
		 ps.frame_width);
	VExtendPSBoundingBox (& ps, ps.left_margin, ymin,
			      ps.right_margin, ps.top_margin);
    }

    /* Shrink the margins to reflect the actual page space remaining: */
    ps.bottom_margin = ymin;
    if (fw > 0.0f) {
	fw += gutter / 2.0f;	/* inset within frame by half gutter */
	ps.left_margin += fw;
	ps.right_margin -= fw;
	ps.bottom_margin += fw;
	ps.top_margin -= fw;
    }

    if (! eps)
	fputs ("%%EndPageSetup\n", outfile);
}


/*
 *  IncludeEPSFile
 *
 *  Include an EPS file, scaling it to box_height by box_width, and
 *  positioning it at llx, lly.
 *
 *  ==> Check efficiency of use of linked list.
 */

static void IncludeEPSFile (VStringConst filename, float llx, float lly)
{
    FILE *f;
    int i, lineno = 0;
    enum { InitState, LoadState, SkipState } state = InitState;
    float bb_left, bb_bottom, bb_right, bb_top;
    float xscale, yscale, scale, act_width, act_height;
    VList lines;
    char *cp, line[1024];

    /* Open the file: */
    if (strcmp (filename, "-") == 0) {
	f = stdin;
	filename = "<stdin>";
    } else if (! (f = fopen (filename, "r"))) {
	VWarning ("Unable to open %s", filename);
	return;
    }

    /* Read it entirely into memory while (a) noting its bounding box,
       (b) eliminating any duplication of the Vista proc set we've already
       included in the output file, and (c) eliminating certain Document
       Structuring Convention comments that may confuse other software: */
    lines = VListCreate ();
    bb_left = bb_bottom = bb_right = bb_top = 0.0f;
    while (lineno++, fgets (line, sizeof(line), f) != NULL) {
	switch (state) {

	case InitState:
	    if (! SubstrEqu (line, "%!PS-Adobe"))
		VWarning ("%s may not be a PostScript file", filename);
	    state = LoadState;
	    /* (no break) */

	case LoadState:
	    if (SubstrEqu (line, "%%BoundingBox:") &&
		! strstr (line, "(atend")) {

		/* Parse a %%BoundingBox comment: */
		if (sscanf (line, "%%%%BoundingBox: %g %g %g %g", & bb_left,
			    & bb_bottom, & bb_right, & bb_top) != 4)
		    VWarning ("Bad %%BoundingBox comment at line %d of %s",
			      lineno, filename);

	    } else if (strcmp (line, "%%BeginResource: procset "
			       VPSProcSetName " "
			       VPSProcSetVersion " "
			       VPSProcSetRevision "\n") == 0) {
		state = SkipState;
		continue;
	    } else if (line[0] == '%' && line[1] == '%') {

		/* Skip a DSC comment that's on dsc_trim_list: */
		for (i = 0; i < VNumber (dsc_trim_list); i++)
		    if (SubstrEqu (line + 2, dsc_trim_list[i]))
			break;
		if (i < VNumber (dsc_trim_list))
		    continue;	/* skip a DSC comment */
	    }
	    break;

	case SkipState:
	    if (SubstrEqu (line, "%%EndResource"))
		state = LoadState;
	    continue;
	}
	VListAppend (lines, VNewString (line));
    }
    fclose (f);

    /* Check that a valid bounding box was obtained: */
    if (bb_left >= bb_right || bb_bottom >= bb_top) {
	VWarning ("No valid %%BoundingBox comment found in %s", filename);
	goto Done;
    }

    /* Center the EPS document's bounding box inside the space available
       for it on the page: */
    xscale = box_width / (bb_right - bb_left);
    yscale = box_height / (bb_top - bb_bottom);
    scale = VMin (xscale, yscale);
    act_width = scale * (bb_right - bb_left);
    act_height = scale * (bb_top - bb_bottom);
    llx += (box_width - act_width) / 2.0f;
    lly += (box_height - act_height) / 2.0f;

    /* Prepare for including the EPS document: */
    fputs ("BeginEPSF\n", outfile);
    fprintf (outfile, "%g %g %g %g cliprect\n",
	     llx, lly, act_width, act_height);
    VExtendPSBoundingBox (& ps, llx, lly, llx + act_width, lly + act_height);

    /* Transform user space to fit EPS file inside the
       layout box: */
    fprintf (outfile, "%g %g translate %g %g scale %g %g translate\n",
	     llx, lly, scale, scale, - bb_left, - bb_bottom);

    /* Include the text of the EPS document: */
    fprintf (outfile, "%%%%BeginDocument: %s\n", filename);
    for (cp = VListFirst (lines); cp; cp = VListNext (lines))
	fputs (cp, outfile);
    fputs ("%%EndDocument\n", outfile);
    fputs ("EndEPSF\n", outfile);

Done:
    VListDestroy (lines, VFree);
}


/*
 *  SubstrEqu
 *
 *  Return TRUE if b is an initial substring of a.
 */

static VBoolean SubstrEqu (VStringConst a, VStringConst b)
{
    while (*a && *a == *b) {
	a++;
	b++;
    }
    return *b == 0;
}
