/*
 *  $Id: vlink.c,v 1.4 1994/06/15 23:14:49 pope Exp pope $
 *
 *  The vlink program, which creates a set of edges from an image.
 */

/*
 *  Copyright 1993, 1994 University of British Columbia
 *
 *  Permission to use, copy, modify, distribute, and sell this software and its
 *  documentation for any purpose is hereby granted without fee, provided that
 *  the above copyright notice appears in all copies and that both that
 *  copyright notice and this permission notice appear in supporting
 *  documentation. UBC makes no representations about the suitability of this
 *  software for any purpose. It is provided "as is" without express or
 *  implied warranty.
 *
 *  Author: David Lowe, UBC Laboratory for Computational Intelligence
 */

/* From the Vista library: */
#include "vista/Vlib.h"
#include "vista/mu.h"
#include "vista/option.h"
#include "vista/os.h"
#include "vista/VEdges.h"
#include "vista/VImage.h"

/* File identification string: */
VRcsId ("$Id: vlink.c,v 1.4 1994/06/15 23:14:49 pope Exp pope $");

/* Default thresholds: */
#define HighDefault		30
#define LowDefault		15
#define MinLengthDefault	4


/* Where to put the results: */
enum {
    ResultAfter,
    ResultInsert,
    ResultReplace
};

VDictEntry ResultDict[] = {
    { "after", ResultAfter },
    { "insert", ResultInsert },
    { "replace", ResultReplace },
    { NULL }
};


/*
 *  main
 *
 *  Program entry point.
 */

int main (int argc, char *argv[])
{
    static VLong high_opt = HighDefault, low_opt = LowDefault;
    static VLong minlength_opt = MinLengthDefault, result_opt = ResultReplace;
    static VBoolean high_found, low_found;
    static VOptionDescRec options[] = {
	{ "low", VLongRepn, 1, & low_opt, & low_found, NULL,
	      "Low threshold value" },
	{ "high", VLongRepn, 1, & high_opt, & high_found, NULL,
	      "High threshold value" },
	{ "minlength", VLongRepn, 1, & minlength_opt, VOptionalOpt, NULL,
	      "Minimum length of each edge in pixels" },
	{ "result", VLongRepn, 1, & result_opt, VOptionalOpt, ResultDict,
	      "Where to place result" }
    };
    FILE *in_file, *out_file;
    VAttrList list;
    VAttrListPosn posn;
    VImage image, copy;
    VEdges edge_set = NULL;
    VLong high, low;
    int nimages = 0;

    /* Parse command line arguments and identify the input and output files: */
    VParseFilterCmd (VNumber (options), options, argc, argv,
		     & in_file, & out_file);

    /* Read the input file: */
    if (! (list = VReadFile (in_file, NULL)))
	exit (EXIT_FAILURE);

    /* For each single-band gradient image: */
    for (VFirstAttr (list, & posn); VAttrExists (& posn); VNextAttr (& posn)) {
	if (VGetAttrRepn (& posn) != VImageRepn)
	    continue;
	VGetAttrValue (& posn, NULL, VImageRepn, & image);
	if (VImageComponentInterp (image) != VBandInterpGradient ||
	    VImageNComponents (image) != 1)
	    continue;
	nimages++;

	/* Only band 0 is processed: */
	if (VImageNBands (image) > 1)
	    VWarning ("Only processing band 0 of %d-band source image",
		      VImageNBands (image));

	/* If neither high nor low threshold was supplied, and the image
	   has high_threshold and low_threshold attributes, use the
	   values they supply: */
	if (high_found || low_found ||
	    VGetAttr (VImageAttrList (image), "low_threshold", NULL,
		      VLongRepn, & low) != VAttrFound ||
	    VGetAttr (VImageAttrList (image), "high_threshold", NULL,
		      VLongRepn, & high) != VAttrFound) {
	    high = high_opt;
	    low = low_opt;
	}

	/* Link a copy of the image, since VLinkImage destroys its src (!): */
	if (! (copy = VCopyImage (image, NULL, VAllBands)) ||
	    ! (edge_set = VLinkImage (copy, low, high, minlength_opt)))
	    exit (EXIT_FAILURE);
	VDestroyImage (copy);

	/* Add attributes recording the thresholds actually used: */
	VSetAttr (VEdgesAttrList (edge_set), "low_threshold", NULL,
		  VLongRepn, (VLong) low);
	VSetAttr (VEdgesAttrList (edge_set), "high_threshold", NULL,
		  VLongRepn, (VLong) high);
	VSetAttr (VEdgesAttrList (edge_set), "minlength", NULL,
		  VLongRepn, (VLong) minlength_opt);

	/* Store the result back in the list of file objects: */
	switch (result_opt) {

	case ResultAfter:
	    VInsertAttr (& posn, TRUE, "edges", NULL, VEdgesRepn, edge_set);
	    VNextAttr (& posn);
	    break;

	case ResultInsert:
	    VAppendAttr (VImageAttrList (image), "edges", NULL, VEdgesRepn,
			 edge_set);
	    break;
	    
	case ResultReplace:
	    VSetAttrValue (& posn, NULL, VEdgesRepn, edge_set);
	    VDestroyImage (image);
	}
    }

    /* Write the output file: */
    if (! VWriteFile (out_file, list))
	exit (EXIT_FAILURE);

    fprintf (stderr, "%s: Linked %d image%s.\n",
	     argv[0], nimages, nimages == 1 ? "" : "s");
    return EXIT_SUCCESS;
}
