/*
 * $Id: vop.c,v 1.8 1995/05/01 23:28:18 pope Exp pope $
 * 
 * The vop program performs arithmetic and logical operations on images.
 */

/*
 *  Copyright 1993, 1994 University of British Columbia
 *
 *  Permission to use, copy, modify, distribute, and sell this software and its
 *  documentation for any purpose is hereby granted without fee, provided that
 *  the above copyright notice appears in all copies and that both that
 *  copyright notice and this permission notice appear in supporting
 *  documentation. UBC makes no representations about the suitability of this
 *  software for any purpose. It is provided "as is" without express or
 *  implied warranty.
 *
 *  Authors:
 *  Ralph Horstmann and Art Pope, UBC Laboratory for Computational Intelligence
 */

/* From the Vista library: */
#include "vista/Vlib.h"
#include "vista/mu.h"
#include "vista/option.h"
#include "vista/os.h"
#include "vista/VImage.h"

/* From the standard C library: */
#include <math.h>

#ifdef sgi
#undef _POSIX_SOURCE
#undef _XOPEN_SOURCE
#include <sigfpe.h>
#include <signal.h>
#endif

#ifdef SunOS_4
#include <signal.h>
#endif

#if defined (SunOS_5) && defined (HasSunC)
#include <floatingpoint.h>
#include <siginfo.h>
#include <sunmath.h>
#endif

/* File identification string: */
VRcsId ("$Id: vop.c,v 1.8 1995/05/01 23:28:18 pope Exp pope $");

/* Names for the operations supported by VImageOp: */
static VDictEntry op_dict[] = {
    { "abs", VImageOpAbs },
    { "add", VImageOpAdd },
    { "and", VImageOpAnd },
    { "dist", VImageOpDist },
    { "div", VImageOpDiv },
    { "exp", VImageOpExp },
    { "log", VImageOpLog },
    { "max", VImageOpMax },
    { "min", VImageOpMin },
    { "mult", VImageOpMult },
    { "not", VImageOpNot },
    { "or", VImageOpOr },
    { "sqrt", VImageOpSqrt },
    { "square", VImageOpSquare },
    { "sub", VImageOpSub },
    { "xor", VImageOpXor },
    { NULL }
};

#define Unary(op)							\
    (op == VImageOpAbs || op == VImageOpExp || op == VImageOpLog ||	\
     op == VImageOpNot || op == VImageOpSqrt || op == VImageOpSquare)

/* Record of errors signalled by exceptions: */
static struct {
    int int_division;
    int flt_division;
    int flt_overflow;
    int flt_underflow;
    int flt_inexact;
    int flt_invalid;
} error_counts;

/* Later in this file: */
#ifdef sgi
void SgiTrapHandler (int, int, void *);
#endif
#ifdef SunOS_4
static void SunFPEHandler (int, int, void *, char *);
#endif
#if defined (SunOS_5) && defined (HasSunC)
static void SunFPEHandler (int, siginfo_t *, void *);
#endif


/*
 *  Program entry point.
 */

int main (int argc, char *argv[])
{
    static VImageOpKind op;
    static VBoolean value_found, image_found, min_found, max_found;
    static VString image_filename;
    static VDouble value, min, max;
    static VOptionDescRec  options[] = {
        { "op", VLongRepn, 1, & op, VRequiredOpt, op_dict,
	      "Operation to be performed" },
        { "value", VDoubleRepn, 1, & value, & value_found, NULL,
	      "Constant operand for binary operation" },
        { "image", VStringRepn, 1, & image_filename, & image_found, NULL,
	      "Image operand for binary operation" },
        { "min", VDoubleRepn, 1, & min, & min_found, NULL,
	      "Clipping lower bound" },
        { "max", VDoubleRepn, 1, & max, & max_found, NULL,
	      "Clipping upper bound" }
    };
    FILE *in_file, *out_file, *image_file = NULL;
    VAttrList list;
    VAttrListPosn posn;
    VImage *images, src;
    int nimages = 0;
    VDouble *minp, *maxp;

    /* Parse command line arguments and identify the input and output files: */
    VParseFilterCmd(VNumber (options), options, argc, argv,
		    & in_file, & out_file);
    minp = min_found ? & min : NULL;
    maxp = max_found ? & max : NULL;

    /* Check consistency between type of operation and operands provided: */
    if (Unary (op)) {
	if (value_found || image_found) {
	    VWarning ("Unary operation requires neither -value nor -image");
	    value_found = image_found = FALSE;
	}
    } else {
	if (! (value_found ^ image_found))
	    VError ("Binary operation requires either -value or -image");
	if (image_found) {

	    /* Load the image serving as the second operand: */
	    if (strcmp (image_filename, "-") == 0)
		image_file = stdin;
	    else if (! (image_file = fopen (image_filename, "r")))
		VError ("Failed to open image file %s", image_filename);
	    if ((nimages = VReadImages (image_file, & list, & images)) == 0)
		exit (EXIT_FAILURE);
	    if (nimages > 1)
		VWarning ("Using only the first image in %s", image_filename);
	    VDestroyAttrList (list);
	    fclose (image_file);
	}
    }

    /* Read source image(s): */
    if (! (list = VReadFile (in_file, NULL)))
	exit (EXIT_FAILURE);

    /* Set up a handlers for arithmetic and floating point exceptions: */
#ifdef sgi
    {
	static struct sigaction act = { 0 };
	act.sa_handler = SgiTrapHandler;
	if (sigaction (SIGTRAP, & act, NULL))
	    VError ("sigaction() failed");
    }
    sigfpe_[_UNDERFL].count = INT_MAX;
    sigfpe_[_OVERFL].count = INT_MAX;
    sigfpe_[_DIVZERO].count = INT_MAX;
    sigfpe_[_INVALID].count = INT_MAX;
    handle_sigfpes (_ON, _EN_UNDERFL | _EN_OVERFL | _EN_DIVZERO | _EN_INVALID,
		    0, _TURN_OFF_HANDLER_ON_ERROR, 0);
#endif

#ifdef SunOS_4
    if (ieee_handler ("set", "all", SunFPEHandler))
	VError ("ieee_handler() failed");
    SunFPEHandler (0, 0, NULL, NULL);
#endif

#if defined (SunOS_5) && defined (HasSunC)
    #ifndef BADSIG
    #define BADSIG (void (*)())-1
    #endif
    if (ieee_handler ("set", "all", SunFPEHandler))
	VError ("ieee_handler() failed");
    if (sigfpe (FPE_INTDIV, SunFPEHandler) == BADSIG)
	VError ("sigfpe() failed");
#endif /* SunOS_5 && HasSunC */

    /* Operate on each source image: */
    nimages = 0;
    for (VFirstAttr (list, & posn); VAttrExists (& posn); VNextAttr (& posn)) {
	if (VGetAttrRepn (& posn) != VImageRepn)
	    continue;
	VGetAttrValue (& posn, NULL, VImageRepn, & src);
	memset (& error_counts, 0, sizeof (error_counts));
	if (value_found) {

	    /* Operation between image and constant: */
	    src = VImageOpV (src, src, VAllBands, op, value, minp, maxp);
	} else if (image_found) {

	    /* Operation between two images: */
	    src = VImageOpI (src, src, VAllBands, op, images[0],
			     VImageNBands (src) == VImageNBands (images[0]) ?
			     VAllBands : 0, minp, maxp);
	} else {

	    /* Operation on single image: */
	    src = VImageOpU (src, src, VAllBands, op, minp, maxp);
	}
	if (! src)
	    exit (EXIT_FAILURE);

	/* Report any errors detected: */
	#define ReportError(n,type)					\
	    if (n)							\
		VWarning ("%d %s%s on image %d", n, type,		\
			  n == 1 ? "" : "s", nimages + 1)

	ReportError (error_counts.int_division, "integer zero division");
	ReportError (error_counts.flt_division, "f.p. zero division");
	ReportError (error_counts.flt_overflow, "f.p. overflow");
	ReportError (error_counts.flt_underflow, "f.p. underflow");
	ReportError (error_counts.flt_inexact, "inexact f.p. result");
	ReportError (error_counts.flt_invalid, "invalid f.p. operation");

	#undef ReportError

	nimages++;
    }

    /* Write the results to the output file: */
    if (! VWriteFile (out_file, list))
	exit (EXIT_FAILURE);
    fprintf (stderr, "%s: Processed %d image%s.\n",
	     argv[0], nimages, nimages == 1 ? "" : "s");
    return EXIT_SUCCESS;
}


/*
 *  Floating point and arithmetic exception handlers.
 */

#ifdef SunOS_4
static void SunFPEHandler (int sig, int code, void *scp, char *addr)
{
    switch (code) {
    case FPE_INTDIV_TRAP:   error_counts.int_division++; break;
    case FPE_FLTINEX_TRAP:  error_counts.flt_inexact++; break;
    case FPE_FLTDIV_TRAP:   error_counts.flt_division++; break;
    case FPE_FLTUND_TRAP:   error_counts.flt_underflow++; break;
    case FPE_FLTOPERR_TRAP: error_counts.flt_invalid++; break;
    case FPE_FLTOVF_TRAP:   error_counts.flt_overflow++; break;
    }
    if ((int) signal (SIGFPE, SunFPEHandler) == -1)
	VSystemError ("Call to signal() failed");
}

int matherr (struct exception *exc)
{
    switch (exc->type) {
    case DOMAIN:    error_counts.flt_invalid++; break;
    case SING:	    error_counts.flt_division++; break;
    case OVERFLOW:  error_counts.flt_overflow++; break;
    case UNDERFLOW: error_counts.flt_underflow++; break;
    }
    return 1;
}
#endif

#if defined (SunOS_5) && defined (HasSunC)
static void SunFPEHandler (int sig, siginfo_t *sip, void *uap)
{
    switch (sip->si_code) {

    case FPE_INTDIV:
	VError ("Divide by zero");

    case FPE_FLTDIV: error_counts.flt_division++; break;
    case FPE_FLTOVF: error_counts.flt_overflow++; break;
    case FPE_FLTUND: error_counts.flt_underflow++; break;
    case FPE_FLTRES: error_counts.flt_inexact++; break;
    case FPE_FLTINV: error_counts.flt_invalid++; break;
    }
}
#endif

#ifdef sgi
void SgiTrapHandler (int sig, int code, void *sc)
{
    VError (code == BRK_DIVZERO ? "Divide by zero" : "SIGTRAP signal raised");
}
#endif

#ifdef hpux
int matherr (struct exception *exc)
{
    switch (exc->type) {
    case DOMAIN:    error_counts.flt_invalid++; break;
    case SING:	    error_counts.flt_division++; break;
    case OVERFLOW:  error_counts.flt_overflow++; break;
    case UNDERFLOW: error_counts.flt_underflow++; break;
    case TLOSS:
    case PLOSS:	    error_counts.flt_inexact++; break;
    }
    return 1;
}
#endif
