/*
 *  $Id: vpdc.c,v 1.3 1994/10/01 00:21:25 pope Exp pope $
 *
 *  The compiler for Vista parameter definition files.
 */

/*
 *  Copyright 1994 University of British Columbia
 *
 *  Permission to use, copy, modify, distribute, and sell this software and its
 *  documentation for any purpose is hereby granted without fee, provided that
 *  the above copyright notice appears in all copies and that both that
 *  copyright notice and this permission notice appear in supporting
 *  documentation. UBC makes no representations about the suitability of this
 *  software for any purpose. It is provided "as is" without express or
 *  implied warranty.
 *
 *  Author: Arthur Pope, UBC Laboratory for Computational Intelligence
 */

/* From the Vista library: */
#include "vista/Vlib.h"
#include "vista/file.h"
#include "vista/mu.h"
#include "vista/option.h"
#include "vista/os.h"

/* From the standard C library: */
#include <ctype.h>

/* File identification string: */
VRcsId ("$Id: vpdc.c,v 1.3 1994/10/01 00:21:25 pope Exp pope $");

/* Later in this file: */
static VWarningHandler WarningHandler;
static VBoolean LoadParameterDefinitions (void);
static void GenerateHeaderFile (void);
static int GenerateString (VStringConst);

/* Constants: */
#define maxLineLength			256
#define defaultCommentColumn		36
#define defaultContinuationColumn	78

/* Program options: */
static VShort commentColumn = defaultCommentColumn;
static VShort continuationColumn = defaultContinuationColumn;
static VBoolean lineDirectives = TRUE;

static VOptionDescRec options[] = {
    { "comment_column", VShortRepn, 1, & commentColumn, VOptionalOpt,
	  NULL, "Column at which comments should begin" },
    { "continuation_column", VShortRepn, 1, & continuationColumn, VOptionalOpt,
	  NULL, "Column that should contain continuation character" },
    { "line", VBooleanRepn, 1, & lineDirectives, VOptionalOpt,
	  NULL, "Include preprocessor #line directives" }
};

/* Input and output files: */
static VStringConst inFilename;
#if defined (PATH_MAX) && PATH_MAX < 1024
static char outFilename[PATH_MAX + 1];
#else
static char outFilename[1024];
#endif
static FILE *inFile, *outFile;
static int lineno = 0;			/* current line in input file */

/* Buffer capable of containing a line of the input file: */
typedef char Buffer[maxLineLength];

/* Information about one parameter definition: */
typedef struct ParDefnRec *ParDefn;
typedef struct ParDefnRec {
    ParDefn next;
    int lineno;
    VStringConst name;
    VRepnKind repn;
    VStringConst value;
    VStringConst comment;
} ParDefnRec;

/* Information read from the input file: */
static Buffer parSetName;		/* parameter set name */
static VBoolean parSetNameFound = FALSE;
static ParDefn firstPar = NULL, lastPar;  /* list of parameter definitions */
		

/*
 *  Program entry point.
 */

int main (int argc, char *argv[])
{
    char *cp;

    /* Parse command line arguments: */
    if (! VParseCommand (VNumber (options), options, & argc, argv) ||
	(argc != 2 && argc != 3)) {
Usage:
	VReportUsage (argv[0], VNumber (options), options,
		      "<infile> [<outfile>]");
	exit (EXIT_FAILURE);
    }

    /* Open input and output files: */
    if (! (inFile = fopen (inFilename = argv[1], "r"))) {
	VWarning ("Unable to open input file %s", inFilename);
	goto Usage;
    }
    if (argc == 3)
	strcpy (outFilename, argv[2]);
    else {
	strcpy (outFilename, argv[1]);
	if (cp = strrchr (outFilename, '.')) {
	    cp[1] = 'h';
	    cp[2] = 0;
	}
	fprintf (stderr, "%s: Output file defaulting to %s.\n",
		 argv[0], outFilename);
    }
    if (strcmp (inFilename, outFilename) == 0)
	VError ("Cannot use %s for both input and output file", inFilename);
    if (! (outFile = fopen (outFilename, "w"))) {
	VWarning ("Unable to open output file %s", outFilename);
	goto Usage;
    }

    /* Use our own warning handler so that error messages are output
       in a format that message-parsing utilities understand: */
    VSetWarningHandler (WarningHandler);

    /* Load information from the input file: */
    if (! LoadParameterDefinitions ())
	exit (EXIT_FAILURE);

    /* Check that a parameter set name and at least one parameter
       definition were encountered: */
    if (! parSetNameFound)
	VError ("No parameter set name found");
    if (firstPar == NULL)
	VError ("No parameter definitions found");

    /* Output the header file: */
    GenerateHeaderFile ();
    return EXIT_SUCCESS;
}


/*
 *  WarningHandler
 *
 *  Special handler for VWarning invocations. It formats error messages
 *  so that they include the input filename and line number for the benefit
 *  of message-parsing utilities.
 */

static void WarningHandler (VStringConst msg)
{
    char *cp;

    /* Strip off the vpdc: Warning: VRoutine: prefixes: */
    msg = (cp = strchr (msg, ':')) ? cp + 2 : msg;
    msg = (cp = strchr (msg, ':')) ? cp + 2 : msg;
    msg = (cp = strchr (msg, ':')) ? cp + 2 : msg;

    /* Print the rest prefixed by the filename and line number: */
    fprintf (stderr, "%s: %d: %s", inFilename, lineno, msg);
}


/*
 *  LoadParameterDefinitions
 *
 *  Parse a parameter definition file, storing its contents in global
 *  structures.
 */

static VBoolean LoadParameterDefinitions (void)
{
    int i;
    VRepnKind repn;
    VBoolean result = TRUE;
    ParDefn p;
    Buffer buf, name, value, comment;

    /* For each line of the input file: */
    while (fgets (buf, sizeof (buf), inFile)) {
	lineno++;
	
	/* Skip comment lines and empty lines: */
	if (buf[0] == '#')
	    continue;
	for (i = 0; buf[i] && isspace (buf[i]); i++) ;
	if (! buf[i])
	    continue;
	    
	/* The first non-comment line contains the name of the parameter
	   set: */
	if (! parSetNameFound) {
	    parSetNameFound = TRUE;
	    if (sscanf (buf, " %s %s", parSetName, comment) == 1)
		continue;
	    VWarning ("Expected parameter set name not found");
	    result = FALSE;
	    continue;
	}

	/* Extract the parameter name, type, value, and comment: */
	if (! VParseParamDefn (buf, name, & repn, value, comment)) {
	    result = FALSE;
	    continue;
	}

	/* Store the information in a newly-allocated record: */
	p = VNew (ParDefnRec);
	if (firstPar == NULL)
	    firstPar = p;
	else lastPar->next = p;
	lastPar = p;
	p->next = NULL;
	p->lineno = lineno;
	p->name = VNewString (name);
	p->repn = repn;
	p->value = VNewString (value);
	p->comment = comment[0] ? VNewString (comment) : NULL;
    }

    /* Check that we're at EOF: */
    if (result && ! (result &= feof (inFile)))
	VWarning ("EOF expected");

    return result;
}


/*
 *  GenerateHeaderFile
 *
 *  Output a header file based on the parameter definitions stored in
 *  global structures.
 */

static void GenerateHeaderFile (void)
{
    ParDefn p;
    int i, col;

    #define Repn(root) { "V" #root, V ## root ## Repn }
    static struct {
	VStringConst name;
	VRepnKind repn;
    } repn_names[] = {
	Repn (Bit), Repn (UByte), Repn (SByte), Repn (Short), Repn (Long),
	Repn (Float), Repn (Double), Repn (Boolean), Repn (String),
    };
    #undef Repn

    #define SkipTo(goal) \
	do {						\
	    fputc (' ', outFile);			\
	} while (++col < goal);

    #define Continue() \
        SkipTo (continuationColumn);			\
	fprintf (outFile, "\\\n");

    /* Output file prologue: */
    fprintf (outFile, "/*\n *  >> Created automatically by vpdc <<\n */\n\n");
    fprintf (outFile, "#ifndef %s_included\n", parSetName);
    fprintf (outFile, "#define %s_included 1\n\n", parSetName);
    fprintf (outFile, "/* From the Vista library: */\n");
    fprintf (outFile, "#include <vista/Vlib.h>\n");
    fprintf (outFile, "#include <vista/mu.h>\n");
    fprintf (outFile, "#include <vista/option.h>\n\n");
    if (lineDirectives) {
	fprintf (outFile, "#line 1 ");
	GenerateString (inFilename);
	fprintf (outFile, "\n\n");
    }    

    /* Output the structure definition: */ 
    fprintf (outFile, "typedef struct {\n");
    for (p = firstPar; p; p = p->next) {
	for (i = 0; i < VNumber (repn_names); i++)
	    if (repn_names[i].repn == p->repn)
		break;
	if (i == VNumber (repn_names))
	    VError ("Error in GenerateHeaderFile");
	if (lineDirectives)
	    fprintf (outFile, "#line %d\n", p->lineno);
	col = fprintf (outFile, "    %s %s;", repn_names[i].name, p->name);
	do {
	    fputc (' ', outFile);
	} while (++col < commentColumn);
	fprintf (outFile, "/* %s */\n", p->comment);
    }
    fprintf (outFile, "} %s;\n\n", parSetName);

    /* Output a macro defining the default values: */
    col = fprintf (outFile, "#define %s_Defaults {", parSetName);
    Continue ();
    for (p = firstPar; p; p = p->next) {
	col = fprintf (outFile, "    ");
	if (p->repn == VStringRepn)
	    col += GenerateString (p->value);
	else col += fprintf (outFile, "%s", p->value);
	if (p->next) {
	    fputc (',', outFile);
	    col++;
	}
	SkipTo (commentColumn);
	col += fprintf (outFile, "/* %s */", p->name);
	Continue ();
    }
    fprintf (outFile, "}\n\n");

    /* Output a macro defining an initializer for a table of option
       descriptors: */
    col = fprintf (outFile, "#define %s_Defns {", parSetName);
    Continue ();
    for (p = firstPar; p; p = p->next) {
	for (i = 0; i < VNumber (repn_names); i++)
	    if (repn_names[i].repn == p->repn)
		break;
	if (i == VNumber (repn_names))
	    VError ("Error in GenerateHeaderFile");
	col = fprintf (outFile, "    { \"%s\", %sRepn, 1,",
		       p->name, repn_names[i].name);
	Continue ();
	col = fprintf (outFile, "\t(VPointer) VOffsetOf (%s, %s),",
		       parSetName, p->name) + 7;
	Continue ();
	col = fprintf (outFile, "\tVOptionalOpt, NULL, ") + 7;
	if (p->comment)
	    col += GenerateString (p->comment);
	else col += fprintf (outFile, "NULL");
	col += fprintf (outFile, " }");
	if (p->next) {
	    fputc (',', outFile);
	    col++;
	}
	Continue ();
    }
    fprintf (outFile, "}\n\n");
    
    /* Output a macro defining the number of parameters: */
    for (p = firstPar, i = 0; p; p = p->next, i++) ;
    fprintf (outFile, "#define %s_N %d\n\n", parSetName, i);

    /* Output the epilogue: */
    fprintf (outFile, "#endif /* %s_included */\n", parSetName);

    #undef SkipTo
    #undef Continue
}


/*
 *  GenerateString
 *
 *  Emit a C string constant with appropriate escape characters.
 */

static int GenerateString (VStringConst cp)
{
    int ch, col = 0;

    fputc ('"', outFile);
    while (ch = *cp++) {
	if (ch == '\\' || ch == '"') {
	    fputc ('\\', outFile);
	    col++;
	}
	fputc (ch, outFile);
	col++;
    }
    fputc ('"', outFile);
    return col + 2;
}
