/*
 *  $Id: vscale.c,v 1.5 1994/06/15 23:14:49 pope Exp pope $
 *
 *  The vscale program, which scales images by various methods.
 */

/*
 *  Copyright 1993, 1994 University of British Columbia
 *
 *  Permission to use, copy, modify, distribute, and sell this software and its
 *  documentation for any purpose is hereby granted without fee, provided that
 *  the above copyright notice appears in all copies and that both that
 *  copyright notice and this permission notice appear in supporting
 *  documentation. UBC makes no representations about the suitability of this
 *  software for any purpose. It is provided "as is" without express or
 *  implied warranty.
 *
 *  Author: Daniel Ko, UBC Laboratory for Computational Intelligence
 */

/* From the Vista library: */
#include "vista/Vlib.h"
#include "vista/mu.h"
#include "vista/option.h"
#include "vista/os.h"
#include "vista/VEdges.h"
#include "vista/VImage.h"

/* File identification string: */
VRcsId ("$Id: vscale.c,v 1.5 1994/06/15 23:14:49 pope Exp pope $");

/* Command line options: */
static VLong reduce_factor, expand_factor;
static VFloat scale_factor, rscale_factor = 1.0, cscale_factor = 1.0;
static VBoolean scale_found, rscale_found, cscale_found,
		reduce_found, sample_found, expand_found, aspect_found;
static VOptionDescRec options[] = {
    { "reduce", VLongRepn, 1, & reduce_factor, & reduce_found, NULL,
	  "Scale down by integer factor (averaging)" },
    { "sample", VLongRepn, 1, & reduce_factor, & sample_found, NULL,
	  "Scale down by integer factor (subsampling)" },
    { "expand", VLongRepn, 1, & expand_factor, & expand_found, NULL,
	  "Scale up by integer factor" },
    { "scale", VFloatRepn, 1, & scale_factor, & scale_found, NULL,
	  "Scale by any factor" },
    { "rscale", VFloatRepn, 1, & rscale_factor, & rscale_found, NULL,
	  "Scale row dimension by any factor" },
    { "cscale", VFloatRepn, 1, & cscale_factor, & cscale_found, NULL,
	  "Scale column dimension by any factor" },
    { "aspect", VBooleanRepn, 1, & aspect_found, VOptionalOpt, NULL,
	  "Scale to make pixels square" }
};

/* Later in this file: */
static void ScaleImage (VAttrListPosn *);
static void ScaleEdges (VAttrListPosn *);


/*
 *  Program entry point.
 */

int main (int argc, char *argv[])
{
    FILE *in_file, *out_file;
    VAttrList list;
    VAttrListPosn posn;
    int nobjects = 0;

    /* Parse command line arguments: */
    VParseFilterCmd (VNumber (options), options, argc, argv,
		     & in_file, & out_file);
    if (reduce_found + sample_found + expand_found + scale_found +
	(rscale_found || cscale_found) + aspect_found > 1)
	VError ("Invalid combination of options");
    if (! (reduce_found || sample_found || expand_found || scale_found ||
	   rscale_found || cscale_found || aspect_found)) {
	reduce_found = TRUE;
	reduce_factor = 2;
    }
      
    /* Read source image(s): */
    if (! (list = VReadFile (in_file, NULL)))
	exit (EXIT_FAILURE);

    /* Scale each object: */
    for (VFirstAttr (list, & posn); VAttrExists (& posn); VNextAttr (& posn)) {
	switch (VGetAttrRepn (& posn)) {

	case VImageRepn:
	    ScaleImage (& posn);
	    break;

	case VEdgesRepn:
	    ScaleEdges (& posn);
	    break;

	default:
	    continue;
	}
	nobjects++;
    }

    /* Write the results to the output file: */
    if (! VWriteFile (out_file, list))
	exit (EXIT_FAILURE);

    /* Report what happened: */
    if (reduce_found || sample_found)
	fprintf(stderr, "%s: Reduced %d object%s by a factor of %d.\n",
		argv[0], nobjects, nobjects == 1 ? "" : "s",
		(int) reduce_factor);
    else if (expand_found)
	fprintf(stderr, "%s: Expanded %d object%s by a factor of %d.\n",
		argv[0], nobjects, nobjects == 1 ? "" : "s", (int)
		expand_factor);
    else if (scale_found)
	fprintf(stderr, "%s: Scaled %d object%s by a factor of %g.\n",
		argv[0], nobjects, nobjects == 1 ? "" : "s", scale_factor);
    else if (rscale_found || cscale_found)
	fprintf(stderr, 
		"%s: Scaled %d object%s by %g row-wise, %g column-wise.\n",
		argv[0], nobjects, nobjects == 1 ? "" : "s",
		rscale_factor, cscale_factor);
    else
	fprintf(stderr, "%s: Squared up pixels in %d object%s.\n",
		argv[0], nobjects, nobjects == 1 ? "" : "s");

    return EXIT_SUCCESS;
}


static void ScaleImage (VAttrListPosn *posn)
{
    VImage src, result;
    VFloat aspect;

    VGetAttrValue (posn, NULL, VImageRepn, & src);
    
    if (reduce_found)
	result = VReduceImage (src, NULL, VAllBands,
			       (int) reduce_factor, (int) reduce_factor);
    else if (sample_found)
	result = VSampleImage (src, NULL, VAllBands,
			       (int) reduce_factor, (int) reduce_factor);
    else if (expand_found)
	result = VExpandImage (src, NULL, VAllBands,
			       (int) expand_factor, (int) expand_factor);
    else if (scale_found)
	result = VScaleImage (src, NULL, VAllBands,
			      (double) scale_factor, (double) scale_factor);
    else {
	
	if (aspect_found) {
	    
	    /* Remove any pixel_aspect_ratio attribute: */
	    aspect = 1.0;
	    VExtractAttr (VImageAttrList (src), VPixelAspectRatioAttr,
			  NULL, VFloatRepn, & aspect, FALSE);
	    if (aspect_found) {
		if (aspect > 1.0) {
		    cscale_factor = aspect;
		    rscale_factor = 1.0;
		} else if (aspect < 1.0) {
		    cscale_factor = 1.0;
		    rscale_factor = 1.0 / aspect;
		} else return;
	    }
	}
	result = VScaleImage (src, NULL, VAllBands,
			      (double) rscale_factor,
			      (double) cscale_factor);
    }
    if (! result)
	exit (EXIT_FAILURE);
    VSetAttrValue (posn, NULL, VImageRepn, result);
    VDestroyImage (src);
}


static void ScaleEdges (VAttrListPosn *posn)
{    
    VEdges edges;
    VEdge e;
    int i;
    
    VGetAttrValue (posn, NULL, VEdgesRepn, & edges);

    if (reduce_found || sample_found)
	rscale_factor = cscale_factor = 1.0 / reduce_factor;
    else if (expand_found)
	rscale_factor = cscale_factor = expand_factor;
    else if (scale_found)
	rscale_factor = cscale_factor = scale_factor;

    for (e = VFirstEdge (edges); VEdgeExists (e); e = VNextEdge (e))
	for (i = 0; i < VEdgeNPoints (e); i++) {
	    VEdgePointArray (e)[i][0] *= cscale_factor;
	    VEdgePointArray (e)[i][1] *= rscale_factor;
	}

    VEdgesNRows (edges) = VEdgesNRows (edges) * rscale_factor + 0.5;
    VEdgesNColumns (edges) = VEdgesNColumns (edges) * cscale_factor + 0.5;
}
