/*
 *  $Id: vsegedges.c,v 1.1 1994/06/10 01:38:23 pope Exp pope $
 *
 *  The vsegedges program segments edges by recursively splitting them
 *  into straight lines.
 */

/*
 *  Copyright 1994 University of British Columbia
 *
 *  Permission to use, copy, modify, distribute, and sell this software and its
 *  documentation for any purpose is hereby granted without fee, provided that
 *  the above copyright notice appears in all copies and that both that
 *  copyright notice and this permission notice appear in supporting
 *  documentation. UBC makes no representations about the suitability of this
 *  software for any purpose. It is provided "as is" without express or
 *  implied warranty.
 *
 *  Authors: David Lowe & Art Pope
 *           UBC Laboratory for Computational Intelligence
 */

/* From the Vista library: */
#include "vista/Vlib.h"
#include "vista/option.h"
#include "vista/mu.h"
#include "vista/os.h"
#include "vista/VEdges.h"

/* From the standard C library: */
#include <math.h>

/* File identification string: */
VRcsId ("$Id: vsegedges.c,v 1.1 1994/06/10 01:38:23 pope Exp pope $");


/* Program options: */
static VFloat accuracy = 2.0, magnitude = 0.0;
static VLong granularity = 4;
enum { pointsProduct, linesProduct, fitsProduct };
static VShort product = fitsProduct;
static VDictEntry productDict[] = {
    { "points", 0 },
    { "lines", 1 },
    { "fits", 2 },
    NULL
};
static VOptionDescRec options[] = {
    { "accuracy", VFloatRepn, 1, & accuracy, VOptionalOpt, NULL,
	  "Min. deviation due to measurement accuracy (pixels)" },
    { "granularity", VLongRepn, 1, & granularity, VOptionalOpt, NULL,
	  "Min. line length during recursive subdivision (pixels)" },
    { "magnitude", VFloatRepn, 1, & magnitude, VOptionalOpt, NULL,
	  "Min. average magnitude of a retained edge segment" },
    { "product", VShortRepn, 1, & product, VOptionalOpt, productDict,
	  "Desired output" }
};


/*
 *  Program entry point.
 */

int main (int argc, char *argv[])
{
    FILE *in_file, *out_file;
    int nedgesets = 0;
    VAttrList list;
    VAttrListPosn posn;
    VEdges src, result;

    /* Parse command line arguments and identify the input and output files: */
    VParseFilterCmd (VNumber (options), options, argc, argv,
		     & in_file, & out_file);
    if (accuracy == 0.0) {
	accuracy = 1.0;
	VWarning ("-accuracy value must be > 0; using 1");
    }
    if (granularity < 2) {
	granularity = 2;
	VWarning ("-granularity value must be >= 2; using 2");
    }

    /* Read the input file: */
    if (! (list = VReadFile (in_file, NULL)))
	exit (EXIT_FAILURE);

    /* Replace each edge set with its segmented version: */
    for (VFirstAttr (list, & posn); VAttrExists (& posn); VNextAttr (& posn)) {
	if (VGetAttrRepn (& posn) != VEdgesRepn)
	    continue;
	VGetAttrValue (& posn, NULL, VEdgesRepn, & src);
	result = VSegEdgesIntoLines (src, accuracy, granularity,
				     magnitude, product);
	if (! result)
	    exit (EXIT_FAILURE);
	VDestroyEdges (src);
	VSetAttrValue (& posn, NULL, VEdgesRepn, result);
	nedgesets++;
    }

    /* Write the output file: */
    if (! VWriteFile (out_file, list))
	exit (EXIT_FAILURE);

    fprintf (stderr, "%s: Processed %d edge set%s.\n",
	     argv[0], nedgesets, nedgesets == 1 ? "" : "s");
    return EXIT_SUCCESS;
}
