/*
 *  $Id: vselbands.c,v 1.1 1994/06/10 01:38:30 pope Exp pope $
 *
 *  The vselbands program, which selects bands from an image.
 */

/*
 *  Copyright 1994 University of British Columbia
 *
 *  Permission to use, copy, modify, distribute, and sell this software and its
 *  documentation for any purpose is hereby granted without fee, provided that
 *  the above copyright notice appears in all copies and that both that
 *  copyright notice and this permission notice appear in supporting
 *  documentation. UBC makes no representations about the suitability of this
 *  software for any purpose. It is provided "as is" without express or
 *  implied warranty.
 *
 *  Author: Arthur Pope, UBC Laboratory for Computational Intelligence
 */

/* From the Vista library: */
#include "vista/Vlib.h"
#include "vista/mu.h"
#include "vista/option.h"
#include "vista/os.h"
#include "vista/VImage.h"

/* File identification string: */
VRcsId ("$Id: vselbands.c,v 1.1 1994/06/10 01:38:30 pope Exp pope $");

/* Later in this file: */
static VImage SelectBands (VImage);
static VBoolean Included (int, VArgVector *, VDictEntry *);
static int Decode (VStringConst, VDictEntry *);

/* Command line options: */
static VArgVector band_args, frame_args, viewpt_args, color_args, comp_args;
static VBoolean band_found, frame_found, viewpt_found, color_found, comp_found;

static VOptionDescRec options[] = {
    { "band", VStringRepn, 0, & band_args, & band_found, NULL,
	  "Bands to be selected" },
    { "frame", VStringRepn, 0, & frame_args, & frame_found, NULL,
	  "Framess to be selected" },
    { "viewpoint", VStringRepn, 0, & viewpt_args, & viewpt_found, NULL,
	  "Viewpoints to be selected" },
    { "color", VStringRepn, 0, & color_args, & color_found, NULL,
	  "Colors to be selected" },
    { "component", VStringRepn, 0, & comp_args, & comp_found, NULL,
	  "Components to be selected" }
};

/* Dictionaries of names that can be used in place of band numbers: */
static VDictEntry viewptDict[] = {
    { "left", 0 }, { "right", 1 },
    NULL
};
static VDictEntry colorDict[] = {
    { "red", 0 }, { "green", 1 }, { "blue", 2 },
    NULL
};
static VDictEntry compDict[] = {
    { "real", 0 }, { "imaginary", 1 },
    { "x", 0 }, { "y", 1 }, { "z", 2 },
    NULL
};


/*
 *  Program entry point.
 */

int main (int argc, char *argv[])
{
    FILE *in_file, *out_file;
    VAttrList list;
    VAttrListPosn posn;
    VImage src, result;
    int nimages = 0;

    /* Parse command line arguments: */
    VParseFilterCmd (VNumber (options), options, argc, argv,
		     & in_file, & out_file);

    /* Require EITHER -band OR some of -frame, -viewpoint, etc.: */
    if (! (band_found ^
	   (frame_found | viewpt_found | color_found | comp_found)))
	VError ("Specify either -band or some of -frame, -viewpoint, etc.");

    /* Read source image(s): */
    if (! (list = VReadFile (in_file, NULL)))
	exit (EXIT_FAILURE);
    
    /* Select specified bands out of each image: */
    for (VFirstAttr (list, & posn); VAttrExists (& posn); ) {
	if (VGetAttrRepn (& posn) != VImageRepn)
	    continue;
	VGetAttrValue (& posn, NULL, VImageRepn, & src);
	result = SelectBands (src);
	VDestroyImage (src);
	if (result) {
	    VSetAttrValue (& posn, NULL, VImageRepn, result);
	    VNextAttr (& posn);
	} else VDeleteAttr (& posn);
	nimages++;
    }

    /* Write the results to the output file: */
    if (! VWriteFile (out_file, list))
	exit (EXIT_FAILURE);

    fprintf (stderr, "%s: Processed %d image%s.\n", argv[0], nimages,
	     nimages == 1 ? "" : "s");

    return EXIT_SUCCESS;
}


/*
 *  SelectBands
 *
 *  Create a new image containing just the bands specified by the
 *  program options.
 */

static VImage SelectBands (VImage src)
{
    int band, *bands, nbands = 0;
    int frame, viewpt, color, comp, nframes, nviewpts, ncolors, ncomps;
    VImage dest;
    VBandInterp frame_interp, viewpt_interp, color_interp, comp_interp;

    /* Make a list of bands to be copied to the destination image: */
    bands = VMalloc (VImageNBands (src) * sizeof (int));
    if (band_found) {
	for (band = 0; band < VImageNBands (src); band++)
	    if (Included (band, & band_args, NULL))
		bands[nbands++] = band;
    } else {
	for (frame = 0; frame < VImageNFrames (src); frame++) {
	    if (! Included (frame, & frame_args, NULL))
		continue;
	    for (viewpt = 0; viewpt < VImageNViewpoints (src); viewpt++) {
		if (! Included (viewpt, & viewpt_args, viewptDict))
		    continue;
		for (color = 0; color < VImageNColors (src); color++) {
		    if (! Included (color, & color_args, colorDict))
			continue;
		    for (comp = 0; comp < VImageNComponents (src); comp++)
			if (Included (comp, & comp_args, compDict))
			    bands[nbands++] =
				VBandIndex (src, frame, viewpt, color, comp);
		}
	    }
	}
    }

    /* It could be that no bands were selected: */
    if (nbands == 0) {
	VFree (bands);
	return NULL;
    }

    /* Create the destination image and copy over the bands: */
    dest = VCreateImage (nbands, VImageNRows (src), VImageNColumns (src),
			 VPixelRepn (src));
    if (! dest)
	exit (EXIT_FAILURE);
    for (band = 0; band < nbands; band++)
	if (! VCopyBand (src, bands[band], dest, band))
	    exit (EXIT_FAILURE);

    /* Copy over image attributes that are still valid: */
    VCopyImageAttrs (src, dest);
    if (! band_found) {

	/* Determine how many frames, viewpoints, etc. have been copied
	   from the source image. If all frames, etc., have been copied,
	   also copy the frame interpretation attribute. */
	for (nframes = frame = 0; frame < VImageNFrames (src); frame++)
	    nframes += Included (frame, & frame_args, NULL);
	frame_interp = (nframes == VImageNFrames (src)) ?
	    VImageFrameInterp (src) : VBandInterpNone;

	for (nviewpts = viewpt = 0; viewpt < VImageNViewpoints (src); viewpt++)
	    nviewpts += Included (viewpt, & viewpt_args, viewptDict);
	viewpt_interp = (nviewpts == VImageNViewpoints (src)) ?
	    VImageViewpointInterp (src) : VBandInterpNone;

	for (ncolors = color = 0; color < VImageNColors (src); color++)
	    ncolors += Included (color, & color_args, colorDict);
	color_interp = (ncolors == VImageNColors (src)) ?
	    VImageColorInterp (src) : VBandInterpNone;

	for (ncomps = comp = 0; comp < VImageNComponents (src); comp++)
	    ncomps += Included (comp, & comp_args, compDict);
	comp_interp = (ncomps == VImageNComponents (src)) ?
	    VImageComponentInterp (src) : VBandInterpNone;

	VSetBandInterp (dest,
			frame_interp, nframes,
			viewpt_interp, nviewpts,
			color_interp, ncolors,
			comp_interp, ncomps);
    }
    VFree (bands);
    return dest;
}


/*
 *  Included
 *
 *  Returns TRUE if a particular index value is, according to the command
 *  line, to be included in the output image.
 */

static VBoolean Included (int v, VArgVector *vec, VDictEntry *dict)
{
    int i, lb, ub;
    VStringConst str;

    if (vec->number == 0)
	return TRUE;
    for (i = 0; i < vec->number; i++) {
	ub = lb = Decode (((VStringConst *) vec->vector)[i], dict);
	if (i < vec->number - 2) {
	    str = ((VStringConst *) vec->vector)[i + 1];
	    if (strcmp (str, "to") == 0 ||
		strcmp (str, "-") == 0 ||
		strcmp (str, ":") == 0) {
		ub = Decode (((VStringConst *) vec->vector)[i + 2], dict);
		if (lb > ub)
		    VError ("\"%s %s %s\" isn't a valid range",
			    ((VStringConst *) vec->vector)[i],
			    ((VStringConst *) vec->vector)[i + 1],
			    ((VStringConst *) vec->vector)[i + 2]);
		i += 2;
	    }
	}
	if (v >= lb && v <= ub)
	    return TRUE;
    }
    return FALSE;
}


/*
 *  Decode
 *
 *  Decode an index value from the command line, with the help of a dictionary.
 */

static int Decode (VStringConst str, VDictEntry *dict)
{
    VLong value;

    if (! VDecodeAttrValue (str, dict, VLongRepn, & value))
	VError ("Invalid index argument \"%s\"", str);
    return value;
}
