/*
 *  $Id: vselect.c,v 1.5 1994/06/15 23:14:49 pope Exp pope $
 *
 *  The vselect program, which selects certain objects from a data file.
 */

/*
 *  Copyright 1993, 1994 University of British Columbia
 *
 *  Permission to use, copy, modify, distribute, and sell this software and its
 *  documentation for any purpose is hereby granted without fee, provided that
 *  the above copyright notice appears in all copies and that both that
 *  copyright notice and this permission notice appear in supporting
 *  documentation. UBC makes no representations about the suitability of this
 *  software for any purpose. It is provided "as is" without express or
 *  implied warranty.
 *
 *  Author: Arthur Pope, UBC Laboratory for Computational Intelligence
 */

/* From the V library: */
#include "vista/Vlib.h"
#include "vista/mu.h"
#include "vista/option.h"
#include "vista/os.h"
#include "vista/VImage.h"

/* File identification string: */
VRcsId ("$Id: vselect.c,v 1.5 1994/06/15 23:14:49 pope Exp pope $");

/* Later in this file: */
static VStringConst ObjectAttribute (VAttrListPosn *, VStringConst);


int main (int argc, char **argv)
{
    static VShort object;
    static VString name_str, type_str, attr_str[2];
    static VBoolean object_found, name_found, type_found, attr_found, not;
    static VOptionDescRec options[] = {
	{ "object", VShortRepn, 1, & object, & object_found, NULL,
	      "Object's order within file (0..n-1)" },
        { "name", VStringRepn, 1, & name_str, & name_found, NULL,
	      "Object's name" },
        { "type", VStringRepn, 1, & type_str, & type_found, NULL,
	      "Object's type" },
        { "attr", VStringRepn, 2, & attr_str, & attr_found, NULL,
	      "Name and value of object's attribute" },
        { "not", VBooleanRepn, 1, & not, VOptionalOpt, NULL,
	      "Invert selection criteria" }
    };
    FILE *in_file, *out_file;
    VAttrList list;
    int i, n;
    VAttrListPosn posn;
    VBoolean selected;
    VStringConst str;
    VBundle b;

    /* Parse command line arguments and identify the input and output files: */
    VParseFilterCmd (VNumber (options), options, argc, argv,
		     & in_file, & out_file);

    /* Ensure that exactly one of the alternate selection criteria was
       specified: */
    if ((int) object_found + name_found + type_found + attr_found != 1)
	VError ("Specify one of -object, -name, -type, or -attr");
    if (object_found && object < 0)
	VError ("Object index must be >= 0");

    /* ==> Special hack: remove registration of standard object types such as
           "image" and "edges" so that VReadFile will leave them
	   as VBundles rather than convert them to VImage, VEdges, etc.
	   This simplifies support for such things as selecting images
	   by their pixel representation. */
    VRepnInfo[VEdgesRepn].name = "";
    VRepnInfo[VImageRepn].name = "";

    /* Read the contents of the input file: */
    if (! (list = VReadFile (in_file, NULL)))
	exit (EXIT_FAILURE);
    
    /* For each object in that file: */
    i = n = 0;
    VFirstAttr (list, & posn);
    while (VAttrExists (& posn)) {

	/* Determine whether it is one of those selected: */
	if (object_found)
	    selected = (i == object);

	else if (name_found)

	    selected = (strcmp (VGetAttrName (& posn), name_str) == 0);

	else if (type_found) {

	    selected = (VGetAttrRepn (& posn) == VBundleRepn &&
			VGetAttrValue (& posn, NULL, VBundleRepn, & b) &&
			strcmp (b->type_name, type_str) == 0);

	} else if (attr_found)

	    selected = (str = ObjectAttribute (& posn, attr_str[0])) &&
		       (strcmp (str, attr_str[1]) == 0);

	else selected = FALSE;

	/* Delete the attribute if its not among those selected: */
	if (selected != not) {
	    n++;
	    VNextAttr (& posn);
	} else VDeleteAttr (& posn);
	i++;
    }

    if (object_found && object >= i)
	VWarning ("-object argument (%d) exceeds no. of objects in file (%d)",
		  object, i);

    /* Write the selected attributes to the output file: */
    if (! VWriteFile (out_file, list))
	exit (EXIT_FAILURE);

    fprintf (stderr, "%s: Selected %d object%s out of %d.\n",
	     argv[0], n, n == 1 ? "" : "s", i);
    return EXIT_SUCCESS;
}


/*
 *  ObjectAttribute
 *
 *  Return the string value of a particular attribute of an object.
 */

static VStringConst ObjectAttribute (VAttrListPosn *posn, VStringConst name)
{
    VAttrList sublist;
    VBundle b;
    VStringConst str;

    switch (VGetAttrRepn (posn)) {

    case VAttrListRepn:

	/* If the object is a general attribute list, just search the list
	   for the desired attribute: */
	VGetAttrValue (posn, NULL, VAttrListRepn, & sublist);
	break;

    case VBundleRepn:

	/* If the object is a bundle, search its attribute list for the
	   desired attribute: */
	VGetAttrValue (posn, NULL, VBundleRepn, & b);
	sublist = b->list;
	break;

    default:
	return NULL;
    }

    return VGetAttr (sublist, name, NULL, VStringRepn, & str) == VAttrFound ?
	str : NULL;
}
