/*
 *  $Id: vsynth.c,v 1.2 1994/06/30 23:03:57 pope Exp pope $
 *
 *  The vsynth program, which synthesizes an image containing one of a
 *  repertoire of possible patterns.
 */

/*
 *  Copyright 1994 University of British Columbia
 *
 *  Permission to use, copy, modify, distribute, and sell this software and its
 *  documentation for any purpose is hereby granted without fee, provided that
 *  the above copyright notice appears in all copies and that both that
 *  copyright notice and this permission notice appear in supporting
 *  documentation. UBC makes no representations about the suitability of this
 *  software for any purpose. It is provided "as is" without express or
 *  implied warranty.
 *
 *  Author: Arthur Pope, UBC Laboratory for Computational Intelligence
 */

/* From the Vista library: */
#include "vista/Vlib.h"
#include "vista/mu.h"
#include "vista/option.h"
#include "vista/os.h"
#include "vista/VImage.h"

/* From the standard C library: */
#include <math.h>

/* File identification string: */
VRcsId ("$Id: vsynth.c,v 1.2 1994/06/30 23:03:57 pope Exp pope $");

/* Types of pixel value patterns: */
enum {
    binomialPattern,
    constantPattern,
    impulsePattern,
    normalPattern,
    rampPattern,
    sineGratingPattern,
    uniformPattern,
    zonePlatePattern
};
static VDictEntry pattern_dict[] = {
    { "binomial", binomialPattern },
    { "constant", constantPattern },
    { "impulse", impulsePattern },
    { "normal", normalPattern },
    { "ramp", rampPattern },
    { "sine", sineGratingPattern },
    { "uniform", uniformPattern },
    { "zone", zonePlatePattern },
    { NULL }
};

/* Program options: */
static VLong type;
static VLong nrows = 256, ncolumns = 256;
static VLong repn = VFloatRepn;

static VStringConst object_name = "image";

static VDouble origin_arg[2] = { 0.0, 0.0 };
static double origin[2];
static VBoolean origin_found;

static VDouble amplitude = 1.0, base = 0.0;
static VDouble frequency = 0.0, orientation = 0.0, period = 8.0, phase = 0.0;
static VDouble freq[2] = { 0.0, 0.0 };
static VBoolean frequency_found, orientation_found, period_found;
static VBoolean freq_found[2];

static VDouble density = 0.5;
static VLong seed;

static VOptionDescRec options[] = {
    { "pattern", VLongRepn, 1, & type, VRequiredOpt, pattern_dict,
	  "Type of pixel value pattern" },
    { "nrows", VLongRepn, 1, & nrows, VOptionalOpt, NULL,
	  "Number of rows" }, 
    { "ncolumns", VLongRepn, 1, & ncolumns, VOptionalOpt, NULL,
	  "Number of columns" }, 
    { "repn", VLongRepn, 1, & repn, VOptionalOpt, VNumericRepnDict,
	  "Pixel representation" },
    { "name", VStringRepn, 1, & object_name, VOptionalOpt, NULL,
	  "Name to be given created image" },
    { "origin", VDoubleRepn, 2, origin_arg, & origin_found, NULL,
	  "Origin of pattern" },
    { "amplitude", VDoubleRepn, 1, & amplitude, VOptionalOpt, NULL,
	  "Amplitude of pattern" },
    { "base", VDoubleRepn, 1, & base, VOptionalOpt, NULL,
	  "Amplitude of background or DC component" },
    { "frequency", VDoubleRepn, 1, & frequency, & frequency_found, NULL,
	  "Frequency of grating pattern (cycles/pixel)" },
    { "orientation", VDoubleRepn, 1, & orientation, & orientation_found, NULL,
	  "Orientation of pattern (degrees, CCW from X axis)" },
    { "period", VDoubleRepn, 1, & period, & period_found, NULL,
	  "Period of grating pattern (pixels/cycle)" },
    { "phase", VDoubleRepn, 1, & phase, VOptionalOpt, NULL,
	  "Phase of grating pattern at origin (in [0,period])" },
    { "xfreq", VDoubleRepn, 1, & freq[0], & freq_found[0], NULL,
	  "Frequency of grating in X direction (cycles/pixel)" },
    { "yfreq", VDoubleRepn, 1, & freq[1], & freq_found[1], NULL,
	  "Frequency of grating in Y direction (cycles/pixel)" },
    { "density", VDoubleRepn, 1, & density, VOptionalOpt, NULL,
	  "Dot density for -type binomial (in [0,1])" },
    { "seed", VLongRepn, 1, & seed, VOptionalOpt, NULL,
	  "Seed for initializing random number generator" }
};


/*
 *  Program entry point.
 */

int main (int argc, char **argv)
{
    time_t now;
    FILE *out_file;
    VImage result;
    VAttrList list;

    /* Choose a random seed: */
    time (& now);
    VRandomSeed ((long) now);
    seed = VRandomLong ();

    /* Parse command line arguments: */
    VParseFilterCmd (VNumber (options), options, argc, argv, NULL, & out_file);

    /* Create an image of the specified size and pixel representation: */
    if (nrows <= 0)
	VError ("%s must be >= 0", "nrows");
    if (ncolumns <= 0)
	VError ("%s must be >= 0", "ncolumns");
    if(! (result = VCreateImage (1, nrows, ncolumns, repn)))
	exit (EXIT_FAILURE);

    /* If origin isn't specified, it defaults to the image centre: */
    if (origin_found) {
	origin[0] = origin_arg[0];
	origin[1] = origin_arg[1];
    } else {
	origin[0] = nrows / 2.0;
	origin[1] = ncolumns / 2.0;
    }

    /* Convert orientation to radians: */
    orientation *= M_PI / 180.0;

    /* Initialize the random number generator: */
    VRandomSeed (seed);

    /* Fill the image with the synthesized pattern: */
    switch (type) {

    case binomialPattern:
	VBinomialNoiseImage (result, 0, base, base + amplitude, 1.0 - density);
	break;

    case constantPattern:
	VFillImage (result, 0, base);
	break;

    case impulsePattern:
	VFillImage (result, 0, base);
	if (origin[0] >= 0 && origin[0] < VImageNRows (result) &&
	    origin[1] >= 0 && origin[1] < VImageNColumns (result))
	    VSetPixel (result, 0, (int) origin[0], (int) origin[1], amplitude);
	break;

    case normalPattern:
	VNormalNoiseImage (result, 0, base, amplitude);
	break;

    case rampPattern:
	VRampImage (result, 0, origin, orientation, base, amplitude);
	break;

    case sineGratingPattern:
	if ((frequency_found + period_found +
	     (freq_found[0] || freq_found[1]) > 1) ||
	    ((freq_found[0] || freq_found[1]) && orientation_found)) {
	    fprintf (stderr, "%s: Specify one of:\n", argv[0]);
	    fprintf (stderr, "\t-frequency and -orientation\n");
	    fprintf (stderr, "\t-period and -orientation\n");
	    fprintf (stderr, "\t-xfreq and -yfreq\n");
	    exit (EXIT_FAILURE);
	}
	if (period_found)
	    frequency = (period == 0) ? 0.0 : 2.0 * M_PI / period;
	else if (freq_found[0] || freq_found[1]) {
	    orientation = atan2 (freq[1], freq[0]);
	    frequency = sqrt (freq[0] * freq[0] + freq[1] * freq[1]);
	}
	VSineGratingImage (result, 0, origin, amplitude, base, frequency,
			   orientation, phase);
	break;

    case uniformPattern:
	VUniformNoiseImage (result, 0, base, base + amplitude);
	break;

    case zonePlatePattern:
	VZonePlateImage (result, 0, origin, amplitude, base,
			 (period == 0) ? DBL_MAX : 1.0 / period, phase);
    }

    /* Write the image to the output file: */
    list = VCreateAttrList ();
    VAppendAttr (list, object_name, NULL, VImageRepn, result);
    if (! VWriteFile (out_file, list))
	exit (EXIT_FAILURE);
    
    fprintf (stderr, "%s: Created %d x %d %s image.\n",
	     argv[0], (int) nrows, (int) ncolumns, VRepnName ((int) repn));

    return EXIT_SUCCESS;
}
