/*
 * $Id: vtops.c,v 1.6 1994/06/15 23:14:49 pope Exp pope $
 *
 * A program for rendering Vista images and edge sets in PostScript.
 */

/*
 *  Copyright 1993, 1994 University of British Columbia
 *
 *  Permission to use, copy, modify, distribute, and sell this software and its
 *  documentation for any purpose is hereby granted without fee, provided that
 *  the above copyright notice appears in all copies and that both that
 *  copyright notice and this permission notice appear in supporting
 *  documentation. UBC makes no representations about the suitability of this
 *  software for any purpose. It is provided "as is" without express or
 *  implied warranty.
 *
 *  Author: Daniel Ko, UBC Laboratory for Computational Intelligence
 */

/* From the Vista library: */
#include "vista/Vlib.h"
#include "vista/mu.h"
#include "vista/option.h"
#include "vista/os.h"
#include "vista/postscript.h"
#include "vista/VEdges.h"
#include "vista/VImage.h"
#include "vista/VList.h"

/* From the standard C library: */
#include <math.h>

/* File identification string: */
VRcsId ("$Id: vtops.c,v 1.6 1994/06/15 23:14:49 pope Exp pope $");

/* Default edge colors and line widths: */
VStringConst defaultColors[] = {
    "green", "blue", "yellow", "cyan", "magenta",
    "purple", "orange", "white", "tomato4", "hot pink"
};
#define defaultLineWidth    0.0f

/* Command line options: */
static VArgVector in_filenames;		/* list of input files */
static VString out_filename;		/* output file */
static VShort band;			/* image band to render */
static VBoolean eps = TRUE;		/* whether EPS or PS format */
static VArgVector colors;		/* list of edge colors */
static VBoolean draw_endpoints;		/* whether to draw endpoints */
static VArgVector linewidths;		/* list of edge line widths */
static VFloat arrow_width = 0.2;	/* scale of gradient arrows */

static VBoolean in_found, out_found;

static VOptionDescRec options[] = {
    { "in", VStringRepn, 0, & in_filenames, & in_found, NULL,
	  "Input images and/or edge sets" },
    { "out", VStringRepn, 1, & out_filename, & out_found, NULL,
	  "PostScript output" },
    { "band", VShortRepn, 1, & band, VOptionalOpt, NULL,
	  "Image band to render" },
    { "eps", VBooleanRepn, 1, & eps, VOptionalOpt, NULL,
	  "Whether to produce Encapsulated PostScript" },
    { "colors", VStringRepn, 0, & colors, VOptionalOpt, NULL,
	  "Edge set colors" },
    { "endpoints", VBooleanRepn, 1, & draw_endpoints, VOptionalOpt, NULL,
	  "Draw edge endpoints" },
    { "linewidths", VFloatRepn, 0, & linewidths, VOptionalOpt, NULL,
	  "Edge set line widths (pixels)" },
    { "arrow_width", VFloatRepn, 1, & arrow_width, VOptionalOpt, NULL,
	  "Width of flow field arrows" }
};

/* PostScript options and state: */
static VPSDescRec ps;

/* Objects to render: */
static VImage image;			/* image to render */
static VList edge_sets;			/* list of edge sets to render */
static float nrows, ncolumns;		/* their max dimensions */

/* Later in this file: */
static void LoadObjects (void);
static void ParseColor (VStringConst, float *, float *, float *);


/*
 *  Program entry point.
 */

int main (int argc, char **argv)
{
    VStringConst title;
    VEdges edge_set;
    float red, green, blue;
    FILE *outfile;
    int i;
    VBoolean result;

    /* Parse command-line arguments: */
    if (! VParseCommand (VNumber (options), options, & argc, argv) ||
	! VParsePSOptions (& argc, argv, & ps) ||
	! VIdentifyFiles (VNumber (options), options, "in",
			  & argc, argv, 0 /* stdin */) ||
	! VIdentifyFiles (VNumber (options), options, "out",
			  & argc, argv, -1))
	goto Usage;
    if (argc > 1) {
	VReportBadArgs (argc, argv);
Usage:  VReportUsage (argv[0], VNumber (options), options, NULL);
	VReportValidPSOptions ();
	exit (EXIT_FAILURE);
    };

    /* Assemble the image and/or collection of edge sets to be
       rendered: */
    LoadObjects ();

    /* Output the PostScript document and page headers: */
    title = ((VStringConst *) in_filenames.vector)[0];
    if (strcmp (title, "-") == 0)
	title = "<stdin>";
    if (! VBeginVPSDoc (outfile = VOpenOutputFile (out_filename, TRUE),
			& ps, eps, title))
	exit (EXIT_FAILURE);
    VBeginVPSPage (outfile, & ps, nrows, ncolumns);

    /* Draw image: */
    if (image) {
	if (VImageColorInterp (image) == VBandInterpRGB &&
	    VImageNComponents (image) == 1)
	    result = VRGBImageToPS (outfile, image, band);
	else if (VImageComponentInterp (image) == VBandInterpGradient &&
		 VImageNComponents (image) == 2)
	    result = VGradientImageToPS (outfile, image, band, arrow_width);
	else result = VImageBandToPS (outfile, image, band);
	if (! result)
	    exit (EXIT_FAILURE);
    }

    /* Draw edge sets: */
    for (edge_set = (VEdges) VListFirst (edge_sets), i = 0;
	 edge_set; edge_set = (VEdges) VListNext (edge_sets), i++) {

	/* Pick a color: */
	ParseColor (i < colors.number ?
		    ((VStringConst *) colors.vector)[i] :
		    defaultColors[i % VNumber (defaultColors)],
		    & red, & green, & blue);
	fprintf (outfile, "%g %g %g setrgbcolor", red, green, blue);

	/* Pick an edge width: */
	fprintf (outfile, " %g setlinewidth\n",
		 i < linewidths.number ?
		    ((VFloat *) linewidths.vector)[i] : defaultLineWidth);

	/* Draw the edge set: */
	VEdgesToPS (outfile, edge_set, draw_endpoints);
    };

    VEndVPSPage (outfile, & ps);
    VEndVPSDoc (outfile, & ps);

    /* Provide feedback: */
    i = VListCount (edge_sets);
    fprintf (stderr, "%s: ", argv[0]);
    if (image)
	fprintf (stderr, "1 image");
    if (i > 0)
	fprintf (stderr, "%s%d edge set%s",
		 image ? " and " : "", i, i > 1 ? "s" : "");
    fprintf (stderr, " converted to PostScript.\n");

    return EXIT_SUCCESS;
}


/*
 *  LoadObjects
 *
 *  Read the objects (an image, any number of edge sets) from the
 *  set of input files provided.
 */

static void LoadObjects (void)
{
    int i;
    FILE *f;
    VAttrList attrs;
    VAttrListPosn posn;
    VBoolean warning_given = FALSE;
    VEdges edge_set;
    VFloat aspect = 1.0;

    image = NULL;
    edge_sets = VListCreate ();
    nrows = ncolumns = 0.0f;

    /* Read the attribute list of each file: */
    for (i = 0; i < in_filenames.number; i++) {

	/* Open next file: */
	f = VOpenInputFile (((VStringConst *) in_filenames.vector)[i], TRUE);

	/* Read attribute list: */
	if (! (attrs = VReadFile (f, NULL)))
	    exit (EXIT_FAILURE);

	/* Scan each attribute: */
	for (VFirstAttr (attrs, & posn); VAttrExists (& posn);
	     VNextAttr (& posn)) {

	    switch (VGetAttrRepn (& posn)) {

	    case VImageRepn:

		/* Read image if none has been read: */
		if (image) {
		    if (! warning_given)
			VWarning ("Ignoring all but first image");
		    warning_given = TRUE;
		    continue;
		}
		VGetAttrValue (& posn, NULL, VImageRepn, & image);
		VGetAttr (VImageAttrList (image), VPixelAspectRatioAttr, NULL,
			  VFloatRepn, & aspect);
		nrows = VMax (nrows, VImageNRows (image));
		ncolumns = VMax (ncolumns, VImageNColumns (image) * aspect);
		break;

	    case VEdgesRepn:
		VGetAttrValue (& posn, NULL, VEdgesRepn, & edge_set);
		VListAppend (edge_sets, edge_set);
		nrows = VMax (nrows, VEdgesNRows (edge_set));
		ncolumns = VMax (ncolumns, VEdgesNColumns (edge_set));
		break;

	    default:
		break;
	    }
	}

	fclose (f);
    }
}


/*
 *  ParseColor
 *
 *  Convert a color specification string into rgb values.
 *  The color specification string follows the X11R5 convention.
 */

static void ParseColor (VStringConst color,
			float *red, float *green, float *blue)
{
    int n, items, i, r, g, b, c;
    float base;
    FILE *f;
    char line[1024], name[1024];

    /* Default is black: */
    *red = *green = *blue = 0.0f;

    /* Parse non-name color specification: */
    if (*color == '#') {
	n = strlen (color);
	color++;
	n--;
	if (n != 3 && n != 6 && n != 9 && n != 12) {
	    VWarning ("Invalid color specification %s", color);
	    return;
	}
	n /= 3;
	g = b = 0;
	do {
	    r = g;
	    g = b;
	    b = 0;
	    for (i = n; --i >= 0; ) {
		c = *color++;
		b <<= 4;
		if (c >= '0' && c <= '9')
		    b |= c - '0';
		else if (c >= 'A' && c <= 'F')
		    b |= c - ('A' - 10);
		else if (c >= 'a' && c <= 'f')
		    b |= c - ('a' - 10);
		else {
		    VWarning ("Invalid color specification %s", color);
		    return;
		}
	    }
	} while (*color);

	for (i = 0, base = 1.0f; i < n; i++, base *= 16.0f) ;
	*red = r / base;
	*green = g / base;
	*blue = b / base;
	return;
    }

    /* Parse color name: */
    if (! (f = fopen (XColorDatabase, "r"))) {
	VWarning ("Cannot find color database " XColorDatabase);
	return;
    }

    /* Scan each line of the database until match: */
    while (fgets (line, sizeof (line), f)) {
	items = sscanf (line, "%d %d %d %[^\n]\n", & r, & g, & b, name);
	if (items != 4) {
	    VWarning ("Invalid color database\n");
	    return;
	}
	if (r < 0 || r > 255 || g < 0 || g > 255 || b < 0 || b > 255) {
	    VWarning ("Invalid entry for %s in " XColorDatabase, color);
	    return;
	}

	/* Convert color is there is a match: */
	if (strcmp (name, color) == 0) {
	    *red = r / 255.0f;
	    *green = g / 255.0f;
	    *blue = b / 255.0f;
	    return;
	}
    }

    VWarning ("Cannot recognize color %s", color);
}
