/*
 *  $Id: vxviewi.c,v 1.11 1994/06/15 23:14:49 pope Exp pope $
 *
 *  Display an image in an X window.
 *  This very basic program is primarily for testing the VImageView widget.
 */

/*
 *  Copyright 1993, 1994 University of British Columbia
 *
 *  Permission to use, copy, modify, distribute, and sell this software and its
 *  documentation for any purpose is hereby granted without fee, provided that
 *  the above copyright notice appears in all copies and that both that
 *  copyright notice and this permission notice appear in supporting
 *  documentation. UBC makes no representations about the suitability of this
 *  software for any purpose. It is provided "as is" without express or
 *  implied warranty.
 *
 *  Author: Arthur Pope, UBC Laboratory for Computational Intelligence
 */

/* From the Vista library: */
#include "vista/Vlib.h"
#include "vista/colormap.h"
#include "vista/mu.h"
#include "vista/option.h"
#include "vista/os.h"
#include "vista/VImage.h"
#include "vista/VImageView.h"

#include "vista/VList.h"

/* From X windows and the Xt intrinsics: */
#include <X11/Intrinsic.h>
#include <X11/keysym.h>
#include <X11/Shell.h>
#include <X11/StringDefs.h>
#include <X11/Xatom.h>

/* File identification string: */
VRcsId ("$Id: vxviewi.c,v 1.11 1994/06/15 23:14:49 pope Exp pope $");

/* Later in this file: */
static void ReadImages (void);
static void ShowView (void);
static void Input (Widget, XtPointer, XtPointer);
static void DoNothing (VPointer);
static void NextFrame (XtPointer, XtIntervalId *);
static void GotoEnd (VBoolean);
static void HandleNonmaskableEvent (Widget, XtPointer, XEvent *, Boolean *);
static void PrintHelp (void);

/* The set of files supplying input images: */
VArgVector in_files;
VBoolean stdin_read = FALSE;		/* whether stdin is among them */

/* The list of images in those files: */
VList images;

/* The list of image views assembled from the input images: */
typedef struct {
    VStringConst filename;		/* name of file containing image */
    int image_idx;			/* index of image within file */
    VImage image;			/* image */
    VStringConst name;			/* optional image name */
    int band;				/* index of band within image */
    VBoolean color;			/* whether image is color */
} View;
VList views;

/* User interface state: */
static XtAppContext app_context;
static VLong delay = 100;		/* ms between movie frames */
static VBoolean noresize = FALSE;	/* don't automatically resize window */
static VBoolean signed_image = FALSE;	/* 1: display 0 as medium grey */
static VLong size = 256;		/* max dimension of displayed window */
static Widget shell_widget, image_widget; /* widgets created */
static Atom proto_atom, delete_atom;	/* atoms associated with WM protocol */
static VBoolean play_forward;		/* TRUE: forward; FALSE: backward */
static enum {
    stopped, running, recycling
} timer_state;				/* state of movie frame timer */
static XtIntervalId timer_id;		/* id of movie frame timer */

/* Command line options: */
static VBoolean in_found;
static VLong palette = None, depth = 0;
static VDictEntry palette_dict[] = {
    { "any", None },
    { "default", XA_RGB_DEFAULT_MAP }, { "best", XA_RGB_BEST_MAP },
    { "gray", XA_RGB_GRAY_MAP }, { "grey", XA_RGB_GRAY_MAP },
    NULL
};
static VOptionDescRec options[] = {
    { "in", VStringRepn, 0, & in_files, & in_found, NULL,
	  "Files of images to be displayed" },
    { "delay", VLongRepn, 1, & delay, VOptionalOpt, NULL,
	  "Milliseconds between movie frames" },
    { "depth", VLongRepn, 1, & depth, VOptionalOpt, NULL,
	  "Standard colormap to use" },
    { "noresize", VBooleanRepn, 1, & noresize, VOptionalOpt, NULL,
	  "Disable automatic resizing of window" },
    { "palette", VLongRepn, 1, & palette, VOptionalOpt,
	  palette_dict, "Color palette to use" },
    { "signed", VBooleanRepn, 1, & signed_image, VOptionalOpt, NULL,
	  "Display 0 as medium grey" },
    { "size", VLongRepn, 1, & size, VOptionalOpt, NULL,
	  "Maximum window dimension" }
};

/* Default resources: */
static String fallback_resources[] = {
    "*image.background: black",
    NULL
};
    

/* Denominator used for specifying aspect ratio to window manager: */
#define aspectDenominator 10000


/*
 *  Program entry point.
 */

int main (int argc, char *argv[])
{
    VColormap vc;
    XVisualInfo vinfo_template;
    long vinfo_mask;

    /* Parse command line arguments, initializing X Windows with them: */
    shell_widget = XtVaAppInitialize (& app_context, "Vxviewi", NULL, 0,
#if ! defined(XtSpecificationRelease) || XtSpecificationRelease < 5
				      (Cardinal *)
#endif				      
				      & argc, argv, fallback_resources,
				      XtNallowShellResize, TRUE,
				      XtNbaseHeight, 0, XtNbaseWidth, 0,
				      XtNheightInc, 1, XtNwidthInc, 1,
				      XtNminAspectY, aspectDenominator,
				      XtNmaxAspectY, aspectDenominator,
				      (char *) NULL);
    if (! VParseCommand (VNumber (options), options, & argc, argv))
	goto Usage;

    /* Identify the files to be displayed: */
    if (! VIdentifyFiles (VNumber (options), options, "in", & argc, argv, 0))
	goto Usage;

    /* Any remaining unparsed arguments are erroneous: */
    if (argc > 1) {
	VReportBadArgs (argc, argv);
Usage:  VReportUsage (argv[0], VNumber (options), options, "file ...");
        fprintf (stderr, "    plus X Windows options.\n");
	PrintHelp ();
        exit (EXIT_FAILURE);
    }

    /* Read the images: */
    ReadImages ();

    /* Select the requested standard colormap: */
    vinfo_mask = 0;
    if (depth > 0) {
	vinfo_mask |= VisualDepthMask;
	vinfo_template.depth = depth;
    }
    vc = VCreateColormap (XtScreen (shell_widget), (Atom) palette,
			  vinfo_mask, & vinfo_template);

    /* ==> Fortunately, it seems that we can set the shell widget's visual,
       colormap, and depth resources even though the widget has already
       been created -- because it hasn't been realized yet. This allows us
       to use XtVaAppInitialize to create the shell widget, rather than do
       things the hard way (calling XtToolkitInitialize, XtOpenDisplay, etc.
       individually so that we can insert shell resources at the appropriate
       point) */

    if (vc)
	XtVaSetValues (shell_widget,
		       XtNcolormap, VColormapColormap (vc),
		       XtNdepth, VColormapDepth (vc),
		       XtNvisual, VColormapVisual (vc), (char *) NULL);
    else VWarning ("Failed to get %s standard colormap",
		   XGetAtomName (XtDisplay (shell_widget), (Atom) palette));

    /* Create the enclosed VImageView widget: */
    image_widget = XtVaCreateManagedWidget ("image", vImageViewWidgetClass,
					    shell_widget,
					    VxNabsolute, ! signed_image,
					    VxNvColormap, vc,
					    (char *) NULL);
    XtAddCallback (image_widget, VxNinputCallback, Input, 0);

    /* Bring up the first image: */
    ShowView ();

    /* Realize the widgets: */
    XtRealizeWidget (shell_widget);

    /* Register our support for the WM_DELETE_WINDOW protocol: */
    proto_atom = XInternAtom (XtDisplay (shell_widget),
			      "WM_PROTOCOLS", FALSE);
    delete_atom = XInternAtom (XtDisplay (shell_widget),
			       "WM_DELETE_WINDOW", FALSE);
    if (proto_atom != None && delete_atom != None)
	XChangeProperty (XtDisplay (shell_widget), XtWindow (shell_widget),
			 proto_atom, XA_ATOM, 32, PropModePrepend,
			 (unsigned char *) & delete_atom, 1);
    XtAddEventHandler (shell_widget, 0, TRUE, HandleNonmaskableEvent, NULL);

    /* Handle events: */
    XtAppMainLoop (app_context);

    /* Just to make lint happy: */
    return EXIT_SUCCESS;
}


/*
 *  ReadImages
 *
 *  Read the images from one or more files and place them in a list of
 *  image views. Each view is a single band of an image, or a 3-band color
 *  image.
 */

static void ReadImages (void)
{
    int file_idx, image_idx, band;
    VStringConst filename, image_name;
    FILE *f;
    VAttrList attributes;
    VAttrListPosn posn;
    VImage image;
    VBoolean color;
    View *view;

    /* Create lists of images and views of those images: */
    images = VListCreate ();
    views = VListCreate ();

    /* For each input file: */
    for (file_idx = 0; file_idx < in_files.number; file_idx++) {

	/* Read all attributes from the file: */
	filename = ((VStringConst *) in_files.vector)[file_idx];
	if (strcmp (filename, "-") != 0) {
	    f = fopen (filename, "r");
	    if (f == NULL) {
		VWarning ("Unable to open image file %s", filename);
		continue;
	    }
	} else {
	    f = stdin;
	    filename = "<stdin>";
	    stdin_read = TRUE;
	}
	if (! (attributes = VReadFile (f, NULL))) {
	    VWarning ("%s is not a valid Vista data file", filename);
	    continue;
	}
	if (f != stdin)
	    fclose (f);

	/* For each image in the file: */
	for (VFirstAttr (attributes, & posn), image_idx = 0;
	     VAttrExists (& posn); VNextAttr (& posn)) {

	    if (VGetAttrRepn (& posn) != VImageRepn)
		continue;
	    VGetAttrValue (& posn, NULL, VImageRepn, & image);
	    VSetAttrValue (& posn, NULL, VPointerRepn, NULL);

	    /* Add it to the list of images: */
	    VListAppend (images, image);

	    /* Note its name, if it has one: */
	    if (VGetAttr (VImageAttrList (image), VNameAttr, NULL,
			  VStringRepn, & image_name) != VAttrFound)
		image_name = NULL;

	    /* If it is a single-component RGB color image, create a view
	       for each RGB triple of bands; otherwise create a view
	       for each image band: */
	    color = (VImageColorInterp (image) == VBandInterpRGB &&
		     VImageNComponents (image) == 1);
	    for (band = 0; band < VImageNBands (image);
		 band += color ? 3 : 1) {
		view = VNew (View);
		view->filename = filename;
		view->image_idx = image_idx;
		view->image = image;
		view->name = image_name;
		view->band = band;
		view->color = color;
		VListAppend (views, view);
	    }
	    image_idx++;
	}

	VDestroyAttrList (attributes);
    }

    if (VListCount (views) == 0)
	VError ("No input images");
    VListFirst (views);
}


/*
 *  ShowView
 *
 *  Display the current image view.
 */

static void ShowView (void)
{
    View *view = VListCurr (views);
    Dimension widget_width, widget_height;
    int image_width, image_height, aspect;

    /* Choose the image and band for display: */
    XtVaSetValues (image_widget,
		   VxNimage, view->image, VxNband, view->band,
		   (char *) NULL);

    /* Resize the shell if we're tracking image size changes: */
    if (! noresize) {

	/* Determine the current maximum dimension of the widget: */
	if (XtIsRealized (image_widget)) {
	    XtVaGetValues (image_widget,
			   XtNwidth, & widget_width,
			   XtNheight, & widget_height,
			   (char *) NULL);
	    size = VMax (widget_width, widget_height);
	}

	/* Compute the size and aspect ratio that fits within: */
	VImageWindowSize (view->image, size, size,
			  & image_width, & image_height);
	aspect = aspectDenominator * image_width / image_height;
	XtVaSetValues (shell_widget,
		       XtNheight, image_height, XtNwidth, image_width,
		       XtNminAspectX, aspect, XtNmaxAspectX, aspect,
		       (char *) NULL);
    }

    /* Report what it is we're now showing: */
    fprintf (stderr, "Image %d", view->image_idx);
    if (view->name)
	fprintf (stderr, " (%s)", view->name);
    fprintf (stderr, ", band %d of %s.\n", view->band, view->filename);
}


/*
 *  Input
 *
 *  Called when a mouse button is hit in the image.
 */

static void Input (Widget w, XtPointer clientData, XtPointer callData)
{
    XEvent *event = (XEvent *) callData;
    View *view;
    double row, column;
    char buf[10];
    KeySym keysym;

    switch (event->type) {

    case ButtonPress:
	if (event->xbutton.button != Button1)
	    break;
	view = VListCurr (views);
	if (VImageViewWindowToImage (image_widget, event->xbutton.x,
				     event->xbutton.y, & row, & column)) {
	    fprintf (stderr, "Row %d, Column %d: ", (int) row, (int) column);
	    
	    if (view->color)
		fprintf (stderr, "%g %g %g\n", 
			 VGetPixel (view->image, view->band,
				    (int) row, (int) column),
			 VGetPixel (view->image, view->band + 1,
				    (int) row, (int) column),
			 VGetPixel (view->image, view->band + 2,
				    (int) row, (int) column));
	    else fprintf (stderr, "%g\n",
			  VGetPixel (view->image, view->band,
				     (int) row, (int) column));
	}
	break;

    case KeyPress:
	if (XLookupString ((XKeyEvent *) event, buf, 2, & keysym, NULL) != 1)
	    break;
	switch (keysym) {

	case XK_a:
	    GotoEnd (FALSE);
	    break;

	case XK_b:
	    play_forward = FALSE;
	    NextFrame (NULL, & timer_id);
	    break;

	case XK_e:
	    GotoEnd (TRUE);
	    break;

	case XK_f:
	    play_forward = TRUE;
	    NextFrame (NULL, & timer_id);
	    break;

	case XK_n:
	    if (VListNext (views))
		ShowView ();
	    else {
		VListLast (views);
		fprintf (stderr, "No more images.\n");
	    }
	    break;

	case XK_p:
	    if (VListPrev (views))
		ShowView ();
	    else {
		VListFirst (views);
		fprintf (stderr, "No previous images.\n");
	    }
	    break;

	case XK_q:
	    exit (EXIT_SUCCESS);

	case XK_r:
	    if (stdin_read) {
		VWarning ("Can't reread images from <stdin>.");
		break;
	    }
	    fprintf (stderr, "Rereading images.\n");
	    XtVaSetValues (image_widget, VxNimage, NULL, (char *) NULL);
	    VListDestroy (images, (void (*) (VPointer)) VDestroyImage);
	    VListDestroy (views, DoNothing);
	    ReadImages ();
	    ShowView ();
	    break;

	case XK_s:
	    if (timer_state == running)
		XtRemoveTimeOut (timer_id);
	    if (timer_state != stopped) {
		ShowView ();
		timer_state = stopped;
	    }
	    break;

	default:
	    if (buf[0] == '?' || buf[0] == 0x1f)
		PrintHelp ();
	}
    }
}


/*
 *  NextFrame
 *
 *  Display the next frame in the current image.
 */

/* ARGSUSED */
static void NextFrame (XtPointer closure, XtIntervalId *id)
{
    View *view = VListCurr (views);

    timer_state = recycling;

    /* Step to the next band in the current image: */
    if (! (play_forward ? VListNext : VListPrev) (views) ||
	view->image != ((View *) VListCurr (views))->image) {
	(play_forward ? VListPrev : VListNext) (views);
	timer_state = stopped;
	ShowView ();
	return;
    }
    view = VListCurr (views);
    XtVaSetValues (image_widget, VxNband, view->band, (char *) NULL);

    /* Update the display: */
    XSync (XtDisplay (image_widget), 0);
    while (XtAppPending (app_context) & XtIMXEvent)
	XtAppProcessEvent (app_context, XtIMXEvent);

    /* Start the timer for the next frame: */
    if (timer_state == recycling) {
	timer_id = XtAppAddTimeOut (app_context, (long) delay,
				    NextFrame, NULL);
	timer_state = running;
    }
}


/*
 *  GotoEnd
 *
 *  Goto the first or last frame of an image.
 */

static void GotoEnd (VBoolean forward)
{
    VImage image = ((View *) VListCurr (views))->image;

    /* Step forward or backward while remaining in the same image: */
    while ((forward ? VListNext : VListPrev) (views) &&
	   image == ((View *) VListCurr (views))->image)
	/* Do nothing. */ ;
    (forward ? VListPrev : VListNext) (views);

    ShowView ();
}



static void DoNothing (VPointer p)
{
}


/*
 *  HandleNonmaskableEvent
 *
 *  Process a non-maskable event associated with the shell window.
 */

/* ARGSUSED */
static void HandleNonmaskableEvent (Widget w, XtPointer x, XEvent *ev,
			     Boolean *continue_to_dispatch)
{
    if (ev->type == ClientMessage && ev->xclient.data.l[0] == delete_atom)
	exit (EXIT_SUCCESS);
}


/*
 *  PrintHelp
 *
 *  Print summary of mouse and keyboard commands.
 */

static void PrintHelp (void)
{
    fprintf (stderr, "\nMouse input:\n");
    fprintf (stderr, "    Left        - report pixel value\n");
    fprintf (stderr, "    Middle      - zoom in\n");
    fprintf (stderr, "    Right       - zoom out\n");
    fprintf (stderr, "    Ctrl Middle - pan\n");

    fprintf (stderr, "\nKeyboard input:\n");
    fprintf (stderr, "    Ctrl-?      - display help\n");
    fprintf (stderr, "    Ctrl-A      - goto first frame\n");
    fprintf (stderr, "    Ctrl-B      - play movie backward\n");
    fprintf (stderr, "    Ctrl-E      - goto last frame\n");
    fprintf (stderr, "    Ctrl-F      - play movie foward\n");
    fprintf (stderr, "    Ctrl-N      - next image\n");
    fprintf (stderr, "    Ctrl-P      - previous image\n");
    fprintf (stderr, "    Ctrl-Q      - quit\n");
    fprintf (stderr, "    Ctrl-R      - reread images\n");
    fprintf (stderr, "    Ctrl-S      - stop movie\n");
    fprintf (stderr, "\n");
}
