/*
 *  $Id: Calibrate.c,v 1.2 1994/06/15 23:14:49 pope Exp pope $
 *
 *  Routines for estimating the parameters of a camera calibration model
 *  from a set of data points for which both world  and image coordinates are
 *  provided.
 *
 *  The method, due to Roger Tsai, is described in `A versatile camera
 *  calibration technique for high-accuracy 3d machine vision metrology
 *  using off-the-shelf TV cameras and lenses', IEEE Journal of Robotics
 *  and Automation, Vol RA-3, No 4, Aug 1987, pp 323-344.
 */

/*
 *  Copyright 1994 University of British Columbia
 *
 *  Permission to use, copy, modify, distribute, and sell this software and its
 *  documentation for any purpose is hereby granted without fee, provided that
 *  the above copyright notice appears in all copies and that both that
 *  copyright notice and this permission notice appear in supporting
 *  documentation. UBC makes no representations about the suitability of this
 *  software for any purpose. It is provided "as is" without express or
 *  implied warranty.
 *
 *  Adapted by Arthur Pope, UBC Laboratory for Computational Intelligence,
 *  from implementation by Reg Willson, Carnegie Mellon University.
 */

/* From the local directory: */
#include "FuncMin.h"
#include "LinSysMin.h"

/* From the Vista library: */
#include "vista/Vlib.h"
#include "vista/os.h"
#include "vista/VCamera.h"

/* From the standard C library: */
#include <math.h>
#include <setjmp.h>

/* File identification string: */
VRcsId ("$Id: Calibrate.c,v 1.2 1994/06/15 23:14:49 pope Exp pope $");

/* Information about a camera calibration problem: */
typedef struct {
    VCameraRec cam;		    /* camera info and current estimate */
    VBoolean extrinsics_only;	    /* only estimate extrinsic parameters */
    VBoolean coplanar;		    /* whether data points are coplanar */
    VBoolean early_distortion_removal; /* controls (Tz,f,k1,Cx,Cy) opt */
    int npoints;		    /* number of data pts */
    VDouble (*pw)[3];		    /* world coords of data pts */
    VDouble (*pid)[2];		    /* distorted image coords of data pts */
    VDouble (*ps)[2];		    /* sensor coords of data pts */
    Estimator est;		    /* least-squares estimator */
    int max_iterations;		    /* first optimization termination cond. */
    double tolerance;		    /* second optimization termination cond. */
    jmp_buf error_env;		    /* longjmp destination on error */
    int parmset;		    /* parameters currently being optimized */
} CCState;

/* Factors involving Ty^-1 computed at step b by Tsai's procedure: */
typedef struct {
    VDouble ty_sx_r1, ty_sx_r2;
    VDouble ty_sx_tx;
    VDouble ty_r4, ty_r5;
    VDouble ty_sx_r3, ty_r6;	    /* for noncoplanar case only */
} TyFactors;

/* List of parameters we optimize iteratively: */
enum {
    parm_Rx, parm_Ry, parm_Rz, parm_Tx, parm_Ty, parm_Tz,
    parm_f, parm_sx, parm_Cx, parm_Cy, parm_k1,
    maxNParms
};

/* Mask bits for those parameters: */
enum {
    parm_Rx_mask = 1 << parm_Rx,
    parm_Ry_mask = 1 << parm_Ry,
    parm_Rz_mask = 1 << parm_Rz,
    parm_Tx_mask = 1 << parm_Tx,
    parm_Ty_mask = 1 << parm_Ty,
    parm_Tz_mask = 1 << parm_Tz,
    parm_f_mask = 1 << parm_f,
    parm_sx_mask = 1 << parm_sx,
    parm_Cx_mask = 1 << parm_Cx,
    parm_Cy_mask = 1 << parm_Cy,
    parm_k1_mask = 1 << parm_k1
};

/* The set of extrinsic parameters: */
#define extrinsic_parameters \
    (parm_Rx_mask | parm_Ry_mask | parm_Rz_mask | \
     parm_Tx_mask | parm_Ty_mask | parm_Tz_mask)

/* Later in this file: */
static void AlgebraicOptimization (CCState *);
static void ComputeTyFactors (CCState *, TyFactors *);
static VDouble ComputeTy (CCState *, TyFactors *);
static void ComputeRotation (CCState *, TyFactors *);
static void ComputeTranslation (CCState *);
static void ComputeFocalLength (CCState *);

static void IterativeOptimization (CCState *);
static VDouble ErrorFunction (CCState *, VDouble []);
static void LoadParms (VDouble [], VCamera, int);
static void StoreParms (VDouble [], VCamera, int);

#ifdef DEBUG
static void DumpState (CCState *, VStringConst);
#endif

/* Handy macros: */
#define Sqr(a) ((a) * (a))


/*
 *  VCalibrateCamera
 *
 *  Calibrate camera parameters using a specified set of data points. If
 *  extrinsics_only is TRUE, only rotation and translation are estimated.
 */

VCamera VCalibrateCamera (VCamera src, VBoolean extrinsics_only,
			  VBoolean coplanar, int npoints,
			  VDouble pw[][3], VDouble pid[][2],
			  int max_iterations, double tolerance)
{
    VCamera result = NULL;
    CCState s;
    int i;

    /* For coplanar data, the Z world coordinate should be zero: */
    if (coplanar)
	for (i = 0; i < npoints; i++)
	    if (pw[i][2]) {
		VWarning ("VCalibrateCamera: "
			  "Z coordinate supplied for coplanar data");
		return NULL;
	    }

    /* If only extrinsic parameters are being estimated, the focal length
       must be defined: */
    if (extrinsics_only && ! src->focal_length_defd) {
	VWarning ("VCalibrateCamera: Focal length undefined");
	return NULL;
    }

    /* Package up the initial estimate and other information: */
    s.cam = *src;
    s.cam.extrinsics_defd = s.cam.focal_length_defd = TRUE;
    if (! (s.extrinsics_only = extrinsics_only))
	s.cam.k1 = 0.0;
    s.coplanar = coplanar;
    s.early_distortion_removal = FALSE;
    s.npoints = npoints;
    s.pw = pw;
    s.pid = pid;
    s.ps = VMalloc (npoints * 2 * sizeof (VDouble));
    s.max_iterations = max_iterations;
    s.tolerance = tolerance;
    if (setjmp (s.error_env))
	goto Done;

    /* Solve algebraically for initial estimates of R, T, and, if
       solving for intrinsic parameters, also sx and f: */
    AlgebraicOptimization (& s);
#ifdef DEBUG
    DumpState (& s, "algebraic optimization");
#endif

    if (extrinsics_only) {

	/* Iteratively improve the R and T estimates: */
	s.parmset = extrinsic_parameters;
	IterativeOptimization (& s);
#ifdef DEBUG
	DumpState (& s, "[R,T] optimization");
#endif

    } else {

	/* Iteratively improve the f, Tz, and k1 estimates: */
	s.parmset = parm_f_mask | parm_Tz_mask | parm_k1_mask;
	IterativeOptimization (& s);
#ifdef DEBUG
	DumpState (& s, "[f,Tz,k1] optimization");
#endif

	if (coplanar) {
	    
	    /* Do a 5 parameter (Tz, f, k1, Cx, Cy) optimization: */
	    s.parmset |= parm_Cx_mask | parm_Cy_mask;
	    IterativeOptimization (& s);
#ifdef DEBUG
	    DumpState (& s, "first [f,Tz,k1,Cx,Cy] optimization");
#endif

	    /* Do a better 5 parameter (Tz, f, k1, Cx, Cy) optimization: */
	    s.early_distortion_removal = TRUE;
	    IterativeOptimization (& s);
	    s.early_distortion_removal = FALSE;
#ifdef DEBUG
	    DumpState (& s, "second [f,Tz,k1,Cx,Cy] optimization");
#endif
	}
	
	/* Do a full optimization minus the image center: */
	s.parmset = extrinsic_parameters | parm_f_mask | parm_k1_mask;
	if (! coplanar)
	    s.parmset |= parm_sx_mask;
	IterativeOptimization (& s);
#ifdef DEBUG
	DumpState (& s, "optimization of everything but center");
#endif
	
	/* Do a full optimization including the image center: */
	s.parmset |= parm_Cx_mask | parm_Cy_mask;
	IterativeOptimization (& s);
#ifdef DEBUG
	DumpState (& s, "full optimization");
#endif	
    }

    /* Allocate a VCamera in which to return the calibration data: */
    result = VCreateCamera ();
    *result = s.cam;
    result->attributes = VCopyAttrList (src->attributes);

Done:
    VFree (s.ps);
    return result;
}


/*
 *  AlgebraicOptimization
 *
 *  Solves algebraically for T, R, f, and (in the non-coplanar case) sx
 *  according to the current estimates of C (image center) and k1
 *  (radial distortion).
 */

static void AlgebraicOptimization (CCState *s)
{
    int i;
    TyFactors u;
    VBoolean compute_sx;

    /* Only when calibrating intrinsic parameters from non-coplanar data
       is sx computed:
    compute_sx = ! s->coplanar && ! s->extrinsics_only;

    /* Compute the sensor coordinates of each data point according to the
       initial estimate of the frame buffer center and, for coplanar
       calibration, the X scale factor: */
    for (i = 0; i < s->npoints; i++)
	VImageToUndistortedSensorCoords (& s->cam, s->pid[i], s->ps[i]);

    /* Compute Ty^-1 sx r1, ..., Ty^-1 sx Tx, Ty^-1 r4: */
    ComputeTyFactors (s, & u);

    /* Compute Ty and, from it, Tx: */
    s->cam.t[1] = ComputeTy (s, & u);
    s->cam.t[0] = u.ty_sx_tx * s->cam.t[1];

    /* For non-coplanar calibration, compute sx and use it to update the
       computed sensor coordinates of each data point: */
    if (compute_sx) {
	s->cam.sx = fabs (s->cam.t[1]) *
	    sqrt (Sqr (u.ty_sx_r1) + Sqr (u.ty_sx_r2) + Sqr (u.ty_sx_r3));
	for (i = 0; i < s->npoints; i++)
	    VImageToUndistortedSensorCoords (& s->cam, s->pid[i], s->ps[i]);
    }

    /* Compute one solution for the 3d rotation matrix: */
    ComputeRotation (s, & u);

    /* Depending on whether we're estimating all parameters, or just
       extrinsic ones... */

    if (s->extrinsics_only) {

	/* Solve for a best estimate of Tx, Ty, and Tz: */
	ComputeTranslation (s);

    } else {

	/* Compute an approximation for f and Tz: */
	ComputeFocalLength (s);
	if (s->cam.f < 0) {
	    /* Try the other solution for the orthonormal matrix: */
	    s->cam.r[0][2] = - s->cam.r[0][2];
	    s->cam.r[1][2] = - s->cam.r[1][2];
	    s->cam.r[2][0] = - s->cam.r[2][0];
	    s->cam.r[2][1] = - s->cam.r[2][1];
	    VComputeCameraRotationAngles (& s->cam);
	    ComputeFocalLength (s);
	    if (s->cam.f < 0) {
		VWarning ("VCalibrate: Possible handedness problem with data");
		longjmp (s->error_env, 1);
	    }
	}
    }
    s->cam.extrinsics_defd = TRUE;
}


/*
 *  ComputeTyFactors
 *
 *  Solve for several factors of Ty^-1 using an overconstrained system of
 *  linear equations. This is step b of Tsai's procedure.
 */

static void ComputeTyFactors (CCState *s, TyFactors *u)
{
    int i, nvars = s->coplanar ? 5 : 7;
    VDouble a[7];

    /* Find the least-squares solution to a system of equations: */
    V_InitEstimator (& s->est, nvars);
    for (i = 0; i < s->npoints; i++) {
	a[0] = s->ps[i][1] * s->pw[i][0];
	a[1] = s->ps[i][1] * s->pw[i][1];
	a[2] = s->ps[i][1];
	a[3] = - s->ps[i][0] * s->pw[i][0];
	a[4] = - s->ps[i][0] * s->pw[i][1];
	if (! s->coplanar) {
	    a[5] = s->ps[i][1] * s->pw[i][2];
	    a[6] = - s->ps[i][0] * s->pw[i][2];
	}
	V_ExtendEstimator (& s->est, a, s->ps[i][0]);
    }
    if (! V_UpdateEstimator (& s->est)) {
	VWarning ("VCalibrate: Numerically ill-conditioned problem");
	longjmp (s->error_env, 1);
    }
    memcpy (u, s->est.x, nvars * sizeof (VDouble));
}


/*
 *  ComputeTy
 *
 *  Compute's Ty from Ty^-1 factors.
 *  This is steps c(1) and c(2) of Tsai's procedure.
 */

static VDouble ComputeTy (CCState *s, TyFactors *u)
{
    VDouble ty, ty2, sr, distance, far_distance, x, y;
    int i, far_point;
    #define Epsilon 1.0E-8

    /* First find |Ty| from Ty^2: */
    if (s->coplanar) {

	/* Coplanar case: */
	if (fabs (u->ty_sx_r1) < Epsilon && fabs (u->ty_sx_r2) < Epsilon)
	    ty2 = 1.0 / (Sqr (u->ty_r4) + Sqr (u->ty_r5));
	else if (fabs (u->ty_r4) < Epsilon && fabs (u->ty_r5) < Epsilon)
	    ty2 = 1.0 / (Sqr (u->ty_sx_r1) + Sqr (u->ty_sx_r2));
	else if (fabs (u->ty_sx_r1) < Epsilon && fabs (u->ty_r4) < Epsilon)
	    ty2 = 1.0 / (Sqr (u->ty_sx_r2) + Sqr (u->ty_r5));
	else if (fabs (u->ty_sx_r2) < Epsilon && fabs (u->ty_r5) < Epsilon)
	    ty2 = 1.0 / (Sqr (u->ty_sx_r1) + Sqr (u->ty_r4));
	else {
	    sr = Sqr (u->ty_sx_r1) + Sqr (u->ty_sx_r2) +
		 Sqr (u->ty_r4) + Sqr (u->ty_r5);
	    ty2 = (sr - sqrt (Sqr (sr) - 4.0 *
		   Sqr (u->ty_sx_r1 * u->ty_r5 - u->ty_r4 * u->ty_sx_r2))) /
	     (2.0 * Sqr (u->ty_sx_r1 * u->ty_r5 - u->ty_r4 * u->ty_sx_r2));
	}

    } else {

	/* Non-coplanar case: */
	ty2 = 1.0 / (Sqr (u->ty_r4) + Sqr (u->ty_r5) + Sqr (u->ty_r6));
    }
    ty = sqrt (ty2);

    /* Find a point that is far from the image center: */
    far_distance = 0.0;
    far_point = 0;
    for (i = 0; i < s->npoints; i++)
	if ((distance = Sqr (s->ps[i][0]) + Sqr (s->ps[i][1])) >
	    far_distance) {
	    far_point = i;
	    far_distance = distance;
	}

    /* Now find the sign for Ty. Assume it is positive: */
    x = u->ty_sx_r1 * s->pw[far_point][0] +
	u->ty_sx_r2 * s->pw[far_point][1] + u->ty_sx_tx;
    y = u->ty_r4 * s->pw[far_point][0] +
	u->ty_r5 * s->pw[far_point][1] + 1.0;
    if (! s->coplanar) {
	x += u->ty_sx_r3 * s->pw[far_point][2];
	y += u->ty_r6 * s->pw[far_point][2];
    }
    x *= ty;
    y *= ty;

    /* Flip Ty if the assumption was wrong: */
    if ((x >= 0.0) != (s->ps[far_point][0] >= 0.0) ||
	(y > 0.0) != (s->ps[far_point][1] >= 0.0))
	ty = -ty;

    return ty;
    #undef Epsilon
}


/*
 *  ComputeRotation
 *
 *  Computes the rotation matrix from Ty^-1 factors.
 *  This is step c(3) of Tsai's procedure.
 */

static void ComputeRotation (CCState *s, TyFactors *u)
{
    VCamera c = & s->cam;
    VDouble t;

    /* Compute the first two rows of the rotation matrix: */
    c->r[0][0] = u->ty_sx_r1 * c->t[1];
    c->r[0][1] = u->ty_sx_r2 * c->t[1];
    c->r[1][0] = u->ty_r4 * c->t[1];
    c->r[1][1] = u->ty_r5 * c->t[1];
    if (s->coplanar) {
	c->r[0][2] = sqrt (1.0 - Sqr (c->r[0][0]) - Sqr (c->r[0][1]));
	c->r[1][2] = sqrt (1.0 - Sqr (c->r[1][0]) - Sqr (c->r[1][1]));
	t = c->r[0][0] * c->r[1][0] + c->r[0][1] * c->r[1][1];
	if (t >= 0.0)
	    c->r[1][2] = - c->r[1][2];
    } else {
	c->r[0][0] /= c->sx;
	c->r[0][1] /= c->sx;
	c->r[0][2] = u->ty_sx_r3 * c->t[1] / c->sx;
	c->r[1][2] = u->ty_r6 * c->t[1];
    }

    /* Use the outer product of the first two rows to get the last row: */
    c->r[2][0] = c->r[0][1] * c->r[1][2] - c->r[0][2] * c->r[1][1];
    c->r[2][1] = c->r[0][2] * c->r[1][0] - c->r[0][0] * c->r[1][2];
    c->r[2][2] = c->r[0][0] * c->r[1][1] - c->r[0][1] * c->r[1][0];

    /* Find the Euler angles corresponding to this rotation matrix: */
    VComputeCameraRotationAngles (c);

    /* Generate a more orthonormal rotation matrix from the Euler angles: */
    VComputeCameraRotationMatrix (c);
}


/*
 *  ComputeTranslation
 *
 *  Solve for Tx, Ty, and Tz using an overconstrained system of linear
 *  equations. This routine gets involved only when we're solving only
 *  for extrinsic parameters (VCalibrateExtrinsics).
 */

static void ComputeTranslation (CCState *s)
{
    VCamera c = & s->cam;
    int i;
    VDouble pc[3], ps[2], a[3];

    /* Assume no translation for now: */
    c->t[0] = c->t[1] = c->t[2] = 0.0;

    /* Find the least-squares solution to a system of linear equations
       in Tx, Ty, Tz: */
    V_InitEstimator (& s->est, 3);

    /* For each data point: */
    for (i = 0; i < s->npoints; i++) {

	/* Convert from world coordinates to
	   untranslated camera coordinates: */
	VWorldToCameraCoords (c, s->pw[i], pc);

	/* Convert from image coordinates to undistorted sensor coordinates: */
	VImageToUndistortedSensorCoords (& s->cam, s->pid[i], ps);

	/* This gives two equations in terms of Tx, Ty, Tz: */
	a[0] = c->f;
	a[1] = 0.0;
	a[2] = - ps[0];
	V_ExtendEstimator (& s->est, a, ps[0] * pc[2] - c->f * pc[0]);

	a[0] = 0.0;
	a[1] = c->f;
	a[2] = - ps[1];
	V_ExtendEstimator (& s->est, a, ps[1] * pc[2] - c->f * pc[1]);
    }
    if (! V_UpdateEstimator (& s->est)) {
	VWarning ("VCalibrate: Numerically ill-conditioned problem");
	longjmp (s->error_env, 1);
    }

    /* Adopt the solution: */
    c->t[0] = s->est.x[0];
    c->t[1] = s->est.x[1];
    c->t[2] = s->est.x[2];
}


/*
 *  ComputeFocalLength
 *
 *  Solve for f and Tz using an overconstrained system of linear equations.
 *  This is step (d) of Tsai's procedure.
 */

static void ComputeFocalLength (CCState *s)
{
    VCamera c = & s->cam;
    int i;
    VDouble a[2], b;

    /* Find the least-squares solution to a system of equations: */
    V_InitEstimator (& s->est, 2);
    for (i = 0; i < s->npoints; i++) {
	a[0] = c->r[1][0] * s->pw[i][0] + c->r[1][1] * s->pw[i][1] + c->t[1];
	a[1] = - s->ps[i][1];
	b = c->r[2][0] * s->pw[i][0] + c->r[2][1] * s->pw[i][1];
	if (! s->coplanar) {
	    a[0] += c->r[1][2] * s->pw[i][2];
	    b += c->r[2][2] * s->pw[i][2];
	}
	V_ExtendEstimator (& s->est, a, b * s->ps[i][1]);
    }
    if (! V_UpdateEstimator (& s->est)) {
	VWarning ("VCalibrate: Numerically ill-conditioned problem");
	longjmp (s->error_env, 1);
    }
    c->f = s->est.x[0];
    c->t[2] = s->est.x[1];
    c->k1 = 0.0;	    /* (the assumption used in the calculation) */
}


/*
 *  IterativeOptimization
 *
 *  Iteratively optimize a specified subset of parameters.
 */

static void IterativeOptimization (CCState *s)
{
    VDouble x[maxNParms];

    /* Start the iteration from the current set of calibration parameters: */
    LoadParms (x, & s->cam, ~0);

    /* Optimize the specified set of parameters: */
    if (! V_PowellMin ((FMFunctionN *) ErrorFunction, s, s->parmset, x,
		     s->max_iterations, s->tolerance))
	longjmp (s->error_env, 1);

    /* Adopt the new, optimized parameters: */
    StoreParms (x, & s->cam, ~0);

    /* If the parameters we're optimizing include the image center (Cx, Cy)
       but not all extrinsic parameters, solve algebraically for the extrinsic
       parameters using optimized image center: */
    if ((s->parmset & parm_Cx_mask) &&
	(s->parmset & extrinsic_parameters) != extrinsic_parameters) {

	/* (The first time we optimize (Tz, f, k1, Cx, Cy), k1 is
	   taken to be zero for this step.) */
	if (! s->early_distortion_removal)
	    s->cam.k1 = 0.0;

	AlgebraicOptimization (s);

	/* Restore the parameters being optimized iteratively that may have
	   been overwritten by AlgebraicOptimization(): */
	StoreParms (x, & s->cam, s->parmset);
    }
}


/*
 *  ErrorFunction
 *
 *  Serves as the function to be minimized by the iterative optimization
 *  procedure. It returns the L2 distance between the projection of the
 *  world points onto the sensor plane, and the projection of the image
 *  points onto the sensor plane.
 */

static double ErrorFunction (CCState *s, VDouble x[])
{
    int i;
    VDouble pc[3], psu1[2], psu2[2], e = 0.0;

    StoreParms (x, & s->cam, s->parmset);

    /* If the parameters we're optimizing include the image center (Cx, Cy)
       but not all extrinsic parameters, solve algebraically for the extrinsic
       parameters using the latest image center: */
    if ((s->parmset & parm_Cx_mask) &&
	(s->parmset & extrinsic_parameters) != extrinsic_parameters) {

	/* (The first time we optimize (Tz, f, k1, Cx, Cy), k1 is
	   taken to be zero for this step.) */
	if (! s->early_distortion_removal)
	    s->cam.k1 = 0.0;

	AlgebraicOptimization (s);

	/* Restore the parameters being optimized iteratively that may have
	   been overwritten by AlgebraicOptimization(): */
	StoreParms (x, & s->cam, s->parmset);
    }

    /* Otherwise, if rotation is among the parameters we're optimizing,
       reconstruct the rotation matrix from the latest Euler angles: */
    else if (s->parmset & parm_Rx_mask)
	VComputeCameraRotationMatrix (& s->cam);

    /* For each data point... */
    for (i = 0; i < s->npoints; i++) {

	/* Convert point from world coordinates to undistorted sensor
	   plane coordinates: */
	VWorldToCameraCoords (& s->cam, s->pw[i], pc);
	psu1[0] = pc[0] * s->cam.f / pc[2];
	psu1[1] = pc[1] * s->cam.f / pc[2];

	/* Convert point from image coordinates to undistorted sensor
	   plane coordinates: */
	VImageToUndistortedSensorCoords (& s->cam, s->pid[i], psu2);

	/* Accumulate the squared error in undistorted sensor coordinates: */
	e += Sqr (psu1[0] - psu2[0]) + Sqr (psu1[1] - psu2[1]);
    }
    return e;
}


/*
 *  LoadParms, StoreParms
 *
 *  Move parameter values between a vector (the form preferred by the
 *  optimization procedure) and a VCamera (ther form preferred by everything
 *  else).
 */

static void LoadParms (VDouble x[], VCamera c, int parmset)
{
    #define LoadParm(i, field) \
	if (parmset & (1 << i))	\
	    x[i] = c->field;

    LoadParm (parm_Rx, angles[0]);
    LoadParm (parm_Ry, angles[1]);
    LoadParm (parm_Rz, angles[2]);
    LoadParm (parm_Tx, t[0]);
    LoadParm (parm_Ty, t[1]);
    LoadParm (parm_Tz, t[2]);
    LoadParm (parm_f, f);
    LoadParm (parm_Cx, c[0]);
    LoadParm (parm_Cy, c[1]);
    LoadParm (parm_k1, k1);

    #undef LoadParm
}

static void StoreParms (VDouble x[], VCamera c, int parmset)
{
    #define StoreParm(i, field)	\
	if (parmset & (1 << i))	\
	    c->field = x[i];

    StoreParm (parm_Rx, angles[0]);
    StoreParm (parm_Ry, angles[1]);
    StoreParm (parm_Rz, angles[2]);
    StoreParm (parm_Tx, t[0]);
    StoreParm (parm_Ty, t[1]);
    StoreParm (parm_Tz, t[2]);
    StoreParm (parm_f, f);
    StoreParm (parm_Cx, c[0]);
    StoreParm (parm_Cy, c[1]);
    StoreParm (parm_k1, k1);

    #undef StoreParm
}


#ifdef DEBUG

static void DumpState (CCState *s, VStringConst step)
{
    VCamera c = & s->cam;
    VDouble mean, stddev, max;

    fprintf (stderr, "\nResult of %s:\n", step);

    fprintf (stderr, "\tf = %.6f  [mm]\n\n", c->f);
    fprintf (stderr, "\tk1 = %.6e  [1/mm^2]\n\n", c->k1);
    fprintf (stderr, "\tTx = %.6f,  Ty = %.6f,  Tz = %.6f  [mm]\n\n",
	     c->t[0], c->t[1], c->t[2]);
    fprintf (stderr, "\tRx = %.6f,  Ry = %.6f,  Rz = %.6f  [deg]\n\n",
	     c->angle[0] * 180 / M_PI, c->angle[1] * 180 / M_PI,
	     c->angle[2] * 180 / M_PI);
    fprintf (stderr, "\tR\n");
    fprintf (stderr, "\t%9.6f  %9.6f  %9.6f\n",
	     c->r[0][0], c->r[0][1], c->r[0][2]);
    fprintf (stderr, "\t%9.6f  %9.6f  %9.6f\n",
	     c->r[1][0], c->r[1][1], c->r[1][2]);
    fprintf (stderr, "\t%9.6f  %9.6f  %9.6f\n\n",
	     c->r[2][0], c->r[2][1], c->r[2][2]);
    fprintf (stderr, "\tsx = %.6f\n", c->sx);
    fprintf (stderr, "\tCx = %.6f,  Cy = %.6f  [pixels]\n\n",
	     c->c[0], c->c[1]);
    fprintf (stderr, "\tTz / f = %.6f\n\n",
	     c->t[2] / c->f);

    VNormalizedCameraCalibrationError (c, s->npoints, s->pw, s->pid,
				       & mean, & stddev, & max);
    fprintf (stderr, "Normalized calibration error:\n"
	     "\tmean = %g  stddev = %g  max = %g\n",
	     mean, stddev, max);
    
    VObjectSpaceCameraCalibrationError (c, s->npoints, s->pw, s->pid,
					& mean, & stddev, & max);
    fprintf (stderr, "Object space calibration error:\n"
	     "\tmean = %g  stddev = %g  max = %g\n",
	     mean, stddev, max);
    fprintf (stderr, "\n");
}

#endif
