/*
 *  $Id: Camera.c,v 1.1 1994/06/10 01:39:16 pope Exp pope $
 *
 *  This file contains methods for the camera model (VCamera) type.
 */

/*
 *  Copyright 1994 University of British Columbia
 *
 *  Permission to use, copy, modify, distribute, and sell this software and its
 *  documentation for any purpose is hereby granted without fee, provided that
 *  the above copyright notice appears in all copies and that both that
 *  copyright notice and this permission notice appear in supporting
 *  documentation. UBC makes no representations about the suitability of this
 *  software for any purpose. It is provided "as is" without express or
 *  implied warranty.
 *
 *  Author: David Lowe, UBC Laboratory for Computational Intelligence
 */

/* From the Vista library: */
#include "vista/Vlib.h"
#include "vista/mu.h"
#include "vista/os.h"
#include "vista/VCamera.h"

/* From the standard C library: */
#include <math.h>

/* File identification string: */
VRcsId ("$Id: Camera.c,v 1.1 1994/06/10 01:39:16 pope Exp pope $");

/* Later in this file: */
static VDecodeMethod VCameraDecodeMethod;
static VEncodeAttrMethod VCameraEncodeAttrMethod;
static VEncodeDataMethod VCameraEncodeDataMethod;

/* Table of methods: */
static VTypeMethods VCameraMethods = {
    (VCopyMethod *) VCopyCamera,	/* copy a VCamera */
    (VDestroyMethod *) VDestroyCamera,	/* destroy a VCamera */
    VCameraDecodeMethod,		/* decode a VCamera's value */
    VCameraEncodeAttrMethod,		/* encode a VCamera's attr list */
    VCameraEncodeDataMethod		/* encode a VCamera's binary data */
};

/* Representation code obtained on registering the VCamera type: */
VRepnKind VCameraRepn;


/*
 *  VRegisterCameraType
 *
 *  Register the VCamera type.
 */

void VRegisterCameraType (void)
{
    VCameraRepn = VRegisterType ("camera", & VCameraMethods);
}


/*
 *  VCreateCamara
 *
 *  Create a VCamera object.
 */

VCamera VCreateCamera (void)
{
    VCamera c = VNew (VCameraRec);

    memset (c, 0, sizeof (VCameraRec));
    c->attributes = VCreateAttrList ();
    return c;
}


/*
 *  VCopyCamera
 *
 *  Copies a VCamera object.
 */

VCamera VCopyCamera (VCamera src)
{
    VCamera result = VNew (VCameraRec);

    *result = *src;
    result->attributes = VCopyAttrList (src->attributes);
    return result;
}


/*
 *  VDestroyCamera
 *
 *  Frees memory used by a VCamera object.
 */

void VDestroyCamera (VCamera c)
{
    VDestroyAttrList (c->attributes);
    VFree (c);
}


/*
 *  VSetCameraRotation
 *
 *  Set a camera's rotation, computing whichever of Euler angles or rotation
 *  matrix isn't supplied.
 */

void VSetCameraRotation (VCamera c, VDouble angles[3], VDouble matrix[3][3])
{
    if (angles)
	memcpy (c->angles, angles, sizeof (c->angles));
    if (matrix)
	memcpy (c->r, matrix, sizeof (c->r));
    if (angles && ! matrix)
	VComputeCameraRotationMatrix (c);
    if (matrix && ! angles)
	VComputeCameraRotationAngles (c);
}


/*
 *  VComputeCameraRotationAngles
 *
 *  Fill in Euler angles from the rotation matrix.
 */

void VComputeCameraRotationAngles (VCamera c)
{
    double cg, sg;

    c->angles[2] = atan2 (c->r[1][0], c->r[0][0]);
    cg = cos (c->angles[2]);
    sg = sin (c->angles[2]);
    c->angles[1] = atan2 (- c->r[2][0], c->r[0][0] * cg + c->r[1][0] * sg);
    c->angles[0] = atan2 (c->r[0][2] * sg - c->r[1][2] * cg,
			  c->r[1][1] * cg - c->r[0][1] * sg);
}


/*
 *  VComputeCameraRotationMatrix
 *
 *  Fill in a camera model's rotation matrix from its Euler angles.
 */

void VComputeCameraRotationMatrix (VCamera c)
{
    double ca, sa, cb, sb, cg, sg;

    ca = cos (c->angles[0]);
    sa = sin (c->angles[0]);
    cb = cos (c->angles[1]);
    sb = sin (c->angles[1]);
    cg = cos (c->angles[2]);
    sg = sin (c->angles[2]);

    c->r[0][0] = cb * cg;
    c->r[0][1] = cg * sa * sb - ca * sg;
    c->r[0][2] = sa * sg + ca * cg * sb;
    c->r[1][0] = cb * sg;
    c->r[1][1] = sa * sb * sg + ca * cg;
    c->r[1][2] = ca * sb * sg - cg * sa;
    c->r[2][0] = -sb;
    c->r[2][1] = cb * sa;
    c->r[2][2] = ca * cb;
}


/*
 *  VCameraDecodeMethod
 *
 *  The "decode" method registered for the "camera" type.
 *  Convert an attribute list to a VCamera object.
 */

static VPointer VCameraDecodeMethod (VStringConst name, VBundle b)
{
    VCamera c = VCreateCamera ();
    VAttrList list;
    VAttrListPosn posn;

    #define Extract(name, repn, field, required) \
	VExtractAttr (b->list, name, NULL, repn ## Repn, & c->field, required)

    /* Set some defaults: */
    c->npixels[0] = 512;	/* default number of columns */
    c->npixels[1] = 480;	/* default number of rows */

    /* Extract some attributes: */
    if (! Extract (VNSelsXAttr, VLong, nsels[0], TRUE) ||
	! Extract (VNSelsYAttr, VLong, nsels[1], TRUE) ||
	! Extract (VNPixelsXAttr, VLong, npixels[0], FALSE) ||
	! Extract (VNPixelsYAttr, VLong, npixels[1], FALSE) ||
	! Extract (VSelSizeXAttr, VDouble, sel_size[0], TRUE) ||
	! Extract (VSelSizeYAttr, VDouble, sel_size[1], TRUE)) {
	VDestroyCamera (c);
	return NULL;
    }

    /* Extract extrinsic parameters, which are optional: */
    c->extrinsics_defd = VLookupAttr (b->list, VRotationXAttr, & posn);
    if (c->extrinsics_defd) {
	if (! Extract (VRotationXAttr, VDouble, angles[0], TRUE) ||
	    ! Extract (VRotationYAttr, VDouble, angles[1], TRUE) ||
	    ! Extract (VRotationZAttr, VDouble, angles[2], TRUE) ||
	    ! Extract (VTranslationXAttr, VDouble, t[0], TRUE) ||
	    ! Extract (VTranslationYAttr, VDouble, t[1], TRUE) ||
	    ! Extract (VTranslationZAttr, VDouble, t[2], TRUE)) {
	    VDestroyCamera (c);
	    return NULL;
	}
	VComputeCameraRotationMatrix (c);
    }

    /* Extract the focal length, which is optional: */
    c->focal_length_defd = VLookupAttr (b->list, VFocalLengthAttr, & posn);
    if (c->focal_length_defd &&
	! Extract (VFocalLengthAttr, VDouble, f, TRUE)) {
	VDestroyCamera (c);
	return NULL;
    }

    /* Extract the remaining standard attributes, all of which
       have defaults: */
    c->pixel_size[0] = c->sel_size[0] * c->nsels[0] / c->npixels[0];
    c->pixel_size[1] = c->sel_size[1] * c->nsels[1] / c->npixels[1];
    c->c[0] = c->npixels[0] / 2.0;
    c->c[1] = c->npixels[1] / 2.0;
    c->sx = 1.0;
    c->k1 = 0.0;
    if (! Extract (VPixelSizeXAttr, VDouble, pixel_size[0], FALSE) ||
	! Extract (VPixelSizeYAttr, VDouble, pixel_size[1], FALSE) ||
	! Extract (VImageCenterXAttr, VDouble, c[0], FALSE) ||
	! Extract (VImageCenterYAttr, VDouble, c[1], FALSE) ||
	! Extract (VScaleXAttr, VDouble, sx, FALSE) ||
	! Extract (VRadialDistortionAttr, VDouble, k1, FALSE)) {
	VDestroyCamera (c);
	return NULL;
    }

    /* Give it the list of whatever attributes remain: */
    list = c->attributes;
    c->attributes = b->list;
    b->list = list;

    return c;

    #undef Extract
}


/*
 *  VCameraEncodeAttrMethod
 *
 *  The "encode_attrs" method registered for the "edges" type.
 *  Encode an attribute list value for a VCamera object.
 */

static VAttrList VCameraEncodeAttrMethod (VPointer value, size_t *lengthp)
{
    VCamera c = value;
    VAttrList list;

    #define Prepend(name, repn, field) \
        VPrependAttr (list, name, NULL, repn ## Repn, c->field)

    /* Temporarily prepend several attributes to the edge set's list: */
    if ((list = c->attributes) == NULL)
	list = c->attributes = VCreateAttrList ();

    /* Some intrinsic parameters are output only if they have
       non-trivial values: */
    if (c->sx != 1.0)
	Prepend (VScaleXAttr, VDouble, sx);
    if (c->k1 != 0.0)
	Prepend (VRadialDistortionAttr, VDouble, k1);
    Prepend (VImageCenterYAttr, VDouble, c[1]);
    Prepend (VImageCenterXAttr, VDouble, c[0]);

    /* Focal length is output only if it is defined: */
    if (c->focal_length_defd)
	Prepend (VFocalLengthAttr, VDouble, f);

    /* Extrinsic parameters are output only if they are defined: */
    if (c->extrinsics_defd) {
	Prepend (VTranslationZAttr, VDouble, t[2]);
	Prepend (VTranslationYAttr, VDouble, t[1]);
	Prepend (VTranslationXAttr, VDouble, t[0]);
	Prepend (VRotationZAttr, VDouble, angles[2]);
	Prepend (VRotationYAttr, VDouble, angles[1]);
	Prepend (VRotationXAttr, VDouble, angles[0]);
    }

    /* The remaining parameters are always output: */
    Prepend (VPixelSizeYAttr, VDouble, pixel_size[1]);
    Prepend (VPixelSizeXAttr, VDouble, pixel_size[0]);
    Prepend (VSelSizeYAttr, VDouble, sel_size[1]);
    Prepend (VSelSizeXAttr, VDouble, sel_size[0]);
    Prepend (VNPixelsYAttr, VLong, npixels[1]);
    Prepend (VNPixelsXAttr, VLong, npixels[0]);
    Prepend (VNSelsYAttr, VLong, nsels[1]);
    Prepend (VNSelsXAttr, VLong, nsels[0]);

    /* There's no binary data for a camera object: */
    *lengthp = 0;

    return list;

    #undef Prepend
}


/*
 *  VCameraEncodeDataMethod
 *
 *  The "encode_data" method registered for the "edges" type.
 *  Encode the edge and point fields for a VCamera object.
 */

static VPointer VCameraEncodeDataMethod (VPointer value, VAttrList list,
					size_t length, VBoolean *free_itp)
{
    VDouble d;

    #define Remove(name) \
        VExtractAttr (list, name, NULL, VDoubleRepn, & d, FALSE)

    /* Remove the attributes prepended by the VCameraEncodeAttrsMethod: */
    Remove (VImageCenterXAttr);
    Remove (VImageCenterYAttr);
    Remove (VFocalLengthAttr);
    Remove (VRadialDistortionAttr);
    Remove (VNPixelsXAttr);
    Remove (VNPixelsYAttr);
    Remove (VNSelsXAttr);
    Remove (VNSelsYAttr);
    Remove (VPixelSizeXAttr);
    Remove (VPixelSizeYAttr);
    Remove (VRotationXAttr);
    Remove (VRotationYAttr);
    Remove (VRotationZAttr);
    Remove (VScaleXAttr);
    Remove (VSelSizeXAttr);
    Remove (VSelSizeYAttr);
    Remove (VTranslationXAttr);
    Remove (VTranslationYAttr);
    Remove (VTranslationZAttr);

    *free_itp = FALSE;
    return list;		/* (return anything but NULL) */

    #undef Remove
}
