/*
 *  $Id: Coord.c,v 1.2 1994/10/01 00:10:25 pope Exp pope $
 *
 *  Routines that use a camera calibration model to perform various coordinate
 *  system conversions.
 */

/*
 *  Copyright 1994 University of British Columbia
 *
 *  Permission to use, copy, modify, distribute, and sell this software and its
 *  documentation for any purpose is hereby granted without fee, provided that
 *  the above copyright notice appears in all copies and that both that
 *  copyright notice and this permission notice appear in supporting
 *  documentation. UBC makes no representations about the suitability of this
 *  software for any purpose. It is provided "as is" without express or
 *  implied warranty.
 *
 *  Adapted by Arthur Pope, UBC Laboratory for Computational Intelligence,
 *  from implementation by Reg Willson, Carnegie Mellon University.
 */

/* From the Vista library: */
#include "vista/Vlib.h"
#include "vista/os.h"
#include "vista/VCamera.h"

/* From the standard C library: */
#include <math.h>

/* File identification string: */
VRcsId ("$Id: Coord.c,v 1.2 1994/10/01 00:10:25 pope Exp pope $");


/*
 *  VWorldToCameraCoords
 *
 *  Takes the position of a point in world coordinates (mm) and returns its
 *  position in camera coordinates (mm) according to a set of camera
 *  calibration parameters.
 */

VBoolean VWorldToCameraCoords (VCamera c, VDouble pw[3], VDouble pc[3])
{
    int i, j;

    if (! c->extrinsics_defd) {
	VWarning ("VWorldToCameraCoords: Extrinsic parameters undefined");
	return FALSE;
    }

    for (i = 0; i < 3; i++)
	for (j = 0, pc[i] = c->t[i]; j < 3; j++)
	    pc[i] += c->r[i][j] * pw[j];

    return TRUE;
}


/*
 *  VCameraToWorldCoords
 *
 *  Takes the position of a point in camera coordinates (mm) and returns its
 *  position in world coordinates (mm) according to a set of camera
 *  calibration parameters.
 */

VBoolean VCameraToWorldCoords (VCamera c, VDouble pc[3], VDouble pw[3])
{
    VDouble det;

    if (! c->extrinsics_defd) {
	VWarning ("VCameraToWorldCoords: Extrinsic parameters undefined");
	return FALSE;
    }

    /* (These equations were found by simple inverting the previous
       routine using Macsyma.) */
    det = c->r[2][0] * (c->r[0][1] * c->r[1][2] - c->r[0][2] * c->r[1][1]) +
	  c->r[2][1] * (c->r[0][2] * c->r[1][0] - c->r[0][0] * c->r[1][2]) +
	  c->r[2][2] * (c->r[0][0] * c->r[1][1] - c->r[0][1] * c->r[1][0]);

    pw[0] = (pc[0] * (c->r[1][1] * c->r[2][2] - c->r[1][2] * c->r[2][1]) +
	     pc[1] * (c->r[0][2] * c->r[2][1] - c->r[0][1] * c->r[2][2]) +
	     pc[2] * (c->r[0][1] * c->r[1][2] - c->r[0][2] * c->r[1][1]) +
	     c->t[0] * (c->r[1][2] * c->r[2][1] - c->r[1][1] * c->r[2][2]) +
	     c->t[1] * (c->r[0][1] * c->r[2][2] - c->r[0][2] * c->r[2][1]) +
	     c->t[2] * (c->r[0][2] * c->r[1][1] - c->r[0][1] * c->r[1][2])) /
		 det;

    pw[1] = (pc[0] * (c->r[1][0] * c->r[2][2] - c->r[1][2] * c->r[2][0]) +
	     pc[1] * (c->r[0][2] * c->r[2][0] - c->r[0][0] * c->r[2][2]) +
	     pc[2] * (c->r[0][0] * c->r[1][2] - c->r[0][2] * c->r[1][0]) +
	     c->t[0] * (c->r[1][2] * c->r[2][0] - c->r[1][0] * c->r[2][2]) +
	     c->t[1] * (c->r[0][0] * c->r[2][2] - c->r[0][2] * c->r[2][0]) +
	     c->t[2] * (c->r[0][2] * c->r[1][0] - c->r[0][0] * c->r[1][2])) /
		 -det;

    pw[2] = (pc[0] * (c->r[1][0] * c->r[2][1] - c->r[1][1] * c->r[2][0]) +
	     pc[1] * (c->r[0][1] * c->r[2][0] - c->r[0][0] * c->r[2][1]) +
	     pc[2] * (c->r[0][0] * c->r[1][1] - c->r[0][1] * c->r[1][0]) +
	     c->t[0] * (c->r[1][1] * c->r[2][0] - c->r[1][0] * c->r[2][1]) +
	     c->t[1] * (c->r[0][0] * c->r[2][1] - c->r[0][1] * c->r[2][0]) +
	     c->t[2] * (c->r[0][1] * c->r[1][0] - c->r[0][0] * c->r[1][1])) /
		det;

    return TRUE;
}


/*
 *  VWorldToImageCoords
 *
 *  Takes the position of a point in world coordinates (mm) and returns its
 *  position in images coordinates (pix) according to a set of camera
 *  calibration parameters.
 */

VBoolean VWorldToImageCoords (VCamera c, VDouble pw[3], VDouble pid[2])
{
    VDouble pc[3], psu[2], psd[2];

    if (! c->extrinsics_defd) {
	VWarning ("VWorldToImageCoords: Extrinsic parameters undefined");
	return FALSE;
    }
    if (! c->focal_length_defd) {
	VWarning ("VWorldToImageCoords: Focal length undefined");
	return FALSE;
    }

    /* Convert from world coordinates to camera coordinates: */
    VWorldToCameraCoords (c, pw, pc);

    /* Convert from camera coordinates to undistorted sensor
       plane coordinates: */
    psu[0] = c->f * pc[0] / pc[2];
    psu[1] = c->f * pc[1] / pc[2];

    /* Convert from undistorted to distorted sensor plane coordinates: */
    VUndistortedToDistortedSensorCoords (c, psu, psd);

    /* Convert from distorted sensor plane coordinates to image coordinates: */
    pid[0] = psd[0] * c->sx / c->pixel_size[0] + c->c[0];
    pid[1] = - psd[1] / c->pixel_size[1] + c->c[1];

    return TRUE;
}


/*
 *  VImageToWorldCoords
 *
 *  Takes the position of a point in image coordinates (pix) and returns its
 *  position in world coordinates (mm) according to a set of camera
 *  calibration parameters.
 */

VBoolean VImageToWorldCoords (VCamera c, VDouble pid[2], VDouble pw[3],
			      VDouble pwz)
{
    VDouble psd[2], psu[2], den;

    if (! c->extrinsics_defd) {
	VWarning ("VImageToWorldCoords: Extrinsic parameters undefined");
	return FALSE;
    }
    if (! c->focal_length_defd) {
	VWarning ("VImageToWorldCoords: Focal length undefined");
	return FALSE;
    }


    /* Convert from image to distorted sensor plane coordinates: */
    psd[0] = c->pixel_size[0] * (pid[0] - c->c[0]) / c->sx;
    psd[1] = - c->pixel_size[1] * (pid[1] - c->c[1]);

    /* Convert from distorted to undistorted sensor plane coordinates: */
    VDistortedToUndistortedSensorCoords (c, psd, psu);

    /* Calculate the corresponding world coordinates: */
    /* (These equations were found by simple inverting the perspective
       projection equations using Macsyma.) */
    den = psu[0] * (c->r[1][1] * c->r[2][0] - c->r[1][0] * c->r[2][1]) +
	  psu[1] * (c->r[0][0] * c->r[2][1] - c->r[0][1] * c->r[2][0]) -
	  c->f * (c->r[0][0] * c->r[1][1] + c->r[0][1] * c->r[1][0]);

    pw[0] = (psu[0] * (c->r[1][2] * c->r[2][1] - c->r[1][1] * c->r[2][2]) +
	     psu[1] * (c->r[0][1] * c->r[2][2] - c->r[0][2] * c->r[2][1]) -
	     c->f * (c->r[0][1] * c->r[1][2] + c->r[0][2] * c->r[1][1])) * pwz;
    pw[0] += psu[0] * (c->r[2][1] * c->t[1] - c->r[1][1] * c->t[2]) +
	     psu[1] * (c->r[0][1] * c->t[2] - c->r[2][1] * c->t[0]) -
	     c->f * (c->r[0][1] * c->t[1] + c->r[1][1] * c->t[0]);
    pw[0] /= den;

    pw[1] = (psu[0] * (c->r[1][2] * c->r[2][0] - c->r[1][0] * c->r[2][2]) +
	     psu[1] * (c->r[0][0] * c->r[2][2] - c->r[0][2] * c->r[2][0]) -
	     c->f * (c->r[0][0] * c->r[1][2] + c->r[0][2] * c->r[1][0])) * pwz;
    pw[1] += psu[0] * (c->r[2][0] * c->t[1] - c->r[1][0] * c->t[2]) +
	     psu[1] * (c->r[0][0] * c->t[2] - c->r[2][0] * c->t[0]) -
	     c->f * (c->r[0][0] * c->t[1] + c->r[1][0] * c->t[0]);
    pw[1] /= -den;

    return TRUE;
}


/*
 *  VDistortedToUndistortedImageCoords
 *
 *  Takes the position of a point as found in the image and returns the
 *  position it would have if there were no camera-caused distortion.
 */

void VDistortedToUndistortedImageCoords (VCamera c,
					 VDouble pid[2], VDouble piu[2])
{
    VDouble psd[2], psu[2];

    /* Convert from image to distorted sensor plane coordinates: */
    psd[0] = c->pixel_size[0] * (pid[0] - c->c[0]) / c->sx;
    psd[1] = - c->pixel_size[1] * (pid[1] - c->c[1]);

    /* Convert from distorted to undistorted sensor plane coordinates: */
    VDistortedToUndistortedSensorCoords (c, psd, psu);

    /* Convert from sensor to image coordinates: */
    piu[0] = psu[0] * c->sx / c->pixel_size[0] + c->c[0];
    piu[1] = - psu[1] / c->pixel_size[1] + c->c[1];
}


/*
 *  VUndistortedToDistortedImageCoords
 *
 *  Takes the position of a point would have if there were no camera-caused
 *  distortion and returns the position it actually has in the image
 */

void VUndistortedToDistortedImageCoords (VCamera c,
					 VDouble piu[2], VDouble pid[2])
{
    VDouble psu[2], psd[2];

    /* Convert from image to sensor coordinates */
    psu[0] = c->pixel_size[0] * (piu[0] - c->c[0]) / c->sx;
    psu[1] = - c->pixel_size[1] * (piu[1] - c->c[1]);

    /* Convert from undistorted to distorted sensor plane coordinates: */
    VUndistortedToDistortedSensorCoords (c, psu, psd);

    /* Convert from distorted sensor plane coordinates to image coordinates: */
    pid[0] = psd[0] * c->sx / c->pixel_size[0] + c->c[0];
    pid[1] = - psd[1] / c->pixel_size[1] + c->c[1];
}


/*
 *  VDistortedToUndistortedSensorCoords
 *
 *  Convert from distorted to undistorted sensor plane coordinates.
 */

void VDistortedToUndistortedSensorCoords (VCamera c,
					  VDouble psd[2], VDouble psu[2])
{
    VDouble distortion_factor;

    psu[0] = psd[0];
    psu[1] = psd[1];
    if (c->k1 != 0.0) {
	distortion_factor = 1.0 + c->k1 * (psd[0] * psd[0] + psd[1] * psd[1]);
	psu[0] *= distortion_factor;
	psu[1] *= distortion_factor;
    }
}


/*
 *  VUndistortedToDistortedSensorCoords
 *
 *  Convert from undistorted to distorted sensor plane coordinates.
 *
 *  Frederic Devernay's (<Frederic.Devernay@sophia.inria.fr>)
 *  significantly improved routine for converting from undistorted to
 *  distorted sensor coordinates. It algebraically solves a cubic
 *  polynomial in Rd (using the Cardan method). See, for example,
 *  Numerical Recipes in C.
 */

void VUndistortedToDistortedSensorCoords (VCamera c,
					  VDouble psu[2], VDouble psd[2])
{
    double ru, rd, a, b, q, r, s, t, u, cost, sint;
    #define Sqrt3 1.732050807568877293527446341505872366943
    #define CubeRoot(t) \
        ((t) >= 0.0 ? pow ((t), 1.0 / 3.0) : - pow (-(t), 1.0 / 3.0))

    /* First check for boundary conditions: */
    if ((psu[0] == 0.0 && psu[1] == 0.0) || c->k1 == 0.0) {
	psd[0] = psu[0];
	psd[1] = psu[1];
	return;
    }

    ru = sqrt (psu[0] * psu[0] + psu[1] * psu[1]);

    /* Find the "good" real root of the equation
	 Ru = (1 + k1 * Rd^2) * Rd,
       which is a polynomial in rd of the form x^3 + a x + b = 0. */
    a = 1.0 / c->k1;
    b = -a * ru;
    q = a / 3.0;
    r = -b / 2.0;
    u = q * q * q + r * r;

    if (u >= 0) {		/* in most cases, u >= 0 */
	u = sqrt (u);
	rd = CubeRoot (r + u) + CubeRoot (r - u);
    } else {			/* u < 0 => 3 real roots */
	u = sqrt (-u);
	s = CubeRoot (hypot (r, u));
	t = atan2 (u, r) / 3.0;
	cost = cos (t);
	sint = sin (t);
	rd = s * (Sqrt3 * sint - cost);
	/* (Other roots: 2 s cos(t) and -s cos(t) - Sqrt3 s sin(t).) */
    }

    t = rd / ru;

    psd[0] = psu[0] * t;
    psd[1] = psu[1] * t;

    #undef Sqrt3
    #undef CubeRoot
}


/*
 *  VImageToUndistortedSensorCoords
 *
 *  Convert distorted image coordinates to undistorted sensor coordinates.
 */

void VImageToUndistortedSensorCoords (VCamera c,
				      VDouble pid[2], VDouble psu[2])
{
    VDouble psd[2];

    psd[0] = c->pixel_size[0] * (pid[0] - c->c[0]);
    psd[1] = - c->pixel_size[1] * (pid[1] - c->c[1]);
    if (c->sx != 1.0)
	psd[0] /= c->sx;
    VDistortedToUndistortedSensorCoords (c, psd, psu);
}
