/*
 *  $Id: Error.c,v 1.1 1994/06/10 01:39:20 pope Exp pope $
 *
 *  Routines for computing various camera calibration error measures.
 */

/*
 *  Copyright 1994 University of British Columbia
 *
 *  Permission to use, copy, modify, distribute, and sell this software and its
 *  documentation for any purpose is hereby granted without fee, provided that
 *  the above copyright notice appears in all copies and that both that
 *  copyright notice and this permission notice appear in supporting
 *  documentation. UBC makes no representations about the suitability of this
 *  software for any purpose. It is provided "as is" without express or
 *  implied warranty.
 *
 *  Adapted by Arthur Pope, UBC Laboratory for Computational Intelligence,
 *  from implementation by Reg Willson, Carnegie Mellon University.
 */

/* From the Vista library: */
#include "vista/Vlib.h"
#include "vista/os.h"
#include "vista/VCamera.h"

/* From the standard C library: */
#include <math.h>

/* File identification string: */
VRcsId ("$Id: Error.c,v 1.1 1994/06/10 01:39:20 pope Exp pope $");

/* Later in this file: */
static void CalcStats (int, VDouble [], VDouble *, VDouble *, VDouble *);


/*
 *  VNormalizedCameraCalibrationError
 *
 *  Computes a measure of camera calibration accuracy, the `normalized
 *  calibration error', described by J Weng, P Cohen, and M Herniou in
 *  `Camera calibration with distortion models and accuracy evaluation',
 *  IEEE T-PAMI 14(10), Oct 1992, pp 965-980.
 *
 *  The 3d coordinates of each calibration data point are estimated by
 *  back-projecting its measured image location through the model to the
 *  plane formed by the original z world component.
 */

void VNormalizedCameraCalibrationError (VCamera c, int npoints,
					VDouble pw[][3], VDouble pid[][2],
					VDouble *mean, VDouble *stddev,
					VDouble *max)
{
    int i;
    VDouble *error, pc[3], psu[2], dx, dy, fu, fv;

    error = VMalloc (npoints * sizeof (VDouble));
    for (i = 0; i < npoints; i++) {

	/* Calculate the location of the data point in camera coordinates: */
	VWorldToCameraCoords (c, pw[i], pc);

	/* Convert the 2d image coordinates of the data point to undistorted
	   sensor plane coordinates: */
	VImageToUndistortedSensorCoords (c, pid[i], psu);

	/* Measure the discrepancy in the back-projected x and y
	   coordinates: */
	dx = pc[0] - pc[2] * psu[0] / c->f;
	dy = pc[1] - pc[2] * psu[1] / c->f;

	fu = c->sx * c->f / c->pixel_size[0];
	fv = c->f / c->pixel_size[1];

	error[i] = sqrt ((dx * dx + dy * dy) /
			 (pc[2] * pc[2] *
			  (1.0 / (fu * fu) + 1.0 / (fv * fv)) / 12.0));
    }
    CalcStats (npoints, error, mean, stddev, max);
    VFree (error);
}


/*
 *  VObjectSpaceCameraCalibrationError
 *
 *  Reports statistics for the magnitude of the distance of closest
 *  approach (i.e. 3d error) between points in object space and the line of
 *  sight formed by back-projecting the measured 2d coordinates out through
 *  the camera model. From Reg Willson, Reg_Willson@ius4.ius.cs.cmu.edu.
 */

void VObjectSpaceCameraCalibrationError (VCamera c, int npoints,
					 VDouble pw[][3], VDouble pid[][2],
					 VDouble *mean, VDouble *stddev,
					 VDouble *max)
{
    int i;
    VDouble *error, pc[3], psu[2], t, dx, dy, dz;

    error = VMalloc (npoints * sizeof (VDouble));
    for (i = 0; i < npoints; i++) {

	/* Calculate the location of the data point in camera coordinates: */
	VWorldToCameraCoords (c, pw[i], pc);

	/* Convert the 2d image coordinates of the data point to undistorted
	   sensor plane coordinates: */
	VImageToUndistortedSensorCoords (c, pid[i], psu);

	/* Find the magnitude of the distance (error) of closest approach
	   between the undistorted line of sight and the point in 3 space: */
	t = (pc[0] * psu[0] + pc[1] * psu[1] + pc[2] * c->f) /
	    (psu[0] * psu[0] + psu[1] * psu[1] + c->f * c->f);
	dx = pc[0] - psu[0] * t;
	dy = pc[1] - psu[1] * t;
	dz = pc[2] - c->f * t;
	error[i] = sqrt (dx * dx + dy * dy + dz * dz);
    }
    CalcStats (npoints, error, mean, stddev, max);
    VFree (error);
}


/*
 *  CalcStats
 *
 *  Local routine for finding mean, standard deviation, and max of a list
 *  of numbers.
 */

static void CalcStats (int n, VDouble data[],
		       VDouble *meanp, VDouble *stddevp, VDouble *maxp)
{
    VDouble max, sum, mean, d;
    int i;

    if (n == 0)
	return;
    for (i = 0, max = data[0], sum = 0.0; i < n; i++) {
	if (data[i] > max)
	    max = data[i];
	sum += data[i];
    }
    mean = sum / n;
    if (meanp)
	*meanp = mean;
    if (stddevp) {
	for (i = 0, sum = 0.0; i < n; i++) {
	    d = data[i] - mean;
	    sum += d * d;
	}
	*stddevp = n > 1 ? sqrt (sum / (n - 1)) : 0.0;
    }
    if (maxp)
	*maxp = max;
}
