/*
 *  $Id: FuncMin.c,v 1.1 1994/06/10 01:39:21 pope Exp pope $
 *
 *  The routines in this file search for a minimum of a multidimensional
 *  function. They're all based on ideas and routines in Press, et al.,
 *  `Numerical Recipes in C', Cambridge University Press, 1988.
 */

/*
 *  Copyright 1994 University of British Columbia
 *
 *  Permission to use, copy, modify, distribute, and sell this software and its
 *  documentation for any purpose is hereby granted without fee, provided that
 *  the above copyright notice appears in all copies and that both that
 *  copyright notice and this permission notice appear in supporting
 *  documentation. UBC makes no representations about the suitability of this
 *  software for any purpose. It is provided "as is" without express or
 *  implied warranty.
 *
 *  Adapted by Arthur Pope, UBC Laboratory for Computational Intelligence
 */

/* From the local directory: */
#include "FuncMin.h"

/* From the Vista library: */
#include <vista/mu.h>
#include <vista/os.h>

/* From the standard library */
#include <math.h>
#include <setjmp.h>

/* File identification string: */
VRcsId ("$Id: FuncMin.c,v 1.1 1994/06/10 01:39:21 pope Exp pope $");

/* Type declaration for a one-dimensional function minimized: */
typedef VDouble FMFunction1 (VPointer, VDouble);

/* Information about an n-dimensional function minimization problem: */
typedef struct {
    FMFunctionN *f;			/* function to minimize */
    VPointer data;			/* passed to f */
    VDouble *x;				/* current guess */
    VDouble xi[fmMaxNVars];		/* current search direction */
    VDouble fx;				/* f(x) for current guess */
    jmp_buf error_env;			/* beam up on error */
} FMState;

/* Later in this file: */
static void LineMinimize (FMState *);
static VDouble LineMinimizeHelper (FMState *, VDouble);
static void BracketMinimum (FMFunction1 *, VPointer,
			    VDouble *, VDouble *, VDouble *);
static VBoolean BrentMin (FMFunction1 *, VPointer, VDouble, VDouble, VDouble,
			  VDouble, VDouble *, VDouble *);


/*
 *  V_PowellMin
 *
 *  Minimizes a function of several variables using Powell's method.
 *
 *  Arguments:
 *	f - the function to be minimized
 *	data - just gets conveyed to f
 *	dims - a bit mask indicating which dimensions to minimize over
 *	x - specifies an initial guess and returns the final result
 *	max_iterations, tolerance - stopping conditions.
 *
 *  The minimization is carried out over all dimensions x[i] such that
 *  (dims & (1 << i)) != 0.
 *
 *  Successful termination occurs when f fails to decrease by more than
 *  tolerance * f(x) during an iteration.
 */

VBoolean V_PowellMin (FMFunctionN *f, VPointer data, int dims, VDouble x[],
		      int max_iterations, VDouble tolerance)
{
    int i, ibig, j, iter;
    VDouble t1, t2, fxtt, fx, del;
    VDouble xt[fmMaxNVars], xtt[fmMaxNVars], xi[fmMaxNVars][fmMaxNVars];
    FMState s;

    if (dims >> (fmMaxNVars + 1))
	VError ("PowellMin: too many parameters");

    /* Package up the problem: */
    s.f = f;
    s.data = data;
    s.x = x;
    s.fx = f (data, x);
    if (setjmp (s.error_env))
	return FALSE;
    tolerance *= 0.5;

    /* Initialize a set of minimization directions: */
    memset (xi, 0, sizeof (xi));
    for (i = 0; i < fmMaxNVars; i++)
	xi[i][i] = 1.0;

    /* Save the initial guess: */
    memcpy (xt, x, sizeof (xt));

    /* For each iteration... */
    for (iter = 0; iter < max_iterations; iter++) {
	fx = s.fx;
	ibig = 0;
	del = 0.0;

	/* Loop over all directions: */
	for (i = 0; i < fmMaxNVars; i++) {
	    if (! (dims & (1 << i)))
		continue;

	    /* Copy the direction: */
	    memcpy (s.xi, xi[i], sizeof (s.xi));
	    fxtt = s.fx;

	    /* Minimize along it: */
	    LineMinimize (& s);

	    /* Record it if it is the largest decrease so far: */
	    if (fabs (fxtt - s.fx) > del) {
		del = fabs (fxtt - s.fx);
		ibig = i;
	    }
	}

	/* Check termination condition: */
	if (fabs (fx - s.fx) <= tolerance * (fabs (fx) + fabs (s.fx)))
	    return TRUE;

	/* Construct the extrapolated point and the average direction moved: */
	for (j = 0; j < fmMaxNVars; j++) {
	    xtt[j] = 2.0 * x[j] - xt[j];
	    s.xi[j] = x[j] - xt[j];
	}
	memcpy (xt, x, sizeof (xt));

	/* Evaluate f at the extrapolated point: */
	fxtt = f (data, xtt);
	if (fxtt >= fx)
	    continue;

	t1 = fx - s.fx - del;
	t2 = fx - fxtt;
	if (2.0 * (fx - 2.0 * s.fx + fxtt) * t1 * t1 - del * t2 * t2 >= 0.0)
	    continue;

	/* Move to the minimum of the new direction: */
	LineMinimize (& s);

	/* Save the new direction: */
	memcpy (xi[ibig], s.xi, sizeof (s.xi));
    }
    VWarning ("Powell: Too many iterations in Powell()");
    return FALSE;
}


/*
 *  LineMinimize
 *
 *  Minimizes the function s->fnc along the line passing through s->p
 *  in the direction s->xi. Uses LineMinimizeHelper, below.
 */

static void LineMinimize (FMState *s)
{
    int j;
    VDouble a = 0.0, b = 1.0, c = 2.0, xmin;

    BracketMinimum ((FMFunction1 *) LineMinimizeHelper, s, & a, & b, & c);
    if (! BrentMin ((FMFunction1 *) LineMinimizeHelper, s, a, b, c, 2.0e-4,
		    & xmin, & s->fx))
	longjmp (s->error_env, 1);
    for (j = 0; j < fmMaxNVars; j++) {
	s->xi[j] *= xmin;
	s->x[j] += s->xi[j];
    }
}

static VDouble LineMinimizeHelper (FMState *s, VDouble x)
{
    int j;
    VDouble xt[fmMaxNVars];

    for (j = 0; j < fmMaxNVars; j++)
	xt[j] = s->x[j] + x * s->xi[j];
    return (s->f) (s->data, xt);
}


/*
 *  BracketMinimum
 *
 *  Given a one-dimensional function, f, and initial points, a and b,
 *  BracketMinimum returns new points a, b, and c, which bracket a minimum
 *  of f.
 */


static void BracketMinimum (FMFunction1 *f, VPointer data,
			    VDouble *ap, VDouble *bp, VDouble *cp)
{
    VDouble a = *ap, b = *bp, c;
    VDouble fa = f (data, a), fb = f (data, b), fc;
    VDouble r, q, u, fu, ulim;
    #define Gold 1.618034	/* default ratio of successive intervals */

    /* Order a and b so that we can go downhill from a to b: */
    if (fa < fb) {
	r = a; a = b; b = r;
	r = fa; fa = fb; fb = r;
    }

    /* Choose a first guess for c: */
    c = b + Gold * (b - a);
    fc = f (data, c);

    /* Until f(b) is less than both f(a) and f(b): */
    while (fb > fc) {

	/* Compute u by parabolic extrapolation from a, b, and c: */
	r = (b - a) * (fb - fc);
	q = (b - c) * (fb - fa);
	u = q - r;
	if (fabs (u) < 1e-20)
	    u = (u >= 0.0) ? 1e-20 : -1e-20;
	u = b - ((b - c) * q - (b - a) * r) / (2.0 * u);

	/* If u is between b and c... */
	if ((b > u && u > c) || (b < u && u < c)) {
	    fu = f (data, u);

	    /* Check whether it reveals a minimum: */
	    if (fu < fc) {
		a = b;	    /* u is a minimum between b and c */
		b = u;
		break;
	    }
	    if (fu > fb) {
		c = u;	    /* b is a minimum between a and u */
		break;
	    }

	    /* If not, use the default magnification instead of the
	       parabolic one: */
	    u = c + Gold * (c - b);
	    fu = f (data, u);

	} else {

	    /* Set a (somewhat arbitrary ) limit on how far we'll go
	       in choosing a new point: */
	    ulim = b + 100.0 * (c - b);

	    /* If u is between c and ulim... */
	    if ((c > u && u > ulim) || (c < u && u < ulim)) {
		fu = f (data, u);
		if (fu < fc) {
		    b = c; c = u;
		    fb = fc; fc = fu;
		    u = c + Gold * (c - b);
		    fu = f (data, u);
		}
	    }

	    /* Otherwise, if ulim is between u and c, use it as the new
	       point: */
	    else if ((u >= ulim && ulim >= c) || (u <= ulim && ulim <= c)) {
		u = ulim;
		fu = f (data, u);
	    }

	    /* Otherwise use the default magnification: */
	    else {
		u = c + Gold * (c - b);
		fu = f (data, u);
	    }
	}

	/* Eliminate the oldest point and continue: */
	a = b; b = c; c = u;
	fa = fb; fb = fc; fc = fu;
    }
    *ap = a;
    *bp = b;
    *cp = c;

    #undef Gold
}


/*
 *  BrentMin
 *
 *  Finds a minimum of a function, in 1-D, using Brent's method. Points a and
 *  b bracketing the minimum must be supplied, and also c between a and c
 *  such that f(c) < f(a) and f(c) < f(b). If successful, returns TRUE, along
 *  with the minimum parameter and function values.
 */

static VBoolean BrentMin (FMFunction1 *f, VPointer data,
			  VDouble a, VDouble x, VDouble b,
			  VDouble tol, VDouble *xmin, VDouble *fxmin)
{
    int iter;
    VDouble d = 0.0, e = 0.0, etemp, u, v, w, fu, fv, fw, fx;
    VDouble p, q, r, tol1, tol2, xm;

    #define CGold 0.3819660

    /* Order a and b so that a < x < b: */
    if (a > b) {
	u = a; a = b; b = u;
    }
    w = v = x;
    fw = fv = fx = f (data, x);

    for (iter = 0; iter < 100; iter++) {
	xm = 0.5 * (a + b);
	tol1 = tol * fabs (x) + 1.0e-10;
	tol2 = 2.0 * tol1;
	if (fabs (x - xm) <= (tol2 - 0.5 * (b - a))) {
	    *xmin = x;
	    *fxmin = fx;
	    return TRUE;
	}
	if (fabs (e) > tol1) {
	    r = (x - w) * (fx - fv);
	    q = (x - v) * (fx - fw);
	    p = (x - v) * q - (x - w) * r;
	    q = 2.0 * (q - r);
	    if (q > 0.0)
		p = -p;
	    q = fabs (q);
	    etemp = e;
	    e = d;
	    if (fabs (p) >= fabs (0.5 * q * etemp) ||
		p <= q * (a - x) ||
		p >= q * (b - x)) {
		e = (x >= xm ? a : b) - x;
		d = CGold * e;
	    } else {
		d = p / q;
		u = x + d;
		if (u - a < tol2 || b - u < tol2)
		    d = (xm >= x) ? tol1 : -tol1;
	    }
	} else {
	    e = (x >= xm ? a : b) - x;
	    d = CGold * e;
	}
	u = ((fabs (d) >= tol1) ? d : ((d >= 0.0) ? tol1 : -tol1)) + x;
	fu = f (data, u);
	if (fu <= fx) {
	    if (u >= x)
		a = x;
	    else b = x;
	    v = w; w = x; x = u;
	    fv = fw; fw = fx; fx = fu;
	} else {
	    if (u < x)
		a = u;
	    else b = u;
	    if (fu <= fw || w == x) {
		v = w; w = u;
		fv = fw; fw = fu;
	    } else if (fu <= fv || v == x || v == w) {
		v = u; fv = fu;
	    }
	}
    }
    VWarning ("BrentMin: Too many iterations");
    return FALSE;

    #undef CGold
}
