/*
 *  $Id: LinSysMin.c,v 1.1 1994/06/10 01:39:25 pope Exp pope $
 *
 *  This file implements a recursive least-squares estimator.
 *
 *  The estimator is implemented using the square-root information filter
 *  algorith as described in Gerald J Bierman's book "Factorization Methods 
 *  for Discrete Sequential Estimation", Academic Press, 1977.
 */

/*
 *  Copyright 1994 University of British Columbia
 *
 *  Permission to use, copy, modify, distribute, and sell this software and its
 *  documentation for any purpose is hereby granted without fee, provided that
 *  the above copyright notice appears in all copies and that both that
 *  copyright notice and this permission notice appear in supporting
 *  documentation. UBC makes no representations about the suitability of this
 *  software for any purpose. It is provided "as is" without express or
 *  implied warranty.
 *
 *  Author: Arthur Pope, UBC Laboratory for Computational Intelligence
 */

/* From the local directory: */
#include "LinSysMin.h"

/* From the Vista library: */
#include <vista/Vlib.h>
#include <vista/os.h>

/* From the standard C library: */
#include <math.h>

/* File identification string: */
VRcsId ("$Id: LinSysMin.c,v 1.1 1994/06/10 01:39:25 pope Exp pope $");

/* Later in this file: */
static void ProcessBatch (Estimator *);


/*
 *  V_InitEstimator
 *
 *  Initialize a least squares estimator.
 */

void V_InitEstimator (Estimator *est, int nvars)
{
    if (nvars > lsmMaxNVars)
	VError ("InitEstimator: nvars too large");
    memset (est, 0, sizeof (Estimator));    /* for debugging clarity */
    est->nvars = nvars;
    est->nrows = 0;
}


/*
 *  V_ExtendEstimator
 *
 *  Incorporate one scalar measurement.
 */

void V_ExtendEstimator (Estimator *est, VDouble a[], VDouble b)
{
    /* Process batched measurements if necessary: */
    if (est->nrows == lsmMaxBatch)
	ProcessBatch (est);

    /* Add the new measurement to the batch: */
    memcpy (est->s[est->nrows], a, est->nvars * sizeof (VDouble));
    est->s[est->nrows++][est->nvars] = b;
}


/*
 *  UpdateEstimator
 *
 *  Update estimated variables using all measurements supplied so far.
 *  Returns FALSE if the estimate has diverged.
 */

VBoolean V_UpdateEstimator (Estimator *est)
{
    int j, k;
    VDouble s;

    /* Incorporate any new measurements: */
    if (est->nrows > est->nvars)
	ProcessBatch (est);

    /* Solve for the state vector, x, by back substitution: */
    for (j = est->nvars - 1; j >= 0; j--) {
	for (k = j + 1, s = 0.0; k < est->nvars; k++)
	    s += est->s[j][k] * est->x[k];
	est->x[j] = (est->s[j][est->nvars] - s) / est->s[j][j];
    }

    return finite (est->e);
}


/*
 *  ProcessBatch
 *
 *  Process batched measurements to update state vector estimate.
 */

static void ProcessBatch (Estimator *est)
{
    int i, j, k;
    VDouble s, a, v[lsmMaxBatch];

    /* Transform s into an upper triangular matrix by means of Householder
       transformations: */
    for (j = 0; j < est->nvars; j++) {
	for (i = j, s = 0.0; i < est->nrows; i++) {
	    v[i] = est->s[i][j];
	    est->s[i][j] = 0.0;
	    s += v[i] * v[i];
	}
	if (s <= 0.0)
	    continue;
	s = sqrt (s);
	if (v[j] > 0)
	    s = - s;
	est->s[j][j] = s;
	v[j] -= s;
	s = 1.0 / (s * v[j]);
	for (k = j + 1; k <= est->nvars; k++) {
	    for (i = j, a = 0.0; i < est->nrows; i++)
		a += est->s[i][k] * v[i];
	    a *= s;
	    for (i = j; i < est->nrows; i++)
		est->s[i][k] += a * v[i];
	}
    }

    /* Pull out the residual: */
    for (j = est->nvars; j < est->nrows; j++) {
	s = est->s[j][est->nvars];
	est->e += s * s;
    }

    /* Reset the array of measurements: */
    est->nrows = est->nvars;
}
