/*
 *  $Id: VCamera.h,v 1.2 1995/05/01 23:29:59 pope Exp pope $
 *
 *  Definitions associated with cameras and camera calibration.
 */

#ifndef V_VCamera_h
#define V_VCamera_h 1

/*
 *  Copyright 1994 University of British Columbia
 *
 *  Permission to use, copy, modify, distribute, and sell this software and its
 *  documentation for any purpose is hereby granted without fee, provided that
 *  the above copyright notice appears in all copies and that both that
 *  copyright notice and this permission notice appear in supporting
 *  documentation. UBC makes no representations about the suitability of this
 *  software for any purpose. It is provided "as is" without express or
 *  implied warranty.
 *
 *  Author: Arthur Pope, UBC Laboratory for Computational Intelligence
 */

/* From the Vista library: */
#include "vista/Vlib.h"

/* From the standard C library: */
#include <stdio.h>

/* For portability: */
#include <X11/Xfuncproto.h>

#ifdef __cplusplus
extern "C" {
#endif


/*
 *  Conventions.
 *
 *  The camera's X axis runs along increasing column coordinates in the
 *  image/frame. The Y axis runs along decreasing row coordinates. The Z
 *  axis points toward the scene.
 *
 *  A 2D and 3D coordinates are represented as 2- or 3-element vectors
 *  carrying X and Y, or X, Y and Z coordinates, in that order.
 *
 *  Distance are expressed in millimeters.
 */


/*
 *  Data structures.
 */

/* Description of a camera/frame grabber combination: */
typedef struct V_CameraRec {

    /* Hardware specifications: */
    VLong nsels[2];			/* no. of camera sensor elements */
    VLong npixels[2];			/* no. of frame grabber pixels */
    VDouble sel_size[2];		/* size of a sensor element (mm) */
    VDouble pixel_size[2];		/* effective size of pixel (mm) */

    /* Which optional parameters are defined: */
    VBoolean extrinsics_defd;		/* rotation and translation defined */
    VBoolean focal_length_defd;		/* focal length defined */

    /* Extrinsic parameters: */
    VDouble angles[3];			/* world->camera Euler angles (rad) */
    VDouble r[3][3];			/* world->camera rotation matrix */
    VDouble t[3];			/* world->camera translation (mm) */

    /* Intrinsic non-distortion parameters: */
    VDouble c[2];			/* camera coord sys Z axis intercept */
    VDouble f;				/* focal length (mm) */
    VDouble sx;				/* X scale factor */

    /* Intrinsic distortion parameters: */
    VDouble k1;				/* radial distortion coefficient */

    VAttrList attributes;		/* list of other attributes */
} VCameraRec, *VCamera;

/* Representation code for the VCamera type: */
extern VRepnKind VCameraRepn;


/*
 *  Macros for accessing data structures.
 */

#define VCameraAttrList(c)		((c)->attributes)
#define VCameraRotation(c)		((c)->angles)
#define VCameraExtrinsicsDefd(c)	((c)->extrinsics_defd)
#define VCameraFocalLength(c)		((c)->f)
#define VCameraFocalLengthDefd(c)	((c)->focal_length_defd)
#define VCameraImageCenter(x)		((x)->c)
#define VCameraNPixels(c)		((c)->npixels)
#define VCameraNSels(c)			((c)->nsels)
#define VCameraPixelSize(c)		((c)->pixel_size)
#define VCameraRadialDistortion(c)	((c)->radial_distortion)
#define VCameraRotMatrix(c)		((c)->r)
#define VCameraScaleX(c)		((c)->sx)
#define VCameraSelSize(c)		((c)->sel_size)
#define VCameraTranslation(c)		((c)->t)


/*
 *  Attributes used to describe cameras and camera calibrations.
 */

#define VCameraNameAttr 	"camera_name"
#define VImageCenterXAttr	"image_center_x"
#define VImageCenterYAttr	"image_center_y"
#define VFocalLengthAttr	"focal_length"
#define VFrameGrabberNameAttr 	"frame_grabber_name"
#define VRadialDistortionAttr	"radial_distortion"
#define VNPixelsXAttr		"npixels_x"
#define VNPixelsYAttr		"npixels_y"
#define VNSelsXAttr		"nsels_x"
#define VNSelsYAttr		"nsels_y"
#define VPixelSizeXAttr		"pixel_size_x"
#define VPixelSizeYAttr		"pixel_size_y"
#define VRotationXAttr		"rotation_x"
#define VRotationYAttr		"rotation_y"
#define VRotationZAttr		"rotation_z"
#define VScaleXAttr		"scale_x"
#define VSelSizeXAttr		"sel_size_x"
#define VSelSizeYAttr		"sel_size_y"
#define VTranslationXAttr	"translation_x"
#define VTranslationYAttr	"translation_y"
#define VTranslationZAttr	"translation_z"


/*
 *  Declarations of library routines.
 */

/* From Calibrate.c: */

extern VCamera VCalibrateCamera (
#if NeedFunctionPrototypes
    VCamera 		/* src */,
    VBoolean 		/* extrinsics_only */,
    VBoolean 		/* coplanar */,
    int 		/* npoints */,
    VDouble [][3]	/* pw */,
    VDouble [][2]	/* pid */,
    int 		/* max_iterations */,
    double 		/* tolerance */
#endif
);

/* From Camera.c: */

extern void VRegisterCameraType (
#if NeedFunctionPrototypes
    void
#endif
);

extern VCamera VCreateCamera (
#if NeedFunctionPrototypes
    void
#endif
);

extern VCamera VCopyCamera (
#if NeedFunctionPrototypes
    VCamera		/* src */
#endif
);

extern void VDestroyCamera (
#if NeedFunctionPrototypes
    VCamera		/* c */
#endif
);

extern void VSetCameraRotation (
#if NeedFunctionPrototypes
    VCamera		/* c */,
    VDouble [3]		/* angles */,
    VDouble [3][3]	/* matrix */
#endif
);

extern void VComputeCameraRotationAngles (
#if NeedFunctionPrototypes
    VCamera		/* c */
#endif
);

extern void VComputeCameraRotationMatrix (
#if NeedFunctionPrototypes
    VCamera		/* c */
#endif
);

/* From Coord.c: */

extern VBoolean VWorldToCameraCoords (
#if NeedFunctionPrototypes
    VCamera		/* c */,
    VDouble [3]		/* pw */,
    VDouble [3]		/* pc */
#endif
);

extern VBoolean VCameraToWorldCoords (
#if NeedFunctionPrototypes
    VCamera		/* c */,
    VDouble [3]		/* pc */,
    VDouble [3]		/* pw */
#endif
);

extern VBoolean VWorldToImageCoords (
#if NeedFunctionPrototypes
    VCamera		/* c */,
    VDouble [3]		/* pw */,
    VDouble [2]		/* pid */
#endif
);

extern VBoolean VImageToWorldCoords (
#if NeedFunctionPrototypes
    VCamera		/* c */,
    VDouble [2]		/* pid */,
    VDouble [3]		/* pw */,
    VDouble		/* pwz */
#endif
);

extern void VDistortedToUndistortedImageCoords (
#if NeedFunctionPrototypes
    VCamera		/* c */,
    VDouble [2]		/* pid */,
    VDouble [2]		/* piu */
#endif
);

extern void VUndistortedToDistortedImageCoords (
#if NeedFunctionPrototypes
    VCamera		/* c */,
    VDouble [2]		/* piu */,
    VDouble [2]		/* pid */
#endif
);

extern void VDistortedToUndistortedSensorCoords (
#if NeedFunctionPrototypes
    VCamera		/* c */,
    VDouble [2]		/* psd */,
    VDouble [2]		/* psu */
#endif
);

extern void VUndistortedToDistortedSensorCoords (
#if NeedFunctionPrototypes
    VCamera		/* c */,
    VDouble [2]		/* psu */,
    VDouble [2]		/* psd */
#endif
);

extern void VImageToUndistortedSensorCoords (
#if NeedFunctionPrototypes
    VCamera		/* c */,
    VDouble [2]		/* pid */,
    VDouble [2]		/* psu */
#endif
);

/* In Error.c: */

extern void VNormalizedCameraCalibrationError (
#if NeedFunctionPrototypes
    VCamera		/* c */,
    int			/* npoints */,
    VDouble [][3]	/* pw */,
    VDouble [][2]	/* pid */,
    VDouble *		/* mean */,
    VDouble *		/* stddev */,
    VDouble *		/* max */
#endif
);

extern void VObjectSpaceCameraCalibrationError (
#if NeedFunctionPrototypes
    VCamera		/* c */,
    int			/* npoints */,
    VDouble [][3]	/* pw */,
    VDouble [][2]	/* pid */,
    VDouble *		/* mean */,
    VDouble *		/* stddev */,
    VDouble *		/* max */
#endif
);

#ifdef __cplusplus
}
#endif

#endif /* V_VCamera_h */
