/*
 *  $Id: vcalsyn.c,v 1.1 1994/06/10 01:39:37 pope Exp pope $
 *
 *  The vcalsyn program, which generates synthetic camera calibration
 *  data according to a specified camera model.
 */

/*
 *  Copyright 1994 University of British Columbia
 *
 *  Permission to use, copy, modify, distribute, and sell this software and its
 *  documentation for any purpose is hereby granted without fee, provided that
 *  the above copyright notice appears in all copies and that both that
 *  copyright notice and this permission notice appear in supporting
 *  documentation. UBC makes no representations about the suitability of this
 *  software for any purpose. It is provided "as is" without express or
 *  implied warranty.
 *
 *  Adapted by Arthur Pope, UBC Laboratory for Computational Intelligence,
 *  from implementation by Reg Willson, Carnegie Mellon University.
 */

/* From the Vista library: */
#include "vista/Vlib.h"
#include "vista/file.h"
#include "vista/mu.h"
#include "vista/option.h"
#include "vista/os.h"
#include "vista/VCamera.h"

/* From the standard C library: */
#include <math.h>

/* File identification string: */
VRcsId ("$Id: vcalsyn.c,v 1.1 1994/06/10 01:39:37 pope Exp pope $");

/* Later in this file: */
static void GenerateCalibrationData (VCamera, int *, VDouble (**)[3],
				     VDouble (**)[2]);
static VAttrList EncodeCalibrationData (int, VDouble (*)[3], VDouble (*)[2]);
static VDouble GaussianDeviate (void);


/*
 *  Program options.
 */

static VBoolean coplanar = FALSE;
static VShort count[3] = { 10, 10, 3 };
static VDouble origin[3] = { 10, 10, 0 };
static VDouble spacing[3] = { 20, 20, 20 };
static VDouble world_stddev = 0.04;
static VDouble image_stddev = 0.02;
static VStringConst object_name = "points";
static VBoolean report = TRUE;
static VLong seed;
static VOptionDescRec options[] = {
    { "coplanar", VBooleanRepn, 1, & coplanar, VOptionalOpt, NULL,
	  "Data points are coplanar" },
    { "count", VShortRepn, 3, & count, VOptionalOpt, NULL,
	  "Number of points in each dimension" },
    { "origin", VDoubleRepn, 3, & origin, VOptionalOpt, NULL,
	  "Origin of first point in world coordinates (mm)" },
    { "spacing", VDoubleRepn, 3, & spacing, VOptionalOpt, NULL,
	  "Spacing between points in world coordinates (mm)" },
    { "world_stddev", VDoubleRepn, 1, & world_stddev, VOptionalOpt, NULL,
	  "World measurenment noise standard deviation (mm)" },
    { "image_stddev", VDoubleRepn, 1, & image_stddev, VOptionalOpt, NULL,
	  "Image measurenment noise standard deviation (pixels)" },
    { "name", VStringRepn, 1, & object_name, VOptionalOpt, NULL,
	  "Name to be given data point collection" },
    { "report", VBooleanRepn, 1, & report, VOptionalOpt, NULL,
	  "Report error measures" },
    { "seed", VLongRepn, 1, & seed, VOptionalOpt, NULL,
	  "Seed for initializing random number generator" }
};


/*
 *  Program entry point.
 */

int main (int argc, char **argv)
{
    time_t now;
    FILE *in_file, *out_file;
    int ncameras, npoints;
    VCamera *cameras, c;
    VAttrList list;
    VDouble (*pw)[3], (*pid)[2];

    /* Register the VCamera type: */
    VRegisterCameraType ();

    /* Choose a random seed: */
    time (& now);
    VRandomSeed ((long) now);
    seed = VRandomLong ();

    /* Parse command line arguments: */
    VParseFilterCmd (VNumber (options), options, argc, argv,
		     & in_file, & out_file);

    /* Coplanar points lie on a Z=0 plane: */
    if (coplanar) {
	count[2] = 1;
	origin[2] = 0.0;
    }

    /* Read the camera model: */
    ncameras = VReadObjects (in_file, VCameraRepn,
			     & list, (VPointer **) & cameras);
    if (ncameras == 0)
	exit (EXIT_FAILURE);
    if (ncameras > 1)
	VWarning ("Ignoring all but first camera model");
    c = cameras[0];
    if (! VCameraExtrinsicsDefd (c))
	VError ("Camera model's extrinsic parameters must be defined");
    if (! VCameraFocalLengthDefd (c))
	VError ("Camera model's focal length must be defined");
    VDestroyAttrList (list);	/* don't care about other attributes */

    /* Initialize the random number generator: */
    VRandomSeed (seed);

    /* Generate a synthetic set of calibration data: */
    GenerateCalibrationData (c, & npoints, & pw, & pid);

    /* Write the calibration data to the output file: */
    list = VCreateAttrList ();
    VSetAttr (list, object_name, NULL, VAttrListRepn,
	      EncodeCalibrationData (npoints, pw, pid));
    if (! VWriteFile (out_file, list))
	exit (EXIT_FAILURE);

    /* Evaluate the camera model: */
    if (report) {
	VDouble mean, stddev, max;
	
	VNormalizedCameraCalibrationError (c, npoints, pw, pid,
					   & mean, & stddev, & max);
	fprintf (stderr, "%s: Normalized calibration error:\n"
		 "\tmean = %g  stddev = %g  max = %g\n",
		 argv[0], mean, stddev, max);
	
	VObjectSpaceCameraCalibrationError (c, npoints, pw, pid,
					    & mean, & stddev, & max);
	fprintf (stderr, "%s: Object space calibration error:\n"
		 "\tmean = %g  stddev = %g  max = %g\n",
		 argv[0], mean, stddev, max);
    }

    return EXIT_SUCCESS;
}


/*
 *  GenerateCalibrationData
 *
 *  Generate a set of calibration data points for a specified camera
 *  model.
 */

static void GenerateCalibrationData (VCamera c, int *npointsp,
				     VDouble (**pwp)[3], VDouble (**pidp)[2])
{
    int ix, iy, iz, i = 0;
    VDouble pwn[3];

    /* Allocate storage for the points: */
    *npointsp = count[0] * count[1] * count[2];
    *pwp = VMalloc (*npointsp * 3 * sizeof (VDouble));
    *pidp = VMalloc (*npointsp * 2 * sizeof (VDouble));

    for (iz = 0; iz < count[2]; iz++)
	for (ix = 0; ix < count[0]; ix++)
	    for (iy = 0; iy < count[1]; iy++, i++) {

		/* Generate a point in 3d world coordinates: */
		(*pwp)[i][0] = ix * spacing[0] + origin[0];
		(*pwp)[i][1] = iy * spacing[1] + origin[1];
		(*pwp)[i][2] = iz * spacing[2] + origin[2];

		/* Add Gaussian dimensional noise: */
		pwn[0] = (*pwp)[i][0] + GaussianDeviate () * world_stddev;
		pwn[1] = (*pwp)[i][1] + GaussianDeviate () * world_stddev;
		pwn[2] = (*pwp)[i][2] + GaussianDeviate () * world_stddev;

		/* Generate the corresponding 2d image coordinates: */
		VWorldToImageCoords (c, pwn, (*pidp)[i]);

		/* Add measurement noise: */
		(*pidp)[i][0] += GaussianDeviate () * image_stddev;
		(*pidp)[i][1] += GaussianDeviate () * image_stddev;
	    }
}


/*
 *  EncodeCalibrationData
 *
 *  Encode a set of calibration points as an attribute list for output
 *  to a Vista data file.
 */

static VAttrList EncodeCalibrationData (int npoints, VDouble (*pw)[3],
					VDouble (*pid)[2])
{
    VAttrList ptlist, sublist;
    int i;
    char buf[20];

    ptlist = VCreateAttrList ();

    for (i = 0; i < npoints; i++) {

	/* Construct an attribute list description of the points: */
	sublist = VCreateAttrList ();
	VAppendAttr (sublist, "w_x", NULL, VDoubleRepn, pw[i][0]);
	VAppendAttr (sublist, "w_y", NULL, VDoubleRepn, pw[i][1]);
	VAppendAttr (sublist, "w_z", NULL, VDoubleRepn,
		     coplanar ? 0.0 : pw[i][2]);
	VAppendAttr (sublist, "i_x", NULL, VDoubleRepn, pid[i][0]);
	VAppendAttr (sublist, "i_y", NULL, VDoubleRepn, pid[i][1]);

	/* Append it to the list of points: */
	sprintf (buf, "%d", i);
	VAppendAttr (ptlist, buf, NULL, VAttrListRepn, sublist);
    }
    return ptlist;
}


/*
 *  GaussianDeviate
 *
 *  Return a normally-distributed random number with mean 0, std dev 1.
 */

static VDouble GaussianDeviate (void)
{

    static VDouble u1, u2, s = 0.0, t = 0.0, v;
    static VBoolean phase = FALSE;

    /* Numbers drawn from N(0,1) are generated in pairs from
       pairs of numbers in U(0,1). */
    if (phase)
	v = s * sin (t);
    else {
	do {
	    u1 = VRandomDouble ();
	} while (u1 == 0.0);
	do {
	    u2 = VRandomDouble ();
	} while (u2 == 0.0);
	s = sqrt (log (u1) * -2.0);
	t = u2 * 2.0 * M_PI;
	v = s * cos (t);
    }
    phase = ! phase;
    return v;
}
