/*
 *  $Id: vcamcal.c,v 1.1 1994/06/10 01:39:40 pope Exp pope $
 *
 *  The vcamcal program, which estimates camera calibration parameters
 *  using a series of data points with known world and image coordinates.
 */

/*
 *  Copyright 1994 University of British Columbia
 *
 *  Permission to use, copy, modify, distribute, and sell this software and its
 *  documentation for any purpose is hereby granted without fee, provided that
 *  the above copyright notice appears in all copies and that both that
 *  copyright notice and this permission notice appear in supporting
 *  documentation. UBC makes no representations about the suitability of this
 *  software for any purpose. It is provided "as is" without express or
 *  implied warranty.
 *
 *  Adapted by Arthur Pope, UBC Laboratory for Computational Intelligence,
 *  from implementation by Reg Willson, Carnegie Mellon University.
 */

/* From the Vista library: */
#include "vista/Vlib.h"
#include "vista/file.h"
#include "vista/mu.h"
#include "vista/option.h"
#include "vista/os.h"
#include "vista/VCamera.h"

/* File identification string: */
VRcsId ("$Id: vcamcal.c,v 1.1 1994/06/10 01:39:40 pope Exp pope $");

/* Later in this file: */
static void ReadDataPoints (VStringConst, VStringConst, int *,
			    VDouble (**)[3], VDouble (**)[2]);

/*
 *  Program options.
 */

static VString data_filename, data_object_name = "points";
static VBoolean coplanar = FALSE;
enum { EstimateAll, EstimateExtrinsics, EstimateNone };
static VDictEntry estimate_dict[] = {
    { "all",		EstimateAll }, 
    { "extrinsics",	EstimateExtrinsics },
    { "none",		EstimateNone },
    { NULL }
};
static VShort estimate = EstimateAll;
static VBoolean report = TRUE;
static VLong max_iterations = 200;
static VDouble tolerance = 1e-10;
static VOptionDescRec options[] = {
    { "data", VStringRepn, 1, & data_filename, VRequiredOpt, NULL,
	  "File containing data points" },
    { "name", VStringRepn, 1, & data_object_name, VOptionalOpt, NULL,
	  "Name of object containing data points" },
    { "coplanar", VBooleanRepn, 1, & coplanar, VOptionalOpt, NULL,
	  "Data points are coplanar" },
    { "estimate", VShortRepn, 1, & estimate, VOptionalOpt, estimate_dict,
	  "Which parameters to estimate" },
    { "report", VBooleanRepn, 1, & report, VOptionalOpt, NULL,
	  "Report error measures" },
    { "iterations", VLongRepn, 1, & max_iterations, VOptionalOpt, NULL,
	  "Max. number of optimization iterations" },
    { "tolerance", VDoubleRepn, 1, & tolerance, VOptionalOpt, NULL,
	  "Desired accuracy of iterative optimization" }
};


/*
 *  Program entry point.
 */

int main (int argc, char **argv)
{
    FILE *in_file, *out_file;
    int npoints, ncameras = 0;
    VDouble (*pw)[3], (*pid)[2];
    VAttrList list;
    VAttrListPosn posn;
    VCamera src, dest;

    /* Register the VCamera type: */
    VRegisterCameraType ();

    /* Parse command line arguments: */
    VParseFilterCmd (VNumber (options), options, argc, argv,
		     & in_file, & out_file);

    /* Read the data points: */
    ReadDataPoints (data_filename, data_object_name, & npoints, & pw, & pid);

    /* Read source camera model: */
    if (! (list = VReadFile (in_file, NULL)))
	exit (EXIT_FAILURE);

    /* Calibrate and/or measure the source camera model: */
    for (VFirstAttr (list, & posn); VAttrExists (& posn); VNextAttr (& posn)) {
	if (VGetAttrRepn (& posn) != VCameraRepn)
	    continue;
	VGetAttrValue (& posn, NULL, VCameraRepn, & src);

	/* Calibrate the camera model: */
	if (estimate != EstimateNone) {
	    dest = VCalibrateCamera (src, estimate == EstimateExtrinsics,
				     coplanar, npoints, pw, pid,
				     max_iterations, tolerance);
	    if (! dest)
		exit (EXIT_FAILURE);
	    VSetAttrValue (& posn, NULL, VCameraRepn, dest);
	    VDestroyCamera (src);
	} else dest = src;

	/* Evaluate the camera model: */
	if (report) {
	    VDouble mean, stddev, max;

	    VNormalizedCameraCalibrationError (dest, npoints, pw, pid,
					       & mean, & stddev, & max);
	    fprintf (stderr, "%s: Normalized calibration error:\n"
		     "\tmean = %g  stddev = %g  max = %g\n",
		     argv[0], mean, stddev, max);

	    VObjectSpaceCameraCalibrationError (dest, npoints, pw, pid,
						& mean, & stddev, & max);
	    fprintf (stderr, "%s: Object space calibration error:\n"
		     "\tmean = %g  stddev = %g  max = %g\n",
		     argv[0], mean, stddev, max);
	}
	ncameras++;
    }

    /* Write the results to the output file: */
    if (! VWriteFile (out_file, list))
	exit (EXIT_FAILURE);
    fprintf (stderr, "%s: Processed %d camera model%s.\n",
	     argv[0], ncameras, ncameras == 1 ? "" : "s");
    return EXIT_SUCCESS;
}

static void ReadDataPoints (VStringConst filename, VStringConst object_name,
			    int *npointsp, VDouble (**pwp)[3],
			    VDouble (**pidp)[2])
{
    FILE *f;
    VAttrList list, ptlist, sublist;
    VAttrListPosn posn;
    int i;

    #define Get(name,var) \
        (VGetAttr (sublist, name, NULL, VDoubleRepn, & var) == VAttrFound)

    /* Open and read the contents of the data file: */
    f = VOpenInputFile (filename, TRUE);
    if (! (list = VReadFile (f, NULL)))
	exit (EXIT_FAILURE);
    fclose (f);

    /* Get the specified object: */
    if (VGetAttr (list, object_name, NULL, VAttrListRepn, & ptlist) !=
	VAttrFound)
	VError ("\"%s\" attribute (data point set) is missing or incorrect",
		object_name);

    /* Count the number of data points: */
    for (VFirstAttr (ptlist, & posn), i = 0; VAttrExists (& posn);
	 VNextAttr (& posn), i++)
	if (VGetAttrRepn (& posn) != VAttrListRepn)
	    VError ("Error at data point %d", i);
    *npointsp = i;

    /* Allocate storage for the data points: */
    *pwp = VMalloc (*npointsp * 3 * sizeof (VDouble));
    *pidp = VMalloc (*npointsp * 2 * sizeof (VDouble));

    /* Extract data point coordinates from the attribute list: */
    for (VFirstAttr (ptlist, & posn), i = 0; VAttrExists (& posn);
	 VNextAttr (& posn), i++) {
	VGetAttrValue (& posn, NULL, VAttrListRepn, & sublist);
	if (! Get ("w_x", (*pwp)[i][0]) || ! Get ("w_y", (*pwp)[i][1]) ||
	    ! Get ("w_z", (*pwp)[i][2]) || ! Get ("i_x", (*pidp)[i][0]) ||
	    ! Get ("i_y", (*pidp)[i][1]))
	    VError ("Error at data point %d", i);
    }	

    VDestroyAttrList (list);

    #undef Get
}
