/*
 *  $Id: vxadjust.c,v 1.5 1994/06/15 23:14:49 pope Exp pope $
 *
 *  This file contains vxadjust, a program to adjust the contrast
 *  and brightness of an image interactively.
 */

/*
 *  Copyright 1993, 1994 University of British Columbia
 *
 *  Permission to use, copy, modify, distribute, and sell this software and its
 *  documentation for any purpose is hereby granted without fee, provided that
 *  the above copyright notice appears in all copies and that both that
 *  copyright notice and this permission notice appear in supporting
 *  documentation. UBC makes no representations about the suitability of this
 *  software for any purpose. It is provided "as is" without express or
 *  implied warranty.
 *
 *  Author: Daniel Ko, UBC Laboratory for Computational Intelligence
 */

/* From the Vista library: */
#include "vista/Vlib.h"
#include "vista/mu.h"
#include "vista/option.h"
#include "vista/os.h"
#include "vista/VImage.h"
#include "vista/VList.h"
#include "vista/VX.h"

/* From the standard C library: */
#include <math.h>

/* From X11R5 and the Motif widget set: */
#if defined(XtSpecificationRelease) && XtSpecificationRelease >= 5
#include <X11/Xmu/Editres.h>
#endif
#include <Xm/ArrowBG.h>
#include <Xm/BulletinB.h>
#include <Xm/Form.h>
#include <Xm/PushBG.h>
#include <Xm/Scale.h>

#include "vista/VImageView.h"

/* File identification string: */
VRcsId ("$Id: vxadjust.c,v 1.5 1994/06/15 23:14:49 pope Exp pope $");

/* Input and output filenames: */
VString in_filename;
VString out_filename;
FILE *out_file;

/* All attributes read from the input file: */
VAttrList attributes;

/* The list of image views assembled from the input images: */
typedef struct {
    VImage orig;		/* un-adjusted image */
    int image_idx;		/* it's position within input file */
    int band;			/* index of band within image */
    VBoolean color;		/* whether image is color */
    VImage adjusted;		/* adjusted image */
    double contrast;		/* contrast of each band */
    double brightness;		/* brightness of each band */
} View;
VList views;
int nimages;			/* number of distinct VImages */

/* Scale widgets for adjusting brightness and contrast: */
Widget contrast_widget;
Widget brightness_widget;

/* Later in this file: */
void PopupControlPanel (Widget, XtPointer, XtPointer);
void ReadFile (void);

void ShowView (void);
void NextView (Widget, XtPointer, XtPointer);
void PrevView (Widget, XtPointer, XtPointer);
void AdjustView (Widget, XtPointer, XtPointer);
void ResetView (Widget, XtPointer, XtPointer);

void ReportPixel (VXInputData, VPointer);

void Quit (VPointer);

void HelpOnMouseUsage (VPointer);
void HelpOnMenus (VPointer);
void HelpOnCP (VPointer);
void HelpOnVersion (VPointer);


/* 
 * Entry point
 */

int main(int argc, char **argv)
{
    static VBoolean in_found, out_found, signed_view = FALSE;
    static VStringConst resources[] = {

	/* Main window resources: */
	"VXAdjust.messageAreaNLines: 2",

	/* Control panel resources: */
	"VXAdjust*form.dialogTitle: Adjustment",
	"VXAdjust*form.horizontalSpacing: 5",
	"VXAdjust*form.verticalSpacing: 5",
	"VXAdjust*XmScale.decimalPoints: 1",
	"VXAdjust*XmScale.maximum: 100",
	"VXAdjust*XmScale.minimum: -100",
	"VXAdjust*brightnessScale.titleString: Brightness",
	"VXAdjust*contrastScale.titleString: Contrast",
	"VXAdjust*reset.labelString: Reset",
	"VXAdjust*reset.topOffset: 20",
	"VXAdjust*reset.height: 30",
	"VXAdjust*reset.width: 100",
	"VXAdjust*prev.width: 100",
	"VXAdjust*next.width: 100",

	/* For those who haven't learned to use a .motifbind file, this will
	   at least get the arrow keys: */
	"VView*defaultVirtualBindings: \n"
	    "osfLeft  : <Key>Left\n"
	    "osfRight : <Key>Right\n",
	NULL	
    };
    static VOptionDescRec options[] = {
	{ "in", VStringRepn, 1, & in_filename, & in_found, NULL,
	      "File of images to be adjusted" },
	{ "out", VStringRepn, 1, & out_filename, & out_found, NULL,
	      "Destination of adjusted images" },
	{ "signed", VBooleanRepn, 1, & signed_view, VOptionalOpt, NULL,
	      "Display 0 as medium grey" }
    };
    
    VXInit ("VXAdjust", resources, & argc, argv);
    
    if (! VParseCommand (VNumber (options), options, & argc, argv) ||
	! VIdentifyFiles (VNumber (options), options, "in", & argc, argv, 0) ||
	! VIdentifyFiles (VNumber (options), options, "out",
			  & argc, argv, 1) ||
	argc > 1) {
	if (argc > 1)
	    VReportBadArgs (argc, argv);
	VReportUsage (argv[0], VNumber (options), options,
		      "[infile] [outfile]");
	VXReportValidOptions ();
        fprintf (stderr, "    plus X Windows options.\n\n");
	exit (EXIT_FAILURE);
    }
    
    /* Load the input file: */
    ReadFile ();

    /* Open the output file: */
    if (strcmp (out_filename, "-") != 0) {
	out_file = fopen (out_filename, "w");
	if (out_file == NULL)
	    VError ("Unable to open output file %s", out_filename);
    } else {
	out_file = stdout;
	out_filename = "<stdout>";
    }

    /* Create menus: */
    VXAddMenu ("_File",
	       "Save and _Quit", "Ctrl<Key>Q", Quit, NULL,
	       (VStringConst) NULL);
    
    VXAddMenu ("_Help",
	       "On _Mouse Usage", NULL, HelpOnMouseUsage, NULL,
	       "On Me_nus", NULL, HelpOnMenus, NULL,
	       "On _Control Panel", NULL, HelpOnCP, NULL,
	       "On _Version", NULL, HelpOnVersion, NULL,
	       (VStringConst) NULL);
    
    /* Add input callback: */
    VXAddInputCallback (VXIbuttonPress, ReportPixel, NULL);

    /* Handle ``signed'' flag: */
    if (signed_view)
	XtVaSetValues (VXGetImageViewWidget(), VxNabsolute, FALSE,
		       (char *) NULL);
    
    /* Add a callback to popup control panel: */
    XtAddCallback (VXGetImageViewWidget(), VxNexposeCallback,
		   PopupControlPanel, NULL);
    
    /* Begin application event loop: */
    VXAppMainLoop ();

    return EXIT_SUCCESS;
}


/*
 * PopupControlPanel
 *
 * Popups the control panel on first exposure event.
 */

/*ARGSUSED*/
void PopupControlPanel (Widget w, XtPointer client_data, XtPointer call_data)
{
    Widget form, reset, next, prev;
    Dimension width;
    Position x,y;
    static VBoolean first_time = TRUE;
    static Arg form_args[] = {
	{ XmNautoUnmanage, FALSE },
	{ XmNdefaultPosition, FALSE }
    };

    #define XVCMW XtVaCreateManagedWidget

    /* Only respond to first exposure: */
    if (! first_time)
	return;
    first_time = FALSE;
    
    /* Create a form dialog: */
    form = XmCreateFormDialog (VXGetApplicationShell(), "form",
			       form_args, XtNumber (form_args));

    /* Position it 10 pixels to the right of the application shell: */
    XtVaGetValues (VXGetApplicationShell(), XmNx, & x, XmNy, & y,
		   XmNwidth, & width, (char *) NULL);
    XtVaSetValues (form, XmNx, x + width + 10, XmNy, y, (char *) NULL);
    
    /* Participate in the Editres Protocol: */
#if defined(XtSpecificationRelease) && XtSpecificationRelease >= 5
    XtAddEventHandler (XtParent (form), (EventMask) 0,
		       True, _XEditResCheckMessages, NULL);
#endif

    /* Create the brightness and contrast scales: */
    brightness_widget = XVCMW ("brightnessScale", xmScaleWidgetClass, form,
			       XmNorientation, XmHORIZONTAL,
			       XmNshowValue, TRUE,
			       XmNtopAttachment, XmATTACH_FORM,
			       XmNleftAttachment, XmATTACH_FORM,
			       XmNrightAttachment, XmATTACH_FORM,
			       (char *) NULL);

    contrast_widget = XVCMW ("contrastScale", xmScaleWidgetClass, form,
			     XmNorientation, XmHORIZONTAL,
			     XmNshowValue, TRUE,
			     XmNtopAttachment, XmATTACH_WIDGET,
			     XmNtopWidget, brightness_widget,
			     XmNtopOffset, 10,
			     XmNleftAttachment, XmATTACH_FORM,
			     XmNrightAttachment, XmATTACH_FORM,
			     (char *) NULL);

    /* Create various buttons: */
    reset = XVCMW ("reset", xmPushButtonGadgetClass, form,
		   XmNtopAttachment, XmATTACH_WIDGET,
		   XmNtopWidget, contrast_widget,
		   XmNbottomAttachment, XmATTACH_FORM,
		   XmNleftAttachment, XmATTACH_FORM,
		   XmNnavigationType, XmTAB_GROUP, (char *) NULL);
    prev = XVCMW ("prev", xmArrowButtonGadgetClass, form,
		  XmNarrowDirection, XmARROW_LEFT,
		  XmNtopAttachment, XmATTACH_OPPOSITE_WIDGET,
		  XmNtopWidget, reset,
		  XmNbottomAttachment, XmATTACH_FORM,
		  XmNleftAttachment, XmATTACH_WIDGET,
		  XmNleftWidget, reset,
		  XmNnavigationType, XmTAB_GROUP, (char *) NULL);
    next = XVCMW ("next", xmArrowButtonGadgetClass, form,
		  XmNarrowDirection, XmARROW_RIGHT,
		  XmNtopAttachment, XmATTACH_OPPOSITE_WIDGET,
		  XmNtopWidget, reset,
		  XmNbottomAttachment, XmATTACH_FORM,
		  XmNleftAttachment, XmATTACH_WIDGET,
		  XmNleftWidget, prev,
		  XmNrightAttachment, XmATTACH_FORM,
		  XmNnavigationType, XmTAB_GROUP, (char *) NULL);
		   
    /* Add callback functions: */
    XtAddCallback (contrast_widget, 
		   XmNvalueChangedCallback, AdjustView, NULL);
    XtAddCallback (brightness_widget, 
		   XmNvalueChangedCallback, AdjustView, NULL);
    XtAddCallback (reset, XmNactivateCallback, ResetView, NULL);
    XtAddCallback (prev, XmNactivateCallback, PrevView, NULL);
    XtAddCallback (next, XmNactivateCallback, NextView, NULL);
    
    XtManageChild (form);
    
    /* Show the first view: */
    ShowView ();
    VXShowMessageArea();

    #undef XVCMW
}


/*
 * ReadFile
 *
 * Load objects from the input file, and create a list of images to be
 * adjusted.
 */

void ReadFile (void)
{
    FILE *f;
    VAttrListPosn posn;
    VImage image, orig;
    View *view;
    int band;
    VBoolean color;
    VString pathname, prev_filename = NULL, cur_filename;
    char title[1024];
    
    /* Read attributes from the input file: */
    if (strcmp (in_filename, "-") != 0) {
	f = fopen (in_filename, "r");
	if (f == NULL)
	    VError ("Unable to open input file %s", in_filename);
    } else {
	f = stdin;
	in_filename = "<stdin>";
    }
    if (! (attributes = VReadFile (f, NULL)))
	VError ("%s is not a valid Vista data file", in_filename);
    if (f != stdin)
	fclose (f);

    /* For each image attribute in the file: */
    views = VListCreate ();
    for (VFirstAttr (attributes, & posn), nimages = 0;
	 VAttrExists (& posn); VNextAttr (& posn)) {
	if (VGetAttrRepn (& posn) != VImageRepn)
	    continue;
	VGetAttrValue (& posn, NULL, VImageRepn, & image);

	nimages++;
	color = (VImageColorInterp (image) == VBandInterpRGB &&
		 VImageNComponents (image) == 1);

	/* Retain a copy of the original image, without adjustments: */
	if (! (orig = VCopyImage (image, NULL, VAllBands)))
	    VError ("ReadFile: Internal error");

	/* Create a view for each band: */
	for (band = 0; band < VImageNBands (image); band++) {
	    view = VNew (View);
	    view->orig = orig;
	    view->image_idx = nimages;
	    view->band = band;
	    view->color = color;
	    view->adjusted = image;
	    view->brightness = view->contrast = 0.0;
	    VListAppend (views, view);
	}
    }

    if (nimages == 0) {
	VWarning ("No input images");
	Quit (NULL);
    }
    VListFirst (views);

    /* Display filename as top-level window title: */
    pathname = VNewString (in_filename);
    cur_filename = strtok (pathname, "/");
    while (cur_filename) {
	prev_filename = cur_filename;
	cur_filename = strtok (NULL, "/");
    }
    sprintf (title, "vxadjust: %s", prev_filename);
    XtVaSetValues (VXGetApplicationShell(),
		   XmNtitle, title, XmNiconName, title, (char *) NULL);
    VFree (pathname);
}


/*
 *  ShowView
 *
 *  Display the current image view.
 */

void ShowView (void)
{
    View *view = VListCurr (views);
    short pts;

    /* Display the appropriate image and band(s): */
    VXSetImage (view->adjusted,
		view->color ? (view->band / 3) * 3 : view->band, 1.0, 0, 0);
    
    /* Set scale values: */
    XtVaGetValues (brightness_widget, XmNdecimalPoints, & pts, (char *) NULL);
    XmScaleSetValue (brightness_widget,
		     (int) (view->brightness * pow (10.0, (double) pts)));
    XtVaGetValues (contrast_widget, XmNdecimalPoints, & pts, (char *) NULL);
    XmScaleSetValue (contrast_widget,
		     (int) (view->contrast * pow (10.0, (double) pts)));

    VXDisplayMessage (TRUE, "Adjusting band %d of image %d",
		      view->band, view->image_idx);
}


/*
 * NextView
 *
 * Move on to next band.
 */

/*ARGSUSED*/
void NextView (Widget w, XtPointer client_data, XtPointer call_data)
{
    if (VListNext (views))
	ShowView ();
    else {
	VListLast (views);
	VWarning ("No more images.");
    }
}


/* 
 * PrevView
 *
 * Back up to previous band.
 */

/*ARGSUSED*/
void PrevView (Widget w, XtPointer client_data, XtPointer call_data)
{
    if (VListPrev (views))
	ShowView ();
    else {
	VListFirst (views);
	VWarning ("No previous images.");
    }
}


/*
 * AdjustView
 *
 * Respond to change in the brightness or contrast scale.
 */

/*ARGSUSED*/
void AdjustView (Widget w, XtPointer client_data, XtPointer call_data)
{
    View *view = VListCurr (views);
    short pts;
    int i;
    VImage tmp;
    
    /* Determine the new contrast and brightness: */
    XtVaGetValues (brightness_widget,
		   XmNdecimalPoints, & pts, XmNvalue, & i, (char *) NULL);
    view->brightness = i / pow (10.0, (double) pts);
    XtVaGetValues (contrast_widget,
		   XmNdecimalPoints, & pts, XmNvalue, & i, (char *) NULL);
    view->contrast = i / pow (10.0, (double) pts);
    
    /* Adjust the image: */
    if (VImageNBands (view->orig) == 1) {
	if (! VAdjustImage (view->orig, view->adjusted, view->band,
			    view->brightness, view->contrast))
	    VError ("AdjustView: Internal error");
    } else {
	if (! (tmp = VAdjustImage (view->orig, NULL, view->band,
				   view->brightness, view->contrast)) ||
	    ! VCopyBand (tmp, 0, view->adjusted, view->band))
	    VError ("AdjustView: Internal error");
	VDestroyImage (tmp);
    }
    
    /* Display the adjusted image: */
    ShowView ();
}


/*
 * ResetView
 *
 * Undo any adjustment made to an image band.
 */

/*ARGSUSED*/
void ResetView (Widget w, XtPointer client_data, XtPointer call_data)
{
    View *view = VListCurr (views);

    /* Reset contrast and brightness to 0: */
    view->brightness = view->contrast = 0.0;
    VCopyBand (view->orig, view->band, view->adjusted, view->band);
    ShowView ();
}


/*
 * ReportPixel
 *
 * Report the location and value of a pixel.
 */

/*ARGSUSED*/
void ReportPixel (VXInputData input_data, VPointer client_data)
{
    View *view = VListCurr (views);

    /* Only respond to left mouse button with no modifier: */
    if (input_data->value != 1 || input_data->modifiers > 0)
	return;

    /* Report row and column coordinates: */
    VXDisplayMessage (TRUE, 
		      "Adjusting band %d of image %d\n"
		      "Pixel: [%4d,%4d]; Value: %g",
		      view->band, view->image_idx,
		      input_data->row, input_data->column,
		      VGetPixel (view->adjusted, view->band,
				 input_data->row, input_data->column));
}


/*
 * Quit
 *
 * Exit application.
 */

/*ARGSUSED*/
void Quit (VPointer client_data) 
{
    /* Write attributes to the output file: */
    if (! VWriteFile (out_file, attributes))
	VError ("Unable to write to %s", out_filename);

    fprintf (stderr, "vxadjust: adjusted %d image(s)\n", nimages);
    exit (EXIT_SUCCESS);
}


/*
 * HelpOnMouseUsage
 */

/*ARUGSUSED*/
void HelpOnMouseUsage (VPointer client_data)
{
    VXPopupTextBox 
	(15, 70,
	 "Help on mouse usage",
	 "Press...               To...\n"
	 "---------------------------------------------------------------\n"
	 "Left Button            Display the location and value of the\n"
	 "                       pixel under the cursor.\n\n"
	 "Middle Button          Zoom in by a factor of 2, with position\n"
	 "                       of cursor as zoom center.\n\n"
	 "Right Button           Zoom out by a factor of 2, with position\n"
	 "                       of cursor as zoom center.\n\n"
	 "Ctrl Middle Button     Move zoom center to position of cursor."
	 );

    return;
}


/*
 * HelpOnMenus
 *
 * Display help information on menus.
 */

/*ARGSUSED*/
void HelpOnMenus (VPointer client_data)
{
    VXPopupTextBox (15, 70,
	 "Help on Menus",
	 "Menu  Entry                 Usage\n"
	 "------------------------------------------------------------------\n"
	 "File  Save and Quit         Save the adjusted images and quit\n"
	 "                            this program.\n\n"
	 "Help  On Mouse Usage        Help message on how to use the mouse\n"
	 "                            to interact with this program.\n\n"
	 "Help  On Menus              This help message.\n\n"
	 "Help  On Control Panel      Help message on how to use the\n"
	 "                            Control Panel.\n\n"
	 );
}


/*
 * HelpOnCP
 *
 * Display help information on Control Panel.
 */

/*ARGSUSED*/
void HelpOnCP (VPointer client_data)
{
    VXPopupTextBox (16, 70,
	 "Help on Control Panel",
	 "Object                Usage\n"
	 "------------------------------------------------------------------\n"
	 "Brightness Scale      Move slider to left to decrease brightness\n"
	 "                      or to right to increase brightness.\n\n"
	 "Contrast Scale        Move slider to left to decrease contrast\n"
	 "                      or to right to increase contrast.\n\n"
	 "Reset Button          Reset the brightness and contrast of a band\n"
	 "                      of an image to their original values.\n\n"
	 "Left-Arrow Button     Move backward to adjust previous band.\n\n"
	 "Right-Arrow Button    Move forward to adjust next band.\n\n"
	 );
}


/*
 * HelpOnVersion
 */

/*ARGSUSED*/
void HelpOnVersion (VPointer client_data)
{
    #define versionString \
        "\n\n\n\n  vxadjust version " ## Version ## ", compiled " ## __DATE__

    VXPopupMessageBox ("Version", versionString);

    #undef versionString
}
