/*
 *  $Id: vxcrop.c,v 1.4 1994/06/15 23:14:49 pope Exp pope $
 *
 *  This file contains the implementation of vxcrop, a program to crop
 *  images interactively.
 */

/*
 *  Copyright 1993, 1994 University of British Columbia
 *
 *  Permission to use, copy, modify, distribute, and sell this software and its
 *  documentation for any purpose is hereby granted without fee, provided that
 *  the above copyright notice appears in all copies and that both that
 *  copyright notice and this permission notice appear in supporting
 *  documentation. UBC makes no representations about the suitability of this
 *  software for any purpose. It is provided "as is" without express or
 *  implied warranty.
 *
 *  Author: Daniel Ko, UBC Laboratory for Computational Intelligence
 */

/* From the Vista library: */
#include "vista/Vlib.h"
#include "vista/mu.h"
#include "vista/option.h"
#include "vista/os.h"
#include "vista/VImage.h"
#include "vista/VList.h"
#include "vista/VX.h"

#include "vista/VImageView.h"

/* File identification string: */
VRcsId ("$Id: vxcrop.c,v 1.4 1994/06/15 23:14:49 pope Exp pope $");

/* Input and output filenames: */
VString in_filename;
VString out_filename;
FILE *out_file;

/* All attributes read from the input file: */
VAttrList attributes;

/* The list of image views assembled from the input images: */
typedef struct {
    VAttrListPosn posn;		/* position in attribute list */
    VImage image;		/* image at that attribute */
    int image_idx;		/* it's position within input file */
    int top, left;		/* cropping box origin and dimensions */
    int width, height;
} View;
VList views;
int nimages;			/* number of images */

/* Color and current coordinates of the cropping box: */
VStringConst box_color = "green";
int start_row, start_column;


/* Later in this file: */
void ReadFile (void);

void ShowView (void);
void DrawCropRect (void);

void KeyPressed (VXInputData, VPointer);

void RubberBandStart (VXInputData, VPointer);
void RubberBandEnd (VXInputData, VPointer);
void RubberBandMove (VXInputData, VPointer);

void Quit (VPointer);

void HelpOnRegion (VPointer);
void HelpOnMenus (VPointer);
void HelpOnMouseUsage (VPointer);
void HelpOnKeyboard (VPointer);
void HelpOnVersion (VPointer);


/*
 *  Entry point:
 */

int main (int argc, char **argv)
{
    static VBoolean in_found, out_found, signed_view = FALSE;
    static VStringConst resources[] = {
	"VXCrop.messageAreaNLines: 2",
	NULL
    };
    static VOptionDescRec options[] = {
	{ "in", VStringRepn, 1, & in_filename, & in_found, NULL,
	      "File of images to be adjusted" },
	{ "out", VStringRepn, 1, & out_filename, & out_found, NULL,
	      "Destination of adjusted images" },
	{ "signed", VBooleanRepn, 1, & signed_view, VOptionalOpt, NULL,
	      "Display 0 as medium grey" },
	{ "color", VStringRepn, 1, & box_color, VOptionalOpt, NULL,
	      "Color of cropping rectangle" }
    };

    VXInit ("VXCrop", resources, & argc, argv);

    if (! VParseCommand (VNumber (options), options, & argc, argv) ||
	! VIdentifyFiles (VNumber (options), options, "in", & argc, argv, 0) ||
	! VIdentifyFiles (VNumber (options), options, "out",
			  & argc, argv, 1) ||
	argc > 1) {
	if (argc > 1)
	    VReportBadArgs (argc, argv);
	VReportUsage (argv[0], VNumber (options), options,
		      "[infile] [outfile]");
	VXReportValidOptions ();
	fprintf (stderr, "    plus X Windows options.\n\n");
	exit (EXIT_FAILURE);
    }

    /* Create menus: */
    VXAddMenu ("_File",
	       "Save and _Quit", "Ctrl<Key>Q", Quit, NULL,
	       (VStringConst) NULL);

    VXAddMenu ("_Help",
	       "On _Selecting a Region", NULL, HelpOnRegion, NULL,
	       "On _Mouse Usage", NULL, HelpOnMouseUsage, NULL,
	       "On _Keyboard Usage", NULL, HelpOnKeyboard, NULL,
	       "On Me_nus", NULL, HelpOnMenus, NULL,
	       "On _Version", NULL, HelpOnVersion, NULL,
	       (VStringConst) NULL);

    /* Add input callbacks: */
    VXAddInputCallback (VXIbuttonPress, RubberBandStart, NULL);
    VXAddInputCallback (VXIbuttonRelease, RubberBandEnd, NULL);
    VXAddInputCallback (VXIpointerMotion, RubberBandMove, NULL);
    VXAddInputCallback (VXIkeyPress, KeyPressed, NULL);

    /* Handle ``signed'' flag: */
    if (signed_view)
	XtVaSetValues (VXGetImageViewWidget(), VxNabsolute, FALSE,
		       (char *) NULL);

    /* Set color of cropping box: */
    if (! VXSetLineColor (box_color))
	exit (EXIT_FAILURE);

    /* Load the input file: */
    ReadFile ();

    /* Open the output file: */
    if (strcmp (out_filename, "-") != 0) {
	out_file = fopen (out_filename, "w");
	if (out_file == NULL)
	    VError ("Unable to open output file %s", out_filename);
    } else {
	out_file = stdout;
	out_filename = "<stdout>";
    }

    /* Display the first image: */
    ShowView ();
    VXShowMessageArea();

    /* Begin application event loop: */
    VXAppMainLoop ();

    return EXIT_SUCCESS;
}


/*
 *  ReadFile
 *
 *  Load objects from the input file, and create a list of images to be
 *  adjusted.
 */

void ReadFile (void)
{
    FILE *f;
    VAttrListPosn posn;
    VImage image;
    View *view;
    VString pathname, prev_filename = NULL, cur_filename;
    char title[1024];

    /* Read attributes from the input file: */
    if (strcmp (in_filename, "-") != 0) {
	f = fopen (in_filename, "r");
	if (f == NULL)
	    VError ("Unable to open input file %s", in_filename);
    } else {
	f = stdin;
	in_filename = "<stdin>";
    }
    if (! (attributes = VReadFile (f, NULL)))
	VError ("%s is not a valid Vista data file", in_filename);
    if (f != stdin)
	fclose (f);

    /* For each image attribute in the file: */
    views = VListCreate ();
    for (VFirstAttr (attributes, & posn);
	 VAttrExists (& posn); VNextAttr (& posn)) {
	if (VGetAttrRepn (& posn) != VImageRepn)
	    continue;
	VGetAttrValue (& posn, NULL, VImageRepn, & image);

	/* Create a view for the image: */
	view = VNew (View);
	view->posn = posn;
	view->image = image;
	view->image_idx = nimages++;
	view->top = view->left = 0;
	view->height = VImageNRows (image);
	view->width = VImageNColumns (image);
	VListAppend (views, view);
    }

    if (nimages == 0) {
	VWarning ("No input images");
	Quit (NULL);
    }
    VListFirst (views);

    /* Display filename as top-level window title: */
    pathname = VNewString (in_filename);
    cur_filename = strtok (pathname, "/");
    while (cur_filename) {
	prev_filename = cur_filename;
	cur_filename = strtok (NULL, "/");
    }
    sprintf (title, "vxcrop: %s", prev_filename);
    XtVaSetValues (VXGetApplicationShell(),
		   XmNtitle, title, XmNiconName, title, (char *) NULL);
    VFree (pathname);
}


/*
 *  ShowView
 *
 *  Display the current image view with its cropping box.
 */

void ShowView (void)
{
    View *view = VListCurr (views);

    if (! VXSetImage (view->image, 0, 1.0, 0, 0))
	exit (EXIT_FAILURE);
    DrawCropRect ();
}


/*
 *  DrawCropRect
 *
 *  Draw the current image's cropping box.
 */

void DrawCropRect (void)
{
    View *view = VListCurr (views);
    double t, l, b, r;

    t = view->top;
    l = view->left;
    b = view->top + view->height;
    r = view->left + view->width;
    VXClearLines ();
    if (! VXDrawLine (t, l, t , r) || ! VXDrawLine (b, l, b, r) ||
	! VXDrawLine (t, l, b, l) || ! VXDrawLine (t, r, b, r))
	exit (EXIT_FAILURE);
    VXDisplayMessage (TRUE,
		      "Cropping image %d.\n"
		      "Region: [%d, %d]; %d x %d",
		      view->image_idx, view->top, view->left,
		      view->height, view->width);
}


/*
 *  KeyPressed
 *
 *  Respond to a keypress over the image.
 */

/*ARGSUSED*/
void KeyPressed (VXInputData input_data, VPointer client_data)
{
    switch (input_data->value) {

    case 0x0e:				/* Ctrl-N */
    case 'N':
    case 'n':
	if (VListNext (views))
	    ShowView ();
	else {
	    VListLast (views);
	    VWarning ("No more images.");
	}
	break;

    case 0x10:				/* Ctrl-P */
    case 'P':
    case 'p':
	if (VListPrev (views))
	    ShowView ();
	else {
	    VListFirst (views);
	    VWarning ("No previous images.");
	}
    }
}


/*
 *  RubberBandStart
 *
 *  Mark starting point of rubberband.
 */

/*ARGSUSED*/
void RubberBandStart (VXInputData input_data, VPointer client_data)
{
    View *view = VListCurr (views);

    /* Only respond to left button with no modifier: */
    if (input_data->value != 1 || input_data->modifiers != 0)
	return;

    /* Store starting point. */
    view->top = start_row = input_data->row;
    view->left = start_column = input_data->column;
    view->height = view->width = 1;

    /* Clear overlays. */
    DrawCropRect ();
}


/*
 *  RubberBandMove
 *
 *  Extend rubberband rectangle.
 */

/*ARGSUSED*/
void RubberBandMove (VXInputData input_data, VPointer client_data)
{
    View *view = VListCurr (views);

    /* Only respond to left button being pressed: */
    if (! (input_data->modifiers & VXMbutton1))
	return;

    /* Grab pointer: */
    XtGrabPointer (VXGetImageViewWidget(), True, PointerMotionMask,
		   GrabModeAsync, GrabModeAsync,
		   XtWindow (VXGetImageViewWidget()), None, CurrentTime);

    /* Determine the cropping parameters: */
    view->top = VMin (start_row, input_data->row);
    view->left = VMin (start_column, input_data->column);
    view->height = VMax ((start_row - input_data->row + 1),
			 (input_data->row - start_row + 1));
    view->width = VMax ((start_column - input_data->column + 1),
			(input_data->column - start_column + 1));

    /* Update the display: */
    DrawCropRect ();
}


/*
 *  RubberBandEnd
 *
 *  Set cropping region.
 */

/*ARGSUSED*/
void RubberBandEnd (VXInputData input_data, VPointer client_data)
{
    /* Only respond to left button with no modifier. */
    if (input_data->value != 1 || input_data->modifiers != VXMbutton1)
	return;

    /* Ungrab pointer: */
    XtUngrabPointer (VXGetImageViewWidget(), CurrentTime);
}


/*
 *  Quit
 *
 *  Save cropped images and exit.
 */

void Quit (VPointer client_data)
{
    View *view;
    VImage result;

    /* Replace each image with its cropped version: */
    VListFirst (views);
    do {
	view = VListCurr (views);
	if (view->top == 0 && view->left == 0 &&
	    view->height == VImageNRows (view->image) &&
	    view->width == VImageNColumns (view->image))
	    continue;
	result = VCropImage (view->image, NULL, VAllBands,
			     view->top, view->left, view->height, view->width);
	if (! result)
	    exit (EXIT_FAILURE);
	VSetAttrValue (& view->posn, NULL, VImageRepn, result);
	VDestroyImage (view->image);
    } while (VListNext (views));

    /* Write attributes to the output file: */
    if (! VWriteFile (out_file, attributes))
	VError ("Unable to write to %s", out_filename);

    fprintf (stderr, "vxcrop: cropped %d image%s.\n",
	     nimages, (nimages == 1) ? "" : "s");
    exit (EXIT_SUCCESS);
}


/*
 * HelpOnRegion
 *
 * Help on selecting a region.
 */

/*ARGSUSED*/
void HelpOnRegion (VPointer client_data)
{
    VXPopupTextBox (12, 70,
	 "Help on Selecting a Region",
	 "A cropping region is specified by any two corners across the\n"
	 "diagonal of the cropping region. To specify the first corner,\n"
	 "move the cursor to that corner, and then press and HOLD the left\n"
	 "mouse button. To specify the second corner, move the cursor to\n"
	 "that corner while holding down the left mouse button. You will\n"
	 "see a box outlining the current region as you move the cursor.\n"
	 "After you have reached the second corner, just release the left\n"
	 "mouse button and the region selected will be outlined by a box.\n"
	 "If you made an error while selecting a region, just release the\n"
	 "left mouse button and start all over again.");
}


/*
 * HelpOnMenus
 *
 * Help on using the menus.
 */

/*ARGSUSED*/
void HelpOnMenus (VPointer client_data)
{
    VXPopupTextBox (15, 70,
	 "Help on Menus",
	 "Menu  Entry                   Usage\n"
	 "------------------------------------------------------------------\n"
	 "File  Save and Quit           Save the cropped images and quit\n"
	 "                              this program.\n\n"
	 "Help  On Selecting a Region   Help message on how to select a\n"
	 "                              region of an image.\n\n"
	 "Help  On Mouse Usage          Help message on how to use the mouse\n"
	 "                              to interact with this program.\n\n"
	 "Help  On Keyboard Usage       Help message on how to use the\n"
	 "                              keyboard to interact with this\n"
         "                              program.\n\n"
	 "Help  On Menus                This help message.\n\n"
	 );

}


/*
 * HelpOnMouseUsage
 *
 * Help on using the mouse to interact with the program.
 */

/*ARGSUSED*/
void HelpOnMouseUsage (VPointer client_data)
{
    VXPopupTextBox (10, 70,
	 "Help on mouse usage",
	 "Press...               To...\n"
	 "---------------------------------------------------------------\n"
	 "Middle Button          Zoom in by a factor of 2, with position\n"
	 "                       of cursor as zoom center.\n"
	 "Right Button           Zoom out by a factor of 2, with position\n"
	 "                       of cursor as zoom center.\n"
	 "Ctrl Middle Button     Move zoom center to position of cursor.\n\n\n"
	 "See also: Help - On Selecting a Region."
	 );
}


/*
 * HelpOnKeyboard
 *
 * Help on using the keyboard to interact with the program.
 */

/*ARGSUSED*/
void HelpOnKeyboard (VPointer client_data)
{
    VXPopupTextBox (5, 70,
	 "Help on keyboard usage",
	 "Press...               To...\n"
	 "---------------------------------------------------------------\n"
	 "Ctrl-N                 Move forward to next image.\n\n"
	 "Ctrl-P                 Move backward to previous image."
	 );
}


/*
 * HelpOnVersion
 *
 * Display version number.
 */

/*ARGSUSED*/
void HelpOnVersion (VPointer client_data)
{
    #define versionString \
        "\n\n\n\n  vxcrop version " ## Version ## ", compiled " ## __DATE__

    VXPopupMessageBox ("Version", versionString);

    #undef versionString
}
