/*
 *  $Id: vxviewe.c,v 1.8 1994/06/15 23:14:49 pope Exp pope $
 *
 *  This file contains the implementation of vxviewe, a program for
 *  displaying Vista image and edge sets.
 */

/*
 *  Copyright 1993, 1994 University of British Columbia
 *
 *  Permission to use, copy, modify, distribute, and sell this software and its
 *  documentation for any purpose is hereby granted without fee, provided that
 *  the above copyright notice appears in all copies and that both that
 *  copyright notice and this permission notice appear in supporting
 *  documentation. UBC makes no representations about the suitability of this
 *  software for any purpose. It is provided "as is" without express or
 *  implied warranty.
 *
 *  Author: Daniel Ko, UBC Laboratory for Computational Intelligence
 */

/* From the Vista library: */
#include "vista/Vlib.h"
#include "vista/option.h"
#include "vista/mu.h"
#include "vista/os.h"
#include "vista/VEdges.h"
#include "vista/VImage.h"
#include "vista/VImageView.h"
#include "vista/VList.h"
#include "vista/VX.h"

/* From the standard C library: */
#include <math.h>

/* File identification string: */
VRcsId ("$Id: vxviewe.c,v 1.8 1994/06/15 23:14:49 pope Exp pope $");


/*
 * ImageDescRec
 *
 * Record for storing image-related data.
 */
typedef struct {
    VStringConst filename;	/* file containing current image */
    int idx;			/* index of current image in file */
    VImage image;		/* current image */
    VImage image_disp;          /* current image being displayed
				 * - image_disp maybe blank when
				 *   image is NULL
				 */
    VBand band;			/* current band */
    VBoolean freeze_band;       /* TRUE iff no further band should be
				 * displayed
				 */
} ImageDescRec;


/*
 * EdgesDescRec
 *
 * Record for storing edges-related data.
 */
typedef struct {
    VStringConst filename;	/* file containing edges */
    int idx;			/* index of edges in file */
    VEdges edges;		/* edges set */
} EdgesDescRec;


/*
 * AppRec
 *
 * Record for storing application-related data
 */
typedef struct {
    ImageDescRec im;		/* current image data */
    VList es;			/* list of edges data */
    VArgVector colors;		/* custom colors */
    VArgVector widths;		/* custom widths */
    VArgVector infiles;		/* list of files */
    int file_idx;               /* index of current file in file list */
    VStringConst filename;      /* current file */
    VAttrList attrs;		/* attribute list of current file*/
    VAttrListPosn posn;		/* current position in attribute list */
    VBoolean background_valid;	/* TRUE iff current backgound valid */
    int obj_idx;                /* index of current object (start from 1) */
    VBoolean command_line_mode; /* TRUE if using files in command-line */
    VBoolean graphics_states_changed; /* TRUE iff graphics states changed */ 
} AppRec;


/* Variable(s): */

static AppRec app;
static VStringConst dcolors[] = {
    "green", "blue", "yellow", "cyan", "magenta",
    "purple", "orange", "white", "tomato4", "hot pink",
};


/*
 * An image is displayed in color if it has 3 color dimensions, an RGB
 * color interpretation, and each color has a single component.
 */

#define IsColor(image) \
    (VImageColorInterp (image) == VBandInterpRGB && \
     VImageNComponents (image) == 1)


/* Function prototypes: */

static VBoolean InitAppRec (void);
static VBoolean NextObjectInFileList (void);
static VBoolean NextObjectInFile (void);
static void DisplayCurrentObjects (void);

static void Quit          (VPointer client_data);
static void NextObject    (VPointer client_data);
static void OpenFile      (VPointer client_data);
static void ClearImage    (VPointer client_data);
static void ClearEdges    (VPointer client_data);
static void InfoOnImage   (VPointer client_data);
static void InfoOnEdges   (VPointer client_data);
static void ToggleSigned  (VPointer client_data);

static void OnMouseUsage  (VPointer client_data);
static void OnFileMenu    (VPointer client_data);
static void OnViewMenu    (VPointer client_data);
static void OnInfoMenu    (VPointer client_data);
static void OnOptionMenu  (VPointer client_data);
static void OnVersion     (VPointer client_data);

static void PresentInfo (VXInputData input_data, VPointer client_data);

static VImage CreateBlankImage (int nrows, int ncolumns);
static FILE *OpenFileForRead (VStringConst filename);
static VBoolean CloseFile (FILE *f);
static float Distance (float r1, float c1, float r2, float c2, int r3, int c3);
static void FreeEdgesDescRec (EdgesDescRec *es_desc_rec_ptr);


/* Function definition(s): */

/*
 * InitAppRec
 *
 * Initialize fields in app.
 */
static VBoolean InitAppRec (void) 
{
    /* Create the (initially empty) list of edges data: */
    app.es = VListCreate();

    /* If app.attrs is not NULL, it will be deallocated: */
    app.attrs = NULL;

    /* If app.im.image is not NULL, it will be assummed to be valid: */
    app.im.image = NULL;

    app.graphics_states_changed = TRUE;
    app.file_idx = -1;

    return TRUE;
}


/*
 * NextObjectInFileList
 *
 * Search for next object in app.infiles.
 */
static VBoolean NextObjectInFileList (void)
{
    FILE *f;

    while (1) {

	/* Search for next object in current file: */
	if (NextObjectInFile() == TRUE)
	    break;
	
	/* No more objects in current file, so read next file: */
	
	/* Read the next file: */
	app.attrs = NULL;
	do {
	    
	    /* Check that there is another file: */
	    if (++app.file_idx >= app.infiles.number)
		return FALSE;
	    
	    /* Open the new file and read its contents: */
	    app.filename = VNewString (((VStringConst *) app.infiles.vector)
				       [app.file_idx]);
	    f = OpenFileForRead (app.filename);
	    if (f != NULL)

		/* Read attribute list of new file: */
		app.attrs = VReadFile (f, NULL);  
	    
	} while (app.attrs == NULL);
	
	/* Initialize posn to first attribute: */
	VFirstAttr (app.attrs, &(app.posn));;

	/* app.posn points to first top-level object: */
	app.obj_idx = 1;
    }

    return TRUE;
}


static VBoolean NextObjectInFile (void) 
{
    EdgesDescRec *es_desc;

    /* If there are more bands in current image, select the next 1 or 3: */
    if (app.im.image != NULL && !app.im.freeze_band) {
	if (IsColor (app.im.image) &&
	    app.im.band + 5 < VImageNBands (app.im.image)) {
	    app.im.band += 3;		/* next RGB triple */
	    app.background_valid = FALSE;
	    return TRUE;
	}
	if (app.im.band + 1 < VImageNBands (app.im.image)) {
	    app.im.band++;		/* next single band */
	    app.background_valid = FALSE;
	    return TRUE;
	}
    }
    
    /* Pick the next selection from the current file: */
    if (app.attrs != NULL)
	
	/* Scan each attribute: */
	for (/* keep current value of app.posn */; 
	     VAttrExists (&(app.posn)); 
	     VNextAttr (&(app.posn)), (app.obj_idx)++) {

	    switch (VGetAttrRepn (& app.posn)) {

	    case VImageRepn:
		
		/* Set image filename: */
		if (app.im.filename != NULL)
		    VFree ((VString) app.im.filename);
		app.im.filename = XtNewString (app.filename);
		
		/* Set image index: */
		app.im.idx = app.obj_idx;
		
		/* Set image value: */
		if (app.im.image != NULL)
		    VDestroyImage (app.im.image);
		VGetAttrValue (& app.posn, NULL, VImageRepn, & app.im.image);
		app.im.band = 0;
		app.im.image_disp = app.im.image;
		app.im.freeze_band = FALSE;

		/* Increment app.posn and app.obj_idx: */
		VNextAttr (&(app.posn));
		(app.obj_idx)++;
		
		/* Remove any edges overlays */
		VListDestroy (app.es, (void (*) (VPointer)) FreeEdgesDescRec);
		app.es = VListCreate();
	
		app.background_valid = FALSE;
		app.graphics_states_changed = TRUE;
		
		return TRUE;
	    
	    case VEdgesRepn:
		
		/* Create a new edges description record: */
		es_desc = VMalloc (sizeof (EdgesDescRec));

		/* Set edges filename: */
		es_desc->filename = XtNewString (app.filename);

		/* Set edges index: */
		es_desc->idx = app.obj_idx;

		/* Set the value of es_desc->edges: */
		VGetAttrValue (& app.posn, NULL, VEdgesRepn, & es_desc->edges);
		
		/* Add es_desc to app.es: */
		VListAppend (app.es, es_desc);

		/* Increment app.posn and app.obj_idx: */
		VNextAttr (&(app.posn));
		(app.obj_idx)++;
		
		return TRUE;

	    default:
		break;
	    }
	}
    
    return FALSE;
}


/*
 * DisplayCurrentObjects
 *
 * Display current image and edges.
 */
static void DisplayCurrentObjects (void)
{
    int i;
    EdgesDescRec *es_desc;
    char *pathname, *filename, *prev_filename = NULL;
    char title[1024];
    
    VXClearLines ();

    app.graphics_states_changed = TRUE;

    /* Set image if image is new: */
    if (app.im.image != NULL && !app.background_valid) {
	VXSetImage (app.im.image, app.im.band, 1.0, 0, 0);
	app.background_valid = TRUE;
    }
    
    /* Display edges on blank image if no current image: */
    if (VListCount (app.es) != 0 && app.im.image == NULL) {
	es_desc = (EdgesDescRec *) VListFirst (app.es);
	app.im.image_disp = CreateBlankImage(VEdgesNRows(es_desc->edges),
					     VEdgesNColumns(es_desc->edges));
	if (app.im.image_disp == NULL) {
	    VWarning ("Unable to display next object");
	    return;
	}
	app.im.band = 0;
	VXSetImage (app.im.image_disp, app.im.band, 1.0, 0, 0);
    }

    /* Dispaly blank image if no current image and edges: */
    if (VListCount (app.es) == 0 && app.im.image == NULL) {
	app.im.image_disp = CreateBlankImage(1,1);
	if (app.im.image_disp == NULL) {
	    VWarning ("Unable to display next object");
	    return;
	}
	app.im.band = 0;
	VXSetImage (app.im.image_disp, app.im.band, 1.0, 0, 0);
    }
    
    /* Draw edges: */
    es_desc = (EdgesDescRec *) VListFirst (app.es);
    i = 0;
    while (es_desc != NULL) {
	VXSetLineWidth (i >= app.widths.number 
			? 0.0
			: (double) ((VFloat *) app.widths.vector)[i]);
	VXSetLineColor (i >= app.colors.number
			? dcolors[i % VNumber(dcolors)]
			: ((VStringConst *) app.colors.vector)[i]);
	VXDrawEdges (es_desc->edges);
	es_desc = (EdgesDescRec *) VListNext (app.es);
	i++;
    }
    
    /* Display filename as top-level window title: */

    if (VListCount (app.es) > 0)
	pathname = XtNewString (((EdgesDescRec *) VListLast (app.es))->
				filename);
    else if (app.im.image != NULL)
	pathname = XtNewString (app.filename);
    else
	pathname = NULL;
    
    if (pathname != NULL) {
	filename = strtok (pathname, "/");
	while (filename != NULL) {
	    prev_filename = filename;
	    filename = strtok (NULL, "/");
	}
    } else prev_filename = "nothing";
    
    sprintf (title, "vxviewe: %s", prev_filename);
    XtVaSetValues (VXGetApplicationShell(),
		   XmNtitle, title,
		   XmNiconName, title,
		   NULL);
    if (pathname != NULL)
	VFree (pathname);

    /* Clear message area: */
    VXDisplayMessage (TRUE, "");
}


/*
 * Quit
 *
 * Exit application.
 */
static void Quit(VPointer client_data) 
{
    exit (EXIT_SUCCESS);
}


/*
 * NextObject
 *
 * Find and display next object.
 */
static void NextObject(VPointer client_data)
{
    VBoolean can_display;

    /* Search for next object: */
    if (app.command_line_mode) 
	can_display = NextObjectInFileList ();
    else
	can_display = NextObjectInFile ();

    /* Display next object: */
    if (can_display)
	DisplayCurrentObjects();
    else VWarning ("No more images or edges");
}


/*
 * OpenFile
 *
 * Open a new file and display first object.
 */
static void OpenFile(VPointer client_data)
{
    FILE *f;

    /* Let user select a file: */
    if (app.filename != NULL)
	VFree ((VString) app.filename);
    app.filename = VXPopupFileBox("Open Vista file");
    if (app.filename == NULL) {  /* no file is selected */
	return;
    }
    
    /* Open the new file and read its contents: */
    f = OpenFileForRead (app.filename);
    if (f == NULL)
	return;
    app.attrs = VReadFile (f, NULL);
    if (app.attrs == NULL) {
	VWarning ("%s is not a Vista file", app.filename);
	return;
    }
    
    /* Initialize posn to first attribute. */
    VFirstAttr (app.attrs, &(app.posn));

    app.obj_idx = 1;
    
    app.im.freeze_band = TRUE;
    app.command_line_mode = FALSE;

    CloseFile (f);

    NextObject (NULL);
}


/*
 * ClearImage
 *
 * Clear current image.
 */
static void ClearImage (VPointer client_data)
{
    if (app.im.image == NULL)
	return;
    
    VDestroyImage (app.im.image);
    app.im.image = NULL;
    
    app.background_valid = FALSE;
    app.graphics_states_changed = TRUE;
    
    DisplayCurrentObjects();
}


/*
 * ClearEdges
 *
 * Clear the last set of edges.
 */
static void ClearEdges (VPointer client_data)
{
    EdgesDescRec *es_desc;
    
    if (VListCount (app.es) == 0)
	return;
    
    es_desc = VListTrim (app.es);
    FreeEdgesDescRec (es_desc);

    app.graphics_states_changed = TRUE;

    DisplayCurrentObjects();
}


/*
 * InfoOnImage
 *
 * Diplay current image info.
 */
static void InfoOnImage (VPointer client_data)
{
    char str[1024], *cp;
    float aspect = 1.0;

    if (app.im.image == NULL) {
	sprintf (str, "Information unavailable.");
    } else {
	VGetAttr (VImageAttrList (app.im.image), VPixelAspectRatioAttr, NULL,
		  VFloatRepn, &aspect);

	strcpy (str, "Current image is ");
	cp = str + strlen (str);
	if (IsColor (app.im.image))
	    sprintf (cp, "bands %d-%d", app.im.band, app.im.band + 2);
	else sprintf (cp, "band %d", app.im.band);
	cp += strlen (cp);
	sprintf (cp,
		 " of object %d\n"
		 "of %s.\n\n"
		 "Number of rows       : %d\n"
		 "Number of columns    : %d\n"
		 "Pixel representation : %s\n"
		 "Pixel aspect ratio   : %g\n",
		 app.im.idx,
		 strcmp (app.im.filename, "-") == 0 
		 ? "<stdin>"
		 : app.im.filename,
		 VImageNRows (app.im.image), 
		 VImageNColumns (app.im.image),
		 (VString) VPixelRepnName (app.im.image),
		 aspect);
    }
    
    VXPopupTextBox (10,50,
		    "Image Information", 
		    str);
}


/*
 * InfoOnEdges
 *
 * Display current edges info.
 */
static void InfoOnEdges (VPointer client_data)
{
    char msg[102400];
    char tmp[1024];
    int i;
    EdgesDescRec *es_desc;

    if (VListCount (app.es) == 0) {
	sprintf (msg, "Information unavailable.");
    } else {
	sprintf (msg, "Displaying %d sets of edges.\n\n", VListCount (app.es));
	
	es_desc = (EdgesDescRec *) VListFirst (app.es);
	i = 0;
	while (es_desc != NULL) {
	    sprintf (tmp,
		     "Edges set %d is object %d of\n"
		     "%s\n"
		     " color                      : %s\n"
		     " width                      : %g pixel(s)\n"
		     " number of edges            : %d\n"
		     " number of rows             : %d\n"
		     " number of columns          : %d\n"
		     " number of fields per edge  : %d\n"
		     " number of fields per point : %d\n\n",
		     i,
		     es_desc->idx,
		     strcmp (es_desc->filename, "-") == 0 ? "<stdin>" :
		     es_desc->filename,
		     i >= app.colors.number ? dcolors[i % VNumber(dcolors)] :
		     ((VStringConst *) app.colors.vector)[i],
		     i >= app.widths.number ? 0.0 :
		     ((VFloat *) app.widths.vector)[i],
		     VNEdges (es_desc->edges),
		     VEdgesNRows (es_desc->edges),
		     VEdgesNColumns (es_desc->edges),
		     VNEdgeFields (es_desc->edges),
		     VNPointFields (es_desc->edges));
	    strncat (msg, tmp, sizeof (msg) - strlen (tmp));

	    es_desc = (EdgesDescRec *) VListNext (app.es);
	    i++;
	}
    }
    
    VXPopupTextBox (10, 50,
		    "Edges Information",
		    msg);
}


/*
 * ToggleSigned
 *
 * Toggle between signed and unsigned interpretation of image.
 */
static void ToggleSigned (VPointer client_data)
{
    Boolean not_signed;
    
    XtVaGetValues (VXGetImageViewWidget(),
		   VxNabsolute, &not_signed,
		   (char *) NULL);
    
    XtVaSetValues (VXGetImageViewWidget(),
		   VxNabsolute, !not_signed,
		   (char *) NULL);
    
    if (not_signed == TRUE)
	VXDisplayMessage (FALSE, "Display mode is changed to signed\n");
    else
	VXDisplayMessage (FALSE, "Display mode is changed to unsigned\n");
}


static void OnMouseUsage (VPointer client_data)
{
    VXPopupTextBox 
	(10, 70,
	 "Help on mouse usage",
	 "Press...               To...\n"
	 "---------------------------------------------------------------\n"
	 "Left Button            Report information of the pixel and edge\n"
	 "                       closest to the cursor; highlight the\n"
	 "                       edge closest to the cursor.\n"
	 "Middle Button          Zoom in by a factor of 2, with position\n"
	 "                       of cursor as zoom center.\n"
	 "Right Button           Zoom out by a factor of 2, with position\n"
	 "                       of cursor as zoom center.\n"
	 "Ctrl Middle Button     Move zoom center to position of cursor."
	 );
}

static void OnFileMenu (VPointer client_data)
{
    VXPopupTextBox
	(10, 70,
	 "Help on File menu",
	 "Choose...  To...\n"
	 "---------------------------------------------------------\n"
	 "Open       Open a new Vista file and display its first\n"
	 "           displayable object.\n"
	 "Quit       Quit this program."
	 );
}

static void OnViewMenu (VPointer client_data)
{
    VXPopupTextBox
	(10, 70,
	 "Help on View menu",
         "Choose...    To...\n"
	 "----------------------------------------------------------------\n"
	 "Next Object  Display next displayable object in current file(s).\n"
	 "Clear Image  Clear current image and leave the edges intact.\n"
	 "Clear Edges  Clear the last set of edges and leave the current\n"
	 "             image and other edges sets intact.\n"
	 );
}

static void OnVersion (VPointer client_data)
{
    #define versionString \
        "\n\n\n\n  vxviewe version " ## Version ## ", compiled " ## __DATE__

    VXPopupMessageBox ("Version", versionString);

    #undef versionString
}

static void OnInfoMenu (VPointer client_data)
{
    VXPopupTextBox
	(10, 70,
	 "Help on Info menu",
	 "Choose...  To...\n"
	 "------------------------------------------------\n"
	 "Image      Display information of current image.\n"
	 "Edges      Display information of current edges.\n"
	 );
}

static void OnOptionMenu (VPointer client_data)
{
    VXPopupTextBox
	(10, 70,
	 "Help on Option menu",
	 "Choose...        To...\n"
	 "---------------------------------------------------\n"
	 "Signed/Unsigned  Toggle between signed and unsigned\n"
	 "                 interpretation of the image."
	 );
}


/* 
 * PresentInfo
 *
 * Display info of cursor position.
 */
static void PresentInfo(VXInputData input_data, VPointer client_data)
{
    VEdges es, min_edges = NULL;
    VEdge e, min_edge = NULL;
    int i, j, edge_count, min_edge_idx = 0, min_point_idx, min_edges_idx = 0;
    int min_point1_idx = 0, min_point2_idx = 0;
    int nrows, np, nfields;
    float min_dist, dist;
    float sr = 0.0f, sc = 0.0f, er = 0.0f, ec = 0.0f;
    float p1r = 0.0f, p1c = 0.0f, p2r = 0.0f, p2c = 0.0f, pr, pc;
    VFloat **points;
    char msg[4096], edge_fields[1024], msg_tmp[1024], fields_tmp[256],
         point_fields[1024];
    EdgesDescRec *es_desc;

    /* Only interested in button1 press with no modifiers: */
    if (input_data->value != 1 || input_data->modifiers > 0)
	return;
    
    min_dist = -1;
    
    /* Find the edge that is closest to cursor position: */
    es_desc = (EdgesDescRec *) VListFirst (app.es);
    i = 0;
    while (es_desc != NULL) {
	es = es_desc->edges;
	nrows = VEdgesNRows(es);
	edge_count = 0;
	
	/* For each edge in es: */
	for (e = VFirstEdge(es); VEdgeExists(e); e = VNextEdge(e)) {

	    np = VEdgeNPoints(e);
	    points = VEdgePointArray(e);
	    if (np == 1) {
		dist = Distance ((float) nrows - points[0][1],
				 points[0][0],
				 (float) nrows - points[0][1],
				 points[0][0],
				 input_data->row,
				 input_data->column);
		if (dist < min_dist || min_dist == -1) {
		    min_dist = dist;
		    min_edge = e;
		    min_edge_idx = edge_count;
		    min_edges_idx = i;
		    min_edges = es;
		    min_point1_idx = min_point2_idx = 0;
		    p1r = p2r = (float) nrows - points[0][1];
		    p1c = p2c = points[0][0];
		}
	    } else {
		for (j = 0; j < np - 1; j++) {
		    dist = Distance ((float) nrows - points[j][1],
				     points[j][0],
				     (float) nrows - points[j+1][1],
				     points[j+1][0],
				     input_data->row,
				     input_data->column);
		    if (dist < min_dist || min_dist == -1) {
			min_dist = dist;
			min_edge = e;
			min_edges_idx = i;
			min_edges = es;
			min_edge_idx = edge_count;
			min_point1_idx = j;
			min_point2_idx = j+1;
			p1r = (float) nrows-points[j][1];
			p1c = points[j][0];
			p2r = (float) nrows-points[j+1][1];
			p2c = points[j+1][0];
		    }   
		} /* for */
	    } /* else */
	    
	    edge_count++;

	} /* for each edge */

	es_desc = (EdgesDescRec *) VListNext (app.es);
	i++;
    } /* for each edges */
    
    /* Display info: */
    if (min_dist != -1) { /* closest edge found */
	
	/* Clear any highlighted edge: */
	if (!app.graphics_states_changed) 
	    VXRestoreOverlays ();
	VXStoreOverlays ();
	app.graphics_states_changed = FALSE;
	
	/* Create an instance of VEdges containing only the
	   closest edge: */
	es = VCreateEdges (VEdgesNRows(min_edges),
			   VEdgesNColumns(min_edges),
			   VNEdgeFields(min_edges),
			   VNPointFields(min_edges));
	VAddEdge (es, VEdgeFields(min_edge), VEdgeNPoints(min_edge),
		  *VEdgePointArray(min_edge), VEdgeClosed(min_edge),
		  TRUE);

	/* Highlight the closest edge in red: */
	VXSetLineColor ("red");
	VXSetLineWidth(min_edges_idx >= app.widths.number ? 0.0 :
		       (double) ((VFloat *) app.widths.vector)[min_edges_idx]);
	VXDrawEdges (es);
	VDestroyEdges(es);

	/* Get the end points of the closest edge: */
	np = VEdgeNPoints(min_edge);
	points = VEdgePointArray(min_edge);
	nrows = (float) VEdgesNRows(min_edges);
	sr = nrows - points[0][1];
	sc = points[0][0];
	er = nrows - points[np-1][1];
	ec = points[np-1][0];
    }
    
    /* Display message: */
    msg[0] = '\0';
    if (app.im.image != NULL) { /* image exists */

	if (! IsColor (app.im.image))
	    sprintf (msg_tmp, 
		     "Image: %3d x %3d; Type: %s; Band %d of %d\n"
		     "Pixel: [%3d, %3d]; Value: %g\n",
		     VImageNRows (app.im.image),
		     VImageNColumns (app.im.image),
		     VPixelRepnName (app.im.image),
		     app.im.band,
		     VImageNBands (app.im.image),
		     input_data->row,
		     input_data->column,
		     VGetPixel (app.im.image, app.im.band,
				input_data->row, input_data->column)
		     );
	else 
	    sprintf (msg_tmp, 
		     "Image: %3d x %3d; Type: %s; Band %d-%d of %d\n"
		     "Pixel: [%3d, %3d]; Value: %g, %g, %g\n",
		     VImageNRows (app.im.image),
		     VImageNColumns (app.im.image),
		     VPixelRepnName (app.im.image),
		     app.im.band, app.im.band + 2,
		     VImageNBands (app.im.image),
		     input_data->row, input_data->column,
		     VGetPixel (app.im.image, app.im.band,
				input_data->row, input_data->column),
		     VGetPixel (app.im.image, app.im.band + 1,
				input_data->row, input_data->column),
		     VGetPixel (app.im.image, app.im.band + 2,
				input_data->row, input_data->column)
		     );
	    
	strncat (msg, msg_tmp, sizeof (msg) - strlen (msg_tmp));
    }
    if (min_dist != -1) { /* closest edge found */
	if (Distance (p1r, p1c, p1r, p1c, input_data->row, input_data->column)
	    >
	    Distance (p2r, p2c, p2r, p2c, input_data->row, input_data->column))
	    pr = p2r, pc = p2c, min_point_idx = min_point2_idx;
	else
	    pr = p1r, pc = p1c, min_point_idx = min_point1_idx;

	/* Extract the edge fields: */
	es = min_edges;
	nfields = VNEdgeFields (es);
	if (nfields == 0) {
	    sprintf (edge_fields, "(No edge fields)");
	} else {
	    sprintf (edge_fields, "Fields: {");
	    for (i = 0; i < nfields - 1; i++) {
		sprintf (fields_tmp, " %g,", (VEdgeFields (min_edge))[i]);
		strncat (edge_fields, fields_tmp, 
			 sizeof (edge_fields) - strlen (fields_tmp));
	    }
	    sprintf (fields_tmp, " %g}", (VEdgeFields (min_edge))[i]);
	    strncat (edge_fields, fields_tmp,
		     sizeof (edge_fields) - strlen (fields_tmp));
	}
	
	/* Extract the point fields: */
	es = min_edges;
	nfields = VNPointFields (es);
	if (nfields == 2) {
	    sprintf (point_fields, "(No other point fields)");
	} else {
	    sprintf (point_fields, "Fields: {");
	    points = VEdgePointArray (min_edge);
	    for (i = 2; i < nfields - 1; i++) {
		sprintf (fields_tmp, " %g,", points[min_point_idx][i]);
		strncat (point_fields, fields_tmp,
			 sizeof (edge_fields) - strlen (fields_tmp));
	    }
	    sprintf (fields_tmp, " %g}", points[min_point_idx][i]);
	    strncat (point_fields, fields_tmp,
		     sizeof (edge_fields) - strlen (fields_tmp));
	}
	
	sprintf (msg_tmp,
		 "Edge: %3d; #Points: %3d; [%.1f, %.1f]->[%.1f,%.1f]; %s\n"
		 "Point: %3d; [%.1f, %.1f]; %s\n",
		 min_edge_idx, VEdgeNPoints (min_edge),
		 sr, sc, er, ec,
		 edge_fields,
		 min_point_idx, pr, pc,
		 point_fields);
	strcat (msg, msg_tmp);
    }
    
    VXShowMessageArea ();
    VXDisplayMessage (TRUE, msg);
}


/*
 * Distance
 *
 * Return the squared distance between a point (r3,c3) and the line segment 
 * (r1,c1) to (r2,c2).
 */
static float Distance (float r1, float c1, float r2, float c2, int r3, int c3)
{
    float ax, ay, bx, by, cx, cy, 
    len_proj_b_on_a, len_a, tmp;

    #define Tolerance 0.01

    ax = r2 - r1;
    ay = c2 - c1;
    bx = r3 + 0.5f - r1;
    by = c3 + 0.5f - c1;

    len_a = (ax*ax + ay*ay);

    if (len_a < Tolerance)
	return ((bx*bx + by*by));
	
    tmp = ax*bx + ay*by;
    len_proj_b_on_a = tmp*tmp / len_a;

    if (len_proj_b_on_a <= 0)
	 return (bx*bx + by*by);

    if (len_proj_b_on_a >= len_a) {
	cx = r3 + 0.5f - r2;
	cy = c3 + 0.5f - c2;
	return (cx*cx + cy*cy);
    } 
     
    tmp = ax*by - ay*bx;
    return (tmp*tmp / len_a);

    #undef Tolerance
}



/*
 * OpenFileForRead
 *
 * Open a file for reading.
 * OpenFileForRead returns stdin if filename is "-".
 */
static FILE *OpenFileForRead (VStringConst filename)
{
    FILE *f;
    
    if (strcmp (filename, "-") == 0 )
	f = stdin;
    else {
	f = fopen (filename, "r");
	if (f == NULL) {
	    VWarning ("OpenFileForRead: Unable to open %s",
		      strcmp (filename, "-") == 0 ? "<stdin>" : filename);
	}
    }
    
    return f;
}


/*
 * CloseFile
 *
 * Close a file.
 */
static VBoolean CloseFile (FILE *f)
{
    if (fclose (f) != 0) {
	VWarning ("CloseFile: Unable to close file");
	return FALSE;
    }

    return TRUE;
}


static void FreeEdgesDescRec (EdgesDescRec *es_desc_rec_ptr)
{
    VDestroyEdges (es_desc_rec_ptr->edges);
    VFree ((VString) (es_desc_rec_ptr->filename));
    VFree (es_desc_rec_ptr);
}


/*
 * CreateBlankImage
 *
 * Create a blank (black) image of specified dimensions.
 */
static VImage CreateBlankImage (int nrows, int ncolumns)
{
    VImage im;
    
    if (im = VCreateImage (1, nrows, ncolumns, VBitRepn))
	VFillImage (im, VAllBands, 0.0);
    return im;
}


/* Entry point: */
int main(int argc, char **argv)
{
    static VBoolean infiles_found;
    static VStringConst resources[] = {
	NULL,
    };
    static VOptionDescRec options[] = {
	{ "in", VStringRepn, 0, &(app.infiles), &infiles_found, NULL,
	      "Files to be displayed" },
	{ "colors", VStringRepn, 0, &(app.colors), VOptionalOpt, NULL,
	      "Colors of edges" },
	{ "widths", VFloatRepn, 0, &(app.widths), VOptionalOpt, NULL,
	      "Widths of edges" }
    };
    
    InitAppRec();
    
    VXInit ("Vxviewe", resources, &argc, argv);
    
    /* Parse command line arguments and identify input files: */
    if (!VParseCommand (VNumber (options), options, & argc, argv)) {
	VReportUsage (argv[0], VNumber (options), options, "file ...");
	VXReportValidOptions ();
        fprintf (stderr, "    plus X Windows options.\n\n");
        exit (EXIT_FAILURE);
    }
    VIdentifyFiles (VNumber (options), options, "in", & argc, argv, 0);
    
    app.command_line_mode = (app.infiles.number != 0);

    /* Create menus: */
    VXAddMenu ("_File",
	       "_Open...", "Ctrl<Key>O", OpenFile, NULL,
	       "_Quit", "Ctrl<Key>Q", Quit, NULL,
	       (VStringConst) NULL);

    VXAddMenu ("_View",
	       "_Next Object", "Ctrl<Key>N", NextObject, NULL,
	       "Clear I_mage", "Ctrl<Key>M", ClearImage, NULL, 
	       "Clear E_dges", "Ctrl<Key>D", ClearEdges, NULL,
	       (VStringConst) NULL);

    VXAddMenu ("_Info",
	       "I_mage", "Ctrl<Key>I", InfoOnImage, NULL,
	       "E_dges", "Ctrl<Key>E", InfoOnEdges, NULL,
	       (VStringConst) NULL);

    VXAddMenu ("_Option",
	       "_Signed/Unsigned", "Ctrl<Key>S", ToggleSigned, NULL,
	       (VStringConst) NULL);

    VXAddMenu ("_Help",
	       "On _Mouse usage", NULL, OnMouseUsage, NULL,
	       "On _File menu", NULL, OnFileMenu, NULL,
	       "On _View menu", NULL, OnViewMenu, NULL,
	       "On _Info menu", NULL, OnInfoMenu, NULL,
	       "On O_ption Menu", NULL, OnOptionMenu, NULL,
	       "On Ve_rsion", NULL, OnVersion, NULL,
	       (VStringConst) NULL);

    /* Add input callbacks: */
    VXAddInputCallback(VXIbuttonPress, PresentInfo, NULL);

    VXHideMessageArea ();

    if (app.command_line_mode == TRUE)
	NextObject (NULL);
    
    /* Begin application event loop: */
    VXAppMainLoop ();

    return EXIT_SUCCESS;
}
